/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 
// GL_RSURF.C: surface - related refresh code
#include <assert.h>

#include "gl_local.h"

static vec3_t	modelorg;		// relative to viewpoint

#define DYNAMIC_LIGHT_WIDTH 128
#define DYNAMIC_LIGHT_HEIGHT 128

#define LIGHTMAP_BYTES 4

#define	BLOCK_WIDTH		128
#define	BLOCK_HEIGHT	128

#define	MAX_LIGHTMAPS	128

int		c_visible_lightmaps;
int		c_visible_textures;

#define GL_LIGHTMAP_FORMAT GL_RGBA

typedef struct
{
	int internal_format;
	int	current_lightmap_texture;

	qboolean	lightmap_modified[MAX_LIGHTMAPS];

	int			allocated[BLOCK_WIDTH];

	// the lightmap texture data needs to be kept in
	// main memory so texsubimage can update properly
	// lets waste LOADS of RAM...!  8MB but so f---ing what...?
	byte		lightmap_buffer[MAX_LIGHTMAPS][4 * BLOCK_WIDTH * BLOCK_HEIGHT];
} gllightmapstate_t;

static gllightmapstate_t gl_lms;

typedef struct FrameRenderState_s
{
	// auxillary texture chains
	msurface_t *r_alpha_surfaces;
	msurface_t *r_detailsurfaces;

	// render flags
	qboolean FrameHasLuma;
	qboolean FrameHasLava;
} FrameRenderState_t;

static FrameRenderState_t FrameRenderState;


extern image_t *r_bloomtexture;
extern image_t *r_blacktexture;

static void		LM_InitBlock (void);
static void		LM_UploadBlock (qboolean dynamic);
static qboolean	LM_AllocBlock (int w, int h, int *x, int *y);

extern void R_SetCacheState (msurface_t *surf);
extern void R_BuildLightMap (msurface_t *surf, int stride);
void R_SetupViewport (unsigned int vpx, unsigned int vpy, unsigned int vpw, unsigned int vph);
void R_SetGL2D (void);
void R_RenderLavaSmokeParticle (vec3_t org, int offset);

static cvar_t *gl_lumamode = NULL;
static cvar_t *gl_dot3bump = NULL;


/*
============================================================= 

BRUSH MODELS

============================================================= 
*/ 

/*
=============== 
R_TextureAnimation

Returns the proper texture for a given time and base texture
=============== 
*/ 
image_t *R_TextureAnimation (mtexinfo_t *tex) 
{
	int		c;

	if (!tex->next) 
		return tex->image;

	c = currententity->frame % tex->numframes;
	while (c) 
	{
		tex = tex->next;
		c--;
	}

	return tex->image;
}


/*
================ 
R_DrawAlphaSurfaces

Draw water surfaces and windows.
The BSP tree is waled front to back, so unwinding the chain
of alpha_surfaces will draw back to front, giving proper ordering.
================ 
*/ 
void R_DrawAlphaSurfaces (void) 
{
	msurface_t	*s;
	float		intens;
	float scroll;
	float warpscroll;

	// go back to the world matrix
	// (we need this here even if none are being drawn)
	glLoadMatrixf (r_world_matrix);

	if (!FrameRenderState.r_alpha_surfaces) return;

	// only need to calc this the once...
	scroll = -64 * ((r_newrefdef.time / 40.0) - (int) (r_newrefdef.time / 40.0));
	if (scroll == 0.0) scroll = -64.0;

	warpscroll = -64 * ((r_newrefdef.time * 0.5) - (int) (r_newrefdef.time * 0.5));

	glEnable (GL_BLEND);
	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

	R_InitVArrays (VERT_DUAL_TEXTURED);

	// the textures are prescaled up for a better lighting range,
	// so scale it back down
	intens = gl_state.inverse_intensity;

	for (s = FrameRenderState.r_alpha_surfaces; s; s = s->texturechain) 
	{
		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, s->texinfo->image->texnum);
		glActiveTexture (GL_TEXTURE1);
		glBindTexture (GL_TEXTURE_2D, s->texinfo->image->lumatex);
		c_brush_polys++;

		// this is a hack for RF_TRANSLUCENT bmodel entities
		if (s->flags & SURF_DRAWTRANS25)
			glColor4f (intens, intens, intens, 0.25);
		else if (s->texinfo->flags & SURF_TRANS33) 
			glColor4f (intens, intens, intens, 0.33);
		else if (s->texinfo->flags & SURF_TRANS66) 
			glColor4f (intens, intens, intens, 0.66);
		else if (s->flags & SURF_DRAWTURB)
			glColor4f (1, 1, 1, 1);
		else
			glColor4f (intens, intens, intens, 1);

		if (s->flags & SURF_DRAWTURB)
			R_AddTurbulentSurfToVArray (s, (s->texinfo->flags & SURF_FLOWING) ? warpscroll : 0);
		else R_AddTexturedSurfToVArray (s, (s->texinfo->flags & SURF_FLOWING) ? scroll : 0);
	}

	// GL_TEXTURE1 will be active coming out
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glColor4f (1, 1, 1, 1);
	glDisable (GL_BLEND);

	R_KillVArrays ();

	FrameRenderState.r_alpha_surfaces = NULL;
}


cvar_t *r_lightscale = NULL;

void R_DrawTextureChains (void)
{
	int i;
	int j;
	image_t *image;
	msurface_t *surf;
	glpoly_t *p;
	float scroll;

	// init at something invalid cos we always want the test to fail first time
	int OldLightMap = -1;

	//VID_Printf (PRINT_ALL, "[%07.5f]\n", r_newrefdef.frametime);

	// get our cvars first time they're needed
	if (!r_lightscale) r_lightscale = Cvar_Get ("r_lightscale", "0", CVAR_ARCHIVE);

	// set up vertex arrays
	R_InitVArrays (VERT_MULTI_TEXTURED);

	// only need to calc this the once...
	scroll = -64 * ((r_newrefdef.time / 40.0) - (int) (r_newrefdef.time / 40.0));

	if (scroll == 0.0) scroll = -64.0;

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);

	switch ((int) r_lightscale->value)
	{
	default:
	case 0:
		glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 1);	// OpenGL spec only allows 1, 2 or 4
		break;

	case 1:
		glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 2);	// OpenGL spec only allows 1, 2 or 4
		break;

	case 2:
		glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 4);	// OpenGL spec only allows 1, 2 or 4
		break;
	}

	for (j = 0, image = gltextures; j < numgltextures; j++, image++) 
	{
		// not registered
		if (!image->registration_sequence) continue;

		// see have we got a texture chain, skip if not
		surf = image->texturechain;
		if (!surf) continue;
		if (surf->flags & SURF_DRAWTURB) continue;

		// bind the texture (once per texture rather than any of that other messing)
		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, image->texnum);

		// activate TMU1 for lightmaps
		glActiveTexture (GL_TEXTURE1);

		for (; surf; surf = surf->texturechain) 
		{
			c_brush_polys++;

			// if they're equal the surface's lightmap has already been bound and tested for
			// modification, so just re-use that.
			if (surf->lightmaptexturenum != OldLightMap)
			{
				// lightmap bind and test for modification
				glBindTexture (GL_TEXTURE_2D, gl_state.lightmap_textures + surf->lightmaptexturenum);

				// only upload a lightmap the first time it's hit
				if (gl_lms.lightmap_modified[surf->lightmaptexturenum])
				{
					// upload
					glTexSubImage2D (GL_TEXTURE_2D, 
						0, 
						0, 
						0, 
						BLOCK_WIDTH, 
						BLOCK_HEIGHT, 
						GL_RGBA, 
						GL_UNSIGNED_BYTE, 
						gl_lms.lightmap_buffer[surf->lightmaptexturenum]);

					gl_lms.lightmap_modified[surf->lightmaptexturenum] = false;
				}

				// save it out for the next surf
				OldLightMap = surf->lightmaptexturenum;
			}

			R_AddLightMappedSurfToVArray (surf, (surf->texinfo->flags & SURF_FLOWING) ? scroll : 0);
		}
	}

	glActiveTexture (GL_TEXTURE1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// RGB_SCALE is only supposed to apply in COMBINE environment mode, but some drivers seem a bit lax on that
	glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 1);
	glDisable (GL_TEXTURE_2D);
	glActiveTexture (GL_TEXTURE0);

	// shut down vertex arrays
	R_KillVArrays ();
}


void R_DrawLumaChains (void)
{
	int i;
	int j;
	image_t *image;
	msurface_t *surf;
	glpoly_t *p;
	float scroll;
	float activescroll;

	// set up vertex arrays
	R_InitVArrays (VERT_SINGLE_TEXTURED);

	// only need to calc this the once...
	scroll = -64 * ((r_newrefdef.time / 40.0) - (int) (r_newrefdef.time / 40.0));

	if (scroll == 0.0) scroll = -64.0;

	glDepthMask (0);
	glEnable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	for (j = 0, image = gltextures; j < numgltextures; j++, image++) 
	{
		// not registered
		if (!image->registration_sequence) continue;
		if (!image->lumatex) continue;

		// see have we got a texture chain, skip if not
		surf = image->texturechain;
		if (!surf) continue;

		if (gl_lumamode->value == 1 && !(surf->texinfo->flags & SURF_LIGHT)) continue;

		// bind the texture (once per texture rather than any of that other messing)
		glBindTexture (GL_TEXTURE_2D, image->lumatex);

		for (; surf; surf = surf->texturechain) 
		{
			if (!(surf->flags & SURF_DRAWTURB))
			{
				if (surf->texinfo->flags & SURF_FLOWING)
					activescroll = scroll;
				else activescroll = 0;

				R_AddTexturedSurfToVArray (surf, activescroll);
			}
		}
	}

	glDepthMask (1);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// shut down vertex arrays
	R_KillVArrays ();
}


void R_InitBumpyTMUs (void)
{
	// TMU0 writes the normalisation cubemap as the default primary colour
	glActiveTexture (GL_TEXTURE0);
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_CUBE_MAP);
	glBindTexture (GL_TEXTURE_CUBE_MAP, r_cubemap->texnum);

	// now rotate the texture matrix by the same transforms as the modelview matrix to get the correct orientation.
	// this whole setup was derived via good old practical experimentation: printing the viewangles to the console
	// until viewangles of 0 0 0 gave us the correctly center positioned bright spot
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();

	// rotate around Y to bring blue/pink to the front
	glRotatef (90, 0, 1, 0);

	// now reposition so that the bright spot is center screen
	glRotatef (-45, 1, 0, 0);

	// rotate by viewangles
	glRotatef (-r_newrefdef.viewangles[2], 1, 0, 0);
	glRotatef (-r_newrefdef.viewangles[0], 0, 1, 0);
	glRotatef (-r_newrefdef.viewangles[1], 0, 0, 1);

	// the next 2 statements will move the cmstr calculations into hardware so that we don;t
	// have to evaluate them once per vert...

	// translate after rotation
	glTranslatef (r_newrefdef.vieworg[0], r_newrefdef.vieworg[1], r_newrefdef.vieworg[2]);

	// now flip everything by -1 again to mimic the software calculation
	// we won't bother batching this part up...
	glScalef (-1, -1, -1);

	glMatrixMode (GL_MODELVIEW);

	// basic replace texenv
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// TMU1 holds the normal map
	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_DOT3_RGB);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
}


void R_KillBumpyTMUs (void)
{
	glActiveTexture (GL_TEXTURE1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);

	glActiveTexture (GL_TEXTURE0);
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
	glDisable (GL_TEXTURE_CUBE_MAP);
	glEnable (GL_TEXTURE_2D);
}


// use the same kill func as for bumpy
void R_InitSpecularTMUs (void)
{
	// leave this here but commented out cos it can be useful to have...
	//VID_Printf (PRINT_ALL, "[%f] [%f] [%f]\n", r_newrefdef.viewangles[0], r_newrefdef.viewangles[1], r_newrefdef.viewangles[2]);

	// TMU0 writes the normalisation cubemap as the default primary colour
	glActiveTexture (GL_TEXTURE0);
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_CUBE_MAP);
	glBindTexture (GL_TEXTURE_CUBE_MAP, r_cubemap->texnum);

	// now rotate the texture matrix by the same transforms as the modelview matrix to get the correct orientation.
	// this whole setup was derived via good old practical experimentation: printing the viewangles to the console
	// until viewangles of 0 0 0 gave us the correctly center positioned bright spot
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();

	// position directly in front when viewangles are 0
	glRotatef (90, 0, 1, 0);
	glRotatef (180, 0, 0, 1);

	// rotate by viewangles
	glRotatef (-r_newrefdef.viewangles[2], 1, 0, 0);
	glRotatef (-r_newrefdef.viewangles[0], 0, 1, 0);
	glRotatef (-r_newrefdef.viewangles[1], 0, 0, 1);

	// the next 2 statements will move the cmstr calculations into hardware so that we don;t
	// have to evaluate them once per vert...

	// translate after rotation
	glTranslatef (r_newrefdef.vieworg[0], r_newrefdef.vieworg[1], r_newrefdef.vieworg[2]);

	// now flip everything by -1 again to mimic the software calculation
	// we won't bother batching this part up...
	glScalef (-1, -1, -1);

	glMatrixMode (GL_MODELVIEW);

	// basic replace texenv
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// TMU1 holds the normal map
	// this, plus the attenuation map, gives almost the same result as a dot3 on the half angle
	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}


void R_DrawSpecularMaps (void)
{
	int i;
	int j;
	image_t *image;
	msurface_t *surf;
	glpoly_t *p;
	float scroll;
	float activescroll;
	float cmstr[3];

	// only need to calc this the once...
	scroll = -64 * ((r_newrefdef.time / 40.0) - (int) (r_newrefdef.time / 40.0));

	if (scroll == 0.0) scroll = -64.0;

	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);
	glBlendFunc (GL_ONE, GL_ONE);

	// set up vertex arrays and tmus
	R_InitVArrays (VERT_BUMPMAPPED);
	R_InitSpecularTMUs ();

	for (j = 0, image = gltextures; j < numgltextures; j++, image++) 
	{
		// not registered
		if (!image->registration_sequence) continue;

		// no normal tex
		if (!image->normtex) continue;

		// see have we got a texture chain, skip if not
		surf = image->texturechain;
		if (!surf) continue;

		// bind the texture (once per texture rather than any of that other messing)
		if (surf->flags & SURF_DRAWTURB)
			glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);
		else glBindTexture (GL_TEXTURE_2D, image->lumatex);

		for (; surf; surf = surf->texturechain) 
		{
			c_brush_polys++;

			if (surf->texinfo->flags & SURF_FLOWING)
				activescroll = scroll;
			else activescroll = 0;

			R_AddTexturedSurfToVArray (surf, activescroll);
		}
	}

	// shut down tmus and vertex arrays
	// uses the same kill func as bumpies
	R_KillBumpyTMUs ();
	R_KillVArrays ();

	glDepthMask (GL_TRUE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
}


void R_DrawDOT3Bumps (void)
{
	int i;
	int j;
	image_t *image;
	msurface_t *surf;
	glpoly_t *p;
	float scroll;
	float activescroll;
	float cmstr[3];

	// only need to calc this the once...
	scroll = -64 * ((r_newrefdef.time / 40.0) - (int) (r_newrefdef.time / 40.0));

	if (scroll == 0.0) scroll = -64.0;

	// set up vertex arrays and tmus
	R_InitVArrays (VERT_BUMPMAPPED);
	R_InitBumpyTMUs ();

	for (j = 0, image = gltextures; j < numgltextures; j++, image++) 
	{
		// not registered
		if (!image->registration_sequence) continue;

		// no normal tex
		if (!image->normtex) continue;

		// see have we got a texture chain, skip if not
		surf = image->texturechain;
		if (!surf) continue;

		// bind the texture (once per texture rather than any of that other messing)
		if (surf->flags & SURF_DRAWTURB)
			glBindTexture (GL_TEXTURE_2D, r_blacktexture->normtex);
		else glBindTexture (GL_TEXTURE_2D, image->normtex);

		for (; surf; surf = surf->texturechain) 
		{
			c_brush_polys++;

			if (surf->texinfo->flags & SURF_FLOWING)
				activescroll = scroll;
			else activescroll = 0;

			R_AddTexturedSurfToVArray (surf, activescroll);
		}
	}

	// shut down tmus and vertex arrays
	R_KillBumpyTMUs ();
	R_KillVArrays ();
}


void R_PrepareTextureChains (void)
{
	int j;
	image_t *image;

	for (j = 0, image = gltextures; j < numgltextures; j++, image++) 
	{
		// NULL the texture chain
		image->texturechain = NULL;
	}

	// NULL the auxillary chains and render flags
	memset (&FrameRenderState, 0, sizeof (FrameRenderState_t));

	// clear down the accumulated skybox
	R_ClearSkyBox ();
}


void R_AddSurfaceToTextureChains (msurface_t *surf)
{
	int maps;
	qboolean is_dynamic = false;

	// get the correct animation
	image_t *image = R_TextureAnimation (surf->texinfo);

	// no heat haze yet at this point
	surf->hazemask = false;

	if (surf->texinfo->flags & SURF_LIGHT)
		FrameRenderState.FrameHasLuma = true;

	if (surf->texinfo->flags & (SURF_TRANS33 | SURF_TRANS66)) 
	{
		// add to the translucent chain
		surf->texturechain = FrameRenderState.r_alpha_surfaces;
		FrameRenderState.r_alpha_surfaces = surf;

		return;
	}

	if (surf->flags & SURF_DRAWTRANS25)
	{
		// hack - also add to the translucent chain
		surf->texturechain = FrameRenderState.r_alpha_surfaces;
		FrameRenderState.r_alpha_surfaces = surf;

		return;
	}

	if (surf->flags & SURF_DRAWTURB)
	{
		// water goes into the alpha surfaces chain too, until i build a better place for it
		surf->texturechain = FrameRenderState.r_alpha_surfaces;
		FrameRenderState.r_alpha_surfaces = surf;

		return;
	}

	// add the surface to the texture chain
	surf->texturechain = image->texturechain;
	image->texturechain = surf;

	// don't accumulate the rest for turbulent surfs
	if (surf->flags & SURF_DRAWTURB) return;

	// add to the detail chain
	surf->detailchain = FrameRenderState.r_detailsurfaces;
	FrameRenderState.r_detailsurfaces = surf;

	// no light on this surf!!!
	if (!surf->samples) return;

	for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
	{
		if (r_newrefdef.lightstyles[surf->styles[maps]].white != surf->cached_light[maps])
			goto dynamic;
	}

	// dynamic this frame or dynamic previously
	if ((surf->dlightframe == r_framecount))
	{
		// dirty the cache
		surf->cached_light[0] = -1;
		surf->cached_light[1] = -1;
		surf->cached_light[2] = -1;
		surf->cached_light[3] = -1;

dynamic:;
		if (gl_dynamic->value)
		{
			// this condition should never eval to false in this part of the code...
			if (!(surf->texinfo->flags & (SURF_SKY | SURF_TRANS33 | SURF_TRANS66 | SURF_WARP)))
			{
				is_dynamic = true;
			}
		}
	}

	if (is_dynamic)
	{
		if ((surf->styles[maps] >= 32 || surf->styles[maps] == 0) && (surf->dlightframe != r_framecount))
			R_SetCacheState (surf);

		// rebuild the lightmap
		R_BuildLightMap (surf, BLOCK_WIDTH * LIGHTMAP_BYTES);

		// flag as modified
		gl_lms.lightmap_modified[surf->lightmaptexturenum] = true;
	}
}


void R_TransformSurface (msurface_t *surf, float *matrix)
{
	int i;
	float *v;
	float *tv;
	glpoly_t *p;

	// do the full poly chain cos turb surfs can go through this too!!!
	for (p = surf->polys; p; p = p->next)
	{
		// transform the midpoint
		v = p->midpoint.v;
		tv = p->midpoint.tv;

		// i used hexadecimal so it would line up nicely.  how sad...
		tv[0] = matrix[0x0] * v[0] + matrix[0x4] * v[1] + matrix[0x8] * v[2] + matrix[0xc];
		tv[1] = matrix[0x1] * v[0] + matrix[0x5] * v[1] + matrix[0x9] * v[2] + matrix[0xd];
		tv[2] = matrix[0x2] * v[0] + matrix[0x6] * v[1] + matrix[0xa] * v[2] + matrix[0xe];

		// now transform the verts
		for (i = 0; i < p->numverts; i++)
		{
			// keep the loop clean
			v = p->verts[i].v;
			tv = p->verts[i].tv;

			// i used hexadecimal so it would line up nicely.  how sad...
			tv[0] = matrix[0x0] * v[0] + matrix[0x4] * v[1] + matrix[0x8] * v[2] + matrix[0xc];
			tv[1] = matrix[0x1] * v[0] + matrix[0x5] * v[1] + matrix[0x9] * v[2] + matrix[0xd];
			tv[2] = matrix[0x2] * v[0] + matrix[0x6] * v[1] + matrix[0xa] * v[2] + matrix[0xe];
		}
	}
}


// MH - detail textures begin
// here we draw the detail textures!
// we could do this as a multitextured render along with the other stuff, but my own tests
// have indicated that for something like this there's no real benefit to be had from doing
// so. the difference is in the order of 1 or 2 FPS - the real hit from detail texturing
// comes from scrunching large images into small spaces
static void R_DrawDetailSurfaces (void) 
{
	msurface_t *surf = FrameRenderState.r_detailsurfaces;
	int		i;
	glpoly_t *p;

	// nothing to draw!
	if (!surf) return;

	// we do this rather than checking for 0 (checking for 0 is always bad with floats) 
	// standard 6 decimal place precision
	if (gl_detailtextures->value < 0.00001) return;

	// set up vertex array
	R_InitVArrays (VERT_SINGLE_TEXTURED);

	// we *could* use the standard glBindTexture here as we don't need to do any bind checking
	// but again we'll maintain consistency
	glBindTexture (GL_TEXTURE_2D, r_detailtexture->texnum);

	// set the correct blending mode
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);
	glEnable (GL_BLEND);

	// here we get clever and do our texture scaling in the texture matrix. this saves
	// us from more runtime calculations in software
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();

	// we only need to scale the X and Y co-ords, corresponding to texture s and t.
	// everyone will have their own favourite scaling amount, so i did a cvar
	glScalef (gl_detailtextures->value, gl_detailtextures->value, 1.0);

	// we can safely leave the texture matrix as the active mode while we're doing the draw
	for (; surf; surf = surf->detailchain)
		R_AddDetailSurfToVArray (surf);

	// restore original texture matrix
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	// shut down vertex array
	R_KillVArrays ();

	// restore original blend
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
}
// MH - detail textures end


cvar_t *r_bloomradius = NULL;
cvar_t *r_bloomalpha = NULL;

void R_DrawBloomChains (void)
{
	int i;
	int j;
	image_t *image;
	msurface_t *surf;
	glpoly_t *p;
	float *v;
	float scroll;
	float activescroll;
	float tcmods, tcmodt;
	int bloomrad;

	// let the user control the bloom radius and alpha.
	if (!r_bloomradius) r_bloomradius = Cvar_Get ("r_bloomradius", "2", CVAR_ARCHIVE);
	if (!r_bloomalpha) r_bloomalpha = Cvar_Get ("r_bloomalpha", "0.16", CVAR_ARCHIVE);

	// prevent silly values
	if (r_bloomradius->value < 1) r_bloomradius->value = 1;
	if (r_bloomalpha->value < 0) r_bloomalpha->value = 0;
	if (r_bloomalpha->value > 1) r_bloomalpha->value = 1;

	// convert the bloom radius to an int for loop control
	bloomrad = (int) r_bloomradius->value;

	if (!FrameRenderState.FrameHasLuma) return;

	// render into and read from a non-displayable auxillary buffer
	// to prevent screen and image flickering
	glDrawBuffer (GL_AUX0);
	glReadBuffer (GL_AUX0);

	// set up a viewport to blend into.
	// rendering into a pbuffer may be preferable
	// (although having seen some sample code, I think I'll stick with this thanks very much.  I don't see
	// how it would offer much performance advantage, would rule out use of the imaging subset to blur in hardware,
	// and would be restricted to hardware that supported it.  not for me.)
	R_SetupViewport (0, 0, BLOOM_SIZE, BLOOM_SIZE);

	// only need to calc this the once...
	scroll = -64 * ((r_newrefdef.time / 40.0) - (int) (r_newrefdef.time / 40.0));

	if (scroll == 0.0) scroll = -64.0;

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glColor4f (1, 1, 1, 1);

	// set up vertex arrays
	R_InitVArrays (VERT_SINGLE_TEXTURED);

	for (j = 0, image = gltextures; j < numgltextures; j++, image++) 
	{
		// not registered
		if (!image->registration_sequence) continue;

		// see have we got a texture chain, skip if not
		surf = image->texturechain;
		if (!surf) continue;

		// if it's not a light emitting texture we will bind a generic black texture, otherwise we
		// won't get proper depth culling if the bloom is behind any other surfs
		if (!(surf->texinfo->flags & SURF_LIGHT))
			glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);
		else glBindTexture (GL_TEXTURE_2D, image->lumatex);

		for (; surf; surf = surf->texturechain) 
		{
			if (!(surf->flags & SURF_DRAWTURB))
			{
				if (surf->texinfo->flags & SURF_FLOWING)
					activescroll = scroll;
				else activescroll = 0;

				R_AddTexturedSurfToVArray (surf, activescroll);
			}
		}
	}

	// shut down vertex arrays
	R_KillVArrays ();

	// now make a texture out of it
	glBindTexture (GL_TEXTURE_2D, r_bloomtexture->texnum);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// copy in what we've got in the framebuffer to create a baseline to work from
	glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, 0, BLOOM_SIZE, BLOOM_SIZE);

	// now accumulate the texture
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glOrtho (0, BLOOM_SIZE, BLOOM_SIZE, 0, -99999, 99999);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glColor4f (1, 1, 1, 1);

	glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);

	// first accumulation is unblended and uses the black texture to derive a base
	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	glVertex2f (1, 1);

	glTexCoord2f (1, 1);
	glVertex2f (BLOOM_SIZE, 1);

	glTexCoord2f (1, 0);
	glVertex2f (BLOOM_SIZE, BLOOM_SIZE);

	glTexCoord2f (1, 1);
	glVertex2f (1, BLOOM_SIZE);

	glEnd ();

	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE);

	glBindTexture (GL_TEXTURE_2D, r_bloomtexture->texnum);
	glBegin (GL_QUADS);

	glColor4f (1, 1, 1, r_bloomalpha->value);

	for (i = -bloomrad; i < (bloomrad + 1); i++)
	{
		tcmods = (float) i * 0.004f;

		for (j = -bloomrad; j < (bloomrad + 1); j++)
		{
			tcmodt = (float) j * 0.004f;

			glTexCoord2f (0 + tcmods, 1 + tcmodt);
			glVertex2f (0, 0);

			glTexCoord2f (1 + tcmods, 1 + tcmodt);
			glVertex2f (BLOOM_SIZE, 0);

			glTexCoord2f (1 + tcmods, 0 + tcmodt);
			glVertex2f (BLOOM_SIZE, BLOOM_SIZE);

			glTexCoord2f (0 + tcmods, 0 + tcmodt);
			glVertex2f (0, BLOOM_SIZE);
		}
	}

	glEnd ();

	glColor4f (1, 1, 1, 1);

	// copy back the accumulated version
	glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, 0, BLOOM_SIZE, BLOOM_SIZE);

	// we also clear depth here cos we only have 1 depth buffer!!!
	glClear (GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	// restore the original draw and read buffers
	glDrawBuffer (gl_state.CurrentBuffer);
	glReadBuffer (gl_state.CurrentBuffer);
}


void R_RenderBloom (void)
{
	if (!FrameRenderState.FrameHasLuma) return;

	// switch to 2D for the bloom blend
	// set 2D virtual screen size
	glViewport (0, 0, vid.width, vid.height);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glOrtho (0, BLOOM_SIZE, BLOOM_SIZE, 0, -99999, 99999);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glColor4f (1,1,1,1);

	glEnable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE);

	glBindTexture (GL_TEXTURE_2D, r_bloomtexture->texnum);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glBegin (GL_QUADS);

	// invert t tex-coords because OpenGL's "bottom-up" paradigm fucks things up on us (not for the 1st time either...)
	glTexCoord2f (0, 1);
	glVertex2f (0, 0);

	glTexCoord2f (1, 1);
	glVertex2f (BLOOM_SIZE, 0);

	glTexCoord2f (1, 0);
	glVertex2f (BLOOM_SIZE, BLOOM_SIZE);

	glTexCoord2f (0, 0);
	glVertex2f (0, BLOOM_SIZE);

	glEnd ();

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	// restore the 3D viewport
	R_SetupViewport (0, 0, vid.width, vid.height);
}


// blow me down.  this worked first time.  wow.
qboolean warpkill;

void R_HazeMaskSurfacesOnDrawList (vec3_t p1, vec3_t p2, vec3_t p3, vec3_t p4)
{
	vec3_t screenmins = {999999, 999999, 999999};
	vec3_t screenmaxs = {-999999, -999999, -999999};
	vec3_t screenpoints[4];
	msurface_t *surf;

	// convert vertex points to screen coords.  z should be the same in all cases
	mhqgluProject (p1, r_newrefdef.ModelViewMatrix, r_newrefdef.ProjectionMatrix, r_newrefdef.Viewport, screenpoints[0]);
	mhqgluProject (p2, r_newrefdef.ModelViewMatrix, r_newrefdef.ProjectionMatrix, r_newrefdef.Viewport, screenpoints[1]);
	mhqgluProject (p3, r_newrefdef.ModelViewMatrix, r_newrefdef.ProjectionMatrix, r_newrefdef.Viewport, screenpoints[2]);
	mhqgluProject (p4, r_newrefdef.ModelViewMatrix, r_newrefdef.ProjectionMatrix, r_newrefdef.Viewport, screenpoints[3]);

	// get the screen mins and maxs
	if (screenpoints[0][0] < screenmins[0]) screenmins[0] = screenpoints[0][0];
	if (screenpoints[1][0] < screenmins[0]) screenmins[0] = screenpoints[1][0];
	if (screenpoints[2][0] < screenmins[0]) screenmins[0] = screenpoints[2][0];
	if (screenpoints[3][0] < screenmins[0]) screenmins[0] = screenpoints[3][0];

	if (screenpoints[0][1] < screenmins[1]) screenmins[1] = screenpoints[0][1];
	if (screenpoints[1][1] < screenmins[1]) screenmins[1] = screenpoints[1][1];
	if (screenpoints[2][1] < screenmins[1]) screenmins[1] = screenpoints[2][1];
	if (screenpoints[3][1] < screenmins[1]) screenmins[1] = screenpoints[3][1];

	if (screenpoints[0][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[0][0];
	if (screenpoints[1][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[1][0];
	if (screenpoints[2][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[2][0];
	if (screenpoints[3][0] > screenmaxs[0]) screenmaxs[0] = screenpoints[3][0];

	if (screenpoints[0][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[0][1];
	if (screenpoints[1][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[1][1];
	if (screenpoints[2][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[2][1];
	if (screenpoints[3][1] > screenmaxs[1]) screenmaxs[1] = screenpoints[3][1];

	// z should always be the same because these points will always face the player
	screenmins[2] = screenmaxs[2] = screenpoints[0][2];

	// re-use the detail texture chain for these as it saves on unnecessary nested loops and omits surfs we
	// don't need to check
	for (surf = FrameRenderState.r_detailsurfaces; surf; surf = surf->detailchain)
	{
		// already masked
		if (surf->hazemask) continue;

		// surf is in front of the haze.  z should be the same in both screenmins and screenmaxs cases
		// so we only need to check out one of them.  these surfs will only have 1 poly so don't bother walking the chain
		if (surf->polys[0].wincoords[2] < screenmins[2]) continue;

		// surf is behind the haze in the z order - now see is it directly behind
		if (surf->polys[0].wincoords[0] < screenmins[0]) continue;		// too far left
		if (surf->polys[0].wincoords[1] < screenmins[1]) continue;		// too far down
		if (surf->polys[0].wincoords[0] > screenmaxs[0]) continue;		// too far right
		if (surf->polys[0].wincoords[1] > screenmaxs[1]) continue;		// too far up

		// surf is directly behind the haze so mark it
		surf->hazemask = true;
	}

	// check has every surf been warped.
	// assume that they have unless we prove otherwise
	warpkill = true;

	// another stroll through the surfs - ho hum.  shouldn't hurt too much this time though...
	// the data will be in the cpu cache and it's cheaper than a loada unnecessary recalcs
	for (surf = FrameRenderState.r_detailsurfaces; surf; surf = surf->detailchain)
	{
		// already masked
		if (surf->hazemask) continue;

		// not every surf in the view has been warped if we get this far any time
		warpkill = false;

		// don;t need to check any more
		return;
	}
}


// uncomment for haze occlusion tests
// #define HAZE_TEST

// seperated out for ease of use with bmodels
void R_SetupHazeMaskForLavaSurf (msurface_t *surf, vec3_t up, vec3_t right)
{
	glpoly_t *p = surf->polys;
	vec3_t p1;
	vec3_t p2;
	vec3_t p3;
	vec3_t p4;
	int i;
	float offset;
	float scale;

#ifdef HAZE_TEST
	glDisable (GL_TEXTURE_2D);
	glColor3f (1, 0, 1);
	glBegin (GL_QUADS);
#endif

	// this should be the same as in gl_rpart.c
#define LAVA_RENDER_SCALE 72.0f
	// a warped surf is a chain of polys so batch up this part of the render per surf rather than per poly
	for (; p; p = p->next)
	{
		scale = (p->verts[0].tv[0] - r_origin[0]) * vpn[0] + (p->verts[0].tv[1] - r_origin[1]) * vpn[1] + (p->verts[0].tv[2] - r_origin[2]) * vpn[2];

		// too near the screen or behind the current viewpoint
		if (scale < 50) continue;

		// do 2 quads rather than one tall one to retain vertical scale when looking up/down
		for (i = 0; i < 2; i++)
		{
			offset = 0.5f + (float) i; // * 0.666f;

			// calculate vertex points for a virtual particle
			// verts[0] is the midpoint in a warped surface
			p1[0] = p->verts[0].tv[0] - (right[0] + up[0]) * 0.5f;
			p1[1] = p->verts[0].tv[1] - (right[1] + up[1]) * 0.5f;
			p1[2] = p->verts[0].tv[2] - (right[2] + up[2]) * 0.5f + (LAVA_RENDER_SCALE * offset);

			p2[0] = p1[0] + up[0];
			p2[1] = p1[1] + up[1];
			p2[2] = p1[2] + up[2];

			p3[0] = p2[0] + right[0];
			p3[1] = p2[1] + right[1];
			p3[2] = p2[2] + right[2];

			p4[0] = p1[0] + right[0];
			p4[1] = p1[1] + right[1];
			p4[2] = p1[2] + right[2];

#ifdef HAZE_TEST
			glVertex3fv (p1);
			glVertex3fv (p2);
			glVertex3fv (p3);
			glVertex3fv (p4);
#endif

			// now check for surface masking
			R_HazeMaskSurfacesOnDrawList (p1, p2, p3, p4);
		}
	}

#ifdef HAZE_TEST
	glEnd ();
	glEnable (GL_TEXTURE_2D);
	glColor3f (1, 1, 1);
#endif
}


void R_SetupHazeMasks (void)
{
	int i;
	int j;
	msurface_t *surf;
	vec3_t up, right;
	entity_t *ent;

	up[0] = vup[0] * LAVA_RENDER_SCALE;
	up[1] = vup[1] * LAVA_RENDER_SCALE;
	up[2] = vup[2] * LAVA_RENDER_SCALE;

	right[0] = vright[0] * LAVA_RENDER_SCALE;
	right[1] = vright[1] * LAVA_RENDER_SCALE;
	right[2] = vright[2] * LAVA_RENDER_SCALE;
#undef LAVA_RENDER_SCALE

	warpkill = false;

	for (i = 0, surf = r_worldmodel->surfaces; i < r_worldmodel->numsurfaces; i++, surf++)
	{
		// not lava
		if (!(surf->flags & SURF_DRAWLAVA)) continue;

		// not in the pvs for this viewcluster
		if (surf->lavavisframe != r_visframecount) continue;

		// create the haze masks
		R_SetupHazeMaskForLavaSurf (surf, up, right);

		// every surf in the view has been warped
		if (warpkill) return;
	}

	// now we hit the bmodels too
	if (!r_drawentities->value) return;

	for (i = 0; i < r_newrefdef.num_entities; i++) 
	{
		ent = &r_newrefdef.entities[i];

		if (ent->flags & RF_BEAM) continue;

		if (!ent->model) continue;

		// no lava
		if (!ent->model->haslava) continue;

		if (ent->model->type == mod_brush)
		{
			surf = &ent->model->surfaces[ent->model->firstmodelsurface];

			for (j = 0; j < ent->model->nummodelsurfaces; j++, surf++) 
			{
				// not lava
				if (!(surf->flags & SURF_DRAWLAVA)) continue;

				// not in the pvs for this viewcluster
				if (surf->lavavisframe != r_visframecount) continue;

				// don;t care about backfacing, culling, or anything fancy, just add it
				R_SetupHazeMaskForLavaSurf (surf, up, right);

				// every surf in the view has been warped
				if (warpkill) return;
			}
		}
	}
}


void R_ProjectSurfaceVerts (void)
{
	glpoly_t *p;
	msurface_t *surf;

	// we can re-use the detail chain for this
	for (surf = FrameRenderState.r_detailsurfaces; surf; surf = surf->detailchain)
	{
		// do each poly
		for (p = surf->polys; p; p = p->next)
		{
			// project poly to window coords
			mhqgluProject (p->midpoint.tv, r_newrefdef.ModelViewMatrix, r_newrefdef.ProjectionMatrix, r_newrefdef.Viewport, p->wincoords);
		}
	}
}


// fade the surface warp in and out to prevent popping
void R_FadeSurfaceWarps (void)
{
	msurface_t *surf;

	// we can re-use the detail chain for this
	for (surf = FrameRenderState.r_detailsurfaces; surf; surf = surf->detailchain)
	{
		if (surf->flags & SURF_UNDERWATER)
		{
			// constant warp with no fading
			surf->warpfade = 1.0f;
			continue;
		}

		if (surf->hazemask)
		{
			// fade the warp up if the surf is hazemasked, upper bound it at 1
			surf->warpfade += r_newrefdef.frametime;
			if (surf->warpfade >= 1.0f) surf->warpfade = 1.0f;
		}
		else
		{
			// fade the warp down
			surf->warpfade -= r_newrefdef.frametime;

			// if the warp is still fading down hazemask it, otherwise lower bound it at 0
			if (surf->warpfade > 0)
				surf->hazemask = true;
			else surf->warpfade = 0;
		}
	}
}


extern cvar_t *r_hazewibble;

void R_DrawSurfacesOnList (void)
{
	// gl_dot3bump 2 is a test mode for seeing the naked normal maps
	if (gl_lumamode->value == 3 && gl_dot3bump->value < 2) R_DrawBloomChains ();

	// restore the original viewport
	R_SetupViewport (0, 0, vid.width, vid.height);

	// heat haze - software method
	// blow me down.  this worked first time.  wow.
	if (FrameRenderState.FrameHasLava && r_worldmodel->haslava && r_hazewibble->value > 0.0f)
	{
		// save out the modelview, projection and viewport
		glGetFloatv (GL_MODELVIEW_MATRIX, r_newrefdef.ModelViewMatrix);
		glGetFloatv (GL_PROJECTION_MATRIX, r_newrefdef.ProjectionMatrix);
		glGetIntegerv (GL_VIEWPORT, r_newrefdef.Viewport);

		// get the midpoint of each surf on the drawlist in window coords
		R_ProjectSurfaceVerts ();

		// mark those surfs which are masked by haze
		R_SetupHazeMasks ();
	}

	// always do this even if not needed to ensure consistent state
	// we have to wait until after surfs are hazemasked cos the calculations depend on whether or not they are
	R_FadeSurfaceWarps ();

	if (gl_dot3bump->value)
	{
		R_DrawDOT3Bumps ();

		// switch on blending for the texture render
		glEnable (GL_BLEND);
		glBlendFunc (GL_DST_COLOR, GL_ZERO);
		glDepthMask (GL_FALSE);
	}

	// gl_dot3bump 2 is the test mode
	if (gl_dot3bump->value < 2) R_DrawTextureChains ();

	if (gl_dot3bump->value)
	{
		glDepthMask (GL_TRUE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable (GL_BLEND);
	}

	// gl_dot3bump 2 is the test mode
	if (gl_dot3bump->value < 2) R_DrawDetailSurfaces ();

	// R_DrawSpecularMaps ();

	if (gl_lumamode->value && gl_dot3bump->value < 2) R_DrawLumaChains ();

	// draw the skybox before the bloom otherwise it may depthmask the bloom
	R_DrawSkyBox ();

	// bloom overlay
	if (gl_lumamode->value == 3 && gl_dot3bump->value < 2) R_RenderBloom ();
}


void R_DirtyCachesForModel (model_t *m, msurface_t *firstsurf, int numsurfs)
{
	int i;

	for (i = 0; i < numsurfs; i++, firstsurf++)
	{
		firstsurf->cached_light[0] = -1;
		firstsurf->cached_light[1] = -1;
		firstsurf->cached_light[2] = -1;
		firstsurf->cached_light[3] = -1;
	}
}


void R_DirtyCaches (void)
{
	int i;
	entity_t *e;
	model_t *m;

	R_DirtyCachesForModel (r_worldmodel, r_worldmodel->surfaces, r_worldmodel->numsurfaces);

	for (i = 0; i < r_newrefdef.num_entities; i++) 
	{
		e = &r_newrefdef.entities[i];

		if (e->flags & RF_BEAM) continue;

		m = e->model;

		if (!m) continue;

		if (m->type == mod_brush)
		{
			if (m->nummodelsurfaces == 0) continue;

			R_DirtyCachesForModel (m, &m->surfaces[m->firstmodelsurface], m->nummodelsurfaces);
		}
	}
}


void R_AddBModelToList (entity_t *e)
{
	vec3_t		mins, maxs;
	int			i;
	qboolean	rotated;
	float		matrix[16];
	cplane_t	*plane;
	float		dot;
	msurface_t	*surf;

	if (currentmodel->nummodelsurfaces == 0) return;

	currententity = e;

	if (e->angles[0] || e->angles[1] || e->angles[2]) 
	{
		rotated = true;

		for (i = 0; i < 3; i++) 
		{
			mins[i] = e->origin[i] - currentmodel->radius;
			maxs[i] = e->origin[i] + currentmodel->radius;
		}
	}
	else
	{
		rotated = false;

		VectorAdd (e->origin, currentmodel->mins, mins);
		VectorAdd (e->origin, currentmodel->maxs, maxs);
	}

	if (R_CullBox (mins, maxs)) return;

	// the entity is in the render now
	e->inrender = true;

	VectorSubtract (r_newrefdef.vieworg, e->origin, modelorg);

	if (rotated) 
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}

	// set dynamic lighting
	R_PushDlights (currentmodel->nodes + currentmodel->firstnode);

	e->angles[0] = -e->angles[0];	// stupid quake bug
	e->angles[2] = -e->angles[2];	// stupid quake bug

	// fixme - do this in software...
	// because we're doing an identity load each time, we only need one push/pop pair in R_AddBModelsToDrawLists...
	glLoadIdentity ();
	R_RotateForEntity (e);
	glGetFloatv (GL_MODELVIEW_MATRIX, matrix);

	e->angles[0] = -e->angles[0];	// stupid quake bug
	e->angles[2] = -e->angles[2];	// stupid quake bug

	// add it here...
	surf = &currentmodel->surfaces[currentmodel->firstmodelsurface];

	for (i = 0; i < currentmodel->nummodelsurfaces; i++, surf++) 
	{
		// no lava visibility
		surf->lavavisframe = -1;

		// find which side of the node we are on
		plane = surf->plane;

		dot = DotProduct (modelorg, plane->normal) - plane->dist;

		// draw the polygon
		if (((surf->flags & SURF_PLANEBACK) && (dot < - BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON))) 
		{
			// mark the surf as translucent
			if (currententity->flags & RF_TRANSLUCENT) surf->flags |= SURF_DRAWTRANS25;

			// apply the new matrix to the surface verts
			R_TransformSurface (surf, matrix);

			// mark for lava
			if (surf->flags & SURF_DRAWLAVA)
			{
				FrameRenderState.FrameHasLava = true;
				surf->lavavisframe = r_visframecount;
			}

			// and add them to the list
			R_AddSurfaceToTextureChains (surf);
		}
	}
}


void R_AddBModelsToDrawLists (void)
{
	int		i;

	if (!r_drawentities->value) return;

	// because we're loading an identity matrix each time we only need one push/pop pair...
	glPushMatrix ();

	for (i = 0; i < r_newrefdef.num_entities; i++) 
	{
		currententity = &r_newrefdef.entities[i];

		currententity->inrender = false;

		if (currententity->flags & RF_BEAM) continue;

		currentmodel = currententity->model;

		if (!currentmodel) continue;

		if (currentmodel->type == mod_brush) R_AddBModelToList (currententity);
	}

	glPopMatrix ();
}


/*
============================================================= 

WORLD MODEL

============================================================= 
*/ 

/*
================ 
R_RecursiveWorldNode
================ 
*/ 
void R_RecursiveWorldNode (mnode_t *node) 
{
	int			c, side, sidebit;
	cplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	float		dot;
	image_t		*image;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return; 
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, draw stuff
	if (node->contents != - 1) 
	{
		pleaf = (mleaf_t *) node;

		// check for door connected areas
		if (r_newrefdef.areabits) 
		{
			if (!(r_newrefdef.areabits[pleaf->area >> 3]& (1 << (pleaf->area&7)))) 
				return;		// not visible
		}

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c) 
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (-- c);
		}

		return;
	}

	// node is just a decision point, so go down the apropriate sides

	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type) 
	{
	case PLANE_X:
		dot = modelorg[0]- plane->dist;
		break;
	case PLANE_Y:
		dot = modelorg[1]- plane->dist;
		break;
	case PLANE_Z:
		dot = modelorg[2]- plane->dist;
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0) 
	{
		side = 0;
		sidebit = 0;
	}
	else
	{
		side = 1;
		sidebit = SURF_PLANEBACK;
	}

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	// draw stuff
	for (c = node->numsurfaces, surf = r_worldmodel->surfaces + node->firstsurface; c; c--, surf++) 
	{
		if (surf->visframe != r_framecount) continue;

		// wrong side
		if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

		if (surf->texinfo->flags & SURF_SKY) 
		{
			// just adds to visible sky bounds
			R_AddSkySurface (surf);
		}
		else R_AddSurfaceToTextureChains (surf);
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}


void R_MarkLava (void)
{
	int i;
	int j;
	mleaf_t *leaf;

	// none initially
	FrameRenderState.FrameHasLava = false;

	// do the world
	for (i = 0, leaf = r_worldmodel->leafs; i < r_worldmodel->numleafs; i++, leaf++)
	{
		// not in the pvs for the current viewpoint
		if (leaf->visframe != r_visframecount) continue;

		// q2 contents are bitwise so...
		if (leaf->contents & CONTENTS_SOLID) continue;
		if (leaf->contents & CONTENTS_WINDOW) continue;
		if (leaf->contents & CONTENTS_AUX) continue;
		if (leaf->contents & CONTENTS_LAVA) continue;
		if (leaf->contents & CONTENTS_SLIME) continue;
		if (leaf->contents & CONTENTS_WATER) continue;

		for (j = 0; j < leaf->nummarksurfaces; j++)
		{
			if (leaf->firstmarksurface[j]->flags & SURF_DRAWLAVA)
			{
				// mark the lava surf with the current r_visframecount
				leaf->firstmarksurface[j]->lavavisframe = r_visframecount;

				// got some now
				FrameRenderState.FrameHasLava = true;
			}
		}
	}

	// bmodels are done in the texture chain adder
}


void R_ParticleStateChange (qboolean Init);

void R_RenderLavaSmoke (qboolean StateChange)
{
	int i;
	int j;
	int k;
	msurface_t *surf;
	glpoly_t *p;
	entity_t *ent;

	// no lava so don't bother (save us some state changes)
	if (!FrameRenderState.FrameHasLava) return;

	// init state changes
	if (StateChange) R_ParticleStateChange (true);

	for (i = 0, j = 0, surf = r_worldmodel->surfaces; i < r_worldmodel->numsurfaces; i++, surf++, j += 4)
	{
		// not lava
		if (!(surf->flags & SURF_DRAWLAVA)) continue;

		// not in the pvs for this viewcluster
		if (surf->lavavisframe != r_visframecount) continue;

		// draw the smoke particles
		// because these surfaces warp they will always have a valid midpoint, but we will need to
		// traverse the full poly chain
		for (p = surf->polys; p; p = p->next, j++)
		{
			// verts[0] is the center point, offset by a function of the surface number and poly number
			// have to use poly number too cos a single surf can have many many polys...
			R_RenderLavaSmokeParticle (p->verts[0].tv, (j % 16) * 16);
		}
	}

	// now we hit the bmodels too.
	// this can hit the old render pretty hard in maps like fact1 where a lava bmodel is stacked on top of a
	// regular lava world surf...  temp disabled
	if (r_drawentities->value && false)
	{
		for (i = 0, k = 0; i < r_newrefdef.num_entities; i++) 
		{
			ent = &r_newrefdef.entities[i];

			if (ent->flags & RF_BEAM) continue;

			if (!ent->model) continue;

			if (ent->model->type == mod_brush)
			{
				surf = &ent->model->surfaces[ent->model->firstmodelsurface];

				for (j = 0; j < ent->model->nummodelsurfaces; j++, surf++, k += 4) 
				{
					// not lava
					if (!(surf->flags & SURF_DRAWLAVA)) continue;

					// not in the pvs for this viewcluster
					if (surf->lavavisframe != r_visframecount) continue;

					// don;t care about backfacing, culling, or anything fancy, just add it
					// draw the smoke particles
					// because these surfaces warp they will always have a valid midpoint, but we will need to
					// traverse the full poly chain
					for (p = surf->polys; p; p = p->next, k++)
					{
						// verts[0] is the center point, offset by a function of the surface number and poly number
						// have to use poly number too cos a single surf can have many many polys...
						R_RenderLavaSmokeParticle (p->verts[0].tv, (k % 16) * 16);
					}
				}
			}
		}
	}

	// shut down necessary state changes
	if (StateChange) R_ParticleStateChange (false);
}


/*
============= 
R_DrawWorld
============= 
*/ 
void R_SetSmokeFrameState (void);

void R_DrawWorld (void) 
{
	entity_t	ent;

	if (!gl_lumamode) gl_lumamode = Cvar_Get ("gl_lumamode", "1", CVAR_ARCHIVE);
	if (!gl_dot3bump) gl_dot3bump = Cvar_Get ("gl_dot3bump", "1", CVAR_ARCHIVE);

	if (!r_drawworld->value) return;
	if (r_newrefdef.rdflags & RDF_NOWORLDMODEL) return;

	currentmodel = r_worldmodel;
	VectorCopy (r_newrefdef.vieworg, modelorg);

	// auto cycle the world frame for texture animation
	memset (&ent, 0, sizeof (ent));
	ent.frame = (int) (r_newrefdef.time * 2);
	currententity = &ent;

	gl_state.currenttextures[0] = gl_state.currenttextures[1] = -1;

	glColor3f (1, 1, 1);

	if (gl_modulate->modified)
	{
		// tweak to allow changes to gl_modulate without needing a vid_restart
		R_DirtyCaches ();
		gl_modulate->modified = false;
	}

	if (gl_dot3bump->modified)
	{
		// we need to scale lighting up by a factor of 1.32 for bumpmapping,
		// so dirty the caches and rebuild the lm scale table
		R_DirtyCaches ();
		gl_dot3bump->modified = false;
	}

	// set dynamic lighting on the world
	R_PushDlights (r_worldmodel->nodes);

	// set all texture chains to NULL before accumulating them
	R_PrepareTextureChains ();

	// do before the surface pass so we know if we have to draw haze
	// do before adding bmodels so that we can set if the frame has lava correctly
	R_MarkLava ();

	// add surfaces for the world and for bmodels
	R_RecursiveWorldNode (r_worldmodel->nodes);
	R_AddBModelsToDrawLists ();

	// update the lava smoke now
	R_SetSmokeFrameState ();

	// surface pass
	R_DrawSurfacesOnList ();

	// DEATH TO TRIANGLE OUTLINES!!!
}


/*
=============== 
R_MarkLeaves

Mark the leaves and nodes that are in the PVS for the current
cluster
=============== 
*/ 
void R_MarkLeaves (void) 
{
	byte	*vis;
	byte	fatvis[MAX_MAP_LEAFS / 8];
	mnode_t	*node;
	int		i, c;
	mleaf_t	*leaf;
	int		cluster;

	if (r_oldviewcluster == r_viewcluster && r_oldviewcluster2 == r_viewcluster2 && !r_novis->value && r_viewcluster != - 1) 
		return;

	// development aid to let you run around and see exactly where
	// the pvs ends
	if (gl_lockpvs->value) 
		return;

	r_visframecount++;
	r_oldviewcluster = r_viewcluster;
	r_oldviewcluster2 = r_viewcluster2;

	if (r_novis->value || r_viewcluster == - 1 || !r_worldmodel->vis) 
	{
		// mark everything
		for (i = 0; i<r_worldmodel->numleafs; i++) 
			r_worldmodel->leafs[i].visframe = r_visframecount;
		for (i = 0; i<r_worldmodel->numnodes; i++) 
			r_worldmodel->nodes[i].visframe = r_visframecount;
		return;
	}

	vis = Mod_ClusterPVS (r_viewcluster, r_worldmodel);
	// may have to combine two clusters because of solid water boundaries
	if (r_viewcluster2 != r_viewcluster) 
	{
		memcpy (fatvis, vis, (r_worldmodel->numleafs + 7) / 8);
		vis = Mod_ClusterPVS (r_viewcluster2, r_worldmodel);
		c = (r_worldmodel->numleafs + 31) / 32;
		for (i = 0; i<c; i++) 
			((int *) fatvis)[i]|= ((int *) vis)[i];
		vis = fatvis;
	}

	for (i = 0,leaf = r_worldmodel->leafs; i<r_worldmodel->numleafs; i++ , leaf++) 
	{
		cluster = leaf->cluster;
		if (cluster == - 1) 
			continue;
		if (vis[cluster >> 3]& (1 << (cluster&7))) 
		{
			node = (mnode_t *) leaf;
			do
			{
				if (node->visframe == r_visframecount) 
					break;
				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}



/*
============================================================================= 

LIGHTMAP ALLOCATION

============================================================================= 
*/ 

static void LM_InitBlock (void) 
{
	memset (gl_lms.allocated, 0, sizeof (gl_lms.allocated));
}

static void LM_UploadBlock (qboolean dynamic) 
{
	int texture;
	int height = 0;

	if (dynamic) 
	{
		texture = 0;
	}
	else
	{
		texture = gl_lms.current_lightmap_texture;
	}

	glBindTexture (GL_TEXTURE_2D, gl_state.lightmap_textures + texture);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	if (dynamic) 
	{
		int i;

		for (i = 0; i < BLOCK_WIDTH; i++) 
		{
			if (gl_lms.allocated[i]> height) 
				height = gl_lms.allocated[i];
		}

		glTexSubImage2D (GL_TEXTURE_2D, 
			0,
			0, 0,
			BLOCK_WIDTH, height,
			GL_LIGHTMAP_FORMAT,
			GL_UNSIGNED_BYTE,
			gl_lms.lightmap_buffer[i]);
	}
	else
	{
		glTexImage2D (GL_TEXTURE_2D, 
			0, 
			gl_lms.internal_format,
			BLOCK_WIDTH, BLOCK_HEIGHT, 
			0, 
			GL_LIGHTMAP_FORMAT, 
			GL_UNSIGNED_BYTE, 
			gl_lms.lightmap_buffer[gl_lms.current_lightmap_texture]);
		if (++ gl_lms.current_lightmap_texture == MAX_LIGHTMAPS) 
			VID_Error (ERR_DROP, "LM_UploadBlock () - MAX_LIGHTMAPS exceeded\n");
	}
}

// returns a texture number and the position inside it
static qboolean LM_AllocBlock (int w, int h, int *x, int *y) 
{
	int		i, j;
	int		best, best2;

	best = BLOCK_HEIGHT;

	for (i = 0; i<BLOCK_WIDTH - w; i++) 
	{
		best2 = 0;

		for (j = 0; j<w; j++) 
		{
			if (gl_lms.allocated[i + j]>= best) 
				break;
			if (gl_lms.allocated[i + j]> best2) 
				best2 = gl_lms.allocated[i + j];
		}
		if (j == w) 
		{	// this is a valid spot
			*x = i;
			*y = best = best2;
		}
	}

	if (best + h > BLOCK_HEIGHT) 
		return false;

	for (i = 0; i<w; i++) 
		gl_lms.allocated[*x + i] = best + h;

	return true;
}

/*
================ 
GL_BuildPolygonFromSurface
================ 
*/
void GL_BuildPolygonFromSurface (msurface_t *fa) 
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	int			vertpage;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;
	vec3_t		total;

	vec3_t surfmaxs = {-99999999, -99999999, -99999999};
	vec3_t surfmins = {99999999, 99999999, 99999999};
	vec3_t dists;
	float mindist = 9999999999;

	// reconstruct the polygon
	pedges = currentmodel->edges;
	lnumverts = fa->numedges;
	vertpage = 0;

	VectorClear (total);
	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->verts = Hunk_Alloc (lnumverts * sizeof (polyverts_t));
	poly->next = fa->polys;
	poly->flags = fa->flags;
	fa->polys = poly;
	poly->numverts = lnumverts;

	// set initial midpoint
	poly->midpoint.tv[0] = 0;
	poly->midpoint.tv[1] = 0;
	poly->midpoint.tv[2] = 0;

	poly->midpoint.st[0] = 0;
	poly->midpoint.st[1] = 0;

	poly->midpoint.lm[0] = 0;
	poly->midpoint.lm[1] = 0;

	poly->midpoint.st128[0] = 0;
	poly->midpoint.st128[1] = 0;

	for (i = 0; i < lnumverts; i++) 
	{
		lindex = currentmodel->surfedges[fa->firstedge + i];

		if (lindex > 0) 
		{
			r_pedge = &pedges[lindex];
			vec = currentmodel->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = currentmodel->vertexes[r_pedge->v[1]].position;
		}

		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s /= fa->texinfo->image->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t /= fa->texinfo->image->height;

		VectorAdd (total, vec, total);

		// verts - set both transformed and untransformed to the same for now
		VectorCopy (vec, poly->verts[i].v);
		VectorCopy (vec, poly->verts[i].tv);

		// set bbox for the surf
		// useful to have for testing and stuff...
		if (vec[0] > surfmaxs[0]) surfmaxs[0] = vec[0];
		if (vec[1] > surfmaxs[1]) surfmaxs[1] = vec[1];
		if (vec[2] > surfmaxs[2]) surfmaxs[2] = vec[2];

		if (vec[0] < surfmins[0]) surfmins[0] = vec[0];
		if (vec[1] < surfmins[1]) surfmins[1] = vec[1];
		if (vec[2] < surfmins[2]) surfmins[2] = vec[2];

		// standard texcoords
		poly->verts[i].st[0] = s;
		poly->verts[i].st[1] = t;

		// lightmap texture coordinates
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s -= fa->texturemins[0];
		s += fa->light_s*16;
		s += 8;
		s /= BLOCK_WIDTH*16;// fa->texinfo->texture->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t -= fa->texturemins[1];
		t += fa->light_t*16;
		t += 8;
		t /= BLOCK_HEIGHT*16;// fa->texinfo->texture->height;

		poly->verts[i].lm[0] = s;
		poly->verts[i].lm[1] = t;

		// MH - detail textures begin
		// here we build a set of texture s and t co - ords suitable for a 128 * 128 version
		// of the texture. this is necessary because using the original texture s and t
		// will cause the detail texture (and anything else we may ever consider adding) 
		// to rescale itself to the same aspect ratio as the original texture, which looks
		// bad. NOTE - to enable this we *must* increase the define of VERTEXSIZE to 9.
		poly->verts[i].st128[0] = (DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3]) / 128.0f;
		poly->verts[i].st128[1] = (DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3]) / 128.0f;

		// accumulate midpoints
		poly->midpoint.tv[0] += poly->verts[i].tv[0];
		poly->midpoint.tv[1] += poly->verts[i].tv[1];
		poly->midpoint.tv[2] += poly->verts[i].tv[2];

		poly->midpoint.st[0] += poly->verts[i].st[0];
		poly->midpoint.st[1] += poly->verts[i].st[1];

		poly->midpoint.lm[0] += poly->verts[i].lm[0];
		poly->midpoint.lm[1] += poly->verts[i].lm[1];

		poly->midpoint.st128[0] += poly->verts[i].st128[0];
		poly->midpoint.st128[1] += poly->verts[i].st128[1];
	}

	poly->numverts = lnumverts;

	// average midpoints
	poly->midpoint.tv[0] /= poly->numverts;
	poly->midpoint.tv[1] /= poly->numverts;
	poly->midpoint.tv[2] /= poly->numverts;

	poly->midpoint.st[0] /= poly->numverts;
	poly->midpoint.st[1] /= poly->numverts;

	poly->midpoint.lm[0] /= poly->numverts;
	poly->midpoint.lm[1] /= poly->numverts;

	poly->midpoint.st128[0] /= poly->numverts;
	poly->midpoint.st128[1] /= poly->numverts;

	// store tv back to v so we can transform bmodels
	poly->midpoint.v[0] = poly->midpoint.tv[0];
	poly->midpoint.v[1] = poly->midpoint.tv[1];
	poly->midpoint.v[2] = poly->midpoint.tv[2];

	// set the most efficient primitive type
	switch (lnumverts)
	{
	case 3:
		poly->QGL_PRIMITIVE_TYPE = GL_TRIANGLES;
		break;

	case 4:
		poly->QGL_PRIMITIVE_TYPE = GL_QUADS;
		break;

	default:
		// this can be better than GL_POLYGON but works the same!!!
		poly->QGL_PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		break;
	}

	// heathaze has 2 more verts so it will always be 5 in a fan
	poly->QGL_HAZE_PRIMITIVE_TYPE = GL_TRIANGLE_FAN;

	// store out the completed bbox.  this isn't actually used in the engine right now, but may be at some point in the future
	VectorCopy (surfmins, fa->mins);
	VectorCopy (surfmaxs, fa->maxs);

	// get the distances.  this is used for scaling surface warps so that they don't get too big on
	// small surfs (and don't get to small on big surfs).
	dists[0] = fabs (fa->maxs[0] - fa->mins[0]);
	dists[1] = fabs (fa->maxs[1] - fa->mins[1]);
	dists[2] = fabs (fa->maxs[2] - fa->mins[2]);

	if (dists[0] < mindist && dists[0] > 0) mindist = dists[0];
	if (dists[1] < mindist && dists[1] > 0) mindist = dists[1];
	if (dists[2] < mindist && dists[2] > 0) mindist = dists[2];

	// should never happen but best to be safe just in case...
	if (mindist > 9999999998) mindist = 0;

	// 8 is used for a 64 x 64 surf so taking a sqrt of mindist should do it.
	fa->warpscale = sqrt (mindist);
}


/*
======================== 
GL_CreateSurfaceLightmap
======================== 
*/ 
void GL_CreateSurfaceLightmap (msurface_t *surf) 
{
	if (surf->flags & (SURF_DRAWSKY|SURF_DRAWTURB)) 
		return;

	surf->lmsmax = (surf->extents[0] >> 4) + 1;
	surf->lmtmax = (surf->extents[1] >> 4) + 1;

	if (!LM_AllocBlock (surf->lmsmax, surf->lmtmax, &surf->light_s, &surf->light_t)) 
	{
		LM_UploadBlock (false);
		LM_InitBlock ();
		if (!LM_AllocBlock (surf->lmsmax, surf->lmtmax, &surf->light_s, &surf->light_t)) 
		{
			VID_Error (ERR_FATAL, "Consecutive calls to LM_AllocBlock (%d,%d) failed\n", surf->lmsmax, surf->lmtmax);
		}
	}

	surf->lightmaptexturenum = gl_lms.current_lightmap_texture;

	R_SetCacheState (surf);
	surf->lbase = gl_lms.lightmap_buffer[gl_lms.current_lightmap_texture] + (surf->light_t * BLOCK_WIDTH + surf->light_s) * LIGHTMAP_BYTES;

	R_BuildLightMap (surf, BLOCK_WIDTH * LIGHTMAP_BYTES);
}


/*
================== 
GL_BeginBuildingLightmaps

================== 
*/ 
void GL_BeginBuildingLightmaps (model_t *m) 
{
	static lightstyle_t	lightstyles[MAX_LIGHTSTYLES];
	int				i;
	unsigned		dummy[128*128];

	memset (gl_lms.allocated, 0, sizeof (gl_lms.allocated));

	r_framecount = 1;		// no dlightcache

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);

	/*
	** setup the base lightstyles so the lightmaps won't have to be regenerated
	** the first time they're seen
	*/ 
	for (i = 0; i<MAX_LIGHTSTYLES; i++) 
	{
		lightstyles[i].rgb[0] = 1;
		lightstyles[i].rgb[1] = 1;
		lightstyles[i].rgb[2] = 1;
		lightstyles[i].white = 3;
	}
	r_newrefdef.lightstyles = lightstyles;

	if (!gl_state.lightmap_textures) 
	{
		gl_state.lightmap_textures	 = TEXNUM_LIGHTMAPS;
		// 		gl_state.lightmap_textures	 = gl_state.texture_extension_number;
		// 		gl_state.texture_extension_number = gl_state.lightmap_textures + MAX_LIGHTMAPS;
	}

	gl_lms.current_lightmap_texture = 1;

	/*
	** if mono lightmaps are enabled and we want to use alpha
	** blending (a,1 - a) then we're likely running on a 3DLabs
	** Permedia2. In a perfect world we'd use a GL_ALPHA lightmap
	** in order to conserve space and maximize bandwidth, however 
	** this isn't a perfect world.
	**
	** So we have to use alpha lightmaps, but stored in GL_RGBA format,
	** which means we only get 1 / 16th the color resolution we should when
	** using alpha lightmaps. If we find another board that supports
	** only alpha lightmaps but that can at least support the GL_ALPHA
	** format then we should change this code to use real alpha maps.
	*/ 
	if (toupper (gl_monolightmap->string[0]) == 'A') 
	{
		gl_lms.internal_format = gl_tex_alpha_format;
	}
	/*
	** try to do hacked colored lighting with a blended texture
	*/ 
	else if (toupper (gl_monolightmap->string[0]) == 'C') 
	{
		gl_lms.internal_format = gl_tex_alpha_format;
	}
	else if (toupper (gl_monolightmap->string[0]) == 'I') 
	{
		gl_lms.internal_format = GL_INTENSITY8;
	}
	else if (toupper (gl_monolightmap->string[0]) == 'L') 
	{
		gl_lms.internal_format = GL_LUMINANCE8;
	}
	else
	{
		gl_lms.internal_format = gl_tex_solid_format;
	}

	/*
	** initialize the dynamic lightmap texture
	*/ 
	glBindTexture (GL_TEXTURE_2D, gl_state.lightmap_textures + 0);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D (GL_TEXTURE_2D, 
		0, 
		gl_lms.internal_format,
		BLOCK_WIDTH, BLOCK_HEIGHT, 
		0, 
		GL_LIGHTMAP_FORMAT, 
		GL_UNSIGNED_BYTE, 
		dummy);
}

/*
======================= 
GL_EndBuildingLightmaps
======================= 
*/ 
void GL_EndBuildingLightmaps (void) 
{
	LM_UploadBlock (false);

	glActiveTexture (GL_TEXTURE1);
	glDisable (GL_TEXTURE_2D);
	glActiveTexture (GL_TEXTURE0);
}

