/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "cvar.h"
#include "render.h"
#include "client.h"
#include "progs.h"
#include "server.h"
#include "keys.h"
#include "view.h"
#include "gl_model.h"
#include "glquake.h"
#include "glext.h"

typedef enum
{
	pt_static,
	pt_grav,
	pt_slowgrav,
	pt_fire,
	pt_explode,
	pt_explode_old,
	pt_explode2,
	pt_blob,
	pt_blob2,
} ptype_t;


typedef enum
{
	// P_POINT is default and will be 0 if not set
	// P_FIXME is a developer notification
	// (i need a good explosion effect...)
	P_POINT,
	P_BLOOD,
	P_SMOKE,
	P_BLAST,
	P_FIXME
} pdraw_t;


void R_AddGlowEffect (float red, float green, float blue, float radius, vec3_t origin);

typedef struct particle_s
{
	// driver-usable fields
	vec3_t		org;
	float		color;

	// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t		vel;
	float		ramp;
	float		die;

	ptype_t		type;
	pdraw_t		draw;

	// blood only
	float		sbase;
	float		tbase;

	// chain by type for drawing
	struct particle_s *pointchain;
	struct particle_s *smokechain;
	struct particle_s *bloodchain;
} particle_t;

// default max # of particles at one time
#define MAX_PARTICLES			2048

// no fewer than this no matter what's on the command line
#define ABSOLUTE_MIN_PARTICLES	512


int ramp1[8] = {0x6f, 0x6d, 0x6b, 0x69, 0x67, 0x65, 0x63, 0x61};
int ramp2[8] = {0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x68, 0x66};
int ramp3[8] = {0x6d, 0x6b, 6, 5, 4, 3};

particle_t	*active_particles, *free_particles;

particle_t	*particles;
int			r_numparticles;

particle_t	*point_particles;
particle_t	*blood_particles;
particle_t	*smoke_particles;

extern int smoketexture;
extern int bloodtexture;

float bloodst_base[3] = {0, 0.34375, 0.6875};
#define bloodst_add  0.3125

/*
===============
R_InitParticles
===============
*/
void R_InitParticles (void)
{
	int	i;

	i = COM_CheckParm ("-particles");

	if (i)
	{
		r_numparticles = (int) (Q_atoi (com_argv[i + 1]));

		if (r_numparticles < ABSOLUTE_MIN_PARTICLES) r_numparticles = ABSOLUTE_MIN_PARTICLES;
	}
	else
	{
		r_numparticles = MAX_PARTICLES;
	}

	particles = (particle_t *) Hunk_AllocName (r_numparticles * sizeof (particle_t), "particles");
}


/*
===============
R_EntityParticles
===============
*/

#define NUMVERTEXNORMALS	162

extern	float	r_avertexnormals[NUMVERTEXNORMALS][3];
vec3_t	avelocities[NUMVERTEXNORMALS];
float	beamlength = 16;
vec3_t	avelocity = {23, 7, 3};
float	partstep = 0.01;
float	timescale = 0.01;

void R_EntityParticles (entity_t *ent)
{
	int			count;
	int			i;
	particle_t	*p;
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	vec3_t		forward;
	float		dist;
	
	dist = 64;
	count = 50;

	if (!avelocities[0][0])
		for (i = 0; i < NUMVERTEXNORMALS * 3; i++)
			avelocities[0][i] = (rand () & 255) * 0.01;

	for (i = 0; i < NUMVERTEXNORMALS; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		angle = cl.time * avelocities[i][0];

		sy = sin (angle);
		cy = cos (angle);

		angle = cl.time * avelocities[i][1];

		sp = sin (angle);
		cp = cos (angle);

		angle = cl.time * avelocities[i][2];

		sr = sin (angle);
		cr = cos (angle);

		forward[0] = cp * cy;
		forward[1] = cp * sy;
		forward[2] = -sp;

		p->die = cl.time + 0.01;
		p->color = 0x6f;
		p->type = pt_explode;

		p->org[0] = ent->origin[0] + r_avertexnormals[i][0] * dist + forward[0] * beamlength;			
		p->org[1] = ent->origin[1] + r_avertexnormals[i][1] * dist + forward[1] * beamlength;			
		p->org[2] = ent->origin[2] + r_avertexnormals[i][2] * dist + forward[2] * beamlength;			

		p->draw = P_POINT;
	}
}


/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int	i;

	free_particles = &particles[0];
	active_particles = NULL;

	for (i = 0; i < r_numparticles; i++)
		particles[i].next = &particles[i + 1];

	particles[r_numparticles - 1].next = NULL;
}


void R_ReadPointFile_f (void)
{
	FILE	*f;
	vec3_t	org;
	int		r;
	int		c;
	particle_t	*p;
	char	name[MAX_OSPATH];

	sprintf (name,"maps/%s.pts", sv.name);

	COM_FOpenFile (name, &f);

	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return;
	}
	
	Con_Printf ("Reading %s...\n", name);
	c = 0;

	while (1)
	{
		r = fscanf (f,"%f %f %f\n", &org[0], &org[1], &org[2]);

		if (r != 3) break;

		c++;

		if (!free_particles)
		{
			Con_Printf ("Not enough free particles\n");
			break;
		}

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = 99999;
		p->color = (-c) & 15;
		p->type = pt_static;

		VectorCopy (vec3_origin, p->vel);
		VectorCopy (org, p->org);

		p->draw = P_POINT;
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}

/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t org, dir;
	int i, count, msgcount, color;

	for (i = 0; i < 3; i++)
		org[i] = MSG_ReadCoord ();

	for (i = 0; i < 3; i++)
		dir[i] = MSG_ReadChar () * (1.0 / 16);

	msgcount = MSG_ReadByte ();
	color = MSG_ReadByte ();

	if (msgcount == 255)
		count = 1024;
	else
		count = msgcount;

	R_RunParticleEffect (org, dir, color, count);
}
	
/*
===============
R_ParticleExplosion

===============
*/
void R_ParticleExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;

	for (i = 0; i < 1024; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 5;
		p->color = ramp1[0];
		p->ramp = rand() & 3;

		if (i & 1)
			p->type = pt_explode;
		else
			p->type = pt_explode2;

		p->org[0] = org[0] + ((rand () % 32) - 16);
		p->org[1] = org[1] + ((rand () % 32) - 16);
		p->org[2] = org[2] + ((rand () % 32) - 16);

		p->vel[0] = (rand () % 512) - 256;
		p->vel[1] = (rand () % 512) - 256;
		p->vel[2] = (rand () % 512) - 256;

		p->draw = P_BLAST;
	}
}

/*
===============
R_ParticleExplosion2

===============
*/
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	int			i, j;
	particle_t	*p;
	int			colorMod = 0;

	for (i = 0; i < 512; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.3;
		p->color = colorStart + (colorMod % colorLength);
		colorMod++;

		p->type = pt_blob;

		p->org[0] = org[0] + ((rand () % 32) - 16);
		p->org[1] = org[1] + ((rand () % 32) - 16);
		p->org[2] = org[2] + ((rand () % 32) - 16);

		p->vel[0] = (rand () % 512) - 256;
		p->vel[1] = (rand () % 512) - 256;
		p->vel[2] = (rand () % 512) - 256;

		p->draw = P_BLAST;
	}
}

/*
===============
R_BlobExplosion

===============
*/
void R_BlobExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;

	for (i = 0; i < 1024; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1 + (rand () & 8) * 0.05;

		if (i & 1)
		{
			p->type = pt_blob;
			p->color = 66 + rand () % 6;
		}
		else
		{
			p->type = pt_blob2;
			p->color = 150 + rand () % 6;
		}

		p->org[0] = org[0] + ((rand () % 32) - 16);
		p->org[1] = org[1] + ((rand () % 32) - 16);
		p->org[2] = org[2] + ((rand () % 32) - 16);

		p->vel[0] = (rand () % 512) - 256;
		p->vel[1] = (rand () % 512) - 256;
		p->vel[2] = (rand () % 512) - 256;

		p->draw = P_BLAST;
	}
}

/*
===============
R_RunParticleEffect

===============
*/
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;

	if (count == 1024)
	{
		R_ParticleExplosion (org);
		return;
	}

	for (i = 0; i < count; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.1 * (rand () % 5);
		p->color = (color & ~7) + (rand() & 7);
		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		p->vel[0] = dir[0] * 15;
		p->vel[1] = dir[1] * 15;
		p->vel[2] = dir[2] * 15;

		if (color == 73 || color == 225)
		{
			p->draw = P_BLOOD;
			p->sbase = bloodst_base[rand () & 3];
			p->tbase = bloodst_base[rand () & 3];
		}
		else p->draw = P_POINT;
	}
}


/*
===============
R_LavaSplash

===============
*/
void R_LavaSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i =- 16; i < 16; i++)
	{
		for (j =- 16; j < 16; j++)
		{
			for (k = 0; k < 1; k++)
			{
				if (!free_particles) return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = cl.time + 2 + (rand () & 31) * 0.02;
				p->color = 224 + (rand () & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8 + (rand () & 7);
				dir[1] = i * 8 + (rand () & 7);
				dir[2] = 256;

				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + (rand () & 63);

				VectorNormalize (dir);						
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);

				p->draw = P_POINT;
			}
		}
	}
}

/*
===============
R_TeleportSplash

===============
*/
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i =- 16; i < 16; i += 4)
	{
		for (j =- 16; j < 16; j += 4)
		{
			for (k =- 24; k < 32; k += 4)
			{
				if (!free_particles) return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = cl.time + 0.2 + (rand () & 7) * 0.02;
				p->color = 7 + (rand () & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8;
				dir[1] = i * 8;
				dir[2] = k * 8;

				p->org[0] = org[0] + i + (rand () & 3);
				p->org[1] = org[1] + j + (rand () & 3);
				p->org[2] = org[2] + k + (rand () & 3);

				VectorNormalize (dir);						
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);

				p->draw = P_POINT;
			}
		}
	}
}



void R_ZerstorerRain (vec3_t org)
{
	int i;
	particle_t	*p;

	// don't draw if paused or in the console, or otherwise
	if (key_dest != key_game) return;

	for (i = 0; i < 16; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1;

		p->type = pt_static;

		p->color = 39 + (rand () % 5);
		p->draw = P_POINT;

		p->org[0] = org[0] + (rand () % 256) - 128;
		p->org[1] = org[1] + (rand () % 256) - 128;
		p->org[2] = org[2] + (rand () % 256) - 128;

		p->vel[0] = 0;
		p->vel[0] = 0;
		p->vel[2] = -128;
	}
}


void R_BubbleParticle (vec3_t org)
{
	int i;
	particle_t	*p;

	// don't draw if paused or in the console, or otherwise
	if (key_dest != key_game) return;

	if (!free_particles) return;
	p = free_particles;
	free_particles = p->next;
	p->next = active_particles;
	active_particles = p;

	// only last one frame
	p->die = cl.time + 0.001;

	p->type = pt_static;

	p->color = 246;
	p->draw = P_POINT;

	p->org[0] = org[0];
	p->org[1] = org[1];
	p->org[2] = org[2];

	p->vel[0] = 0;
	p->vel[1] = 0;
	p->vel[2] = 0;
}



void P_VoreTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		p->color = 9 * 16 + 8 + (rand () & 3);
		p->type = pt_static;
		p->die = cl.time + 0.3;

		p->org[0] = start[0] + ((rand () % 15) - 8);
		p->org[1] = start[1] + ((rand () % 15) - 8);
		p->org[2] = start[2] + ((rand () % 15) - 8);

		p->draw = P_POINT;

		VectorAdd (start, vec, start);
	}
}


void P_Tracer5Trail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);

		p->die = cl.time + 0.5;
		p->type = pt_static;

		p->color = 230 + ((tracercount & 4) << 1);

		tracercount++;

		VectorCopy (start, p->org);

		if (tracercount & 1)
		{
			p->vel[0] = 30 * vec[1];
			p->vel[1] = 30 *- vec[0];
		}
		else
		{
			p->vel[0] = 30 *- vec[1];
			p->vel[1] = 30 * vec[0];
		}

		p->draw = P_POINT;

		VectorAdd (start, vec, start);
	}
}


void P_Blood4Trail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		p->type = pt_grav;
		p->color = 67 + (rand () & 3);

		p->org[0] = start[0] + ((rand () % 6) - 3);
		p->org[1] = start[1] + ((rand () % 6) - 3);
		p->org[2] = start[2] + ((rand () % 6) - 3);

		p->draw = P_BLOOD;

		p->sbase = bloodst_base[rand () & 3];
		p->tbase = bloodst_base[rand () & 3];

		// was 3
		len -= 1;

		VectorAdd (start, vec, start);
	}
}


void P_Tracer3Trail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);

		p->die = cl.time + 0.5;
		p->type = pt_static;

		p->color = 52 + ((tracercount & 4) << 1);

		tracercount++;

		VectorCopy (start, p->org);

		if (tracercount & 1)
		{
			p->vel[0] = 30 * vec[1];
			p->vel[1] = 30 *- vec[0];
		}
		else
		{
			p->vel[0] = 30 *- vec[1];
			p->vel[1] = 30 * vec[0];
		}

		p->draw = P_POINT;

		VectorAdd (start, vec, start);
	}
}


void P_Blood2Trail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		p->type = pt_grav;
		p->color = 67 + (rand () & 3);

		p->org[0] = start[0] + ((rand () % 6) - 3);
		p->org[1] = start[1] + ((rand () % 6) - 3);
		p->org[2] = start[2] + ((rand () % 6) - 3);

		p->draw = P_BLOOD;

		p->sbase = bloodst_base[rand () & 3];
		p->tbase = bloodst_base[rand () & 3];

		VectorAdd (start, vec, start);
	}
}


void P_SmokeTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;

	int oldlen;
	vec3_t oldstart;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	oldlen = len;

	oldstart[0] = start[0];
	oldstart[1] = start[1];
	oldstart[2] = start[2];

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		p->ramp = (rand () & 3) + 2;
		p->color = ramp3[(int) p->ramp];
		p->type = pt_fire;

		p->org[0] = start[0] + ((rand () % 6) - 3);
		p->org[1] = start[1] + ((rand () % 6) - 3);
		p->org[2] = start[2] + ((rand () % 6) - 3);

		p->draw = P_SMOKE;

		VectorAdd (start, vec, start);
	}

	// add a spark cloud around the smoke trails
	while (oldlen > 0)
	{
		oldlen -= 3;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2.0;

		p->ramp = (rand () & 3) + 1;
		p->color = ramp3[(int) p->ramp];

		p->type = pt_fire;
		p->draw = P_POINT;

		p->org[0] = oldstart[0] + ((rand () % 12) - 6);
		p->org[1] = oldstart[1] + ((rand () % 12) - 6);
		p->org[2] = oldstart[2] + ((rand () % 12) - 6);

		VectorAdd (oldstart, vec, oldstart);
	}
}


void P_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;

	int oldlen;
	vec3_t oldstart;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	oldlen = len;

	oldstart[0] = start[0];
	oldstart[1] = start[1];
	oldstart[2] = start[2];

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		p->ramp = (rand () & 3);
		p->color = ramp3[(int) p->ramp];
		p->type = pt_fire;

		p->org[0] = start[0] + ((rand () % 6) - 3);
		p->org[1] = start[1] + ((rand () % 6) - 3);
		p->org[2] = start[2] + ((rand () % 6) - 3);

		p->draw = P_SMOKE;

		VectorAdd (start, vec, start);
	}

	// add a spark cloud around the smoke trails
	while (oldlen > 0)
	{
		oldlen -= 3;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2.0;

		p->ramp = (rand () & 3) + 1;
		p->color = ramp3[(int) p->ramp];

		p->type = pt_fire;
		p->draw = P_POINT;

		p->org[0] = oldstart[0] + ((rand () % 12) - 6);
		p->org[1] = oldstart[1] + ((rand () % 12) - 6);
		p->org[2] = oldstart[2] + ((rand () % 12) - 6);

		VectorAdd (oldstart, vec, oldstart);
	}
}


void P_LaserTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	if (type < 128)
		dec = 2;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len >= 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);

		p->die = cl.time + 0.5;
		p->type = pt_static;

		p->color = 230 + ((tracercount & 4) << 1);

		tracercount++;

		VectorCopy (start, p->org);

		/*
		if (tracercount & 1)
		{
			p->vel[0] = 30 * vec[1];
			p->vel[1] = 30 *- vec[0];
		}
		else
		{
			p->vel[0] = 30 *- vec[1];
			p->vel[1] = 30 * vec[0];
		}
		*/

		p->draw = P_POINT;

		VectorAdd (start, vec, start);
	}
}


void R_RocketTrail (vec3_t start, vec3_t end, int type)
{
	switch (type)
	{
	case 0:
	case 128:
		P_RocketTrail (start, end, type);
		break;

	case 1:
	case 129:
		P_SmokeTrail (start, end, type);
		break;

	case 2:
	case 130:
		P_Blood2Trail (start, end, type);
		break;

	case 3:
	case 131:
		// wizard
		P_Tracer3Trail (start, end, type);
		break;

	case 4:
	case 132:
		P_Blood4Trail (start, end, type);
		break;

	case 5:
	case 133:
		// knight
		P_Tracer5Trail (start, end, type);
		break;

	case 6:
	case 134:
		P_VoreTrail (start, end, type);
		break;

	case 7:
	case 135:
		P_LaserTrail (start, end, type);
		// laser bolts
		break;

	default:
		break;
	}
}


extern	cvar_t	sv_gravity;


/*
===============
R_FreeParticles

update the active particle list by clearing dead particles from the linked list
===============
*/
void R_FreeParticles (void)
{
	particle_t *p, *kill;

	point_particles = NULL;
	smoke_particles = NULL;
	blood_particles = NULL;

	// jump to the first active particle in the list
	while (1)
	{
		kill = active_particles;

		if (kill && kill->die < cl.time)
		{
			active_particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;

			continue;
		}

		break;
	}

	for (p = active_particles; p; p = p->next)
	{
		// jump to the next active particle in the list
		while (1)
		{
			kill = p->next;

			if (kill && kill->die < cl.time)
			{
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;

				continue;
			}

			break;
		}

		// sort the particles for more efficient drawing
		switch (p->draw)
		{
		case P_BLOOD:
			p->bloodchain = blood_particles;
			blood_particles = p;
			break;

		case P_SMOKE:
			p->smokechain = smoke_particles;
			smoke_particles = p;
			break;

		case P_BLAST:
		default:
			p->pointchain = point_particles;
			point_particles = p;
			break;
		}
	}
}


/*
===============
R_UpdateParticles

update particle velocity and origin
===============
*/
void R_UpdateParticles (void)
{
	particle_t *p;
	float grav;
	int i;
	float time2, time3;
	float time1;
	float dvel;

	time3 = frametime * 15;
	time2 = frametime * 10;
	time1 = frametime * 5;
	grav = frametime * sv_gravity.value * 0.05;
	dvel = 4 * frametime;

	for (p = active_particles; p; p = p->next)
	{
		// modify the origin by the particle velocity
		if (p->vel[0] || p->vel[1] || p->vel[2])
		{
			p->org[0] += p->vel[0] * frametime;
			p->org[1] += p->vel[1] * frametime;
			p->org[2] += p->vel[2] * frametime;
		}

		switch (p->type)
		{
		case pt_static:
			break;

		case pt_fire:
			p->ramp += time1;

			if (p->ramp >= 6)
				p->die = -1;
			else p->color = ramp3[(int) p->ramp];

			p->vel[2] += grav;

			break;

		case pt_explode:
			p->ramp += time2;

			if (p->ramp >= 8)
				p->die = -1;
			else p->color = ramp1[(int) p->ramp];

			p->vel[0] += p->vel[0] * dvel;
			p->vel[1] += p->vel[1] * dvel;
			p->vel[2] += p->vel[2] * dvel;

			p->vel[2] -= grav;

			break;

		case pt_explode2:
			p->ramp += time3;

			if (p->ramp >= 8)
				p->die = -1;
			else
				p->color = ramp2[(int) p->ramp];

			p->vel[0] += p->vel[0] * frametime;
			p->vel[1] += p->vel[1] * frametime;
			p->vel[2] += p->vel[2] * frametime;

			p->vel[2] -= grav;

			break;

		case pt_blob:
			p->vel[0] += p->vel[0] * dvel;
			p->vel[1] += p->vel[1] * dvel;
			p->vel[2] += p->vel[2] * dvel;

			p->vel[2] -= grav;

			break;

		case pt_blob2:
			p->vel[0] += p->vel[0] * dvel;
			p->vel[1] += p->vel[1] * dvel;

			p->vel[2] -= grav;

			break;

		case pt_grav:
		case pt_slowgrav:
			p->vel[2] -= grav;

			break;
		}
	}
}


/*
===============
R_DrawParticles
===============
*/
void R_DrawParticles (void)
{
	particle_t *p;
	vec3_t 	up, right;
	float scale;
	byte pColor[4];
	vec3_t pUp, pRight, pOrg;
	qboolean ptexturedraw = false;

	R_FreeParticles ();

	if (!active_particles) return;

	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);

	if (point_particles)
	{
		// point parameters are set in GL_Init ()
		glEnable (GL_POINT_SMOOTH);
		glDisable (GL_TEXTURE_2D);

		glBegin (GL_POINTS);

		for (p = point_particles; p; p = p->pointchain)
		{
			// draw the particle
			glColor3ubv ((byte *) &d_8to24table[(int) p->color]);
			glVertex3fv (p->org);
		}

		glEnd ();

		glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
		glEnable (GL_TEXTURE_2D);
		glDisable (GL_POINT_SMOOTH);
	}

	if (smoke_particles || blood_particles)
	{
		ptexturedraw = true;
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		glActiveTexture (GL_TEXTURE0);
	}

	if (blood_particles)
	{
		// we keep the original blendfunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA) for
		// blood cos the other one makes it too bright
		glBindTexture (GL_TEXTURE_2D, bloodtexture);

		// regular GLQuake uses 1.5 - I need to increase the scale cos my blood particles
		// occupy less space in a larger texture
		VectorScale (vup, 4, up);
		VectorScale (vright, 4, right);

		glBegin (GL_QUADS);

		for (p = blood_particles; p; p = p->bloodchain)
		{
			scale = (p->org[0] - r_origin[0]) * vpn[0] + 
					(p->org[1] - r_origin[1]) * vpn[1] + 
					(p->org[2] - r_origin[2]) * vpn[2];

			if (scale < 20)
			{
				p->die = -1;
				continue;
			}

			scale = 1 + scale * 0.004;

			glColor3ubv ((byte *) &d_8to24table[(int) p->color]);

			// these particles will be off-center, so correct that
			pUp[0] = up[0] * scale;
			pUp[1] = up[1] * scale;
			pUp[2] = up[2] * scale;

			pRight[0] = right[0] * scale;
			pRight[1] = right[1] * scale;
			pRight[2] = right[2] * scale;

			pOrg[0] = p->org[0] - (pRight[0] / 2) - (pUp[0] / 2);
			pOrg[1] = p->org[1] - (pRight[1] / 2) - (pUp[1] / 2);
			pOrg[2] = p->org[2] - (pRight[2] / 2) - (pUp[2] / 2);

			glTexCoord2f (p->sbase, p->tbase);
			glVertex3fv (pOrg);

			glTexCoord2f (p->sbase + bloodst_add, p->tbase);
			glVertex3f (pOrg[0] + pUp[0],
						pOrg[1] + pUp[1],
						pOrg[2] + pUp[2]);

			glTexCoord2f (p->sbase + bloodst_add, p->tbase + bloodst_add);
			glVertex3f (pOrg[0] + pRight[0] + pUp[0],
						pOrg[1] + pRight[1] + pUp[1],
						pOrg[2] + pRight[2] + pUp[2]);

			glTexCoord2f (p->sbase, p->tbase + bloodst_add);
			glVertex3f (pOrg[0] + pRight[0],
						pOrg[1] + pRight[1],
						pOrg[2] + pRight[2]);

			// don't stain these - there are 2 types of blood splash: those generated
			// by flying gibs, where there will already be a stain from the gib, and
			// those generated by attacks, which normally occur too far out of the
			// the range of any surrounding walls to generate stains.  in tests, the
			// radius of the stain would have to be increased so high that it creates
			// a huge red splotch far in excess of the amount of blood shown.
		}

		glEnd ();
	}

	if (smoke_particles)
	{
		// this blend func gives the brighter smoke visible in other engines - i don't like it
		// glBlendFunc (GL_SRC_ALPHA, GL_ONE);

		glBindTexture (GL_TEXTURE_2D, smoketexture);

		VectorScale (vup, 9, up);
		VectorScale (vright, 9, right);

		glBegin (GL_QUADS);

		for (p = smoke_particles; p; p = p->smokechain)
		{
			scale = (p->org[0] - r_origin[0]) * vpn[0] + 
					(p->org[1] - r_origin[1]) * vpn[1] + 
					(p->org[2] - r_origin[2]) * vpn[2];

			if (scale < 20)
			{
				p->die = -1;
				continue;
			}

			scale = 1 + scale * 0.004;

			pColor[0] = ((byte *) &d_8to24table[(int) p->color])[0];
			pColor[1] = ((byte *) &d_8to24table[(int) p->color])[1];
			pColor[2] = ((byte *) &d_8to24table[(int) p->color])[2];
			pColor[3] = 64;

			glColor4ubv (pColor);

			// these particles will be off-center, so correct that
			pUp[0] = up[0] * scale;
			pUp[1] = up[1] * scale;
			pUp[2] = up[2] * scale;

			pRight[0] = right[0] * scale;
			pRight[1] = right[1] * scale;
			pRight[2] = right[2] * scale;

			pOrg[0] = p->org[0] - (pRight[0] / 2) - (pUp[0] / 2);
			pOrg[1] = p->org[1] - (pRight[1] / 2) - (pUp[1] / 2);
			pOrg[2] = p->org[2] - (pRight[2] / 2) - (pUp[2] / 2);

			glTexCoord2f (0, 0);
			glVertex3fv (pOrg);

			glTexCoord2f (1, 0);
			glVertex3f (pOrg[0] + pUp[0],
						pOrg[1] + pUp[1],
						pOrg[2] + pUp[2]);

			glTexCoord2f (1, 1);
			glVertex3f (pOrg[0] + pRight[0] + pUp[0],
						pOrg[1] + pRight[1] + pUp[1],
						pOrg[2] + pRight[2] + pUp[2]);

			glTexCoord2f (0, 1);
			glVertex3f (pOrg[0] + pRight[0],
						pOrg[1] + pRight[1],
						pOrg[2] + pRight[2]);
		}

		glEnd ();

		// only need to reset the blendfunc if we're drawing smoke
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	if (ptexturedraw)
	{
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}

	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	R_UpdateParticles ();
}

