/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_mesh.c: triangle model functions

#include "quakedef.h"

/*
=================================================================

ALIAS MODEL DISPLAY LIST GENERATION

=================================================================
*/

model_t		*aliasmodel;
aliashdr_t	*paliashdr;

qboolean	used[8192];

// the command list holds counts and s/t values that are valid for
// every frame
int		commands[8192];
int		numcommands;

// all frames will have their vertexes rearranged and expanded
// so they are in the order expected by the command list
int		vertexorder[8192];
int		numorder;

int		allverts, alltris;

int		stripverts[128];
int		striptris[128];
int		stripcount;

/*
================
StripLength
================
*/
int	StripLength (int starttri, int startv)
{
	int			m1, m2;
	int			j;
	mtriangle_t	*last, *check;
	int			k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+2)%3];
	m2 = last->vertindex[(startv+1)%3];

	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront)
			continue;
		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1)
				continue;
			if (check->vertindex[ (k+1)%3 ] != m2)
				continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j])
				goto done;

			// the new edge
			if (stripcount & 1)
				m2 = check->vertindex[ (k+2)%3 ];
			else
				m1 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = check->vertindex[ (k+2)%3 ];
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2)
			used[j] = 0;

	return stripcount;
}

/*
===========
FanLength
===========
*/
int	FanLength (int starttri, int startv)
{
	int		m1, m2;
	int		j;
	mtriangle_t	*last, *check;
	int		k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+0)%3];
	m2 = last->vertindex[(startv+2)%3];


	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront)
			continue;
		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1)
				continue;
			if (check->vertindex[ (k+1)%3 ] != m2)
				continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j])
				goto done;

			// the new edge
			m2 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = m2;
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2)
			used[j] = 0;

	return stripcount;
}


/*
================
BuildTris

Generate a list of trifans or strips
for the model, which holds for all frames
================
*/
void BuildTris (void)
{
	int		i, j, k;
	int		startv;
	mtriangle_t	*last, *check;
	int		m1, m2;
	int		striplength;
	trivertx_t	*v;
	mtriangle_t *tv;
	float	s, t;
	int		index;
	int		len, bestlen, besttype;
	int		bestverts[1024];
	int		besttris[1024];
	int		type;

	//
	// build tristrips
	//
	numorder = 0;
	numcommands = 0;
	memset (used, 0, sizeof(used));
	for (i=0 ; i<pheader->numtris ; i++)
	{
		// pick an unused triangle and start the trifan
		if (used[i])
			continue;

		bestlen = 0;
		for (type = 0 ; type < 2 ; type++)
//	type = 1;
		{
			for (startv =0 ; startv < 3 ; startv++)
			{
				if (type == 1)
					len = StripLength (i, startv);
				else
					len = FanLength (i, startv);
				if (len > bestlen)
				{
					besttype = type;
					bestlen = len;
					for (j=0 ; j<bestlen+2 ; j++)
						bestverts[j] = stripverts[j];
					for (j=0 ; j<bestlen ; j++)
						besttris[j] = striptris[j];
				}
			}
		}

		// mark the tris on the best strip as used
		for (j=0 ; j<bestlen ; j++)
			used[besttris[j]] = 1;

		if (besttype == 1)
			commands[numcommands++] = (bestlen+2);
		else
			commands[numcommands++] = -(bestlen+2);

		for (j=0 ; j<bestlen+2 ; j++)
		{
			// emit a vertex into the reorder buffer
			k = bestverts[j];
			vertexorder[numorder++] = k;

			// emit s/t coords into the commands stream
			s = stverts[k].s;
			t = stverts[k].t;
			if (!triangles[besttris[0]].facesfront && stverts[k].onseam)
				s += pheader->skinwidth / 2;	// on back side
			s = (s + 0.5) / pheader->skinwidth;
			t = (t + 0.5) / pheader->skinheight;

			*(float *)&commands[numcommands++] = s;
			*(float *)&commands[numcommands++] = t;
		}
	}

	commands[numcommands++] = 0;		// end of list marker

	Con_DPrintf ("%3i tri %3i vert %3i cmd\n", pheader->numtris, numorder, numcommands);

	allverts += numorder;
	alltris += pheader->numtris;
}


/*
================
GL_MakeAliasModelDisplayLists
================
*/
void GL_MakeAliasModelDisplayLists (model_t *m, aliashdr_t *hdr)
{
	int		i, j;
	maliasgroup_t	*paliasgroup;
	int			*cmds;
	drawvertx_t	*verts;
	char	cache[MAX_QPATH], fullpath[MAX_OSPATH], *c;
	FILE	*f;
	int		len;
	byte	*data;

	aliasmodel = m;
	paliashdr = hdr;

	// build every time
	// requires more load time but avoids undying ugliness in the longer term
	BuildTris ();

	// save the data out
	paliashdr->poseverts = numorder;

	cmds = Hunk_Alloc (numcommands * 4);
	paliashdr->commands = (byte *)cmds - (byte *)paliashdr;
	Q_memcpy (cmds, commands, numcommands * 4);

	verts = Hunk_Alloc (paliashdr->numposes * paliashdr->poseverts * sizeof (drawvertx_t));
	paliashdr->posedata = (byte *) verts - (byte *) paliashdr;

	for (i = 0; i < paliashdr->numposes; i++)
	{
		for (j = 0; j < numorder; j++)
		{
			// prescale the verts at load time.  this requires moving over to floats for storing verts, which
			// gives a * 4 memory overhead but cuts down on a LOT of run-time calculations
			// i've tried converting back to bytes but STRANGE things happen...
			verts->v[0] = ((float) poseverts[i][vertexorder[j]].v[0]) * paliashdr->scale[0] + paliashdr->scale_origin[0];
			verts->v[1] = ((float) poseverts[i][vertexorder[j]].v[1]) * paliashdr->scale[1] + paliashdr->scale_origin[1];
			verts->v[2] = ((float) poseverts[i][vertexorder[j]].v[2]) * paliashdr->scale[2] + paliashdr->scale_origin[2];

			verts->lightnormalindex = poseverts[i][vertexorder[j]].lightnormalindex;
			verts->nodraw = false;

			verts++;
		}
	}

	// here's a nasty hack for you!
	// code for elimination of muzzleflashes on viewmodels.
	// is it a view model???
	if ((m->name[6] == 'v' || m->name[6] == 'V') && m->name[7] == '_' && hdr->numframes > 2)
	{
		// get pointers to the verts
		drawvertx_t *vertsf0 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);
		drawvertx_t *vertsf1 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);

		// setthese pointers to the 0th and 1st frames
		vertsf0 += hdr->frames[0].firstpose * hdr->poseverts;
		vertsf1 += hdr->frames[1].firstpose * hdr->poseverts;

		// now go through them and compare.  we expect that (a) the animation is sensible and there's no major
		// difference between the 2 frames to be expected, and (b) any verts that do exhibit a major difference
		// can be assumed to belong to the muzzleflash
		for (j = 0; j < numorder; j++)
		{
			// get difference in front to back movement
			float vdiff = vertsf1->v[0] - vertsf0->v[0];

			// if it's above a certain treshold, assume a muzzleflash and mark for nodraw
			// 10 is the approx lowest range of visible front to back in a view model, so that seems reasonable to work with
			if (vdiff > 10)
				vertsf0->nodraw = true;
			else vertsf0->nodraw = false;

			// next set of verts
			vertsf0++;
			vertsf1++;
		}
	}

	// now we mark every other vert in the model as per the instructions in the first frame
	for (i = 1; i < paliashdr->numframes; i++)
	{
		// get pointers to the verts
		drawvertx_t *vertsf0 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);
		drawvertx_t *vertsfi = (drawvertx_t *) ((byte *) hdr + hdr->posedata);

		// set these pointers to the 0th and i'th frames
		vertsf0 += hdr->frames[0].firstpose * hdr->poseverts;
		vertsfi += hdr->frames[i].firstpose * hdr->poseverts;

		for (j = 0; j < numorder; j++)
		{
			// just copy it across
			vertsfi->nodraw = vertsf0->nodraw;

			// next set of verts
			vertsf0++;
			vertsfi++;
		}
	}
}


/*
==================
GL_CountAliasPolys

Counts the number of verts and polys in our MDLs so that we can dynamically allocate memory for
arrays to hold them at load time.
==================
*/
void GL_CountAliasPolys (aliashdr_t *paliashdr)
{
	int count;
	int PolyCount;
	int VertexCount;
	int *order;

	// order contains (a) type of primitive, and (b) texture st coords
	order = (int *) ((byte *) paliashdr + paliashdr->commands);

	// and so it begins...
	PolyCount = 0;
	VertexCount = 0;

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		// set the correct primitive type on each poly
		if (count < 0) count = -count;

		do
		{
			// increment counters
			VertexCount++;
			order += 2;
		} while (--count);

		// increment the poly counter
		PolyCount++;
	}

	// store out the counts
	paliashdr->numrenderpolys = PolyCount;
	paliashdr->numrenderverts = VertexCount;
}


/*
=============================================================

  ALIAS MODEL RENDER

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = 
{
#include "anorms.h"
};

vec3_t	shadevector;
vec3_t	shadelight;

// the new light scaling means that these go a little darker than regular glquake
#define MINIMUM_PLAYER_LIGHT	4.0f
#define MINIMUM_GUN_LIGHT		8.0f

#define MINIMUM_LIGHT			0.0f
#define MAXIMUM_LIGHT			128.0f
#define LIGHT_DIVIDER			128.0f

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
;

float	*shadedots1 = r_avertexnormal_dots[0];
float	*shadedots2 = r_avertexnormal_dots[0];
float	lightLerpOffset;

// polys and verts used for loading
glpoly_t *AliasPolys = NULL;
polyverts_t *AliasVerts = NULL;
vec3_t *AliasVecs = NULL;

/*
===================
AllocAliasPolyVerts

Allocates memory at load time (in R_NewMap) to hold the verts and polys used for drawing.
===================
*/
void AllocAliasPolyVerts (void)
{
	int i;
	model_t *m;
	aliashdr_t *paliashdr;

	// initial maxes
	int MaxPolys = 0;
	int MaxVerts = 0;

	// start on the model after the world
	for (i = 1; i < MAX_MODELS; i++)
	{
		m = cl.model_precache[i];
		if (!m) break;
		if (m->type != mod_alias) continue;

		// get the header and run the counts on it
		paliashdr = (aliashdr_t *) Mod_Extradata (m);
		GL_CountAliasPolys (paliashdr);

		// now check against the maxes
		if (paliashdr->numrenderpolys > MaxPolys) MaxPolys = paliashdr->numrenderpolys;
		if (paliashdr->numrenderverts > MaxVerts) MaxVerts = paliashdr->numrenderverts;
	}

	// store 1 extra slot for no good reason whatsoever
	AliasPolys = Hunk_Alloc (sizeof (glpoly_t) * (MaxPolys + 1));
	AliasVerts = Hunk_Alloc (sizeof (polyverts_t) * (MaxVerts + 1));
	AliasVecs = Hunk_Alloc (sizeof (vec3_t) * (MaxVerts + 1));
}


/*
===================
GL_MakeAliasPolys

Makes a list of polys from the mesh data so that they can be sent through a common render path.
Because this data changes each frame (aside from s/t) it must be dynamically updated each frame!!!

This is all CPU and RAM muscle, so it ups the specs a bit, but considering Q1 was originally meant to run
on a P60 that's no big deal at all.

People, we cannot cache.
(Well, we could, but it would require several squillion MB of RAM)
(about 36MB for MAX_VISEDICTS...)
===================
*/
void GL_MakeAliasPolys (aliashdr_t *paliashdr, int pose1, int pose2, float backlerp)
{
	vec3_t diff;
	float l;
	float ldiff;
	float lquant[2];

	drawvertx_t *verts1;
	drawvertx_t *verts2;
	int *order;

	int count;
	int PolyCount;
	int VertexCount;

	float frontlerp = 1.0f - backlerp;

	// get the verts for each pose to be blended
	verts1 = (drawvertx_t *) ((byte *) paliashdr + paliashdr->posedata);
	verts2 = verts1;

	// get the correct vert pointer for each pose to be blended
	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	// order contains (a) type of primitive, and (b) texture st coords
	order = (int *) ((byte *) paliashdr + paliashdr->commands);

	// and so it begins...
	PolyCount = 0;
	VertexCount = 0;

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		// set the correct primitive type on each poly
		if (count < 0)
		{
			count = -count;
			AliasPolys[PolyCount].PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		}
		else AliasPolys[PolyCount].PRIMITIVE_TYPE = GL_TRIANGLE_STRIP;

		// set the rest of the base data
		AliasPolys[PolyCount].verts = &AliasVerts[VertexCount];
		AliasPolys[PolyCount].basev = &AliasVecs[VertexCount];
		AliasPolys[PolyCount].vertoffset = VertexCount;
		AliasPolys[PolyCount].numverts = 0;

		// current poly always points to NULL so that there will be a definite end to the list
		AliasPolys[PolyCount].next = NULL;

		// chain previous poly to this one
		if (PolyCount > 0) AliasPolys[PolyCount - 1].next = &AliasPolys[PolyCount];

		do
		{
			// support for fullbright flash elimination
			if (verts1->nodraw || verts2->nodraw) goto NoDrawVerts;

			// set up the lighting
			lquant[0] = (shadedots1[verts1->lightnormalindex] + (backlerp * (shadedots1[verts2->lightnormalindex] - shadedots1[verts1->lightnormalindex])));
			lquant[1] = (shadedots2[verts1->lightnormalindex] + (backlerp * (shadedots2[verts2->lightnormalindex] - shadedots2[verts1->lightnormalindex])));

			if (lquant[0] != lquant[1])
			{
				if (lquant[0] > lquant[1])
				{
					ldiff = lquant[0] - lquant[1];
					ldiff *= lightLerpOffset;
					l = lquant[0] - ldiff;
				}
				else
				{
					ldiff = lquant[1] - lquant[0];
					ldiff *= lightLerpOffset;
					l = lquant[0] + ldiff;
				}
			}
			else l = lquant[0];

			// set texcoords
			AliasVerts[VertexCount].st[0] = ((float *) order)[0];
			AliasVerts[VertexCount].st[1] = ((float *) order)[1];

			// set colours
			AliasVerts[VertexCount].rgba[0] = l * shadelight[0];
			AliasVerts[VertexCount].rgba[1] = l * shadelight[1];
			AliasVerts[VertexCount].rgba[2] = l * shadelight[2];
			AliasVerts[VertexCount].rgba[3] = 1;

			// set vertex data
			// prescaling the verts at load time means that we get to avoid having to do sexy
			// stuff with the interpolation, and can just draw away...!
			AliasVecs[VertexCount][0] = verts1->v[0] * frontlerp + verts2->v[0] * backlerp;
			AliasVecs[VertexCount][1] = verts1->v[1] * frontlerp + verts2->v[1] * backlerp;
			AliasVecs[VertexCount][2] = verts1->v[2] * frontlerp + verts2->v[2] * backlerp;

			// increment counters
			AliasPolys[PolyCount].numverts++;
			VertexCount++;

NoDrawVerts:;
			order += 2;

			// increment vertex pointers
			verts1++;
			verts2++;
		} while (--count);

		// increment the poly counter
		PolyCount++;
	}
}


/*
=================
R_SetupAliasFrame

=================
*/
void R_SetupAliasFrame (entity_t *e, aliashdr_t *paliashdr)
{
	int				pose, numposes;
	float			interval;
	float			blend;
	qboolean		lerpmdl = true;
	float			matrix[16];
	int				anim;
	int				texnum;
	int				i;

	if ((e->frame >= paliashdr->numframes) || (e->frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", e->frame);
		e->frame = 0;
	}

	pose = paliashdr->frames[e->frame].firstpose;
	numposes = paliashdr->frames[e->frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[e->frame].interval;
		pose += (int)(cl.time / interval) % numposes;

		// not really needed for non-interpolated mdls, but does no harm
		e->frame_interval = interval;
	}
	else e->frame_interval = 0.1;

	// conditions for turning lerping off
	if (e->pose2 != pose) lerpmdl = false;					// don't interpolate between different poses
	if (paliashdr->numposes == 1) lerpmdl = false;			// only one pose
	if (e == &cl.viewent && e->frame == 0) lerpmdl = false;	// super-nailgun spinning down bug
	if (r_framecount < 10) lerpmdl = false;					// reliving their dying throes

	// entities with these special flags aren't interpolated either
	if (e->model->flags & EF_GIB) lerpmdl = false;
	if (e->model->flags & EF_ZOMGIB) lerpmdl = false;
	if (e->model->flags & EF_WIZARDGOO) lerpmdl = false;
	if (e->model->flags & EF_KNIGHTFIRE) lerpmdl = false;
	if (e->model->flags & EF_VORE) lerpmdl = false;
	if (e->model->flags & EF_ROCKET) lerpmdl = false;
	if (e->model->flags & EF_GRENADE) lerpmdl = false;
	if (e->model->flags & EF_ROTATE) lerpmdl = false;

	// wait till now to do this so we can correctly switch interpolation off
	// (viewent can switch mdls but keep the same pose, and therefore have invalid data cached in the poses)
	if (!lerpmdl)
	{
		e->frame_start_time = realtime;
		e->pose1 = e->pose2;
		e->pose2 = pose;
		blend = 0;
	}
	else blend = (realtime - e->frame_start_time) / e->frame_interval;

	// don't let blend pass 1
	if (cl.paused || blend > 1.0) blend = 1.0;

	// transform
	R_RotateForEntity (e);

	// make the polys
	GL_MakeAliasPolys (paliashdr, e->pose1, e->pose2, blend);

	// transform verts
	R_TransformPolys (AliasPolys, e->matrix);

	// get the correct texture number
	anim = (int) (cl.time * 10) & 3;
	texnum = paliashdr->glt[currententity->skinnum][anim]->texnum;

	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (currententity->colormap != vid.colormap && !gl_nocolors.value)
	{
		i = currententity - cl_entities;

		if (i >= 1 && i <= cl.maxclients) texnum = playertextures[i];
	}

	// regular skin
	glActiveTexture (GL_TEXTURE0);
	glBindTexture (GL_TEXTURE_2D, texnum);

	// fullbright
	glActiveTexture (GL_TEXTURE1);

	if (paliashdr->glt[currententity->skinnum][anim]->lumatexnum)
		glBindTexture (GL_TEXTURE_2D, paliashdr->glt[currententity->skinnum][anim]->lumatexnum);
	else glBindTexture (GL_TEXTURE_2D, r_blacktexture.texnum);

	GL_DrawVertexArrays (AliasPolys);
}


/*
=================
R_DrawAliasModel

=================
*/
void R_DrawAliasModel (entity_t *e)
{
	int			i, j;
	model_t		*clmodel;
	vec3_t		mins, maxs;
	aliashdr_t	*paliashdr;
	float		an;
	float		ang_ceil, ang_floor;

	clmodel = currententity->model;

	VectorAdd (currententity->origin, clmodel->mins, mins);
	VectorAdd (currententity->origin, clmodel->maxs, maxs);

	if (e != &cl.viewent)
		if (R_CullBox (mins, maxs))
			return;

	VectorCopy (currententity->origin, r_entorigin);
	VectorSubtract (r_origin, r_entorigin, modelorg);

	// get lighting information
	if (e->model->flags & EF_ROTATE)
	{
		// pickup boxes are lit differently, so do something similar for mdls to keep it consistent
		shadelight[0] = 0.8f;
		shadelight[1] = 0.8f;
		shadelight[2] = 0.8f;
	}
	else
	{
		R_LightPoint (currententity->origin, shadelight);

		// don;t go below 0
		if (shadelight[0] < MINIMUM_LIGHT) shadelight[0] = MINIMUM_LIGHT;
		if (shadelight[1] < MINIMUM_LIGHT) shadelight[1] = MINIMUM_LIGHT;
		if (shadelight[2] < MINIMUM_LIGHT) shadelight[2] = MINIMUM_LIGHT;

		// clamp maximum
		if (shadelight[0] > MAXIMUM_LIGHT) shadelight[0] = MAXIMUM_LIGHT;
		if (shadelight[1] > MAXIMUM_LIGHT) shadelight[1] = MAXIMUM_LIGHT;
		if (shadelight[2] > MAXIMUM_LIGHT) shadelight[2] = MAXIMUM_LIGHT;

		// clamp some minimums even higher
		if (e == &cl.viewent || ((currententity - cl_entities) >= 1 && (currententity - cl_entities) <= cl.maxclients))
		{
			// hack to avoid a goto
			while (1)
			{
				float gs = (shadelight[0] * 30.0f + shadelight[1] * 59.0f + shadelight[2] * 11.0f) / 100.0f;
				float Multy;

				// hack to prevent division by 0 or near-zero in almost total blackness
				if (gs > 1)
					e->lastGoodGS = gs;
				else
				{
					gs = e->lastGoodGS;
					shadelight[0] = e->lastGoodShadelight[0];
					shadelight[1] = e->lastGoodShadelight[1];
					shadelight[2] = e->lastGoodShadelight[2];

					// hack to avoid a goto
					break;
				}

				if (e == &cl.viewent && gs < MINIMUM_GUN_LIGHT)
				{
					// allways give the gun some light
					Multy = MINIMUM_GUN_LIGHT / gs;

					shadelight[0] *= Multy;
					shadelight[1] *= Multy;
					shadelight[2] *= Multy;
				}
				else if (((currententity - cl_entities) >= 1 && (currententity - cl_entities) <= cl.maxclients) && gs < MINIMUM_PLAYER_LIGHT)
				{
					// never allow players to go totally black
					Multy = MINIMUM_PLAYER_LIGHT / gs;

					shadelight[0] *= Multy;
					shadelight[1] *= Multy;
					shadelight[2] *= Multy;
				}

				// store the modified version in the last good
				e->lastGoodShadelight[0] = shadelight[0];
				e->lastGoodShadelight[1] = shadelight[1];
				e->lastGoodShadelight[2] = shadelight[2];

				// hack to avoid a goto
				break;
			}
		}

		// 0 to 1 scale
		shadelight[0] /= LIGHT_DIVIDER;
		shadelight[1] /= LIGHT_DIVIDER;
		shadelight[2] /= LIGHT_DIVIDER;
	}

	// light interpolation
	lightLerpOffset = (e->angles[1] + e->angles[0]) * (SHADEDOT_QUANT / 360.0);

	ang_ceil = ceil (lightLerpOffset);
	ang_floor = floor (lightLerpOffset);

	lightLerpOffset = ang_ceil - lightLerpOffset;

	shadedots1 = r_avertexnormal_dots[(int) ang_ceil & (SHADEDOT_QUANT - 1)];
	shadedots2 = r_avertexnormal_dots[(int) ang_floor & (SHADEDOT_QUANT - 1)];

	an = e->angles[1] / 180 * M_PI;
	shadevector[0] = cos (-an);
	shadevector[1] = sin (-an);
	shadevector[2] = 1;
	VectorNormalize (shadevector);

	// locate the proper data
	paliashdr = (aliashdr_t *) Mod_Extradata (currententity->model);

	c_alias_polys += paliashdr->numtris;

	R_SetupAliasFrame (currententity, paliashdr);
}


void R_DrawSbarWidgetModel (model_t *wm, int skinnum)
{
	entity_t we;
	aliashdr_t *wheader;
	int pose;
	int numposes;
	float interval;
	drawvertx_t	*verts;
	int		*order;
	int		count;
	vec3_t	centers;
	float	zscale;
	vec3_t	mins = {9999999, 9999999, 9999999};
	vec3_t	maxs = {-9999999, -9999999, -9999999};
	vec3_t	vert;

	// set up a dummy entity
	memset (&we, 0, sizeof (entity_t));
	we.model = wm;
	we.skinnum = skinnum;

	// get the alias model header
	wheader = (aliashdr_t *) Mod_Extradata (wm);

	// get the correct skin
	glBindTexture (GL_TEXTURE_2D, wheader->glt[we.skinnum][0]->texnum);

	pose = wheader->frames[0].firstpose;
	numposes = wheader->frames[0].numposes;

	// this is pretty redundant cos pose will always += 0
	if (numposes > 1)
	{
		interval = wheader->frames[0].interval;
		pose += (int)(realtime / interval) % numposes;
	}

	verts = (drawvertx_t *)((byte *)wheader + wheader->posedata);
	verts += pose * wheader->poseverts;

	order = (int *)((byte *)wheader + wheader->commands);

	// get the z scale first
	while (1)
	{
		count = *order++;

		if (!count)
			break;

		if (count < 0)
			count = -count;

		do
		{
			if (verts->v[0] > maxs[0]) maxs[0] = verts->v[0];
			if (verts->v[0] < mins[0]) mins[0] = verts->v[0];

			if (verts->v[1] > maxs[1]) maxs[1] = verts->v[1];
			if (verts->v[1] < mins[1]) mins[1] = verts->v[1];

			if (verts->v[2] > maxs[2]) maxs[2] = verts->v[2];
			if (verts->v[2] < mins[2]) mins[2] = verts->v[2];

			order += 2;
			verts++;
		} while (--count);
	}

	// center on the model
	centers[0] = (mins[0] + maxs[0]) / 2;
	centers[1] = (mins[1] + maxs[1]) / 2;
	centers[2] = (mins[2] + maxs[2]) / 2;

	// set the scale to match the size of the regular quake armour models
	zscale = (maxs[2] - centers[2]) / 13.312031f;

	// reset verts and order for drawing
	verts = (drawvertx_t *)((byte *)wheader + wheader->posedata);
	verts += pose * wheader->poseverts;

	order = (int *)((byte *)wheader + wheader->commands);

	// now draw it
	glColor3f (1, 1, 1);

	while (1)
	{
		count = *order++;

		if (!count)
			break;

		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else glBegin (GL_TRIANGLE_STRIP);

		do
		{
			glTexCoord2fv ((float *) order);

			// scale and center the model
			vert[0] = (verts->v[0] - centers[0]) * zscale;
			vert[1] = (verts->v[1] - centers[1]) * zscale;
			vert[2] = (verts->v[2] - centers[2]) * zscale;

			glVertex3fv (vert);

			order += 2;
			verts++;
		} while (--count);

		glEnd ();
	}
}


