/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "quakedef.h"


/*
==================
R_AnimateLight

Added lightstyle interpolation.  Yummy!!!
==================
*/
void R_AnimateLight (void)
{
	int			j, k;
	float		l;
	int			flight;
	int			clight;
	float		lerpfrac;
	float		backlerp;

	// light animations
	// 'm' is normal light, 'a' is no light, 'z' is double bright
	flight = (int) floor (cl.time * 10);
	clight = (int) ceil (cl.time * 10);
	lerpfrac = (cl.time * 10) - flight;
	backlerp = 1.0f - lerpfrac;

	for (j = 0; j < MAX_LIGHTSTYLES; j++)
	{
		if (!cl_lightstyle[j].length)
		{
			// was 256, changed to 264 for consistency
			d_lightstylevalue[j] = 264;
			continue;
		}
		else if (cl_lightstyle[j].length == 1)
		{
			// single length style so don't bother interpolating
			d_lightstylevalue[j] = 22 * (cl_lightstyle[j].map[0] - 'a');
			continue;
		}

		// interpolate animating light
		// frame just gone
		k = flight % cl_lightstyle[j].length;
		k = cl_lightstyle[j].map[k] - 'a';
		l = (float) (k * 22) * backlerp;

		// upcoming frame
		k = clight % cl_lightstyle[j].length;
		k = cl_lightstyle[j].map[k] - 'a';
		l += (float) (k * 22) * lerpfrac;

		d_lightstylevalue[j] = (int) l;
	}
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

void R_InitDLights (void)
{
	// used to contain a bitmask array for dlights.  this is all that's needed now.
	// make sure it fits in a byte array
	if (MAX_DLIGHTS % 8) Sys_Error ("MAX_DLIGHTS %% 8");
}


/*
=============
R_MarkLights
=============
*/
void R_MarkLights (dlight_t *light, int lnum, mnode_t *node)
{
	mplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;
	int			i;
	int			j;

	if (node->contents < 0)
		return;

	splitplane = node->plane;
	dist = DotProduct (light->origin, splitplane->normal) - splitplane->dist;
	
	if (dist > light->radius)
	{
		R_MarkLights (light, lnum, node->children[0]);
		return;
	}

	if (dist < -light->radius)
	{
		R_MarkLights (light, lnum, node->children[1]);
		return;
	}

	// mark the polygons
	surf = cl.worldmodel->surfaces + node->firstsurface;

	for (i=0 ; i<node->numsurfaces ; i++, surf++)
	{
		if (surf->dlightframe != r_framecount)
		{
			// init a new dlight on the surf
			for (j = 0; j < MAX_DLIGHTS; j += 8)
				surf->dlightbits[(j >> 3)] = 0;

			surf->dlightframe = r_framecount;
		}

		// mark it
		surf->dlightbits[(lnum >> 3)] |= (1 << (lnum & 7));
	}

	R_MarkLights (light, lnum, node->children[0]);
	R_MarkLights (light, lnum, node->children[1]);
}


/*
=============
R_PushDlights
=============
*/
void R_PushDlights (mnode_t *headnode)
{
	int		i;
	dlight_t	*l;

	l = cl_dlights;

	for (i = 0; i < MAX_DLIGHTS; i++, l++)
	{
		// MH - if (l->radius <= 0) l->die is set to -1 in CL_DecayLights
		if (l->die < cl.time)
			continue;

		R_MarkLights (l, i, headnode);
	}
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;
static vec3_t	pointcolor;


int RecursiveLightPoint (vec3_t color, mnode_t *node, vec3_t start, vec3_t end)
{
	float		front, back, frac;
	vec3_t		mid;

loc0:;
	// hit nothing
	if (node->contents < 0) return false;

	// calculate mid point
	if (node->plane->type < 3)
	{
		front = start[node->plane->type] - node->plane->dist;
		back = end[node->plane->type] - node->plane->dist;
	}
	else
	{
		front = DotProduct (start, node->plane->normal) - node->plane->dist;
		back = DotProduct (end, node->plane->normal) - node->plane->dist;
	}

	// LordHavoc: optimized recursion
	if ((back < 0) == (front < 0))
	{
		node = node->children[front < 0];
		goto loc0;
	}

	frac = front / (front-back);
	mid[0] = start[0] + (end[0] - start[0]) * frac;
	mid[1] = start[1] + (end[1] - start[1]) * frac;
	mid[2] = start[2] + (end[2] - start[2]) * frac;

	// go down front side
	if (RecursiveLightPoint (color, node->children[front < 0], start, mid))
	{
		// hit something
		return true;
	}
	else
	{
		int i, ds, dt;
		msurface_t *surf;

		// check for impact on this node
		VectorCopy (mid, lightspot);
		lightplane = node->plane;

		surf = cl.worldmodel->surfaces + node->firstsurface;

		for (i = 0; i < node->numsurfaces; i++, surf++)
		{
			// no lightmaps on these
			if (surf->flags & SURF_DRAWTURB) continue;
			if (surf->flags & SURF_DRAWSKY) continue;

			ds = (int) ((float) DotProduct (mid, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]);
			dt = (int) ((float) DotProduct (mid, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]);

			// out of range
			if (ds < surf->texturemins[0] || dt < surf->texturemins[1]) continue;

			ds -= surf->texturemins[0];
			dt -= surf->texturemins[1];

			// out of range
			if (ds > surf->extents[0] || dt > surf->extents[1]) continue;

			if (surf->samples)
			{
				// LordHavoc enhanced to interpolate lighting
				byte *lightmap;
				int maps, line3, dsfrac = ds & 15, dtfrac = dt & 15, r00 = 0, g00 = 0, b00 = 0, r01 = 0, g01 = 0, b01 = 0, r10 = 0, g10 = 0, b10 = 0, r11 = 0, g11 = 0, b11 = 0;
				float scale;
				line3 = surf->smax * 3;

				lightmap = surf->samples + ((dt >> 4) * surf->smax + (ds >> 4)) * 3;

				for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
				{
					scale = (float) d_lightstylevalue[surf->styles[maps]] * ONEOVER255;

					r00 += (float) lightmap[0] * scale;
					g00 += (float) lightmap[1] * scale;
					b00 += (float) lightmap[2] * scale;

					r01 += (float) lightmap[3] * scale;
					g01 += (float) lightmap[4] * scale;
					b01 += (float) lightmap[5] * scale;

					r10 += (float) lightmap[line3 + 0] * scale;
					g10 += (float) lightmap[line3 + 1] * scale;
					b10 += (float) lightmap[line3 + 2] * scale;

					r11 += (float) lightmap[line3 + 3] * scale;
					g11 += (float) lightmap[line3 + 4] * scale;
					b11 += (float) lightmap[line3 + 5] * scale;

					lightmap += surf->smax * surf->tmax * 3;
				}

				// i'm pretty sure that LH added a lot of these extra brackets for no reason other than to obfuscate his code.
				// i've left them in anyway, but spaced things out a bit better
				color[0] += (float) ((int) ((((((((r11 - r10) * dsfrac) >> 4) + r10) - 
					((((r01 - r00) * dsfrac) >> 4) + r00)) * dtfrac) >> 4) + ((((r01 - r00) * dsfrac) >> 4) + r00)));

				color[1] += (float) ((int) ((((((((g11 - g10) * dsfrac) >> 4) + g10) - 
					((((g01 - g00) * dsfrac) >> 4) + g00)) * dtfrac) >> 4) + ((((g01 - g00) * dsfrac) >> 4) + g00)));

				color[2] += (float) ((int) ((((((((b11 - b10) * dsfrac) >> 4) + b10) - 
					((((b01 - b00) * dsfrac) >> 4) + b00)) * dtfrac) >> 4) + ((((b01 - b00) * dsfrac) >> 4) + b00)));
			}

			// success
			return true;
		}

		// go down back side
		return RecursiveLightPoint (color, node->children[front >= 0], mid, end);
	}
}

/*
=============
R_LightPoint

=============
*/
void R_LightPoint (vec3_t p, vec3_t color)
{
	vec3_t		end;
	int			lnum;
	vec3_t		dist;
	float		add;

	if (!cl.worldmodel->lightdata)
	{
		// no light data
		pointcolor[0] = pointcolor[1] = pointcolor[2] = 255;
		return;
	}

	// end point increased to full map scale
	end[0] = p[0];
	end[1] = p[1];
	end[2] = p[2] - 8192;

	// initial colour
	pointcolor[0] = pointcolor[1] = pointcolor[2] = 0;

	// get it
	RecursiveLightPoint (pointcolor, cl.worldmodel->nodes, p, end);

	// copy across
	Vector3Copy (color, pointcolor);

	// add dlights
	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (p, cl_dlights[lnum].origin, dist);
			add = cl_dlights[lnum].radius - Length (dist);

			if (add > 0)
			{
				// divide by 2 to match the new colour scale
				color[0] += add / 2.0f;
				color[1] += add / 2.0f;
				color[2] += add / 2.0f;
			}
		}
	}
}



/*
========================================================================================================================

			LIGHT CODE RELOCATED FROM GL_RSURF.C

========================================================================================================================
*/

// rgba lightmaps
#define LIGHTMAP_BYTES	4

// RGB
int blocklights[18 * 18 * 3];

#define	MAX_LIGHTMAPS	256

typedef struct glRect_s
{
	unsigned char l,t,w,h;
} glRect_t;


// struct to hold all lightmap related stuff
typedef struct lightmapinfo_s
{
	qboolean modified;
	glRect_t rectchange;

	int allocated[LM_BLOCK_WIDTH];

	byte *data;

	int texnum;
} lightmapinfo_t;

// all of our lightmaps
lightmapinfo_t lightmaps[MAX_LIGHTMAPS];


/*
===============
R_AddDynamicLights

===============
*/
void R_AddDynamicLights (msurface_t *surf)
{
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	mtexinfo_t	*tex;
	dlight_t	*dl;
	unsigned int *bl;

	tex = surf->texinfo;

	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		// get the dlight
		dl = &cl_dlights[lnum];

		// dead light
		// id forget this...!!!
		if (dl->die < cl.time) continue;

		// not marked for checking by this light
		if (!(surf->dlightbits[(lnum >> 3)] & (1 << (lnum & 7)))) continue;

		rad = dl->radius;
		dist = DotProduct (dl->origin, surf->plane->normal) - surf->plane->dist;
		rad -= fabs (dist);
		minlight = dl->minlight;

		if (rad < minlight) continue;

		minlight = rad - minlight;

		for (i = 0; i < 3; i++)
			impact[i] = dl->origin[i] - surf->plane->normal[i] * dist;

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];

		bl = blocklights;

		for (t = 0; t < surf->tmax; t++)
		{
			td = local[1] - t * 16;

			if (td < 0) td = -td;

			for (s = 0; s < surf->smax; s++, bl += 3)
			{
				sd = local[0] - s * 16;

				if (sd < 0) sd = -sd;

				if (sd > td)
					dist = sd + (td >> 1);
				else
					dist = td + (sd >> 1);

				if (dist < minlight)
				{
					// accumulate light
					bl[0] += (rad - dist) * dl->Colour[0];
					bl[1] += (rad - dist) * dl->Colour[1];
					bl[2] += (rad - dist) * dl->Colour[2];

					// we know it's dynamic now
					surf->cached_dlight = true;
				}
			}
		}
	}
}


/*
================
R_InitBLTable and blTable

Pre-computed blocklights tables to prevent excessive run-time calculations.
================
*/
// blocklights can go to about 230000 in practice, with a theoretical max of just over 550000, before
// we even factor in dynamic lights.
// ideally it should be a short... if only C data types clamped instead of wrapping...
#define MAX_BLOCKLIGHTS		65536

byte blTable[MAX_BLOCKLIGHTS];

void R_InitBLTable (void)
{
	int i;
	int bl;

	for (i = 0; i < MAX_BLOCKLIGHTS; i++)
	{
		bl = (i >> 8);

		if (bl > 255) bl = 255;

		blTable[i] = (byte) bl;
	}
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
qboolean R_BuildLightMap (msurface_t *surf, byte *dest, int stride, qboolean fullRebuild)
{
	int			t;
	int			i, j, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	int			lightadj[4];
	int			*bl;

	size = surf->smax * surf->tmax;
	lightmap = surf->samples;

	// set to full bright if no light data
	if (!cl.worldmodel->lightdata)
	{
		for (bl = blocklights, i = 0; i < size; i++, bl += 3)
		{
			bl[0] = 255 * 256;
			bl[1] = 255 * 256;
			bl[2] = 255 * 256;
		}

		goto store;
	}

	// clear to no light
	for (bl = blocklights, i = 0; i < size; i++, bl += 3)
	{
		bl[0] = 0;
		bl[1] = 0;
		bl[2] = 0;
	}

	// add dynamic lights first
	if (surf->dlightframe == r_framecount) R_AddDynamicLights (surf);

	// if it wasn't hit by a dlight and we don't need a full rebuild, return that
	// we don't do a rebuild
	if (!surf->cached_dlight && !fullRebuild) return false;

	// add all the lightmaps
	if (lightmap)
	{
		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
		{
			scale = d_lightstylevalue[surf->styles[maps]];
			surf->cached_light[maps] = scale;	// 8.8 fraction

			for (bl = blocklights, i = 0, j = 0; i < size; i++, bl += 3)
			{
				bl[0] += lightmap[j++] * scale;
				bl[1] += lightmap[j++] * scale;
				bl[2] += lightmap[j++] * scale;
			}

			// skip to next lightmap
			lightmap += size * 3;
		}
	}

store:;
	// bound, invert, and shift
	stride -= (surf->smax << 2);

	for (bl = blocklights, i = 0; i < surf->tmax; i++, dest += stride)
	{
		for (j = 0; j < surf->smax; j++, bl += 3, dest += 4)
		{
			// swap to BGR, allow negative so that we can do negative dlights
			dest[0] = (bl[2] < 0) ? 0 : (bl[2] >= MAX_BLOCKLIGHTS ? 255 : blTable[bl[2]]);
			dest[1] = (bl[1] < 0) ? 0 : (bl[1] >= MAX_BLOCKLIGHTS ? 255 : blTable[bl[1]]);
			dest[2] = (bl[0] < 0) ? 0 : (bl[0] >= MAX_BLOCKLIGHTS ? 255 : blTable[bl[0]]);

			// always full alpha
			dest[3] = 255;
		}
	}

	// surface lightmap was modified
	return true;
}


/*
================
R_CheckLightmapModification

Checks for modification on all lightmaps
================
*/
void R_CheckLightmapModification (msurface_t *surf)
{
	byte		*base;
	int			maps;
	glRect_t    *theRect;
	qboolean	fullRebuild;

	if (surf->flags & SURF_DRAWTURB) return;
	if (surf->flags & SURF_DRAWSKY) return;

	// check for lightmap modification
	for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
	{
		if (d_lightstylevalue[surf->styles[maps]] != surf->cached_light[maps])
		{
			// no matter what happens in the dynamics, we want to rebuild this lightmap
			fullRebuild = true;
			goto dynamic;
		}
	}

	// dynamic this frame || dynamic previously
	if (surf->dlightframe == r_framecount || surf->cached_dlight)
	{
		// it can only get to here if it doesn't also require a light style modification
		fullRebuild = false;

dynamic:;
		// test a rebuild of the lightmap before touching any of the modification data
		// get out if we didn't need to rebuild the lightmap
		if (!R_BuildLightMap (surf, surf->lightmapdata, LM_BLOCK_WIDTH * LIGHTMAP_BYTES, fullRebuild))
			return;

		// mark as modified
		lightmaps[surf->lightmaptexturenum].modified = true;

		// get modified rectangle
		theRect = &lightmaps[surf->lightmaptexturenum].rectchange;

		if (surf->light_t < theRect->t)
		{
			if (theRect->h)
				theRect->h += theRect->t - surf->light_t;
			theRect->t = surf->light_t;
		}

		if (surf->light_s < theRect->l)
		{
			if (theRect->w)
				theRect->w += theRect->l - surf->light_s;
			theRect->l = surf->light_s;
		}

		if ((theRect->w + theRect->l) < (surf->light_s + surf->smax)) theRect->w = (surf->light_s - theRect->l) + surf->smax;
		if ((theRect->h + theRect->t) < (surf->light_t + surf->tmax)) theRect->h = (surf->light_t - theRect->t) + surf->tmax;
	}
}


/*
================
R_ModifyLightmaps

Uploads all lightmaps that were modified
================
*/
void R_ModifyLightmaps (void)
{
	int			i;
	glRect_t	*theRect;

	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		if (lightmaps[i].modified)
		{
			lightmaps[i].modified = false;

			theRect = &lightmaps[i].rectchange;

			glBindTexture (GL_TEXTURE_2D, lightmaps[i].texnum);

			glTexSubImage2D (GL_TEXTURE_2D, 
				0, 
				0, 
				theRect->t, 
				LM_BLOCK_WIDTH, 
				theRect->h, 
				GL_BGRA, 
				GL_UNSIGNED_BYTE,
				lightmaps[i].data + (theRect->t * LM_BLOCK_WIDTH * LIGHTMAP_BYTES));

			theRect->l = LM_BLOCK_WIDTH;
			theRect->t = LM_BLOCK_HEIGHT;
			theRect->h = 0;
			theRect->w = 0;
		}
	}
}


/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

// returns a texture number and the position inside it
int AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		bestx;
	int		texnum;

	for (texnum = 0; texnum < MAX_LIGHTMAPS; texnum++)
	{
		best = LM_BLOCK_HEIGHT;

		for (i = 0; i < LM_BLOCK_WIDTH - w; i++)
		{
			best2 = 0;

			for (j = 0; j < w; j++)
			{
				if (lightmaps[texnum].allocated[i + j] >= best) break;
				if (lightmaps[texnum].allocated[i + j] > best2) best2 = lightmaps[texnum].allocated[i + j];
			}

			if (j == w)
			{
				// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > LM_BLOCK_HEIGHT) continue;

		for (i = 0; i < w; i++)
			lightmaps[texnum].allocated[*x + i] = best + h;

		return texnum;
	}

	// game will crash here, so...
	Sys_Error ("AllocBlock: full");

	// ...shut up the compiler
	return 0;
}


/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	// should never happen
	if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) return;

	// precalc and store smax and tmax
	surf->smax = (surf->extents[0] >> 4) + 1;
	surf->tmax = (surf->extents[1] >> 4) + 1;

	// allocate a block and get the position in it
	surf->lightmaptexturenum = AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t);
	surf->gllmtexnum = lightmaps[surf->lightmaptexturenum].texnum;

	// allocate memory for this lightmap if it's currently empty
	if (!lightmaps[surf->lightmaptexturenum].data)
		lightmaps[surf->lightmaptexturenum].data = malloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * LIGHTMAP_BYTES);

	// set the lightmap pointer for this surf
	surf->lightmapdata = lightmaps[surf->lightmaptexturenum].data;
	surf->lightmapdata += (surf->light_t * LM_BLOCK_WIDTH + surf->light_s) * LIGHTMAP_BYTES;

	// build the lightmap.  this always does a full rebuild
	R_BuildLightMap (surf, surf->lightmapdata, LM_BLOCK_WIDTH * LIGHTMAP_BYTES, true);
}


/*
==================
GL_BuildLightmaps

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
void GL_BuildLightmaps (void)
{
	int		i;
	texture_t *tx;
	msurface_t *surf;

	// no dlight cache
	r_framecount = 1;

	// clear out any dlights
	for (i = 0; i < MAX_DLIGHTS; i++)
	{
		cl_dlights[i].die = -1;
		cl_dlights[i].radius = 0;
		cl_dlights[i].key = -1;
	}

	// clear down lightmap data
	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		// clear block allocation
		memset (lightmaps[i].allocated, 0, sizeof (lightmaps[i].allocated));

		// free any allocated data
		if (lightmaps[i].data) free (lightmaps[i].data);

		// set the data to NULL so we can validly test if it's a new lightmap
		lightmaps[i].data = NULL;

		// unmodified and no polys
		lightmaps[i].modified = false;

		// initial changed rectangle
		lightmaps[i].rectchange.l = LM_BLOCK_WIDTH;
		lightmaps[i].rectchange.t = LM_BLOCK_HEIGHT;
		lightmaps[i].rectchange.w = 0;
		lightmaps[i].rectchange.h = 0;
	}

	// only do the worldmodel cos we don't do lightmaps on the bsp bmodels, and * models are inline with the world
	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
	{
		// don't built lightmaps for these surface types
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		// sort surfaces by texture so that surfaces with the same texture will share lightmaps,
		// thus further optimising the world render
		surf->texturechain = surf->texinfo->texture->texturechain;
		surf->texinfo->texture->texturechain = surf;
	}

	// ditto on the worldmodel
	for (i = 0; i < cl.worldmodel->numtextures; i++)
	{
		// get the texture
		// can't do tx++ cos texture_t is a variably sized struct
		tx = cl.worldmodel->textures[i];

		// bad texture
		if (!tx) continue;

		// get the chain
		surf = tx->texturechain;

		// no surfs in the chain
		if (!surf) continue;

		// create the lightmaps for each surf in the chain
		for (; surf; surf = surf->texturechain)
			GL_CreateSurfaceLightmap (surf);

		// no more
		tx->texturechain = NULL;
	}

	// upload all lightmaps that were filled
	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		// no more used
		if (!lightmaps[i].allocated[0]) break;

		// upload as a texture
		glBindTexture (GL_TEXTURE_2D, lightmaps[i].texnum);

		// no mipmaps
		glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE);

		// set the data
		glTexImage2D (GL_TEXTURE_2D, 
			0, 
			GL_RGBA, 
			LM_BLOCK_WIDTH, 
			LM_BLOCK_HEIGHT, 
			0, 
			GL_BGRA, 
			GL_UNSIGNED_BYTE, 
			lightmaps[i].data);

		// linear filtering
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
}


void R_InitLight (void)
{
	int i;
	float j;

	// base properties on startup
	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		// set the initial lightmap info data to NULL
		lightmaps[i].data = NULL;

		GL_AllocTexture (&lightmaps[i].texnum);
	}

	// init the lighting table
	R_InitBLTable ();

	// dynamic lights
	R_InitDLights ();
}
