/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_mesh.c: triangle model functions

#include "quakedef.h"


typedef struct aliasvert_s
{
	float v[3];
	float st[2];
	float rgba[4];
} aliasvert_t;

aliasvert_t *aliasverts = NULL;
int maxaliasverts;

extern gltexture_t *r_blacktexture;

/*
=================================================================

ALIAS MODEL DISPLAY LIST GENERATION

=================================================================
*/

model_t		*aliasmodel;
aliashdr_t	*paliashdr;

qboolean	used[8192];

// the command list holds counts and s/t values that are valid for
// every frame
int		commands[8192];
int		numcommands;

// all frames will have their vertexes rearranged and expanded
// so they are in the order expected by the command list
int		vertexorder[8192];
int		numorder;

int		allverts, alltris;

int		stripverts[128];
int		striptris[128];
int		stripcount;

/*
================
StripLength
================
*/
int	StripLength (int starttri, int startv)
{
	int			m1, m2;
	int			j;
	mtriangle_t	*last, *check;
	int			k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+2)%3];
	m2 = last->vertindex[(startv+1)%3];

	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront)
			continue;
		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1)
				continue;
			if (check->vertindex[ (k+1)%3 ] != m2)
				continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j])
				goto done;

			// the new edge
			if (stripcount & 1)
				m2 = check->vertindex[ (k+2)%3 ];
			else
				m1 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = check->vertindex[ (k+2)%3 ];
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2)
			used[j] = 0;

	return stripcount;
}

/*
===========
FanLength
===========
*/
int	FanLength (int starttri, int startv)
{
	int		m1, m2;
	int		j;
	mtriangle_t	*last, *check;
	int		k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+0)%3];
	m2 = last->vertindex[(startv+2)%3];


	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront)
			continue;
		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1)
				continue;
			if (check->vertindex[ (k+1)%3 ] != m2)
				continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j])
				goto done;

			// the new edge
			m2 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = m2;
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2)
			used[j] = 0;

	return stripcount;
}


/*
================
BuildTris

Generate a list of trifans or strips
for the model, which holds for all frames
================
*/
void BuildTris (void)
{
	int		i, j, k;
	int		startv;
	float	s, t;
	int		len, bestlen, besttype = 0;
	int		bestverts[1024];
	int		besttris[1024];
	int		type;

	//
	// build tristrips
	//
	numorder = 0;
	numcommands = 0;
	memset (used, 0, sizeof(used));
	for (i=0 ; i<pheader->numtris ; i++)
	{
		// pick an unused triangle and start the trifan
		if (used[i])
			continue;

		bestlen = 0;
		for (type = 0 ; type < 2 ; type++)
//	type = 1;
		{
			for (startv =0 ; startv < 3 ; startv++)
			{
				if (type == 1)
					len = StripLength (i, startv);
				else
					len = FanLength (i, startv);
				if (len > bestlen)
				{
					besttype = type;
					bestlen = len;
					for (j=0 ; j<bestlen+2 ; j++)
						bestverts[j] = stripverts[j];
					for (j=0 ; j<bestlen ; j++)
						besttris[j] = striptris[j];
				}
			}
		}

		// mark the tris on the best strip as used
		for (j=0 ; j<bestlen ; j++)
			used[besttris[j]] = 1;

		if (besttype == 1)
			commands[numcommands++] = (bestlen+2);
		else
			commands[numcommands++] = -(bestlen+2);

		for (j=0 ; j<bestlen+2 ; j++)
		{
			// emit a vertex into the reorder buffer
			k = bestverts[j];
			vertexorder[numorder++] = k;

			// emit s/t coords into the commands stream
			s = stverts[k].s;
			t = stverts[k].t;
			if (!triangles[besttris[0]].facesfront && stverts[k].onseam)
				s += pheader->skinwidth / 2;	// on back side
			s = (s + 0.5) / pheader->skinwidth;
			t = (t + 0.5) / pheader->skinheight;

			*(float *)&commands[numcommands++] = s;
			*(float *)&commands[numcommands++] = t;
		}
	}

	commands[numcommands++] = 0;		// end of list marker

	Con_DPrintf ("%3i tri %3i vert %3i cmd\n", pheader->numtris, numorder, numcommands);

	allverts += numorder;
	alltris += pheader->numtris;
}


/*
================
GL_MakeAliasModelDisplayLists
================
*/
void GL_MakeAliasModelDisplayLists (model_t *m, aliashdr_t *hdr)
{
	int		i, j;
	int			*cmds;
	drawvertx_t	*verts;
	int count;
	int		*order;

	aliasmodel = m;
	paliashdr = hdr;

	// build it from scratch
	BuildTris ();

	// save the data out
	paliashdr->poseverts = numorder;

	cmds = Hunk_Alloc (numcommands * 4);
	paliashdr->commands = (byte *) cmds - (byte *) paliashdr;
	memcpy (cmds, commands, numcommands * 4);

	verts = Hunk_Alloc (paliashdr->numposes * paliashdr->poseverts * sizeof (drawvertx_t));
	paliashdr->posedata = (byte *) verts - (byte *) paliashdr;

	for (i = 0; i < paliashdr->numposes; i++)
	{
		for (j = 0; j < numorder; j++)
		{
			// prescale the verts at load time.  this requires moving over to floats for storing verts, which
			// gives a * 4 memory overhead but cuts down on a LOT of run-time calculations
			// i've tried converting back to bytes but STRANGE things happen...
			verts->v[0] = ((float) poseverts[i][vertexorder[j]].v[0]) * paliashdr->scale[0] + paliashdr->scale_origin[0];
			verts->v[1] = ((float) poseverts[i][vertexorder[j]].v[1]) * paliashdr->scale[1] + paliashdr->scale_origin[1];
			verts->v[2] = ((float) poseverts[i][vertexorder[j]].v[2]) * paliashdr->scale[2] + paliashdr->scale_origin[2];

			verts->lightnormalindex = poseverts[i][vertexorder[j]].lightnormalindex;

			verts++;
		}
	}


	// here's a nasty hack for you!
	// code for elimination of muzzleflashes on viewmodels.
	// is it a view model???
	if ((m->name[6] == 'v' || m->name[6] == 'V') && m->name[7] == '_' && hdr->numframes > 2)
	{
		// get pointers to the verts
		drawvertx_t *vertsf0 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);
		drawvertx_t *vertsf1 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);

		// setthese pointers to the 0th and 1st frames
		vertsf0 += hdr->frames[0].firstpose * hdr->poseverts;
		vertsf1 += hdr->frames[1].firstpose * hdr->poseverts;

		// now go through them and compare.  we expect that (a) the animation is sensible and there's no major
		// difference between the 2 frames to be expected, and (b) any verts that do exhibit a major difference
		// can be assumed to belong to the muzzleflash
		for (j = 0; j < numorder; j++)
		{
			// get difference in front to back movement
			float vdiff = vertsf1->v[0] - vertsf0->v[0];

			// if it's above a certain treshold, assume a muzzleflash and mark for nodraw
			// 10 is the approx lowest range of visible front to back in a view model, so that seems reasonable to work with
			if (vdiff > 10)
				vertsf0->nodraw = true;
			else vertsf0->nodraw = false;

			// next set of verts
			vertsf0++;
			vertsf1++;
		}
	}

	// now we mark every other vert in the model as per the instructions in the first frame
	for (i = 1; i < paliashdr->numframes; i++)
	{
		// get pointers to the verts
		drawvertx_t *vertsf0 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);
		drawvertx_t *vertsfi = (drawvertx_t *) ((byte *) hdr + hdr->posedata);

		// set these pointers to the 0th and i'th frames
		vertsf0 += hdr->frames[0].firstpose * hdr->poseverts;
		vertsfi += hdr->frames[i].firstpose * hdr->poseverts;

		for (j = 0; j < numorder; j++)
		{
			// just copy it across
			vertsfi->nodraw = vertsf0->nodraw;

			if (vertsf0->nodraw)
			{
				vertsfi->v[0] = vertsf0->v[0];
				vertsfi->v[1] = vertsf0->v[1];
				vertsfi->v[2] = vertsf0->v[2];
			}

			// next set of verts
			vertsf0++;
			vertsfi++;
		}
	}

	// now we take a walk through the commands and see how many verts we need
	for (order = commands, j = 0; ;)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		// set primitive type
		if (count < 0) count = -count;

		// store in count
		if (count > j) j = count;

		do
		{
			order += 2;
		} while (--count);
	}

	// allocate vertexes
	if (!aliasverts)
	{
		// take at least 9 to begin with and grow as required
		maxaliasverts = j > 9 ? j : 9;
		aliasverts = (aliasvert_t *) malloc (sizeof (aliasvert_t) * maxaliasverts);
	}
	else if (j > maxaliasverts)
	{
		// go to the new max
		maxaliasverts = j;

		// reallocate
		free (aliasverts);
		aliasverts = (aliasvert_t *) malloc (sizeof (aliasvert_t) * maxaliasverts);
	}
}



/*
=============================================================

  ALIAS MODELS RENDER

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
};

vec3_t	shadevector;
vec3_t	shadelight;

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16

float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
;

float	*shadedots1 = r_avertexnormal_dots[0];
float	*shadedots2 = r_avertexnormal_dots[0];
float	lightLerpOffset;

int	lastposenum;


/*
=============
GL_DrawAliasFrame
=============
*/
void GL_DrawAliasFrame (aliashdr_t *paliashdr, int pose1, int pose2, float backlerp, float alpha)
{
	float l;
	float ldiff;
	float lquant[2];

	drawvertx_t	*verts1;
	drawvertx_t	*verts2;

	int		*order;
	int		count;

	float frontlerp = 1.0f - backlerp;

	int i;
	GLenum PRIMITIVE_TYPE;

	lastposenum = pose1;

	verts1 = (drawvertx_t *) ((byte *) paliashdr + paliashdr->posedata);
	verts2 = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	order = (int *) ((byte *) paliashdr + paliashdr->commands);

	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		if (count < 0)
		{
			count = -count;
			PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		}
		else PRIMITIVE_TYPE = GL_TRIANGLE_STRIP;

		// needed for aliasverts indexing
		i = 0;

		do
		{
			// texture coordinates come from the draw list
			aliasverts[i].st[0] = ((float *) order)[0];
			aliasverts[i].st[1] = ((float *) order)[1];

			order += 2;

			// set up the lighting
			lquant[0] = (shadedots1[verts1->lightnormalindex] + (backlerp * (shadedots1[verts2->lightnormalindex] - shadedots1[verts1->lightnormalindex])));
			lquant[1] = (shadedots2[verts1->lightnormalindex] + (backlerp * (shadedots2[verts2->lightnormalindex] - shadedots2[verts1->lightnormalindex])));

			if (lquant[0] != lquant[1])
			{
				if (lquant[0] > lquant[1])
				{
					ldiff = lquant[0] - lquant[1];
					ldiff *= lightLerpOffset;
					l = lquant[0] - ldiff;
				}
				else
				{
					ldiff = lquant[1] - lquant[0];
					ldiff *= lightLerpOffset;
					l = lquant[0] + ldiff;
				}
			}
			else l = lquant[0];

			aliasverts[i].rgba[0] = l * shadelight[0];
			aliasverts[i].rgba[1] = l * shadelight[1];
			aliasverts[i].rgba[2] = l * shadelight[2];
			aliasverts[i].rgba[3] = alpha;

			aliasverts[i].v[0] = verts1->v[0] * frontlerp + verts2->v[0] * backlerp;
			aliasverts[i].v[1] = verts1->v[1] * frontlerp + verts2->v[1] * backlerp;
			aliasverts[i].v[2] = verts1->v[2] * frontlerp + verts2->v[2] * backlerp;

			verts1++;
			verts2++;
			i++;
		} while (--count);

		glDrawArrays (PRIMITIVE_TYPE, 0, i);
	}
}


/*
=============
GL_DrawAliasShadow
=============
*/
extern	vec3_t			lightspot;

void GL_DrawAliasShadow (aliashdr_t *paliashdr, int posenum)
{
}



/*
=================
R_SetupAliasFrame

=================
*/
void R_SetupAliasFrame (entity_t *e, aliashdr_t *paliashdr)
{
	int				pose, numposes;
	float			interval;
	qboolean		lerpmdl = true;
	float			blend;

	// silently revert to frame 0
	if ((e->frame >= paliashdr->numframes) || (e->frame < 0)) e->frame = 0;

	pose = paliashdr->frames[e->frame].firstpose;
	numposes = paliashdr->frames[e->frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[e->frame].interval;
		pose += (int)(cl.time / interval) % numposes;

		// not really needed for non-interpolated mdls, but does no harm
		e->frame_interval = interval;
	}
	else e->frame_interval = 0.1;

	// conditions for turning lerping off
	if (e->pose2 != pose) lerpmdl = false;					// don't interpolate between different poses
	if (paliashdr->numposes == 1) lerpmdl = false;			// only one pose
	if (e == &cl.viewent && e->frame == 0) lerpmdl = false;	// super-nailgun spinning down bug
	if (r_framecount < 10) lerpmdl = false;					// reliving their dying throes

	// wait till now to do this so we can correctly switch interpolation off
	// (viewent can switch mdls but keep the same pose, and therefore have invalid data cached in the poses)
	if (!lerpmdl)
	{
		e->frame_start_time = realtime;
		e->pose1 = e->pose2;
		e->pose2 = pose;
		blend = 0;
	}
	else blend = (realtime - e->frame_start_time) / e->frame_interval;

	// don't let blend pass 1
	if (cl.paused || blend > 1.0) blend = 1.0;

	if (e->model->flags & EF_TRANS66)
	{
		glEnable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	GL_DrawAliasFrame (paliashdr, e->pose1, e->pose2, blend, (e->model->flags & EF_TRANS66) ? 0.80 : 1.0);

	if (e->model->flags & EF_TRANS66)
	{
		glDisable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}



/*
=================
R_DrawAliasModel

=================
*/
void R_DrawAliasModel (entity_t *e)
{
	int			i;
	model_t		*clmodel;
	vec3_t		mins, maxs;
	aliashdr_t	*paliashdr;
	float		an;
	int			anim;
	float		ang_ceil, ang_floor;

	clmodel = currententity->model;

	VectorAdd (currententity->origin, clmodel->mins, mins);
	VectorAdd (currententity->origin, clmodel->maxs, maxs);

	// never cull the view model
	if (R_CullBox (mins, maxs) && e != &cl.viewent)
		return;

	VectorCopy (currententity->origin, r_entorigin);
	VectorSubtract (r_origin, r_entorigin, modelorg);

	// get lighting information
	if (currententity->model->flags & EF_ROTATE)
	{
		shadelight[0] = shadelight[1] = shadelight[2] = 0.333f;
	}
	else
	{
		currententity->origin[2] += 8.0f;
		R_LightPoint (currententity->origin, shadelight);
		currententity->origin[2] -= 8.0f;
	}

	// light interpolation
	lightLerpOffset = (e->angles[1] + e->angles[0]) * (SHADEDOT_QUANT / 360.0);

	ang_ceil = ceil (lightLerpOffset);
	ang_floor = floor (lightLerpOffset);

	lightLerpOffset = ang_ceil - lightLerpOffset;

	shadedots1 = r_avertexnormal_dots[(int) ang_ceil & (SHADEDOT_QUANT - 1)];
	shadedots2 = r_avertexnormal_dots[(int) ang_floor & (SHADEDOT_QUANT - 1)];

	an = e->angles[1] / 180 * M_PI;
	shadevector[0] = cos (-an);
	shadevector[1] = sin (-an);
	shadevector[2] = 1;
	VectorNormalize (shadevector);

	// locate the proper data
	paliashdr = (aliashdr_t *) Mod_Extradata (currententity->model);

	// draw all the triangles
    glPushMatrix ();
	R_RotateForEntity (e);

	anim = (int) (cl.time * 10) & 3;

	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (currententity->colormap != vid.colormap && !gl_nocolors.value)
	{
		i = currententity - cl_entities;

		if (i >= 1 && i<=cl.maxclients)
		{
			glActiveTexture (GL_TEXTURE0);
			glBindTexture (GL_TEXTURE_2D, playertextures[-1 + i]);

			glActiveTexture (GL_TEXTURE1);
			glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);

			goto FrameSetup;
		}
	}

	glActiveTexture (GL_TEXTURE0);
	glBindTexture (GL_TEXTURE_2D, paliashdr->gltexture[currententity->skinnum][anim]->texnum);

	// alias models have their luma automatically set to black if there is none, so we can just bind away.
	// this is based on the assumption that a tmu switch and another bind is less expensive than a tmu switch,
	// taking it down, and switching back.  that might not be true on all implementations...
	glActiveTexture (GL_TEXTURE1);
	glBindTexture (GL_TEXTURE_2D, paliashdr->gltexture[currententity->skinnum][anim]->lumatexnum);

FrameSetup:;
	R_SetupAliasFrame (currententity, paliashdr);

	glPopMatrix ();

	if (r_shadows.value)
	{
	}
}


// separated out as it's also needed for the viewmodel
void R_InitAliasRender (void)
{
	// bring up vertex pointer
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (aliasvert_t), aliasverts[0].v);

	// bring up image texture array
	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (aliasvert_t), aliasverts[0].st);

	glClientActiveTexture (GL_TEXTURE1);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (aliasvert_t), aliasverts[0].st);

	// bring up colour array
	glEnableClientState (GL_COLOR_ARRAY);
	glColorPointer (4, GL_FLOAT, sizeof (aliasvert_t), aliasverts[0].rgba);

	glShadeModel (GL_SMOOTH);
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PRIMARY_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);
}


// separated out as it's also needed for the viewmodel
void R_EndAliasRender (void)
{
	glDisableClientState (GL_COLOR_ARRAY);

	glClientActiveTexture (GL_TEXTURE1);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glClientActiveTexture (GL_TEXTURE0);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glDisableClientState (GL_VERTEX_ARRAY);

	glActiveTexture (GL_TEXTURE1);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glColor4f (1, 1, 1, 1);

	glShadeModel (GL_FLAT);
}


void R_DrawAliasEntities (void)
{
	int i;
	qboolean ClearState = false;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		if (currententity->model->type == mod_alias)
		{
			if (!ClearState)
			{
				R_InitAliasRender ();
				ClearState = true;
			}

			R_DrawAliasModel (currententity);
		}
	}

	if (ClearState) R_EndAliasRender ();
}



