/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __MODEL__
#define __MODEL__

#include "modelgen.h"
#include "spritegn.h"

/*

d*_t structures are on-disk representations
m*_t structures are in-memory

*/

// entity effects

#define	EF_BRIGHTFIELD			1
#define	EF_MUZZLEFLASH 			2
#define	EF_BRIGHTLIGHT 			4
#define	EF_DIMLIGHT 			8



/*
==============================================================================

			TEXTURES

==============================================================================
*/

// texture loading flags
#define TEX_BASIC					0
#define TEX_MIPMAP					1
#define TEX_ALPHAMASK				2
#define TEX_INCACHE					4
#define TEX_LUMA					8
#define TEX_32BPP					16
#define TEX_NOCACHE					32
#define TEX_BSP						64
#define TEX_ALPHAHACK				128		// hipnotic alpha hack
#define TEX_MDL						256
#define TEX_SPR						512
#define TEX_NORMAL					1024	// normal rendering pass (as opposed to a hypothetical future TEX_NORMALMAP)
#define TEX_FONT8X8					2048	// 8 x 8 menu font texture


typedef struct gltexture_s
{
	// texture numbers shared by this texture object
	int		texnum;
	int		lumatexnum;

	char	identifier[MAX_OSPATH];
	int		width, height;

	int			cache;
	int			flags;

	byte md5digest[16];

	// generic colour member for anything that might ever need it
	float	colour[3];

	struct gltexture_s *next;
} gltexture_t;


typedef struct volume_s
{
	gltexture_t *tex;
	int volnum;
	struct volume_s *next;
} volume_t;

typedef struct glpic_s
{
	struct gltexture_s		*gltexture;
	float	sl, tl, sh, th;
} glpic_t;

typedef struct cachepic_s
{
	char		name[MAX_QPATH];
	qpic_t		pic;
	byte		padding[32];	// for appended glpic
} cachepic_t;

//
// in memory representation
//
// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	vec3_t		position;
} mvertex_t;

#define	SIDE_FRONT	0
#define	SIDE_BACK	1
#define	SIDE_ON		2


// plane_t structure
// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct mplane_s
{
	vec3_t	normal;
	float	dist;
	byte	type;			// for texture axis selection and fast side tests
	byte	signbits;		// signx + signy<<1 + signz<<1
	byte	pad[2];
} mplane_t;

typedef struct texture_s
{
	char		name[16];
	unsigned	width, height;

	struct gltexture_s	*gltexture;
	struct msurface_s	*texturechain;	// for sorted drawing

	int			anim_total;				// total tenths in sequence ( 0 = no)
	int			anim_min, anim_max;		// time for this frame min <=time< max

	struct texture_s *anim_next;		// in the animation sequence
	struct texture_s *alternate_anims;	// bmodels in frmae 1 use these

	unsigned	offsets[MIPLEVELS];		// four mip maps stored

	// global texturelist
	struct texture_s *texturelist;

	volume_t *volume;
} texture_t;


#define	SURF_PLANEBACK		2
#define	SURF_DRAWSKY		4
#define SURF_DRAWSPRITE		8
#define SURF_DRAWTURB		16
#define SURF_DRAWTILED		32
#define SURF_DRAWBACKGROUND	64
#define SURF_UNDERWATER		128
#define SURF_BMODEL			256

// defines for special liquid types
#define SURF_DRAWLAVA		512
#define SURF_DRAWSLIME		1024
#define SURF_DRAWTELE		2048

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	unsigned short	v[2];
	unsigned int	cachededgeoffset;
} medge_t;

typedef struct
{
	float		vecs[2][4];
	float		mipadjust;
	texture_t	*texture;
	int			flags;
} mtexinfo_t;

// define the vertex data we're going to use
typedef struct glvertex_s
{
	// vertex positions
	float v[3];
	float tv[3];

	// texture coords
	float st[2];

	// lightmap coords
	float lm[2];

	// colour
	float rgba[4];
} glvertex_t;


typedef struct glpoly_s
{
	struct	glpoly_s	*next;
	struct	glpoly_s	*chain;
	int		numverts;
	int		flags;			// for SURF_UNDERWATER

	vec3_t midpoint;
	vec3_t wincoords;
	float fxofs;
	float warpscale;
	float warpfade;
	qboolean hazemask;

	glvertex_t *verts;
} glpoly_t;

#define	LIGHTMAP_BYTES	4

#define	LM_BLOCK_WIDTH		128
#define	LM_BLOCK_HEIGHT		128


typedef struct glRect_s
{
	unsigned char l,t,w,h;
} glRect_t;


typedef struct lightmap_s
{
	qboolean	modified;
	glRect_t	rectchange;
	int			allocated[LM_BLOCK_WIDTH];
	byte		data[LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * LIGHTMAP_BYTES];

	int			texnum;

	struct lightmap_s *next;
} lightmap_t;

#define LIGHT_MODIFIED_UNMODIFIED	0
#define LIGHT_MODIFIED_STATIC		1
#define LIGHT_MODIFIED_DYNAMIC		2

typedef struct msurface_s
{
	int			visframe;		// should be drawn when node is crossed

	int			lavavisframe;

	mplane_t	*plane;
	int			flags;

	int			firstedge;	// look up in model->surfedges[], negative numbers
	int			numedges;	// are backwards edges
	
	short		texturemins[2];
	short		extents[2];

	int			light_s, light_t;	// gl lightmap coordinates

	glpoly_t	*polys;				// multiple if warped
	struct	msurface_s	*texturechain;

	struct	msurface_s	*genericchain;

	mtexinfo_t	*texinfo;

	// lighting info
	int			dlightframe;

	byte		*dlbits;

	lightmap_t	*lightmap;

	int			lightmodified;

	int			smax;
	int			tmax;
	byte		*base;

	// generic colour stuff
	// this is a pointer to a colour member of a gltexture_t
	float		*colour;

	volume_t	*volume;

	byte		styles[MAXLIGHTMAPS];
	int			cached_light[MAXLIGHTMAPS];	// values currently used in lightmap
	qboolean	cached_dlight;				// true if dynamic light in cache
	byte		*samples;		// [numstyles*surfsize]
} msurface_t;

typedef struct mnode_s
{
	// common with leaf
	int			contents;		// 0, to differentiate from leafs
	int			visframe;		// node needs to be traversed if current

	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

	// node specific
	mplane_t	*plane;
	struct mnode_s	*children[2];	

	unsigned short		firstsurface;
	unsigned short		numsurfaces;
} mnode_t;



typedef struct mleaf_s
{
	// common with node
	int			contents;		// 0, to differentiate from leafs
	int			visframe;		// node needs to be traversed if current

	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

	// leaf specific
	int			leafnum;

	// run off decompressed vis
	byte		*decompressed_vis;

	efrag_t		*efrags;

	// true if the leaf contains any water surfs
	qboolean	haswater;

	msurface_t	**firstmarksurface;
	int			nummarksurfaces;
	int			key;			// BSP sequence number for leaf's contents
	byte		ambient_sound_level[NUM_AMBIENTS];
} mleaf_t;

// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct
{
	dclipnode_t	*clipnodes;
	mplane_t	*planes;
	int			firstclipnode;
	int			lastclipnode;
	vec3_t		clip_mins;
	vec3_t		clip_maxs;
} hull_t;

/*
==============================================================================

SPRITE MODELS

==============================================================================
*/


// FIXME: shorten these?
typedef struct mspriteframe_s
{
	int		width;
	int		height;
	float	up, down, left, right;
	struct gltexture_s *gltexture;
} mspriteframe_t;

typedef struct
{
	int				numframes;
	float			*intervals;
	mspriteframe_t	*frames[1];
} mspritegroup_t;

typedef struct
{
	spriteframetype_t	type;
	mspriteframe_t		*frameptr;
} mspriteframedesc_t;

typedef struct
{
	int					type;
	int					maxwidth;
	int					maxheight;
	int					numframes;
	float				beamlength;		// remove?
	void				*cachespot;		// remove?
	mspriteframedesc_t	frames[1];
} msprite_t;


/*
==============================================================================

ALIAS MODELS

Alias models are position independent, so the cache manager can move them.
==============================================================================
*/

typedef struct
{
	float	v[3];
	byte	lightnormalindex;
	qboolean nodraw;
} drawvertx_t;


typedef struct
{
	int					firstpose;
	int					numposes;
	float				interval;
	trivertx_t			bboxmin;
	trivertx_t			bboxmax;
	int					frame;
	char				name[16];
} maliasframedesc_t;

typedef struct
{
	trivertx_t			bboxmin;
	trivertx_t			bboxmax;
	int					frame;
} maliasgroupframedesc_t;

typedef struct
{
	int						numframes;
	int						intervals;
	maliasgroupframedesc_t	frames[1];
} maliasgroup_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct mtriangle_s {
	int					facesfront;
	int					vertindex[3];
} mtriangle_t;


#define	MAX_SKINS	32
typedef struct aliashdr_s
{
	int			ident;
	int			version;
	vec3_t		scale;
	vec3_t		scale_origin;
	float		boundingradius;
	vec3_t		eyeposition;
	int			numskins;
	int			skinwidth;
	int			skinheight;
	int			numverts;
	int			numtris;
	int			numframes;
	synctype_t	synctype;
	int			flags;
	float		size;

	int					numposes;
	int					poseverts;
	int					posedata;	// numposes*poseverts trivert_t
	int					commands;	// gl command list with embedded s/t
	struct gltexture_s	*gltexture[MAX_SKINS][4];
	int					texels[MAX_SKINS];	// only for player skins
	maliasframedesc_t	frames[1];	// variable sized
} aliashdr_t;

#define	MAXALIASVERTS	2000
#define	MAXALIASFRAMES	256
#define	MAXALIASTRIS	2048
extern	aliashdr_t	*pheader;
extern	stvert_t	stverts[MAXALIASVERTS];
extern	mtriangle_t	triangles[MAXALIASTRIS];
extern	trivertx_t	*poseverts[MAXALIASFRAMES];

//===================================================================

//
// Whole model
//

// worldtypes
#define WORLDTYPE_WIZARD	0
#define WORLDTYPE_METAL		1
#define WORLDTYPE_BASE		2


typedef enum {mod_brush, mod_sprite, mod_alias} modtype_t;

#define	EF_ROCKET	1			// leave a trail
#define	EF_GRENADE	2			// leave a trail
#define	EF_GIB		4			// leave a trail
#define	EF_ROTATE	8			// rotate (bonus items)
#define	EF_TRACER	16			// green split trail
#define	EF_ZOMGIB	32			// small blood trail
#define	EF_TRACER2	64			// orange split trail + rotate
#define	EF_TRACER3	128			// purple trail

// MH additional flags
#define EF_FLARESPRITE		256
#define EF_FLARETORCH		512
#define EF_ROCKETFLARE		1024
#define EF_POWERUPRED		2048
#define EF_POWERUPBLUE		131072
#define EF_BUBBLE			4096
#define EF_SHAMBLER			8192
#define EF_FLICKER			16384
#define EF_SMOKESMOKE		32768
#define EF_NODRAW			65536
#define EF_TRANS66			262144

typedef struct model_s
{
	char		name[MAX_QPATH];
	qboolean	needload;		// bmodels and sprites don't cache normally

	modtype_t	type;
	int			numframes;
	synctype_t	synctype;
	
	int			flags;

//
// volume occupied by the model graphics
//		
	vec3_t		mins, maxs;
	float		radius;

//
// solid volume for clipping 
//
	qboolean	clipbox;
	vec3_t		clipmins, clipmaxs;

//
// brush model
//
	int			firstmodelsurface, nummodelsurfaces;

	int			numsubmodels;
	dmodel_t	*submodels;

	int			numplanes;
	mplane_t	*planes;

	int			numleafs;		// number of visible leafs, not counting 0
	mleaf_t		*leafs;

	int			numvertexes;
	mvertex_t	*vertexes;

	int			numedges;
	medge_t		*edges;

	int			numnodes;
	mnode_t		*nodes;

	int			numtexinfo;
	mtexinfo_t	*texinfo;

	int			numsurfaces;
	msurface_t	*surfaces;

	int			numsurfedges;
	int			*surfedges;

	int			numclipnodes;
	dclipnode_t	*clipnodes;

	int			nummarksurfaces;
	msurface_t	**marksurfaces;

	hull_t		hulls[MAX_MAP_HULLS];

	int			numtextures;
	texture_t	**textures;

	byte		*visdata;

	byte		*lightdata;
	char		*entities;

	// midpoint of the bmodel
	vec3_t		midpoint;

	// world type of the model for setting various blah client-side
	int			worldtype;

	// additional model data - !!!THIS MUST BE THE LAST ITEM IN THE STRUCT!!!
	cache_user_t	cache;		// only access through Mod_Extradata

} model_t;

//============================================================================

void	Mod_Init (void);
void	Mod_ClearAll (void);
model_t *Mod_ForName (char *name, qboolean crash);
void	*Mod_Extradata (model_t *mod);	// handles caching
void	Mod_TouchModel (char *name);

mleaf_t *Mod_PointInLeaf (float *p, model_t *model);
byte	*Mod_LeafPVS (mleaf_t *leaf, model_t *model);

#endif	// __MODEL__
