/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

entity_t	r_worldentity;

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

qboolean	envmap;				// true during envmap command capture 

int			playertextures[16];		// up to 16 color translated skins

float r_frametime;

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float	r_world_matrix[16];
float	r_base_world_matrix[16];

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

texture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value


void R_MarkLeaves (void);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_shadows = {"r_shadows","0"};
cvar_t	r_dynamic = {"r_dynamic","1"};

cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_nocolors = {"gl_nocolors","0"};

void R_SmokeFrame (void);
void R_DrawBrushModel (entity_t *e);
void R_DrawAliasModel (entity_t *e);
void R_DrawSpriteModel (entity_t *e);
void RotatePointAroundVector (vec3_t dst, const vec3_t dir, const vec3_t point, float degrees);
void R_AnimateLight (void);
void V_CalcBlend (void);
void R_DrawWorld (void);
void R_DrawParticles (void);
void R_DrawWaterSurfaces (qboolean Bloom);

qboolean in_mapshot = false;

/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	int		i;

	for (i=0 ; i<4 ; i++)
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
			return true;
	return false;
}


void R_RotateForEntity (entity_t *e)
{
    glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    glRotatef (e->angles[1],  0, 0, 1);
    glRotatef (-e->angles[0],  0, 1, 0);
    glRotatef (e->angles[2],  1, 0, 0);
}


/*
=============
R_DrawEntitiesOnList
=============
*/

void R_InitAliasRender (void);
void R_EndAliasRender (void);
void R_DrawAliasEntities (void);
void R_DrawSpriteEntities (void);

void R_DrawEntitiesOnList (void)
{
	if (!r_drawentities.value)
		return;

	R_DrawAliasEntities ();
	R_DrawSpriteEntities ();
}


/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	if (!r_drawviewmodel.value) return;
	if (chase_active.value) return;
	if (envmap) return;
	if (!r_drawentities.value) return;
	if (cl.stats[STAT_HEALTH] <= 0) return;

	currententity = &cl.viewent;

	if (!currententity->model) return;

	R_InitAliasRender ();

	if (cl.items & IT_INVISIBILITY)
	{
		glEnable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE);
	}

	// hack the depth range to prevent view model from poking into walls
	glDepthRange (0.0, 0.1);
	R_DrawAliasModel (currententity);
	glDepthRange (0.1, 0.8);

	R_EndAliasRender ();

	if (cl.items & IT_INVISIBILITY)
	{
		glDisable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test
	bits = 0;

	for (j = 0; j < 3; j++)
		if (out->normal[j] < 0)
			bits |= 1 << j;

	return bits;
}


// the so-called "simple case" turned out to be not so simple after all.  the world's like that sometimes.  live with it.
void R_SetFrustum (void)
{
	int		i;

	// rotate VPN right by FOV_X/2 degrees
	RotatePointAroundVector (frustum[0].normal, vup, vpn, -(90 - r_refdef.fov_x / 2 ));

	// rotate VPN left by FOV_X/2 degrees
	RotatePointAroundVector (frustum[1].normal, vup, vpn, 90 - r_refdef.fov_x / 2);

	// rotate VPN up by FOV_X/2 degrees
	RotatePointAroundVector (frustum[2].normal, vright, vpn, 90 - r_refdef.fov_y / 2);

	// rotate VPN down by FOV_X/2 degrees
	RotatePointAroundVector (frustum[3].normal, vright, vpn, -(90 - r_refdef.fov_y / 2 ));

	for (i = 0; i < 4; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
	R_AnimateLight ();

	r_framecount++;

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);

	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;
}


/*
=============
R_SetupGL
=============
*/
void R_SetupGL3DViewport (int width, int height)
{
	extern float r_farclip;

	glViewport (0, 0, width, height);

	// set up viewpoint
	glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();

	// retain correct aspect irrespective of size
	gluPerspective (r_refdef.fov_y, (float) vid.width / (float) vid.height, 4, r_farclip);

	glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();

    glRotatef (-90, 1, 0, 0);	    // put Z going up
    glRotatef (90, 0, 0, 1);	    // put Z going up
    glRotatef (-r_refdef.viewangles[2], 1, 0, 0);
    glRotatef (-r_refdef.viewangles[0], 0, 1, 0);
    glRotatef (-r_refdef.viewangles[1], 0, 0, 1);

	glTranslatef (-r_refdef.vieworg[0], -r_refdef.vieworg[1], -r_refdef.vieworg[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	glCullFace (GL_FRONT);

	// set drawing parms
	if (gl_cull.value)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
}


/*
=============
R_Clear
=============
*/
void R_Clear (void)
{
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDepthFunc (GL_LEQUAL);
	glDepthRange (0.1, 0.8);
}

void R_RenderBloom (void);

// split out so that it can be called more than once
void R_RenderScene (void)
{
	R_Clear ();

	R_SetupFrame ();

	R_SetFrustum ();

	R_MarkLeaves ();	// done here so we know if we're in water

	R_DrawWorld ();		// adds static entities to the list

	S_ExtraUpdate ();	// don't let sound get messed up if going slow

	R_DrawEntitiesOnList ();

	if (r_viewleaf->contents == CONTENTS_EMPTY)
	{
		R_DrawParticles ();
		R_DrawWaterSurfaces (false);
	}
	else
	{
		R_DrawWaterSurfaces (false);
		R_DrawParticles ();
	}

	R_SmokeFrame ();
	R_DrawViewModel ();

	R_RenderBloom ();
}


/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	r_frametime = cl.time - cl.oldtime;

	if (r_norefresh.value) return;

	if (!r_worldentity.model || !cl.worldmodel) Sys_Error ("R_RenderView: NULL worldmodel");

	// draw this on frame 3 cos the first few are skipped over
	if (r_framecount == 3)
	{
		// i like this...
		SCR_CenterPrint (cl.levelname);
	}

	R_RenderScene ();
}
