/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"

void R_InitParticles (void);
void R_ClearParticles (void);
void GL_BuildLightmaps (void);
void GetBModelMidPoints (void);

gltexture_t *r_blacktexture = NULL;
gltexture_t *r_bloomtexture = NULL;

unsigned int bloomdata[BLOOM_SIZE * BLOOM_SIZE];

void R_InitNoTexture (void)
{
	int		x,y, m;
	byte	*dest;

	// create a simple checkerboard texture for the default
	r_notexture_mip = (texture_t *) malloc (sizeof (texture_t) + 16 * 16 + 8 * 8 + 4 * 4 + 2 * 2);

	r_notexture_mip->width = r_notexture_mip->height = 16;

	r_notexture_mip->offsets[0] = sizeof (texture_t);
	r_notexture_mip->offsets[1] = r_notexture_mip->offsets[0] + 16 * 16;
	r_notexture_mip->offsets[2] = r_notexture_mip->offsets[1] + 8 * 8;
	r_notexture_mip->offsets[3] = r_notexture_mip->offsets[2] + 4 * 4;

	for (m = 0; m < 4; m++)
	{
		dest = (byte *) r_notexture_mip + r_notexture_mip->offsets[m];

		for (y = 0; y < (16 >> m); y++)
		{
			for (x = 0; x < (16 >> m); x++)
			{
				if ((y < (8 >> m)) ^ (x < (8 >> m)))
					*dest++ = 0;
				else *dest++ = 15;
			}
		}
	}

	r_notexture_mip->anim_max = 0;
	r_notexture_mip->anim_min = 0;
	r_notexture_mip->anim_next = r_notexture_mip;
	r_notexture_mip->anim_total = 0;

	r_notexture_mip->gltexture = GL_LoadTexture ("notexture", 16, 16, (byte *) r_notexture_mip + r_notexture_mip->offsets[0], (TEX_MIPMAP | TEX_INCACHE));
}


void R_InitColouredTexture (gltexture_t **tex, byte PalIndex)
{
	byte data[16];

	memset (data, PalIndex, 16);

	*tex = GL_LoadTexture ("*******", 4, 4, data, (TEX_MIPMAP | TEX_INCACHE));
}


void R_InitBloomTexture (gltexture_t **tex)
{
	int i;

	// make it initially random to avoid MD5 matching
	for (i = 0; i < (BLOOM_SIZE * BLOOM_SIZE); i++)
		bloomdata[i] = (rand () % 256) * (rand () % 256) * (rand () % 256) * (rand () % 256);

	// this won't take all of the data, but no big deal
	*tex = GL_LoadTexture ("*******", BLOOM_SIZE, BLOOM_SIZE, (byte *) bloomdata, TEX_INCACHE);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
}


/*
==================
R_InitTextures
==================
*/
void R_InitTextures (void)
{
	BuildTextureSearchPaths ("textures/", NULL, NULL, NULL);
	R_InitNoTexture ();
	R_InitColouredTexture (&r_blacktexture, 0);

	R_InitBloomTexture (&r_bloomtexture);
}


/*
===============
R_Envmap_f

Grab six views for environment mapping tests
===============
*/
void R_Envmap_f (void)
{
}

/*
===============
R_Init
===============
*/
void R_InitSmokes (void);
extern cvar_t r_skyspeed;
void R_InitSkySphere (void);
void R_LoadSky_f (void);

cvar_t r_bloom = {"r_bloom", "1", true};
cvar_t gl_lockpvs = {"gl_lockpvs", "0"};

void R_Init (void)
{	
	extern byte *hunk_base;
	int i;

	Cmd_AddCommand ("loadsky", R_LoadSky_f);	
	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);	
	Cmd_AddCommand ("envmap", R_Envmap_f);	
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);	

	// bloom control
	Cvar_RegisterVariable (&r_bloom);

	Cvar_RegisterVariable (&r_norefresh);
	Cvar_RegisterVariable (&r_drawentities);
	Cvar_RegisterVariable (&r_drawviewmodel);
	Cvar_RegisterVariable (&r_shadows);
	Cvar_RegisterVariable (&r_dynamic);
	Cvar_RegisterVariable (&gl_lockpvs);

	Cvar_RegisterVariable (&gl_cull);
	Cvar_RegisterVariable (&gl_polyblend);
	Cvar_RegisterVariable (&gl_nocolors);
	Cvar_RegisterVariable (&r_skyspeed);

	R_InitParticles ();
	R_InitSmokes ();
	R_InitSkySphere ();

	for (i = 0; i < 16; i++) GL_AllocTexture (&playertextures[i]);
}


/*
===============
R_TranslatePlayerSkin

Translates a skin texture by the per-player color lookup
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int		top, bottom;
	byte	translate[256];
	int		i, s;
	model_t	*model;
	aliashdr_t *paliashdr;
	byte	*original;
	int			inwidth, inheight;
	extern	byte		**player_8bit_texels_tbl;
	byte	translated[320*200];

	top = cl.scores[playernum].colors & 0xf0;
	bottom = (cl.scores[playernum].colors &15)<<4;

	for (i=0 ; i<256 ; i++)
		translate[i] = i;

	for (i=0 ; i<16 ; i++)
	{
		if (top < 128)	// the artists made some backwards ranges.  sigh.
			translate[TOP_RANGE+i] = top+i;
		else
			translate[TOP_RANGE+i] = top+15-i;
				
		if (bottom < 128)
			translate[BOTTOM_RANGE+i] = bottom+i;
		else
			translate[BOTTOM_RANGE+i] = bottom+15-i;
	}

	//
	// locate the original skin pixels
	//
	currententity = &cl_entities[1+playernum];
	model = currententity->model;
	if (!model)
		return;		// player doesn't have a model yet
	if (model->type != mod_alias)
		return; // only translate skins on alias models

	paliashdr = (aliashdr_t *)Mod_Extradata (model);
	s = paliashdr->skinwidth * paliashdr->skinheight;
	if (currententity->skinnum < 0 || currententity->skinnum >= paliashdr->numskins) {
		Con_Printf("(%d): Invalid player skin #%d\n", playernum, currententity->skinnum);
		original = (byte *)paliashdr + paliashdr->texels[0];
	} else
		original = (byte *)paliashdr + paliashdr->texels[currententity->skinnum];
	if (s & 3)
		Sys_Error ("R_TranslateSkin: s&3");

	inwidth = paliashdr->skinwidth;
	inheight = paliashdr->skinheight;

	// because this happens during gameplay, do it fast
	// instead of sending it through gl_upload 8
    glBindTexture (GL_TEXTURE_2D, playertextures[playernum]);

	for (i=0 ; i<s ; i+=4)
	{
		translated[i] = translate[original[i]];
		translated[i+1] = translate[original[i+1]];
		translated[i+2] = translate[original[i+2]];
		translated[i+3] = translate[original[i+3]];
	}

	// don't mipmap these, because it takes too long
	GL_UploadTexture ((unsigned *) translated, paliashdr->skinwidth, paliashdr->skinheight, TEX_BASIC);
}


/*
===============
R_NewMap
===============
*/
extern vec3_t farmin;
extern vec3_t farmax;
extern float r_farclip;
void R_AddTexturesToList (void);
void R_LoadSky (char *skyname, qboolean verbose);
void R_SetWorldVolumes (void);

void R_NewMap (void)
{
	int		i;
	vec3_t	dist;

	for (i = 0; i < 256; i++)
		d_lightstylevalue[i] = 264;		// normal light value

	memset (&r_worldentity, 0, sizeof (r_worldentity));
	r_worldentity.model = cl.worldmodel;

	// clear out efrags in case the level hasn't been reloaded
	for (i = 0; i < cl.worldmodel->numleafs + 1; i++)
		cl.worldmodel->leafs[i].efrags = NULL;
		 	
	r_viewleaf = NULL;
	R_ClearParticles ();

	GL_BuildLightmaps ();
	GetBModelMidPoints ();

	// identify sky texture
	skytexturenum = -1;

	for (i = 0; i < cl.worldmodel->numtextures; i++)
	{
		if (!cl.worldmodel->textures[i]) continue;
		if (!Q_strncmp (cl.worldmodel->textures[i]->name,"sky", 3)) skytexturenum = i;

 		cl.worldmodel->textures[i]->texturechain = NULL;
	}

	// set far clipping plane - give it an extra 512 for headroom
	dist[0] = farmax[0] - farmin[0];
	dist[1] = farmax[1] - farmin[1];
	dist[2] = farmax[2] - farmin[2];

	r_farclip = sqrt (dist[0] * dist[0] + dist[1] * dist[1] + dist[2] * dist[2]) + 512.0f;

	Con_DPrintf ("Far Clipping Plane: %f\n", r_farclip);

	// reset for next time around
	farmin[0] = farmin[1] = farmin[2] = 999999999;
	farmax[0] = farmax[1] = farmax[2] = -999999999;

	// clear down any unused textures
	GL_DeleteTextures ();

	R_AddTexturesToList ();

	// set up the skybox:
	switch (cl.worldmodel->worldtype)
	{
	case 0:
		// medieval
		R_LoadSky ("medieval", false);
		break;

	case 1:
		// metal
		R_LoadSky ("metal", false);
		break;

	case 2:
		// base
		R_LoadSky ("base", false);
		break;

	default:
		// anything else
		R_LoadSky ("medieval", false);
		break;
	}

	// build world volumes
	R_SetWorldVolumes ();
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
}


void D_FlushCaches (void)
{
}


