/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"

int			skytexturenum;

#ifndef GL_RGBA4
#define	GL_RGBA4	0
#endif


msurface_t  *skychain = NULL;
msurface_t  *genericchain = NULL;

qboolean render_bloom;

typedef struct facevert_s
{
	float v[3];
	float st[2];
	float lm[2];
} facevert_t;

facevert_t *faceverts = NULL;
int maxsurfverts;

void R_DrawSkyChain (qboolean skydraw);
void R_MarkLava (void);
void R_SurfaceHazeWarp (void);
void R_BlendLightmaps (void);
void R_RenderDynamicLightmaps (void);
void MakeLightmapST (msurface_t *surf, vec3_t vec, float *st);
void WarpTexCoords (float SBase, float TBase, float factor, float *stOut);
void R_PushDlights (mnode_t *headnode);

extern gltexture_t *r_blacktexture;

texture_t *texturelist = NULL;

// true if there are liquid surfaces in the PVS
qboolean liquidpvs;

volume_t *worldvolumes = NULL;
int globalflags;

void R_AddTextureToList (texture_t *tex)
{
	texture_t *t;

	// search through the list for it
	for (t = texturelist; t; t = t->texturelist)
		if (t == tex)
			return;

	// didn't find it, so add it in
	tex->texturelist = texturelist;
	texturelist = tex;

	// no associated gltexture_t struct
	if (!tex->gltexture) return;

	// add in normal texture flag
	if (tex->gltexture->flags & TEX_LUMA) return;
	if (tex->gltexture->flags & TEX_ALPHAHACK) return;

	tex->gltexture->flags |= TEX_NORMAL;
}


void R_AddModelTexturesToList (model_t *m)
{
	int i;

	for (i = 0; i < m->numtextures; i++)
	{
		if (!m->textures[i]) continue;

		R_AddTextureToList (m->textures[i]);
	}
}


void R_AddTexturesToList (void)
{
	int i;
	model_t *m;

	texturelist = NULL;

	for (i = 1; i < MAX_MODELS; i++)
	{
		m = cl.model_precache[i];
		if (!m) break;

		R_AddModelTexturesToList (m);
	}

	// remove alphahack from world textures
	for (i = 0; i < cl.worldmodel->numtextures; i++)
	{
		if (!cl.worldmodel->textures[i]) continue;
		if (!cl.worldmodel->textures[i]->gltexture) continue;

		cl.worldmodel->textures[i]->gltexture->flags |= TEX_ALPHAHACK;
		cl.worldmodel->textures[i]->gltexture->flags ^= TEX_ALPHAHACK;
	}
}


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	int		reletive;
	int		count;
	texture_t *original = base;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}
	
	if (!base->anim_total)
		return base;

	reletive = (int)(cl.time*10) % base->anim_total;

	count = 0;	
	while (base->anim_min > reletive || base->anim_max <= reletive)
	{
		base = base->anim_next;

		if (!base) return original;
		if (++count > 100) return original;
	}

	return base;
}


/*
=============================================================

	BRUSH MODELS

=============================================================
*/

/*
================
DrawGLWarpPoly

Warp the vertex coordinates
================
*/
void DrawGLWarpPoly (glpoly_t *p)
{
	int		i;
	glvertex_t	*v;
	float	warpscale = p->warpfade * p->warpscale;

	v = p->verts;

	faceverts[0].st[0] = v->st[0];
	faceverts[0].st[1] = v->st[1];

	faceverts[0].lm[0] = v->lm[0];
	faceverts[0].lm[1] = v->lm[1];

	faceverts[0].v[0] = v->tv[0] + warpscale * sin (v->tv[1] * 0.05f + realtime) * sin (v->tv[2] * 0.05f + realtime);
	faceverts[0].v[1] = v->tv[1] + warpscale * sin (v->tv[0] * 0.05f + realtime) * sin (v->tv[2] * 0.05f + realtime);
	faceverts[0].v[2] = v->tv[2] + warpscale * sin (v->tv[0] * 0.05f + realtime) * sin (v->tv[1] * 0.05f + realtime);

	for (i = 1; i < p->numverts; i++)
	{
		faceverts[i].st[0] = v[i].st[0];
		faceverts[i].st[1] = v[i].st[1];

		faceverts[i].lm[0] = v[i].lm[0];
		faceverts[i].lm[1] = v[i].lm[1];

		faceverts[i].v[0] = v[i].tv[0];
		faceverts[i].v[1] = v[i].tv[1];
		faceverts[i].v[2] = v[i].tv[2];
	}

	glDrawArrays (GL_TRIANGLE_FAN, 0, p->numverts);
}


/*
================
DrawGLPoly

Ignores poly midpoint, only does the edges.
================
*/
void DrawGLPoly (glpoly_t *p)
{
	int		i;
	glvertex_t	*v;

	v = &p->verts[1];

	for (i = 0; i < p->numverts - 1; i++)
	{
		faceverts[i].st[0] = v[i].st[0];
		faceverts[i].st[1] = v[i].st[1];

		faceverts[i].lm[0] = v[i].lm[0];
		faceverts[i].lm[1] = v[i].lm[1];

		faceverts[i].v[0] = v[i].tv[0];
		faceverts[i].v[1] = v[i].tv[1];
		faceverts[i].v[2] = v[i].tv[2];
	}

	glDrawArrays (GL_TRIANGLE_FAN, 0, p->numverts);
}


/*
================
R_DrawTextureChains
================
*/
void R_DrawTextureChain (int flags)
{
	msurface_t	*surf;
	texture_t	*t;
	int oldLightmap = -1;

	if (flags & TEX_ALPHAHACK)
	{
		glEnable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	if (flags & TEX_LUMA)
	{
		glActiveTexture (GL_TEXTURE2);
		glEnable (GL_TEXTURE_2D);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

		glClientActiveTexture (GL_TEXTURE2);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (facevert_t), faceverts[0].st);
	}

	for (t = texturelist; t; t = t->texturelist)
	{
		// no surfs!
		if (!(surf = t->texturechain)) continue;

		// draw these later
		if ((surf->flags & SURF_DRAWSKY)) continue;
		if ((surf->flags & SURF_DRAWTURB)) continue;

		// not in this pass
		if (!(t->gltexture->flags & flags)) continue;

		// bind luma
		if (flags & TEX_LUMA)
		{
			// we've guaranteed that lumatexnum is valid if this is true
			glActiveTexture (GL_TEXTURE2);
			glBindTexture (GL_TEXTURE_2D, t->gltexture->lumatexnum);
		}

		// bind regular texture
		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, t->gltexture->texnum);
		glActiveTexture (GL_TEXTURE1);

		for (; surf; surf = surf->texturechain)
		{
			if (surf->lightmap->texnum != oldLightmap)
			{
				oldLightmap = surf->lightmap->texnum;
				glBindTexture (GL_TEXTURE_2D, surf->lightmap->texnum);
			}

			if ((surf->flags & SURF_UNDERWATER) || surf->polys->hazemask)
				DrawGLWarpPoly (surf->polys);
			else DrawGLPoly (surf->polys);
		}
	}

	if (flags & TEX_LUMA)
	{
		glActiveTexture (GL_TEXTURE2);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable (GL_TEXTURE_2D);
		glActiveTexture (GL_TEXTURE1);

		glClientActiveTexture (GL_TEXTURE2);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	if (flags & TEX_ALPHAHACK) glDisable (GL_BLEND);
}


void R_DrawTextureChains (void)
{
	// bring up vertex pointer
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (facevert_t), faceverts[0].v);

	// bring up image texture arrays
	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (facevert_t), faceverts[0].st);

	glClientActiveTexture (GL_TEXTURE1);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (facevert_t), faceverts[0].lm);

	// bring up lightmaps TMU
	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);

	// only scale if there's light
	if (cl.worldmodel->lightdata) glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);

	// draw chains in multiple passes
	if (globalflags & TEX_NORMAL) R_DrawTextureChain (TEX_NORMAL);
	if (globalflags & TEX_LUMA) R_DrawTextureChain (TEX_LUMA);
	if (globalflags & TEX_ALPHAHACK) R_DrawTextureChain (TEX_ALPHAHACK);

	// GL_RGB_SCALE needs to be explicitly reset.  Even though it's specified to only work with GL_COMBINE,
	// some implementations can be a little lax and set it for all values of GL_TEXTURE_ENV_MODE.
	// I've knocked ATI enough in the past, so this time it's "Hello NVidia!!!"
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);
	glActiveTexture (GL_TEXTURE0);

	glClientActiveTexture (GL_TEXTURE1);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glClientActiveTexture (GL_TEXTURE0);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glDisableClientState (GL_VERTEX_ARRAY);
}


/*
=============================================================

	WORLD MODEL

=============================================================
*/


void R_AddSkySurface (msurface_t *surf);

void R_AddSurfToTextureChains (msurface_t *surf)
{
	texture_t *anim;

	// prevent e2m3 barfs
	if (!surf->texinfo) return;
	if (!surf->texinfo->texture) return;

	if (surf->flags & SURF_DRAWSKY)
	{
		// add to skybox
		R_AddSkySurface (surf);

		surf->texturechain = skychain;
		skychain = surf;
	}
	else
	{
		// animate here and build the chain from the animated version
		anim = R_TextureAnimation (surf->texinfo->texture);

		// prevent e2m3 barfs
		if (!anim->gltexture) return;

		surf->texturechain = anim->texturechain;
		anim->texturechain = surf;

		if (!(surf->flags & SURF_DRAWTURB))
		{
			surf->genericchain = genericchain;
			genericchain = surf;

			surf->polys->hazemask = false;
		}

		// check for bloom
		if (anim->gltexture->lumatexnum) render_bloom = true;
		if (surf->flags & SURF_DRAWLAVA) render_bloom = true;
		if (surf->flags & SURF_DRAWSLIME) render_bloom = true;

		// store flags globally
		globalflags |= anim->gltexture->flags;
	}
}


float *Get_WaterFog (mleaf_t *leaf)
{
	mnode_t *node = leaf->parent;
	int i;
	msurface_t *surf;

	// backtrack up the node chain until we find one that contains water
	for (;;)
	{
		// this should never happen...
		if (!node->parent) return NULL;

		// check all of the surfs
		for (i = 0, surf = cl.worldmodel->surfaces + node->firstsurface; i < node->numsurfaces; i++, surf++)
			if (surf->flags & SURF_DRAWTURB)
				return surf->texinfo->texture->gltexture->colour;

		node = node->parent;
	}
}


/*
================
R_RecursiveWorldNode
================
*/
void R_StoreEfrags (efrag_t **ppefrag);

void R_RecursiveWorldNode (mnode_t *node)
{
	int			c, side;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		dot;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *)node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;

				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		if (pleaf->efrags) R_StoreEfrags (&pleaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;

	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;

	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;

	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0)
		side = 0;
	else
		side = 1;

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	// draw stuff
	c = node->numsurfaces;

	if (c)
	{
		surf = cl.worldmodel->surfaces + node->firstsurface;

		if (dot < 0 -BACKFACE_EPSILON)
			side = SURF_PLANEBACK;
		else if (dot > BACKFACE_EPSILON)
			side = 0;

		for ( ; c ; c--, surf++)
		{
			if (surf->visframe != r_framecount)
				continue;

			// wrong side
			// don't backface underwater surfaces, because they warp
			if (!(surf->flags & SURF_UNDERWATER) && ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)))
				continue;

			// check for liquid surfs in the PVS
			if (surf->flags & SURF_DRAWTURB)
				liquidpvs = true;

			R_AddSurfToTextureChains (surf);
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}


void R_ClearTextureChains (void)
{
	texture_t	*t;

	// clear down all of our texture chains
	for (t = texturelist; t; t = t->texturelist)
		t->texturechain = NULL;

	// and the other chains we use
	skychain = NULL;
	genericchain = NULL;
}


void R_TransformSurface (msurface_t *surf, float *matrix)
{
	glpoly_t *p;
	int i;
	glvertex_t *v;

	if (surf->flags & SURF_DRAWSKY) return;

	for (p = surf->polys; p; p = p->next)
	{
		v = p->verts;

		for (i = 0; i < p->numverts; i++)
		{
			// i used hexadecimal so it would line up nicely.  how sad...
			v[i].tv[0] = matrix[0x0] * v[i].v[0] + matrix[0x4] * v[i].v[1] + matrix[0x8] * v[i].v[2] + matrix[0xc];
			v[i].tv[1] = matrix[0x1] * v[i].v[0] + matrix[0x5] * v[i].v[1] + matrix[0x9] * v[i].v[2] + matrix[0xd];
			v[i].tv[2] = matrix[0x2] * v[i].v[0] + matrix[0x6] * v[i].v[1] + matrix[0xa] * v[i].v[2] + matrix[0xe];
		}
	}
}


qboolean IsSurfaceUnderwater (msurface_t *surf)
{
	mleaf_t *leaf;

	// is there liquid in the PVS?
	if (!liquidpvs)
	{
		// if there is no liquid visible, we can base the result on a simple check of the viewleaf
		if (r_viewleaf->contents == CONTENTS_WATER) return true;
		if (r_viewleaf->contents == CONTENTS_SLIME) return true;
		if (r_viewleaf->contents == CONTENTS_LAVA) return true;

		return false;
	}

	// removed high/low water mark testing - it broke with vertical teleports (and as lava is also there, we can't just omit it for teles)

	// now get the leaf containing the midpoint of the surf - note that 
	// we must use the transformed version here as the surf can move
	leaf = Mod_PointInLeaf (surf->polys->verts->tv, cl.worldmodel);

	// valid for these types
	if (leaf->contents == CONTENTS_WATER) return true;
	if (leaf->contents == CONTENTS_SLIME) return true;
	if (leaf->contents == CONTENTS_LAVA) return true;

	return false;
}


// mark all nodes in a bmodel visible
void R_MakeVisible (mnode_t *node)
{
	node->visframe = r_visframecount;

	if (node->contents < 0) return;

	R_MakeVisible (node->children[0]);
	R_MakeVisible (node->children[1]);
}


void R_AddInlineBModelToDrawLists (entity_t *e)
{
	vec3_t		mins, maxs;
	int			i;
	msurface_t	*surf;
	model_t		*clmodel;
	qboolean	rotated;
	float		matrix[16];

	currententity = e;

	clmodel = e->model;

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		rotated = true;
		for (i = 0; i < 3; i++)
		{
			mins[i] = e->origin[i] - clmodel->radius;
			maxs[i] = e->origin[i] + clmodel->radius;
		}
	}
	else
	{
		rotated = false;
		VectorAdd (e->origin, clmodel->mins, mins);
		VectorAdd (e->origin, clmodel->maxs, maxs);
	}

	if (R_CullBox (mins, maxs)) return;

	VectorSubtract (r_refdef.vieworg, e->origin, modelorg);

	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}

	surf = &clmodel->surfaces[clmodel->firstmodelsurface];

	e->angles[0] = -e->angles[0];	// stupid quake bug

	// removed origin hack - we'll do it in the maps instead ;)
    glLoadIdentity ();
	R_RotateForEntity (e);
	glGetFloatv (GL_MODELVIEW_MATRIX, matrix);

	e->angles[0] = -e->angles[0];	// stupid quake bug

	// force all nodes to visible
	R_MakeVisible (clmodel->nodes + clmodel->hulls[0].firstclipnode);

	// accumulate surfs into texture chains
	for (i = 0; i < clmodel->nummodelsurfaces; i++, surf++)
	{
		R_TransformSurface (surf, matrix);
		R_AddSurfToTextureChains (surf);

		if (IsSurfaceUnderwater (surf))
			surf->flags |= SURF_UNDERWATER;
		else
		{
			surf->flags |= SURF_UNDERWATER;
			surf->flags ^= SURF_UNDERWATER;
		}

		// mark as visible
		surf->visframe = r_framecount;

		// brute force flagging as a bmodel
		surf->flags |= SURF_BMODEL;
	}

	// calculate dynamic lighting for bmodel
	if (clmodel->name[0] == '*') R_PushDlights (clmodel->nodes + clmodel->hulls[0].firstclipnode);
}


void R_AddBModelsToDrawLists (void)
{
	int i;

	if (!r_drawentities.value) return;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		if (currententity->model->type == mod_brush)
		{
			if (currententity->model->name[0] == '*')
				R_AddInlineBModelToDrawLists (currententity);
		}
	}
}


void R_RenderBModelInstances (void)
{
	int			i;
	int			j;
	vec3_t		shadelight;
	qboolean	StateInit = false;
	msurface_t	*surf;
	model_t		*clmodel;
	texture_t	*tex;
	float		matrix[16];

	if (!r_drawentities.value) return;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		if (currententity->model->type == mod_brush)
		{
			if (currententity->model->name[0] != '*')
			{
				if (!StateInit)
				{
					// bring up vertex pointer
					glEnableClientState (GL_VERTEX_ARRAY);
					glVertexPointer (3, GL_FLOAT, sizeof (facevert_t), faceverts[0].v);

					// bring up image texture arrays
					glClientActiveTexture (GL_TEXTURE0);
					glEnableClientState (GL_TEXTURE_COORD_ARRAY);
					glTexCoordPointer (2, GL_FLOAT, sizeof (facevert_t), faceverts[0].st);

					glClientActiveTexture (GL_TEXTURE1);
					glEnableClientState (GL_TEXTURE_COORD_ARRAY);
					glTexCoordPointer (2, GL_FLOAT, sizeof (facevert_t), faceverts[0].st);

					// bring up lightmaps TMU
					glActiveTexture (GL_TEXTURE0);
					glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
					glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PRIMARY_COLOR);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
					glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);

					glActiveTexture (GL_TEXTURE1);
					glEnable (GL_TEXTURE_2D);
					glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

					StateInit = true;
				}

				// rather than lightmapping these, we'll send them through R_LightPoint so that they
				// get the correct ambient light for the world.  we need to hack/unhack the origin as it's in the floor :(
				currententity->origin[2] += 8.0f;
				R_LightPoint (currententity->origin, shadelight);
				currententity->origin[2] -= 8.0f;

				glColor3f (shadelight[0], shadelight[1], shadelight[2]);

				glLoadIdentity ();
				R_RotateForEntity (currententity);
				glGetFloatv (GL_MODELVIEW_MATRIX, matrix);
				glLoadMatrixf (r_world_matrix);

				clmodel = currententity->model;
				surf = &clmodel->surfaces[clmodel->firstmodelsurface];

				// don't bother doing culling or rotations or anything like that, as this is just a simple 6-sided box
				for (j = 0; j < clmodel->nummodelsurfaces; j++, surf++)
				{
					tex = R_TextureAnimation (surf->texinfo->texture);

					// there's no point in chaining these as they change pretty often
					glActiveTexture (GL_TEXTURE0);
					glBindTexture (GL_TEXTURE_2D, tex->gltexture->texnum);

					if (tex->gltexture->lumatexnum)
					{
						glActiveTexture (GL_TEXTURE1);
						glBindTexture (GL_TEXTURE_2D, tex->gltexture->lumatexnum);
					}
					else
					{
						glActiveTexture (GL_TEXTURE1);
						glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);
					}

					R_TransformSurface (surf, matrix);
					DrawGLPoly (surf->polys);
				}
			}
		}
	}

	if (StateInit)
	{
		glActiveTexture (GL_TEXTURE1);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable (GL_TEXTURE_2D);

		// GL_RGB_SCALE needs to be explicitly reset.  Even though it's specified to only work with GL_COMBINE,
		// some implementations can be a little lax and set it for all values of GL_TEXTURE_ENV_MODE.
		// I've knocked ATI enough in the past, so this time it's "Hello NVidia!!!"
		glActiveTexture (GL_TEXTURE0);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);

		glClientActiveTexture (GL_TEXTURE0);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);

		glDisableClientState (GL_VERTEX_ARRAY);
	}

	glColor3f (1, 1, 1);

	glLoadMatrixf (r_world_matrix);
}


extern gltexture_t *r_bloomtexture;
extern unsigned int bloomdata[];
unsigned int bloomblur[BLOOM_SIZE * BLOOM_SIZE];
void R_DrawWaterSurfaces (qboolean Bloom);

extern cvar_t r_bloom;

void R_CaptureBloom (void)
{
	texture_t *t;
	msurface_t *surf;

	render_bloom = true;

	if (r_bloom.value < 1) return;
	if (!render_bloom) return;

	// to do - if there's no lumas, lava or slime, we don't need to bloom!!!
	R_SetupGL3DViewport (BLOOM_SIZE, BLOOM_SIZE);

	// hack the depthrange to prevent having to clear
	glDepthRange (0.8, 1.0);

	// bring up vertex pointer
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (facevert_t), faceverts[0].v);

	// bring up image texture arrays
	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (facevert_t), faceverts[0].st);

	for (t = texturelist; t; t = t->texturelist)
	{
		// no surfs!
		if (!(surf = t->texturechain)) continue;

		// draw these later
		if ((surf->flags & SURF_DRAWSKY)) continue;
		if ((surf->flags & SURF_DRAWTURB)) continue;

		if (t->gltexture->lumatexnum)
			glBindTexture (GL_TEXTURE_2D, t->gltexture->lumatexnum);
		else
			glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);

		for (; surf; surf = surf->texturechain)
		{
			if ((surf->flags & SURF_UNDERWATER) || surf->polys->hazemask)
				DrawGLWarpPoly (surf->polys);
			else DrawGLPoly (surf->polys);
		}
	}

	glClientActiveTexture (GL_TEXTURE0);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glDisableClientState (GL_VERTEX_ARRAY);

	R_DrawWaterSurfaces (true);

	// read it in
	glReadPixels (0, 0, BLOOM_SIZE, BLOOM_SIZE, GL_RGBA, GL_UNSIGNED_BYTE, bloomdata);

	// restore depthrange
	glDepthRange (0.1, 0.8);
}


// fast blur filter - adapted from pseudocode at http://www.blackpawn.com/texts/blur/default.html
int divby[BLOOM_SIZE];

void BlurHorizontal (unsigned int *src, unsigned int *dst, int radius)
{
	int x, y;
	int kx;
	int total[3];
	int row;
	int col;

	for (y = 0; y < BLOOM_SIZE; y++)
	{
		total[0] = total[1] = total[2] = 0;
		row = y * BLOOM_SIZE;

		// start at zero because we can't have negative x!!!
		for (kx = 0; kx <= radius; kx++)
		{
			col = row + kx;

			total[0] += (int) ((unsigned char *) &src[col])[0];
			total[1] += (int) ((unsigned char *) &src[col])[1];
			total[2] += (int) ((unsigned char *) &src[col])[2];
		}

		((unsigned char *) &dst[row])[0] = total[0] / divby[0];
		((unsigned char *) &dst[row])[1] = total[1] / divby[0];
		((unsigned char *) &dst[row])[2] = total[2] / divby[0];

		for (x = 1; x < BLOOM_SIZE; x++)
		{
			// don't bother if too small or too large
			if (x - radius - 1 >= 0)
			{
				col = row + x - radius - 1;

				total[0] -= (int) ((unsigned char *) &src[col])[0];
				total[1] -= (int) ((unsigned char *) &src[col])[1];
				total[2] -= (int) ((unsigned char *) &src[col])[2];
			}

			if (x + radius < BLOOM_SIZE)
			{
				col = row + x + radius;

				total[0] += (int) ((unsigned char *) &src[col])[0];
				total[1] += (int) ((unsigned char *) &src[col])[1];
				total[2] += (int) ((unsigned char *) &src[col])[2];
			}

			((unsigned char *) &dst[row + x])[0] = total[0] / divby[x];
			((unsigned char *) &dst[row + x])[1] = total[1] / divby[x];
			((unsigned char *) &dst[row + x])[2] = total[2] / divby[x];
		}
	}
}


void Transpose (unsigned int *Image)
{
	int x, y;
	unsigned int tmp;

	for (y = 0; y < BLOOM_SIZE; y++)
	{
		for (x = y + 1; x < BLOOM_SIZE; x++)
		{
			tmp = Image[y * BLOOM_SIZE + x];
			Image[y * BLOOM_SIZE + x] = Image[x * BLOOM_SIZE + y];
			Image[x * BLOOM_SIZE + y] = tmp;
		}
	}
}


void DoBlur (unsigned int *src, unsigned int *dst, int radius)
{
	int i, j;

	// division factors for handling edge cases smoothly
	// need to recalc each time because of variable radius
	for (i = BLOOM_SIZE - 1, j = 0; ; i--, j++)
	{
		if (j > i) break;

		divby[i] = radius + 1 + j;
		divby[j] = radius + 1 + j;

		if (divby[i] > radius * 2 + 1) divby[i] = radius * 2 + 1;
		if (divby[j] > radius * 2 + 1) divby[j] = radius * 2 + 1;
	}

	// blur horizontally into dst
	BlurHorizontal (src, dst, radius);

	// flip dst
	Transpose (dst);

	// blur horizontally back into src
	BlurHorizontal (dst, src, radius);

	// flip src
	Transpose (src);
}


void R_RenderBloom (void)
{
	int i;
	int baseradius = 2;

	if (r_bloom.value < 1) return;
	if (!render_bloom) return;

	glViewport (glx, gly, glwidth, glheight);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();

	glOrtho  (0, BLOOM_SIZE, BLOOM_SIZE, 0, -99999, 99999);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	glBindTexture (GL_TEXTURE_2D, r_bloomtexture->texnum);

	for (i = 0; i < (int) r_bloom.value; i++)
	{
		DoBlur (bloomdata, bloomblur, baseradius);
		baseradius++;
	}

	// scale on the gpu - opengl does scale before bias so double the bias.  bias is required because black texture
	// scales upwards to dark grey somewhere along the line (probably palette mangling... :( )
	glPixelTransferf (GL_RED_BIAS, -0.02f);
	glPixelTransferf (GL_RED_SCALE, 2.0f);

	glPixelTransferf (GL_GREEN_BIAS, -0.02f);
	glPixelTransferf (GL_GREEN_SCALE, 2.0f);

	glPixelTransferf (GL_BLUE_BIAS, -0.02f);
	glPixelTransferf (GL_BLUE_SCALE, 2.0f);

	glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, BLOOM_SIZE, BLOOM_SIZE, GL_RGBA, GL_UNSIGNED_BYTE, bloomdata);

	glPixelTransferf (GL_RED_BIAS, 0.0f);
	glPixelTransferf (GL_RED_SCALE, 1.0f);

	glPixelTransferf (GL_GREEN_BIAS, 0.0f);
	glPixelTransferf (GL_GREEN_SCALE, 1.0f);

	glPixelTransferf (GL_BLUE_BIAS, 0.0f);
	glPixelTransferf (GL_BLUE_SCALE, 1.0f);

	glEnable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE);

	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	glVertex2f (0, 0);

	glTexCoord2f (1, 1);
	glVertex2f (BLOOM_SIZE, 0);

	glTexCoord2f (1, 0);
	glVertex2f (BLOOM_SIZE, BLOOM_SIZE);

	glTexCoord2f (0, 0);
	glVertex2f (0, BLOOM_SIZE);

	glEnd ();

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);

	// go back to 3d
	R_SetupGL3DViewport (glwidth, glheight);
}


/*
=============
R_DrawWorld
=============
*/
void R_ClearSkyBox (void);

void R_DrawWorld (void)
{
	entity_t	ent;

	memset (&ent, 0, sizeof (ent));
	ent.model = cl.worldmodel;

	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;

	glColor3f (1, 1, 1);

	R_ClearTextureChains ();
	R_ClearSkyBox ();

	// no liquid in the pvs
	liquidpvs = false;

	// no bloom
	render_bloom = false;

	// no flags
	globalflags = 0;

	// do before the surface pass so we know if we have to draw haze
	R_MarkLava ();

	// mark visible surfs and accumulate into texture chains
	R_RecursiveWorldNode (cl.worldmodel->nodes);

	// left till here as we need the surface visibility info
	R_PushDlights (cl.worldmodel->nodes);

	// add in inline bmodels
	R_AddBModelsToDrawLists ();

	// now that we have the full drawlists and visibility info, we add in the dynamic lights
	R_RenderDynamicLightmaps ();

	// calculate heathaze warps
	R_SurfaceHazeWarp ();

	// re-upload any lightmaps that were modified
	R_BlendLightmaps ();

	// capture the bloom
	R_CaptureBloom ();

	// create our viewport for rendering
	R_SetupGL3DViewport (glwidth, glheight);

	// draw everything in the world
	R_DrawTextureChains ();

	// render instanced bmodels
	R_RenderBModelInstances ();

	// draw sky last so as to let opengl clip away any of it that z-fails
	if (skychain)
		R_DrawSkyChain (true);
	else R_DrawSkyChain (false);

	//R_RenderBloom ();
}


/*
===============
R_MarkLeaves
===============
*/
void R_AddLeafToPVS (mleaf_t *leaf)
{
	byte	*vis;
	mnode_t	*node;
	int		i;

	vis = Mod_LeafPVS (leaf, cl.worldmodel);

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->visframe == r_visframecount)
					break;

				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}


extern cvar_t gl_lockpvs;

void R_MarkLeaves (void)
{
	if (r_oldviewleaf == r_viewleaf) return;
	if (gl_lockpvs.value) return;

	r_visframecount++;
	r_oldviewleaf = r_viewleaf;

	R_AddLeafToPVS (r_viewleaf);

	// no old viewleaf
	if (!r_oldviewleaf) return;

	// same old viewleaf contents
	if (r_viewleaf->contents == r_oldviewleaf->contents) return;

	// smoother transitions between viewleafs: improves contents changes/teleports/etc
	R_AddLeafToPVS (r_oldviewleaf);
}



mvertex_t	*r_pcurrentvertbase;
model_t		*currentmodel;

int	nColinElim;


void MakeTextureST (msurface_t *surf, vec3_t vec, float *st)
{
	float s;
	float t;

	s = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
	s /= surf->texinfo->texture->width;

	t = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
	t /= surf->texinfo->texture->height;

	st[0] = s;
	st[1] = t;
}


/*
================
BuildSurfaceDisplayList
================
*/
vec3_t farmin = {999999999, 999999999, 999999999};
vec3_t farmax = {-999999999, -999999999, -999999999};

void BuildSurfaceDisplayList (msurface_t *surf)
{
	int			i, j, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	float		*vec;
	glpoly_t	*poly;
	vec3_t		mids;
	vec3_t		mins = {999999, 999999, 999999};
	vec3_t		maxs = {-999999, -999999, -999999};
	vec3_t		dists;
	float		mindist = 9999999999;

	// reconstruct the polygon
	pedges = currentmodel->edges;
	lnumverts = surf->numedges;

	// draw texture
	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->verts = Hunk_Alloc (sizeof (glvertex_t) * (lnumverts + 2));
	poly->next = surf->polys;
	poly->flags = surf->flags;
	surf->polys = poly;
	poly->numverts = lnumverts;

	mids[0] = mids[1] = mids[2] = 0;

	for (i = 1; i <= lnumverts; i++)
	{
		lindex = currentmodel->surfedges[surf->firstedge + (i - 1)];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = r_pcurrentvertbase[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = r_pcurrentvertbase[r_pedge->v[1]].position;
		}

		// accumulate the midpoints
		mids[0] += vec[0];
		mids[1] += vec[1];
		mids[2] += vec[2];

		// get min and max
		for (j = 0; j < 3; j++)
		{
			if (vec[j] < mins[j]) mins[j] = vec[j];
			if (vec[j] > maxs[j]) maxs[j] = vec[j];
		}

		// make the verts
		VectorCopy (vec, poly->verts[i].v);
		MakeTextureST (surf, vec, poly->verts[i].st);
		MakeLightmapST (surf, vec, poly->verts[i].lm);

		// check against far clipping plane dimensions
		if (poly->verts[i].v[0] < farmin[0]) farmin[0] = poly->verts[i].v[0];
		if (poly->verts[i].v[1] < farmin[1]) farmin[1] = poly->verts[i].v[1];
		if (poly->verts[i].v[2] < farmin[2]) farmin[2] = poly->verts[i].v[2];

		if (poly->verts[i].v[0] > farmax[0]) farmax[0] = poly->verts[i].v[0];
		if (poly->verts[i].v[1] > farmax[1]) farmax[1] = poly->verts[i].v[1];
		if (poly->verts[i].v[2] > farmax[2]) farmax[2] = poly->verts[i].v[2];
	}

	// take the average as the midpoint
	mids[0] /= (float) lnumverts;
	mids[1] /= (float) lnumverts;
	mids[2] /= (float) lnumverts;

	// copy first vertex to last
	memcpy (&poly->verts[i], &poly->verts[1], sizeof (glvertex_t));

	// set the midpoint verts - ALSO COPY TO MIDPOINT FOR HEAT HAZE
	VectorCopy (mids, poly->verts[0].v);
	VectorCopy (mids, poly->midpoint);
	MakeTextureST (surf, mids, poly->verts[0].st);
	MakeLightmapST (surf, mids, poly->verts[0].lm);

	// each poly now has 2 more verts
	poly->numverts = lnumverts + 2;

	// get the warp scale
	dists[0] = fabs (maxs[0] - mins[0]);
	dists[1] = fabs (maxs[1] - mins[1]);
	dists[2] = fabs (maxs[2] - mins[2]);

	if (dists[0] < mindist && dists[0] > 0) mindist = dists[0];
	if (dists[1] < mindist && dists[1] > 0) mindist = dists[1];
	if (dists[2] < mindist && dists[2] > 0) mindist = dists[2];

	// should never happen but best to be safe just in case...
	if (mindist > 9999999998) mindist = 0;

	// 8 is used for a 64 x 64 surf so taking a sqrt of mindist should do it.
	poly->warpscale = sqrt (mindist);
	poly->hazemask = false;
	poly->warpfade = 0;

	// allocate vertexes
	if (!faceverts)
	{
		// take at least 9 to begin with and grow as required
		maxsurfverts = poly->numverts > 9 ? poly->numverts : 9;
		faceverts = (facevert_t *) malloc (sizeof (facevert_t) * maxsurfverts);
	}
	else if (poly->numverts > maxsurfverts)
	{
		// go to the new max
		maxsurfverts = poly->numverts;

		// reallocate
		free (faceverts);
		faceverts = (facevert_t *) malloc (sizeof (facevert_t) * maxsurfverts);
	}

	// copy v to tv
	for (i = 0; i < poly->numverts; i++)
	{
		poly->verts[i].tv[0] = poly->verts[i].v[0];
		poly->verts[i].tv[1] = poly->verts[i].v[1];
		poly->verts[i].tv[2] = poly->verts[i].v[2];
	}
}


// for satisfactorily reliable determination of whether a bmodel is underwater
void GetBModelMidPoints (void)
{
	int i;
	int j;
	model_t *m;
	msurface_t *surf;
	float numsurfs;

	for (j = 1; j < MAX_MODELS; j++)
	{
		m = cl.model_precache[j];
		if (!m) break;

		if (m->type != mod_brush) continue;

		surf = &m->surfaces[m->firstmodelsurface];
		m->midpoint[0] = m->midpoint[1] = m->midpoint[2];

		numsurfs = 0.0f;

		for (i = 0; i < m->nummodelsurfaces; i++, surf++)
		{
			if (surf->flags & SURF_DRAWSKY) continue;
			if (surf->flags & SURF_DRAWTURB) continue;

			// verts 0 represents the midpoint always
			m->midpoint[0] += surf->polys[0].verts[0].v[0];
			m->midpoint[1] += surf->polys[0].verts[0].v[1];
			m->midpoint[2] += surf->polys[0].verts[0].v[2];

			numsurfs += 1.0f;
		}

		m->midpoint[0] /= numsurfs;
		m->midpoint[1] /= numsurfs;
		m->midpoint[2] /= numsurfs;
	}
}


void R_SetWorldVolumes (void)
{
}


