/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 
// r_misc.c

#include "gl_local.h"

// MH - detail textures begin
image_t *r_detailtexture;
image_t *r_cubemap;
image_t *r_bloomtexture;
image_t *r_hazetexture;
image_t *r_blacktexture;
image_t *r_nonormaltexture;
image_t *r_particlefont;
image_t *r_crosshairfont;


static void transform_point (float out[4], const float m[16], const float in[4])
{
#define MATRIX(row,col)  m[col * 4 + row]
	out[0] = MATRIX(0, 0) * in[0] + MATRIX(0, 1) * in[1] + MATRIX(0, 2) * in[2] + MATRIX(0, 3) * in[3];
	out[1] = MATRIX(1, 0) * in[0] + MATRIX(1, 1) * in[1] + MATRIX(1, 2) * in[2] + MATRIX(1, 3) * in[3];
	out[2] = MATRIX(2, 0) * in[0] + MATRIX(2, 1) * in[1] + MATRIX(2, 2) * in[2] + MATRIX(2, 3) * in[3];
	out[3] = MATRIX(3, 0) * in[0] + MATRIX(3, 1) * in[1] + MATRIX(3, 2) * in[2] + MATRIX(3, 3) * in[3];
#undef MATRIX
}


// version of gluProject from the mesa library converted to use vec3_t's as it's in and out
void mhqgluProject (vec3_t vin, const float modelMatrix[16], const float projMatrix[16], const int viewport[4], vec3_t vout)
{
	float in[4];
	float out[4];

	// save to a local cos we will be modifying it here
	in[0] = vin[0];
	in[1] = vin[1];
	in[2] = vin[2];
	in[3] = 1.0;

	transform_point (out, modelMatrix, in);
	transform_point (in, projMatrix, out);

	// could not project
	if (in[3] == 0.0) return;

	in[0] /= in[3];
	in[1] /= in[3];
	in[2] /= in[3];

	// transform to viewport coords and store back to vout
	vout[0] = viewport[0] + (1 + in[0]) * viewport[2] / 2;
	vout[1] = viewport[1] + (1 + in[1]) * viewport[3] / 2;
	vout[2] = (1 + in[2]) / 2;
}


// wrapper for glDrawBuffer that saves our buffer to the global state structure so that we can
// easily retrieve it without doing a glGet...
void mhglDrawBuffer (GLenum CurrentBuffer)
{
	gl_state.CurrentBuffer = CurrentBuffer;
	glDrawBuffer (CurrentBuffer);
}

// blatantly plagiarized darkplaces code
// no bound in Q2.
// this would be better as a #define but it's not runtime so it doesn't matter
int bound (int smallest, int val, int biggest) 
{
	if (val < smallest) return smallest;
	if (val > biggest) return biggest;

	return val;
}


void fractalnoise (byte *noise, int size, int startgrid) 
{
	int x, y, g, g2, amplitude, min, max, size1 = size - 1, sizepower, gridpower;
	int *noisebuf;
#define n(x,y) noisebuf[((y) &size1) *size + ((x) &size1)]

	for (sizepower = 0; (1 << sizepower) < size;sizepower++);
	if (size != (1 << sizepower)) 
		Sys_Error ("fractalnoise: size must be power of 2\n");

	for (gridpower = 0; (1 << gridpower) < startgrid;gridpower++);
	if (startgrid != (1 << gridpower)) 
		Sys_Error ("fractalnoise: grid must be power of 2\n");

	startgrid = bound (0, startgrid, size);

	amplitude = 0xFFFF;// this gets halved before use
	noisebuf = malloc (size*size*sizeof (int));
	memset (noisebuf, 0, size*size*sizeof (int));

	for (g2 = startgrid;g2;g2 >>= 1) 
	{
		// brownian motion (at every smaller level there is random behavior) 
		amplitude >>= 1;
		for (y = 0;y < size;y += g2) 
			for (x = 0;x < size;x += g2) 
				n (x,y) += (rand () &amplitude);

		g = g2 >> 1;
		if (g) 
		{
			// subdivide, diamond - square algorithm (really this has little to do with squares) 
			// diamond
			for (y = 0;y < size;y += g2) 
				for (x = 0;x < size;x += g2) 
					n (x + g,y + g) = (n (x,y) + n (x + g2,y) + n (x,y + g2) + n (x + g2,y + g2)) >> 2;
			// square
			for (y = 0;y < size;y += g2) 
				for (x = 0;x < size;x += g2) 
				{
					n (x + g,y) = (n (x,y) + n (x + g2,y) + n (x + g,y - g) + n (x + g,y + g)) >> 2;
					n (x,y + g) = (n (x,y) + n (x,y + g2) + n (x - g,y + g) + n (x + g,y + g)) >> 2;
				}
		}
	}
	// find range of noise values
	min = max = 0;
	for (y = 0;y < size;y++) 
		for (x = 0;x < size;x++) 
		{
			if (n (x,y) < min) min = n (x,y);
			if (n (x,y) > max) max = n (x,y);
		}
		max -= min;
		max++;
		// normalize noise and copy to output
		for (y = 0;y < size;y++) 
			for (x = 0;x < size;x++) 
				*noise++= (byte) (((n (x,y) - min) * 256) / max);
		free (noisebuf);
#undef n
}


void R_BuildDetailTexture (void) 
{
	int x, y, light;
	float vc[3], vx[3], vy[3], vn[3], lightdir[3];
	unsigned int tot = 0;

	// increase this if you want
#define DETAILRESOLUTION 256

	byte data[DETAILRESOLUTION][DETAILRESOLUTION][4], noise[DETAILRESOLUTION][DETAILRESOLUTION];

	// this looks odd, but it's necessary cos Q2's uploader will lightscale the texture, which
	// will cause all manner of unholy havoc. So we need to run it through GL_LoadPic before
	// we even fill in the data just to fill in the rest of the image_t struct, then we'll
	// build the texture for OpenGL manually later on.
	r_detailtexture = GL_LoadPic ("***detail***", (byte *) data, DETAILRESOLUTION, DETAILRESOLUTION, it_wall, 32);

	lightdir[0] = 0.5;
	lightdir[1] = 1;
	lightdir[2] = - 0.25;
	VectorNormalize (lightdir);

	fractalnoise (&noise[0][0], DETAILRESOLUTION, DETAILRESOLUTION >> 4);

	for (y = 0;y < DETAILRESOLUTION;y++) 
	{
		for (x = 0;x < DETAILRESOLUTION;x++) 
		{
			vc[0] = x;
			vc[1] = y;
			vc[2] = noise[y][x]* (1.0f / 32.0f);
			vx[0] = x + 1;
			vx[1] = y;
			vx[2] = noise[y][(x + 1) % DETAILRESOLUTION]* (1.0f / 32.0f);
			vy[0] = x;
			vy[1] = y + 1;
			vy[2] = noise[(y + 1) % DETAILRESOLUTION][x]* (1.0f / 32.0f);
			VectorSubtract (vx, vc, vx);
			VectorSubtract (vy, vc, vy);
			CrossProduct (vx, vy, vn);
			VectorNormalize (vn);
			light = 128 - DotProduct (vn, lightdir) * 128;
			light = bound (0, light, 255);

			data[y][x][0] = data[y][x][1] = data[y][x][2] = light;
			data[y][x][3] = 255;

			// accumulate total
			tot += light;
		}
	}

	// compute average into tot
	tot /= (DETAILRESOLUTION * DETAILRESOLUTION);

	// rescale for an average of 128
	for (y = 0;y < DETAILRESOLUTION;y++) 
	{
		for (x = 0;x < DETAILRESOLUTION;x++) 
		{
			float fdata = data[y][x][0];

			fdata *= 128;
			fdata /= (float) tot;

			if (fdata < 0) fdata = 0;
			if (fdata > 255) fdata = 255;

			// rescale to 64 - 192 so that extremes don't fuck up the texture
			fdata *= 0.5;
			fdata += 56;

			data[y][x][0] = (byte) fdata + (rand () % 16);
			data[y][x][1] = (byte) fdata + (rand () % 16);
			data[y][x][2] = (byte) fdata + (rand () % 16);
		}
	}

	// now we build the texture manually. you can reuse this code for auto mipmap generation
	// in other contexts if you wish. defines are in gl.h
	// first, bind the texture. probably not necessary, but it seems like good practice
	glBindTexture (GL_TEXTURE_2D, r_detailtexture->texnum);

	// set the hint for mipmap generation
	glHint (GL_GENERATE_MIPMAP_HINT, GL_NICEST);

	// now switch mipmap generation on.
	// this has been part of the core OpenGL specification since version 1.4, so there's 
	// no real need to check for extension support and other tiresomely silly blather,
	// unless you have a really weird OpenGL implementation. it does, however, even
	// work on ATI cards, so you should have no trouble with it.
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

	// upload the correct texture data without any lightscaling interference
	glTexImage2D (GL_TEXTURE_2D, 
		0, 
		GL_RGBA, 
		DETAILRESOLUTION, 
		DETAILRESOLUTION, 
		0, 
		GL_RGBA, 
		GL_UNSIGNED_BYTE, 
		(byte *) data);

	// sexy filtering
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}
// MH - detail textures end



#define CUBEMAP_MAXSIZE 32
void MakeSpecularAttenuationMap (byte *pixels)
{
	int x, y;
	int LRADIUS = CUBEMAP_MAXSIZE / 2;

	for (y = 0; y < CUBEMAP_MAXSIZE; y++)
	{
		for (x = 0; x < CUBEMAP_MAXSIZE; x++)
		{
			// find x and y as a -LRADIUS to LRADIUS scale
			int TheRX = x - LRADIUS;
			int TheRY = y - LRADIUS;
			float TheXResult;
			float TheYResult;
			float TheFinalResult;
			unsigned char OutResult;

			// divide X by LRADIUS and square it
			TheXResult = ((float) TheRX / (float) LRADIUS);
			TheXResult *= TheXResult;

			// divide Y by LRADIUS and square it
			TheYResult = ((float) TheRY / (float) LRADIUS);
			TheYResult *= TheYResult;

			// add the 2
			TheFinalResult = TheXResult + TheYResult;

			// multiply by 255 for a byte intensity range
			TheFinalResult = TheFinalResult * 255;

			// clamp 0 to 255
			if (TheFinalResult < 0) TheFinalResult = 0;
			if (TheFinalResult > 255) TheFinalResult = 255;

			// store in an unsigned char now - we want the negative of what is generated above
			OutResult = 255 - (unsigned char) TheFinalResult;

			// write to output
			pixels[0] = OutResult;
			pixels[1] = OutResult;
			pixels[2] = OutResult;
			pixels[3] = 255;

			pixels += 4;
		}
	}
}


static void getCubeVector (int i, int cubesize, int x, int y, float *vector) 
{
	float s, t, sc, tc, mag;

	s = ((float) x + 0.5) / (float) cubesize;
	t = ((float) y + 0.5) / (float) cubesize;
	sc = s * 2.0 - 1.0;
	tc = t * 2.0 - 1.0;

	// cleaned manky braces again here...
	switch (i) 
	{
	case 0: vector[0] = 1.0; vector[1] = - tc; vector[2] = - sc; break;
	case 1: vector[0] = - 1.0; vector[1] = - tc; vector[2] = sc; break;
	case 2: vector[0] = sc; vector[1] = 1.0; vector[2] = tc; break;
	case 3: vector[0] = sc; vector[1] = - 1.0; vector[2] = - tc; break;
	case 4: vector[0] = sc; vector[1] = - tc; vector[2] = 1.0; break;
	case 5: vector[0] = - sc; vector[1] = - tc; vector[2] = - 1.0; break;
	}

	mag = 1.0 / sqrt (vector[0]* vector[0]+ vector[1]* vector[1]+ vector[2]* vector[2]);

	vector[0]*= mag;
	vector[1]*= mag;
	vector[2]*= mag;
}


void R_InitCubemapTextures (void) 
{
	// this texture doesn't need to be huge. I've seen examples where it goes up to 128 or higher, but 32
	// is perfectly adequate. Who am I to argue with Mr. Kilgard?
	float vector[3];
	int i, x, y;
	GLubyte pixels[CUBEMAP_MAXSIZE * CUBEMAP_MAXSIZE * 4];

	// set up a dummy for it to get a valid image_t filled in.  we'll be deleting and recreating the texture data
	// using glTexImage2D when we actually set up each individual one.  we generate it as a wall texture to ensure that
	// we get texnums generated for the backup textures as well
	r_cubemap = GL_LoadPic ("***cubemap***", (byte *) pixels, CUBEMAP_MAXSIZE, CUBEMAP_MAXSIZE, it_pic, 32);

	// switch the texture state
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_CUBE_MAP);

	// bind to the regular texture number
	glBindTexture (GL_TEXTURE_CUBE_MAP, r_cubemap->texnum);

	// set up the texture parameters - did a clamp on R as well...
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// this is one area where i'd gladly argue with anyone.
	// corrected the vile braces and added better whitespace...
	// we only go to 5 here - see comment below
	// back to 6 again - see comment below
	for (i = 0; i < 6; i++) 
	{
		for (y = 0; y < CUBEMAP_MAXSIZE; y++) 
		{
			for (x = 0; x < CUBEMAP_MAXSIZE; x++) 
			{
				getCubeVector (i, CUBEMAP_MAXSIZE, x, y, vector);

				pixels[4 * (y * CUBEMAP_MAXSIZE + x) + 0] = 128 + 127 * vector[0];
				pixels[4 * (y * CUBEMAP_MAXSIZE + x) + 1] = 128 + 127 * vector[1];
				pixels[4 * (y * CUBEMAP_MAXSIZE + x) + 2] = 128 + 127 * vector[2];
			}
		}

		// cube map me baybeee
		glTexImage2D (GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGBA, CUBEMAP_MAXSIZE, CUBEMAP_MAXSIZE, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	}

	/*
	// MH - legacy specular cubemap code.  to use, change the outermost loop above to 5, uncomment this block, and uncomment the call to
	// R_DrawSpecularMaps in gl_rsurf.c
	// this code is only incomplete in that i hadn't yet put it on alias models when i removed it.  it accurately enough replicates
	// the specular mapping algorithm from player perspective, but just looked plain wrong.  global static specular maps gave a much
	// better result, so i stuck with those instead.  cos sometimes the right way of doing something ain't always the *RIGHT* way,
	// if ya know what i mean...
	// here we get to cheat and hack like absolute crazy.  ain't life great?
	// rather than waste texture memory with 2 cubemaps, let's just build 1, but put the face that is always pointing away from
	// the camera with bumpmaps as the specular attenuation map.  then when drawing, we can rotate the texture matrix so that
	// this face is pointing *towards* the camera.  this way we also ensure that none of our cubemap faces get wasted.
	MakeSpecularAttenuationMap (pixels);
	glTexImage2D (GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, GL_RGBA, CUBEMAP_MAXSIZE, CUBEMAP_MAXSIZE, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	*/

	// done - restore the texture state
	glDisable (GL_TEXTURE_CUBE_MAP);
	glEnable (GL_TEXTURE_2D);
}
#undef CUBEMAP_MAXSIZE


/*
 ================== 
R_InitParticleTexture
 ================== 
*/
byte	dottexture[8][8] = 
{
	{0,0,0,0,0,0,0,0},
	{0,0,1,1,0,0,0,0},
	{0,1,1,1,1,0,0,0},
	{0,1,1,1,1,0,0,0},
	{0,0,1,1,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
};

byte	badtexture[16][16] =
{
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,1,1,0,0,0,1,0,0,1,1,0,0,0,0,0},
	{0,1,0,1,0,1,0,1,0,1,0,1,0,0,0,0},
	{0,1,1,0,0,1,0,1,0,1,0,1,0,0,0,0},
	{0,1,0,1,0,1,1,1,0,1,0,1,0,0,0,0},
	{0,1,1,0,0,1,0,1,0,1,1,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
	{1,0,0,1,1,1,0,1,1,0,0,1,1,1,1,1},
	{1,0,1,0,1,0,1,0,1,0,1,0,1,1,1,1},
	{1,0,0,1,1,0,1,0,1,0,1,0,1,1,1,1},
	{1,0,1,0,1,0,0,0,1,0,1,0,1,1,1,1},
	{1,0,0,1,1,0,1,0,1,0,0,1,1,1,1,1},
	{1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
	{1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
};


void R_InitParticleTexture (void) 
{
	int		x, y;
	byte	data[8][8][4];
	byte	baddata[16][16][4];
	byte	bloom[BLOOM_SIZE * BLOOM_SIZE * 4];

	// particle texture
	for (x = 0; x < 8; x++) 
	{
		for (y = 0; y < 8; y++) 
		{
			data[y][x][0] = 255;
			data[y][x][1] = 255;
			data[y][x][2] = 255;
			data[y][x][3] = dottexture[x][y] * 255;
		}
	}

	r_particletexture = GL_LoadPic ("***particle***", (byte *) data, 8, 8, it_sprite, 32);

	// also use this for bad textures, but without alpha
	for (x = 0; x < 16; x++) 
	{
		for (y = 0; y < 16; y++) 
		{
			baddata[y][x][0] = badtexture[x][y] * 255;
			baddata[y][x][1] = 255 - badtexture[x][y] * 255;
			baddata[y][x][2] = 0;
			baddata[y][x][3] = 255;
		}
	}

	r_notexture = GL_LoadPic ("***r_notexture***", (byte *) baddata, 16, 16, it_wall, 32);

	// MH - detail textures begin
	// and for detail textures
	// this is just a convenient place to call it from...
	R_BuildDetailTexture ();
	// MH - detail textures end

	// this is a hack to get initial image struct for the bloom texture.  upload it as a pic so that it doesn't mipmap!!!
	r_bloomtexture = GL_LoadPic ("***bloom***", (byte *) bloom, BLOOM_SIZE, BLOOM_SIZE, it_pic, 32);

	// let's make double-sure that it doesn't mipmap...
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE);

	// now do a regular teximage
	memset (bloom, 0, BLOOM_SIZE * BLOOM_SIZE * 4);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGB, BLOOM_SIZE, BLOOM_SIZE, 0, GL_RGB, GL_UNSIGNED_BYTE, bloom);

	r_hazetexture = GL_LoadPic ("***heathaze***", (byte *) bloom, BLOOM_SIZE, BLOOM_SIZE, it_pic, 32);

	// let's make double-sure that it doesn't mipmap...
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE);

	// now do a regular teximage
	memset (bloom, 0, BLOOM_SIZE * BLOOM_SIZE * 4);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGB, HAZE_SIZE, HAZE_SIZE, 0, GL_RGB, GL_UNSIGNED_BYTE, bloom);

	// generate a "black" texture now for use with bloom mapping
	for (x = 0; x < 8; x++) 
	{
		for (y = 0; y < 8; y++) 
		{
			data[y][x][0] = 0;
			data[y][x][1] = 0;
			data[y][x][2] = 0;
			data[y][x][3] = 255;
		}
	}

	r_blacktexture = GL_LoadPic ("***black***", (byte *) data, 8, 8, it_wall, 32);

	// cubemaps for sfx
	R_InitCubemapTextures ();

	for (x = 0; x < 8; x++) 
	{
		for (y = 0; y < 8; y++) 
		{
			data[y][x][0] = 255;
			data[y][x][1] = 255;
			data[y][x][2] = 255;
			data[y][x][3] = 255;
		}
	}

	r_nonormaltexture = GL_LoadPic ("***nonormal***", (byte *) data, 8, 8, it_wall, 32);

	// load the resource bitmap textures now
	r_particlefont = GL_LoadBitmapResourceTexture (IDB_BITMAP1, true);
	r_crosshairfont = GL_LoadBitmapResourceTexture (IDB_BITMAP2, true);
}


/* 
 ============================================================================== 

SCREEN SHOTS 

 ============================================================================== 
*/ 

typedef struct _TargaHeader {
	unsigned char 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	unsigned char	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	unsigned char	pixel_size, attributes;
} TargaHeader;


/* 
 ================== 
GL_ScreenShot_f
 ================== 
*/
extern qboolean inScreenDumpMode;

void GL_ScreenShot_f (void) 
{
	byte		*buffer;
	char		checkname[MAX_OSPATH];
	int			i;
	FILE		*f;

	// create the scrnshots directory if it doesn't exist
	Com_sprintf (checkname, sizeof (checkname) , "%s/scrnshot", FS_Gamedir ());
	Sys_Mkdir (checkname);

	// find a file name to save it to 
	for (i = 0; i <= 9999; i++) 
	{
		Com_sprintf (checkname, sizeof (checkname) , "%s/scrnshot/quake%04i.tga", FS_Gamedir (), i);

		f = fopen (checkname, "rb");

		if (!f) break;	// file doesn't exist

		fclose (f);
	}

	if (i == 10000) 
	{
		VID_Printf (PRINT_ALL, "SCR_ScreenShot_f: Couldn't create a file\n"); 
		return;
	}


	buffer = malloc (vid.width * vid.height * 3 + 18);
	memset (buffer, 0, 18);
	buffer[2] = 2;		// uncompressed type
	buffer[12] = vid.width & 255;
	buffer[13] = vid.width >> 8;
	buffer[14] = vid.height & 255;
	buffer[15] = vid.height >> 8;
	buffer[16] = 24;	// pixel size

	// read as BGR
	glReadPixels (0, 0, vid.width, vid.height, GL_BGR, GL_UNSIGNED_BYTE, buffer + 18); 

	f = fopen (checkname, "wb");
	fwrite (buffer, 1, 18 + vid.width * vid.height * 3, f);
	fclose (f);

	free (buffer);

	// don't feed back when dumping
	if (!inScreenDumpMode) VID_Printf (PRINT_ALL, "Wrote %s\n", checkname);
}


/*
** GL_Strings_f
*/ 
void GL_Strings_f (void) 
{
	VID_Printf (PRINT_ALL, "GL_VENDOR: %s\n", gl_config.vendor_string);
	VID_Printf (PRINT_ALL, "GL_RENDERER: %s\n", gl_config.renderer_string);
	VID_Printf (PRINT_ALL, "GL_VERSION: %s\n", gl_config.version_string);
	VID_Printf (PRINT_ALL, "GL_EXTENSIONS: %s\n", gl_config.extensions_string);
}


float sintable[360];

#define DEG2RAD(a) (a * M_PI) / 180.0F

void InitGLSinTable (void)
{
	int i;

	for (i = 0; i < 360; i++)
	{
		sintable[i] = sin (DEG2RAD ((float) i));
	}
}

float Q_sin (float fin)
{
	return sintable[((int) fin) % 360];
	return 0;
}


/*
** GL_SetDefaultState
*/ 
void GL_SetDefaultState (void) 
{
	InitGLSinTable ();

	glClearColor (0, 0, 0, 1);
	glCullFace (GL_FRONT);
	glEnable (GL_TEXTURE_2D);

	glEnable (GL_ALPHA_TEST);
	glAlphaFunc (GL_GREATER, 0.666);

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glDisable (GL_BLEND);

	glColor4f (1,1,1,1);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel (GL_FLAT);

	GL_TextureMode (gl_texturemode->string);
	GL_TextureAlphaMode (gl_texturealphamode->string);
	GL_TextureSolidMode (gl_texturesolidmode->string);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	if (glPointParameterf) 
	{
		float attenuations[3];

		attenuations[0] = gl_particle_att_a->value;
		attenuations[1] = gl_particle_att_b->value;
		attenuations[2] = gl_particle_att_c->value;

		glEnable (GL_POINT_SMOOTH);
		glPointParameterf (GL_POINT_SIZE_MIN, gl_particle_min_size->value);
		glPointParameterf (GL_POINT_SIZE_MAX, gl_particle_max_size->value);
		glPointParameterfv (GL_POINT_DISTANCE_ATTENUATION, attenuations);
	}

	GL_UpdateSwapInterval ();
}

void GL_UpdateSwapInterval (void) 
{
	if (gl_swapinterval->modified) 
	{
		gl_swapinterval->modified = false;

		if (!gl_state.stereo_enabled) 
		{
#ifdef _WIN32
			if (qwglSwapInterval) 
				qwglSwapInterval (gl_swapinterval->value);
#endif
		}
	}
}