/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// texture management system
// this would probably make a nice C++ class, but I'm a dinosaur, so what the hey?


#include "quakedef.h"
#include "glext.h"
#include "resource.h"



// we keep everything except map textures which are potentially replacable
typedef struct
{
	int			texnum;
	char		identifier[64];

	int			width;
	int			height;

	qboolean	mipmap;
	qboolean	alpha;

	unsigned short crc;

	qboolean	reload;
} gltexture_t;


// MAX_MAP_TEXTURES is 512, double that cos each texture *could* potentially have a fullbright,
// MAX_MAP_ENTITIES is 1024, all of them aside from 2 (worldspawn and info_player_start)
// *could* be alias models (it's highly unlikely though), each of which could also have a
// fullbright, so add 2048 (to be safer), then add a bit for overhead for the 2D stuff
#define	MAX_GLTEXTURES	3328

gltexture_t	gltextures[MAX_GLTEXTURES];

int			numgltextures;

int detailtexture = -1;

int caustictexture = -1;

/*
==========================================================================================

					TEXTURE MANAGEMENT

==========================================================================================
*/

void WriteTGA (char *name, int width, int height, int bpp, byte *data);

void GLT_SetReload (qboolean reload)
{
	int i;

	for (i = 0; i < MAX_GLTEXTURES; i++)
	{
		// always reload at the beginning
		gltextures[i].reload = reload;
		gltextures[i].crc = 0;
		gltextures[i].identifier[0] = 0;
	}

	// no textures
	numgltextures = 0;
	texture_extension_number = 1;
}


void GL_TileTexture (unsigned int *in, unsigned int *out, int inWidth, int inHeight, int tileX, int tileY)
{
	int outPos = 0;
	int i, j, k, l;
	int Row;

	for (l = 0; l < tileY; l++)
	{
		for (k = 0, Row = 0; k < inHeight; k++, Row += inHeight)
		{
			for (j = 0; j < tileX; j++)
			{
				for (i = 0; i < inWidth; i++)
				{
					out[outPos++] = in[i + Row];
				}
			}
		}
	}
}


/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
	int		i;
	gltexture_t	*glt;

	for (i=0, glt=gltextures ; i<numgltextures ; i++, glt++)
	{
		if (!strcmp (identifier, glt->identifier))
			return gltextures[i].texnum;
	}

	return -1;
}


void GL_TexDump_f (void)
{
	int		i;
	gltexture_t	*glt;
	FILE *f;

	f = fopen ("texdump.txt", "w");

	for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
		fprintf (f, "%5i   %s\n", gltextures[i].texnum, glt->identifier);

	fclose (f);
}


#define BLUR_FILTER 0
#define LIGHT_BLUR	1
#define EDGE_FILTER 2

unsigned int FilterMatrix[][3][3] = 
{
	// regular blur
	{
		{1, 1, 1},
		{1, 1, 1},
		{1, 1, 1}
	},
	// light blur
	{
		{1, 1, 1},
		{1, 4, 1},
		{1, 1, 1}
	},
	// find edges
	{
		{-1, -1, -1},
		{-1, 8, -1},
		{-1, -1, -1}
	},
};


void R_FilterTexture (int filterindex, unsigned int *data, int width, int height)
{
	int i;
	int w, h;
	int pos[6];
	unsigned int pix[9];
	int filterweight = 0;
	unsigned int *temp;

	// calculate the filter weight - this is always the sum of the entries in the matrix, although it
	// *could* be a custom value which would have the potential effect of adjusting the brightness of
	// the texture as well as applying a filter.  I'm sticking with the proper average for now though.
	for (h = 0; h < 3; h++)
	{
		for (w = 0; w < 3; w++)
		{
			filterweight += FilterMatrix[filterindex][h][w];
		}
	}

	// don't allow division by zero
	if (!filterweight) filterweight = 1;

	// allocate a temp buffer
	temp = malloc (width * height * 4);

	// copy data to temp
	for (i = 0; i < (width * height); i++)
	{
		temp[i] = data[i];
	}

	// apply the filter
	for (h = 0; h < height; h++)
	{
		// row up
		pos[0] = h - 1;
		if (pos[0] < 0) pos[0] += height;

		// row down
		pos[2] = h + 1;
		if (pos[2] >= height) pos[2] -= height;

		// same row
		pos[4] = h;

		for (w = 0; w < width; w++)
		{
			unsigned int rgba[4] = {0, 0, 0, 0};

			// col left
			pos[1] = w - 1;
			if (pos[1] < 0) pos[1] += width;

			// col right
			pos[3] = w + 1;
			if (pos[3] >= width) pos[3] -= width;

			// same col
			pos[5] = w;

			// base the target pixels on the original data - the order doesn't matter
			pix[0] = data[pos[0] * width + pos[1]];
			pix[1] = data[pos[0] * width + pos[3]];
			pix[2] = data[pos[0] * width + pos[5]];
			pix[3] = data[pos[2] * width + pos[1]];
			pix[4] = data[pos[2] * width + pos[3]];
			pix[5] = data[pos[2] * width + pos[5]];
			pix[6] = data[pos[4] * width + pos[1]];
			pix[7] = data[pos[4] * width + pos[3]];
			pix[8] = data[pos[4] * width + pos[5]];

			// total the pixels
			for (i = 0; i < 9; i++)
			{
				rgba[0] += (unsigned int) ((byte *) &pix[i])[0] * FilterMatrix[filterindex][i / 3][i % 3];
				rgba[1] += (unsigned int) ((byte *) &pix[i])[1] * FilterMatrix[filterindex][i / 3][i % 3];
				rgba[2] += (unsigned int) ((byte *) &pix[i])[2] * FilterMatrix[filterindex][i / 3][i % 3];
				rgba[3] += (unsigned int) ((byte *) &pix[i])[3] * FilterMatrix[filterindex][i / 3][i % 3];
			}

			// set the new values
			((byte *) &temp[pos[4] * width + pos[5]])[0] = (byte) (rgba[0] / filterweight);
			((byte *) &temp[pos[4] * width + pos[5]])[1] = (byte) (rgba[0] / filterweight);
			((byte *) &temp[pos[4] * width + pos[5]])[2] = (byte) (rgba[0] / filterweight);
			((byte *) &temp[pos[4] * width + pos[5]])[3] = (byte) (rgba[0] / filterweight);
		}
	}

	// copy temp back to data
	for (i = 0; i < (width * height); i++)
	{
		data[i] = temp[i];
	}

	// release the temp buffer
	free (temp);
}


/*
===============
BuildIntensityTable

Got this from Quake II and if you don't like it it's just too bad
===============
*/
// we have a gamma table in Quake 1 already
static byte intensitytable[256];
static byte intensitytable2d[256];

void BuildIntensityTable (void)
{
	int i, j;

	for (i = 0; i < 256; i++)
	{
		// intensity table for world texes
		j = i * 2;
		if (j > 255) j = 255;

		intensitytable[i] = j;

		// intensity table for 2d view texes
		j = (i * 4) / 3;
		if (j > 255) j = 255;

		intensitytable2d[i] = j;
	}
}


/*
===============
GL_LightScaleTexture

Got this from Quake II and if you don't like it it's just too bad
===============
*/
void GL_LightScaleTexture (unsigned *texin, int inwidth, int inheight)
{
	int	i, c;
	byte *p;

	p = (byte *) texin;

	c = inwidth * inheight;

	for (i = 0; i < c; i++, p += 4)
	{
		p[0] = gammatable[intensitytable[p[0]]];
		p[1] = gammatable[intensitytable[p[1]]];
		p[2] = gammatable[intensitytable[p[2]]];
	}
}


/*
===============
GL_LightScale2DTexture

Similar to above but used for non-mipped texes, gives a less extreme scaling
===============
*/
void GL_LightScale2DTexture (unsigned *texin, int inwidth, int inheight)
{
	int	i, c;
	byte *p;

	p = (byte *) texin;

	c = inwidth * inheight;

	for (i = 0; i < c; i++, p += 4)
	{
		p[0] = gammatable[intensitytable2d[p[0]]];
		p[1] = gammatable[intensitytable2d[p[1]]];
		p[2] = gammatable[intensitytable2d[p[2]]];
	}
}


/*
===============
GL_ResampleTexture

Got this from Quake II, based on the assumption that the Quake II version of it *must*
necessarily be better (dunno if it actually is, can't really see much difference)
===============
*/
void GL_ResampleTexture (unsigned *texIn, int inwidth, int inheight, unsigned *texOut, int outwidth, int outheight)
{
	int i, j;
	unsigned *inrow, *inrow2;
	unsigned theFrac, fracstep;
	unsigned p1[1024], p2[1024];
	byte *pix1, *pix2, *pix3, *pix4;

	if (inwidth == outwidth && inheight == outheight)
	{
		// do a straight copy
		for (i = 0; i < inwidth * inheight; i++)
			texOut[i] = texIn[i];

		return;
	}

	fracstep = inwidth * 0x10000 / outwidth;
	theFrac = fracstep >> 2;

	for (i = 0; i < outwidth; i++)
	{
		p1[i] = 4 * (theFrac >> 16);
		theFrac += fracstep;
	}

	theFrac = 3 * (fracstep >> 2);

	for (i = 0; i < outwidth; i++)
	{
		p2[i] = 4 * (theFrac >> 16);
		theFrac += fracstep;
	}

	for (i = 0; i < outheight; i++, texOut += outwidth)
	{
		inrow = texIn + inwidth * (int) ((i + 0.25) * inheight / outheight);
		inrow2 = texIn + inwidth * (int) ((i + 0.75) * inheight / outheight);

		// this isn't actually *used* any more
		theFrac = fracstep >> 1;

		for (j = 0; j < outwidth; j++)
		{
			pix1 = (byte *) inrow + p1[j];
			pix2 = (byte *) inrow + p2[j];
			pix3 = (byte *) inrow2 + p1[j];
			pix4 = (byte *) inrow2 + p2[j];

			((byte *) (texOut + j))[0] = (pix1[0] + pix2[0] + pix3[0] + pix4[0]) >> 2;
			((byte *) (texOut + j))[1] = (pix1[1] + pix2[1] + pix3[1] + pix4[1]) >> 2;
			((byte *) (texOut + j))[2] = (pix1[2] + pix2[2] + pix3[2] + pix4[2]) >> 2;
			((byte *) (texOut + j))[3] = (pix1[3] + pix2[3] + pix3[3] + pix4[3]) >> 2;
		}
	}
}


// maximum texture size
extern int max_size;


/*
===============
ColorDiff

Get the difference between 2 colours.  This is accomplished by first converting to
greyscale, then comparing and subtracting.  There are better ways of doing this, but
greyscaling gives the most appropriate visual results for Quake textures.  The
alpha component is generally ignored.
===============
*/
unsigned int ColorDiff (unsigned int p1, unsigned int p2)
{
	byte *rgba1 = (byte *) &p1;
	byte *rgba2 = (byte *) &p2;

	int avg1;
	int avg2;

	// NTSC greyscale conversion standard
	avg1 = ((int) rgba1[0] * 30 + (int) rgba1[1] * 59 + (int) rgba1[2] * 11) / 300;
	avg2 = ((int) rgba2[0] * 30 + (int) rgba2[1] * 59 + (int) rgba2[2] * 11) / 300;

	if (avg1 > avg2)
		return avg1 - avg2;
	else return avg2 - avg1;
}


/*
===============
ColorAvg

Get the average of 2 colours, including alpha component
===============
*/
unsigned int ColorAvg (unsigned int p1, unsigned int p2)
{
	byte *rgba1 = (byte *) &p1;
	byte *rgba2 = (byte *) &p2;
	byte *rgbaavg;
	unsigned int avg;

	rgbaavg = (byte *) &avg;

	rgbaavg[0] = (byte) (((int) rgba1[0] + (int) rgba2[0]) / 2);
	rgbaavg[1] = (byte) (((int) rgba1[1] + (int) rgba2[1]) / 2);
	rgbaavg[2] = (byte) (((int) rgba1[2] + (int) rgba2[2]) / 2);
	rgbaavg[3] = (byte) (((int) rgba1[3] + (int) rgba2[3]) / 2);

	return avg;
}


/*
===============
ChooseLerp

Choose an interpolation function based on the lowest difference (i.e. which colour is
nearest to the original pixel): "colormin" is a confusing name, really...
===============
*/
unsigned int ChooseLerp (unsigned int me, unsigned int row, unsigned int col, unsigned int diag)
{
	unsigned int d1;
	unsigned int d2;
	unsigned int d3;
	unsigned int d4;
	unsigned int colormin;

	d1 = ColorDiff (me, row);
	d2 = ColorDiff (me, col);
	d3 = ColorDiff (me, diag);
	d4 = ColorDiff (row, col);

	colormin = d1;

	if (colormin > d2) colormin = d2;
	if (colormin > d3) colormin = d3;
	if (colormin > d4) colormin = d4;

	if (colormin == d1)
		return ColorAvg (me, row);
	else if (colormin == d2)
		return ColorAvg (me, col);
	else if (colormin == d3)
		return ColorAvg (me, diag);
	else return ColorAvg (me, ColorAvg (row, col));
}


/*
=================================
GL_InterpolatePixels

directional average scaling algorithm from http://www.compuphase.com/graphic/scale2.htm

this is edge-based, so it's very good at bringing out detail in complex textures when viewed
close up.  gives sharper results than bilinear or bicubic, without the "spotty" appearance of
my own algorithm (removed).  it's tolerant of multiple passes so long as MHQuakes internal
1024 x 1024 maximum is adhered to.

the paper at the url above discusses use of bit interleaving to determine colour differences.
i've abandoned that as it gives undesirable results with quake textures (likely the fault of
quake textures rather than anything in the paper).  instead i'm using option 2: converting to
greyscale.  the speed implications mentioned in this paper don't really apply to quake as the 
calculations don't occur at run time and the bottleneck is elsewhere (OpenGL texture uploading).
=================================
*/
void GL_InterpolatePixels (unsigned int *scaled, int w, int h, qboolean alpha)
{
	int hcount;
	int wcount;

	int x;

	// 2D arrays to make life easier for all of us
	static unsigned int temp1[1024][1024];	// copy of original
	static unsigned int temp2[1024][1024];	// upsampled

	// row up, col left, row down, col right, same row, same col
	int pos[6];

	int avg[4];	// rgba

	// copy the original data into our first 2D array
	for (hcount = 0, x = 0; hcount < h; hcount++)
	{
		for (wcount = 0; wcount < w; wcount++, x++)
		{
			temp1[hcount][wcount] = scaled[x];
		}
	}

	for (hcount = 0; hcount < h; hcount++)
	{
		// determine the positions of the 8 surrounding pixels - part one: rows
		// row up
		pos[0] = hcount - 1;
		if (pos[0] < 0) pos[0] += h;

		// row down
		pos[2] = hcount + 1;
		if (pos[2] >= h) pos[2] -= h;

		// same row
		pos[4] = hcount;

		for (wcount = 0; wcount < w; wcount++)
		{
			// determine the positions of the 8 surrounding pixels - part two: columns
			// col left
			pos[1] = wcount - 1;
			if (pos[1] < 0) pos[1] += w;

			// col right
			pos[3] = wcount + 1;
			if (pos[3] >= w) pos[3] -= w;

			// same col
			pos[5] = wcount;

			// each pixel in temp1 maps to 4 pixels in temp2
			temp2[hcount * 2][wcount * 2] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[0]][pos[5]], temp1[pos[4]][pos[1]], temp1[pos[0]][pos[1]]);
			temp2[hcount * 2][wcount * 2 + 1] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[0]][pos[5]], temp1[pos[4]][pos[3]], temp1[pos[0]][pos[3]]);
			temp2[hcount * 2 + 1][wcount * 2] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[2]][pos[5]], temp1[pos[4]][pos[1]], temp1[pos[2]][pos[1]]);
			temp2[hcount * 2 + 1][wcount * 2 + 1] = ChooseLerp (temp1[pos[4]][pos[5]], temp1[pos[2]][pos[5]], temp1[pos[4]][pos[3]], temp1[pos[2]][pos[3]]);
		}
	}

	// copy temp2 back to scaled
	for (hcount = 0, x = 0; hcount < h * 2; hcount++)
	{
		for (wcount = 0; wcount < w * 2; wcount++, x++)
		{
			scaled[x] = temp2[hcount][wcount];
		}
	}
}


/*
===============
GL_Upload32

We make sure in advance that all 2D images coming in here are padded with alpha to make them
a power of 2 and their s and t coords modified accordingly (only if they aren't already a
power of 2) - 2D images are less tolerant of resampling, etc, than the 3D stuff is.

Anything that is not a power of 2 is resampled *up* to the next power of 2, then capped at
either 1024 * 1024 or the max texture size, whichever is lower.  LINEAR_MIPMAP_LINEAR filtering
is always used so as to avoid abrubt transitions between miplevels in the world.  A GL_RGBA 
internal format is preferred for everything - using the full 32 bits should always be faster 
than having to unpack texture data before use.  Even the lowliest budget "price of a pack of 
fags" card has 32MB these days so memory is not a consideration.

Mipmapped textures are lightscaled in advance of uploading to increase the contrast range
and thereby give the illusion that our crappy 8-bit sources are something a little better!!!
===============
*/
qboolean lerptex = false;
qboolean lightscale = true;

void GL_Upload32 (unsigned *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	int w;
	int h;
	static unsigned int scaled[1024 * 1024];

	// get the power of 2 equal to or above the texture size
	for (w = 1; w < width; w *= 2);
	for (h = 1; h < height; h *= 2);

	// cap the size
	if (w > 1024) w = 1024;
	if (h > 1024) h = 1024;

	// cap at max texture size
	if (w > max_size) w = max_size;
	if (h > max_size) h = max_size;

	// either scale it to the new size or do a straight copy
	GL_ResampleTexture (data, width, height, scaled, w, h);

	// don't interpolate cos it's too big
	if (w > 512 || h > 512) lerptex = false;

	if (lerptex)
	{
		GL_InterpolatePixels (scaled, w, h, alpha);

		w *= 2;
		h *= 2;

		lerptex = false;
	}

	if (mipmap)
	{
		if (lightscale) GL_LightScaleTexture (scaled, w, h);

		// hardware mipmapping (OpenGL 1.4 or higher an absolute must)
		glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

		glTexImage2D (GL_TEXTURE_2D, 
					  0, 
					  GL_RGBA, 
					  w, 
					  h, 
					  0, 
					  GL_RGBA, 
					  GL_UNSIGNED_BYTE, 
					  scaled);

		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else
	{
		if (lightscale) GL_LightScale2DTexture (scaled, w, h);

		glTexImage2D (GL_TEXTURE_2D, 
					  0, 
					  GL_RGBA, 
					  w, 
					  h, 
					  0, 
					  GL_RGBA, 
					  GL_UNSIGNED_BYTE, 
					  scaled);

		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
}


typedef struct TGAHeader_s
{
	byte id_length;
	byte colormap_type;
	byte image_type;

	unsigned short colormap_index;
	unsigned short colormap_length;

	byte colormap_size;

	unsigned short x_origin;
	unsigned short y_origin;
	unsigned short width;
	unsigned short height;

	byte pixel_size;
	byte attributes;
} TGAHeader_t;


TGAHeader_t targa_header;
byte *tex_rgba;


int fgetLittleShort (FILE *f)
{
	byte b1 = fgetc (f);
	byte b2 = fgetc (f);

	return (short) (b1 + b2 * 256);
}


int fgetLittleLong (FILE *f)
{
	byte b1 = fgetc (f);
	byte b2 = fgetc (f);
	byte b3 = fgetc (f);
	byte b4 = fgetc (f);

	return b1 + (b2 << 8) + (b3 << 16) + (b4 << 24);
}


qboolean LoadTGA (FILE *fin, qboolean mipmap)
{
	int columns;
	int rows;
	int numPixels;
	byte *pixbuf;
	int row, column;
	qboolean alpha;

	targa_header.id_length = fgetc (fin);
	targa_header.colormap_type = fgetc (fin);
	targa_header.image_type = fgetc (fin);

	targa_header.colormap_index = fgetLittleShort (fin);
	targa_header.colormap_length = fgetLittleShort (fin);
	targa_header.colormap_size = fgetc (fin);
	targa_header.x_origin = fgetLittleShort (fin);
	targa_header.y_origin = fgetLittleShort (fin);
	targa_header.width = fgetLittleShort (fin);
	targa_header.height = fgetLittleShort (fin);
	targa_header.pixel_size = fgetc (fin);
	targa_header.attributes = fgetc (fin);

	if (targa_header.image_type != 2 && targa_header.image_type != 10)
		goto badTGA;

	if (targa_header.colormap_type != 0 || (targa_header.pixel_size != 32 && targa_header.pixel_size != 24))
		goto badTGA;

	// BPP other than 24 or 32 should not happen with TGAs!!!
	// (unless the file is corrupt...)
	if (targa_header.pixel_size == 24)
		alpha = false;
	else if (targa_header.pixel_size == 24)
		alpha = false;
	else goto badTGA;

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;

	tex_rgba = malloc (numPixels * 4);

	// skip TARGA image comment
	if (targa_header.id_length != 0) fseek (fin, targa_header.id_length, SEEK_CUR);
	
	if (targa_header.image_type == 2)
	{
		// Uncompressed, RGB images
		for (row = rows - 1; row >= 0; row--)
		{
			pixbuf = tex_rgba + row * columns * 4;

			for (column = 0; column < columns; column++)
			{
				byte blue = getc (fin);
				byte green = getc (fin);
				byte red = getc (fin);
				byte alphabyte;

				*pixbuf++ = red;
				*pixbuf++ = green;
				*pixbuf++ = blue;

				switch (targa_header.pixel_size)
				{
				case 24:
					*pixbuf++ = 255;
					break;

				case 32:
					alphabyte = getc (fin);
					*pixbuf++ = alphabyte;
					break;
				}
			}
		}
	}
	else if (targa_header.image_type == 10)
	{
		// Runlength encoded RGB images
		byte red;
		byte green;
		byte blue;
		byte alphabyte;
		byte packetHeader;
		byte packetSize;
		byte j;

		for (row = rows - 1; row >= 0; row--)
		{
			pixbuf = tex_rgba + row * columns * 4;

			for (column = 0; column < columns;)
			{
				packetHeader = getc (fin);
				packetSize = 1 + (packetHeader & 0x7f);

				if (packetHeader & 0x80)
				{
					// run-length packet
					blue = getc (fin);
					green = getc (fin);
					red = getc (fin);

					switch (targa_header.pixel_size)
					{
					case 24:
						alphabyte = 255;
						break;

					case 32:
						alphabyte = getc (fin);
						break;
					}
	
					for (j = 0; j < packetSize; j++)
					{
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alphabyte;

						column++;

						if (column == columns)
						{
							// run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							pixbuf = tex_rgba + row * columns * 4;
						}
					}
				}
				else
				{
					// non run-length packet
					for (j = 0; j < packetSize; j++)
					{
						blue = getc (fin);
						green = getc (fin);
						red = getc (fin);

						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;

						switch (targa_header.pixel_size)
						{
						case 24:
							*pixbuf++ = 255;
							break;

						case 32:
							alphabyte = getc (fin);
							*pixbuf++ = alphabyte;
							break;
						}

						column++;

						if (column==columns)
						{
							// pixel packet run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							pixbuf = tex_rgba + row * columns * 4;
						}
					}
				}
			}
			breakOut:;
		}
	}

	// don't interpolate or lightscale these
	lerptex = false;
	lightscale = false;

	// only 32 bit TGA files can have alpha
	GL_Upload32 ((unsigned int *) tex_rgba, targa_header.width, targa_header.height, mipmap, alpha);

	free (tex_rgba);
	fclose (fin);

	return true;

badTGA:;
	fclose (fin);
	return false;
}


qboolean LoadBMP (FILE *f, qboolean mipmap)
{
	int i, j;
	int run;
	int runlength;
	byte bmIn;
	BITMAPINFO bmInfo;
	BITMAPFILEHEADER bmHeader;
	RGBQUAD bmPalette[256];
	byte *bmData;

	// read the file header
	fread (&bmHeader, 1, sizeof (bmHeader), f);

	// validate the type
	if (bmHeader.bfType != 0x4D42) goto badBMP;

	// read the BITMAPINFO
	fread (&bmInfo, 1, sizeof (bmInfo), f);

	if (bmInfo.bmiHeader.biCompression != BI_RGB)
	{
		if (!(bmInfo.bmiHeader.biCompression == BI_RLE8 && bmInfo.bmiHeader.biBitCount == 8))
			goto badBMP;
	}

	// too big
	if (bmInfo.bmiHeader.biWidth > 1024) goto badBMP;
	if (bmInfo.bmiHeader.biHeight > 1024 || bmInfo.bmiHeader.biHeight < -1024) goto badBMP;

	// support 4, 8 and 24 BPP modes
	if (bmInfo.bmiHeader.biBitCount != 4 && 
		bmInfo.bmiHeader.biBitCount != 8 && 
		bmInfo.bmiHeader.biBitCount != 24) goto badBMP;

	tex_rgba = malloc (bmInfo.bmiHeader.biWidth * abs (bmInfo.bmiHeader.biHeight) * 4);
	bmData = tex_rgba;

	if (bmInfo.bmiHeader.biBitCount != 24)
	{
		// read the palette entries
		// the first one is in the BITMAPINFO struct
		bmPalette[0].rgbBlue = bmInfo.bmiColors[0].rgbBlue;
		bmPalette[0].rgbGreen = bmInfo.bmiColors[0].rgbGreen;
		bmPalette[0].rgbRed = bmInfo.bmiColors[0].rgbRed;
		bmPalette[0].rgbReserved = bmInfo.bmiColors[0].rgbReserved;

		// remaining entries follow the struct
		for (i = 1; i < bmInfo.bmiHeader.biClrUsed; i++)
		{
			fread (&bmPalette[i], 1, sizeof (RGBQUAD), f);
		}
	}

	// load the data
	for (i = 0; i < bmInfo.bmiHeader.biWidth * abs (bmInfo.bmiHeader.biHeight); i++)
	{
		if (bmInfo.bmiHeader.biBitCount == 4)
		{
			bmIn = fgetc (f);

			bmData[0] = bmPalette[bmIn >> 4].rgbRed;
			bmData[1] = bmPalette[bmIn >> 4].rgbGreen;
			bmData[2] = bmPalette[bmIn >> 4].rgbBlue;
			bmData[3] = 255;

			bmData[4] = bmPalette[bmIn % 16].rgbRed;
			bmData[5] = bmPalette[bmIn % 16].rgbGreen;
			bmData[6] = bmPalette[bmIn % 16].rgbBlue;
			bmData[7] = 255;

			bmData += 8;

			// increment i again cos we're reading two entires at a time
			i++;

			continue;
		}
		else if (bmInfo.bmiHeader.biBitCount == 8)
		{
			if (bmInfo.bmiHeader.biCompression == BI_RLE8)
			{
				bmIn = fgetc (f);

				if (bmIn == 0)
				{
					bmIn = fgetc (f);
					i++;

					switch (bmIn)
					{
					case 0:
						// End of Line - just do nothing - trust me, it works.
						break;
					case 1:
						// End of Bitmap
						goto bmDone;
						break;
					case 2:
						// Delta
						// (tests indicate this isn't actually used... most likely cos the bitmap format
						// doesn't specify what the skipped pixels are to be filled with!!!)
						bmIn = fgetc (f);
						i++;

						bmData += bmIn * 4;

						bmIn = fgetc (f);
						i++;

						bmData += bmIn * bmInfo.bmiHeader.biWidth * 4;

						break;
					default:
						// absolute mode
						runlength = bmIn;

						for (run = 0; run < runlength; run++)
						{
							bmIn = fgetc (f);
							i++;

							bmData[0] = bmPalette[bmIn].rgbRed;
							bmData[1] = bmPalette[bmIn].rgbGreen;
							bmData[2] = bmPalette[bmIn].rgbBlue;
							bmData[3] = 255;

							bmData += 4;
						}

						if (runlength % 2)
						{
							// align on word boundaries
							bmIn = fgetc (f);
							i++;
						}

						break;
					}
				}
				else
				{
					// encoded mode
					runlength = bmIn;

					bmIn = fgetc (f);
					i++;

					for (run = 0; run < runlength; run++)
					{
						bmData[0] = bmPalette[bmIn].rgbRed;
						bmData[1] = bmPalette[bmIn].rgbGreen;
						bmData[2] = bmPalette[bmIn].rgbBlue;
						bmData[3] = 255;

						bmData += 4;
					}
				}

				continue;
			}

			bmIn = fgetc (f);

			bmData[0] = bmPalette[bmIn].rgbRed;
			bmData[1] = bmPalette[bmIn].rgbGreen;
			bmData[2] = bmPalette[bmIn].rgbBlue;
			bmData[3] = 255;

			bmData += 4;

			continue;
		}
		else	// 24 BPP - GRB format!!!  WHYYYY???
		{
			bmData[1] = fgetc (f);
			bmData[0] = fgetc (f);
			bmData[2] = fgetc (f);
			bmData[3] = 255;

			bmData += 4;
		}
	}

bmDone:;

	if (bmInfo.bmiHeader.biHeight > 0)
	{
		// bottom-up bitmap so we must invert it (sigh)
		for (i = 0; i < abs (bmInfo.bmiHeader.biHeight) / 2; i++)
		{
			for (j = 0; j < bmInfo.bmiHeader.biWidth; j++)
			{
				unsigned int srcpos = (i * bmInfo.bmiHeader.biWidth) + j;
				unsigned int dstpos = (((abs (bmInfo.bmiHeader.biHeight) - 1) - i) * bmInfo.bmiHeader.biWidth) + j;
				unsigned int tmp;

				tmp = ((unsigned int *) tex_rgba)[dstpos];
				((unsigned int *) tex_rgba)[dstpos] = ((unsigned int *) tex_rgba)[srcpos];
				((unsigned int *) tex_rgba)[srcpos] = tmp;
			}
		}
	}

	// don't interpolate or lightscale these
	lerptex = false;
	lightscale = false;

	// BMP files never have alpha
	GL_Upload32 ((unsigned int *) tex_rgba, bmInfo.bmiHeader.biWidth, abs (bmInfo.bmiHeader.biHeight), mipmap, false);

	free (tex_rgba);
	fclose (f);

	return true;

badBMP:;
	fclose (f);
	return false;
}


typedef struct pcx_s
{
	char manufacturer;
	char version;
	char encoding;
	char bits_per_pixel;

	unsigned short xmin;
	unsigned short ymin;
	unsigned short xmax;
	unsigned short ymax;

	unsigned short hres;
	unsigned short vres;

	byte palette[48];

	char reserved;
	char color_planes;

	unsigned short bytes_per_line;
	unsigned short palette_type;

	char filler[58];

	unsigned int data;			// unbounded
} pcx_t;


qboolean LoadPCX (FILE *f, qboolean mipmap)
{
	pcx_t *pcx, pcxbuf;
	byte palette[768];
	byte *pix;
	int x, y;
	int dataByte, runLength;
	int count;

	// parse the PCX file
	fread (&pcxbuf, 1, sizeof (pcxbuf), f);

	pcx = &pcxbuf;

	if (pcx->manufacturer != 0x0a || pcx->version != 5 || pcx->encoding != 1 || pcx->bits_per_pixel != 8 || pcx->xmax >= 1024 || pcx->ymax >= 1024)
	{
		fclose (f);
		return false;
	}

	// seek to palette
	fseek (f, -768, SEEK_END);
	fread (palette, 1, 768, f);

	fseek (f, sizeof (pcxbuf) - 4, SEEK_SET);

	count = (pcx->xmax + 1) * (pcx->ymax + 1);
	tex_rgba = malloc (count * 4);

	for (y = 0; y <= pcx->ymax; y++)
	{
		pix = tex_rgba + 4 * y * (pcx->xmax + 1);

		for (x = 0; x <= pcx->xmax;)
		{
			dataByte = fgetc (f);

			if ((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = fgetc (f);
			}
			else runLength = 1;

			while(runLength-- > 0)
			{
				pix[0] = palette[dataByte * 3];
				pix[1] = palette[dataByte * 3 + 1];
				pix[2] = palette[dataByte * 3 + 2];
				pix[3] = 255;
				pix += 4;
				x++;
			}
		}
	}

	// don't interpolate or lightscale these
	lerptex = false;
	lightscale = false;

	// PCX files never have alpha
	GL_Upload32 ((unsigned int *) tex_rgba, (pcx->xmax + 1), (pcx->ymax + 1), mipmap, false);

	free (tex_rgba);
	fclose (f);

	return true;
}


qboolean GL_UploadExternalTexture (char *txname, qboolean mipmap)
{
	int i;
	FILE *f;
	char texname[1024];

	sprintf (texname, "%s", txname);

	// replace invalid file name characters with _
	for (i = 0; i < strlen (texname); i++)
	{
		if (texname[i] == '*')
		{
			texname[i] = '_';
		}
	}

	COM_FOpenFile (va ("textures/%s.tga", texname), &f);

	if (!f) goto loadBMP;

	if (LoadTGA (f, mipmap)) return true;

loadBMP:;
	COM_FOpenFile (va ("textures/%s.bmp", texname), &f);

	if (!f) goto loadPCX;

	if (LoadBMP (f, mipmap)) return true;

loadPCX:;
	COM_FOpenFile (va ("textures/%s.pcx", texname), &f);

	if (!f) return false;

	if (LoadPCX (f, mipmap)) return true;

	// couldn't load anything
	return false;
}


/*
===============
GL_Upload8
===============
*/
void GL_Upload8 (byte *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	static unsigned trans[1024 * 1024];
	int i;
	int s;
	int p;

	s = width * height;
	alpha = false;

	for (i = 0; i < s; i++)
	{
		p = data[i];

		if (p == 255)
		{
			alpha = true;
			trans[i] = 0;
		}
		else trans[i] = d_8to24table[p];
	}

	GL_Upload32 (trans, width, height, mipmap, alpha);
}


void MakeTGA (char *identifier, byte *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	static unsigned trans[1024 * 1024];
	int i;
	int s;
	int p;

	s = width * height;
	alpha = false;

	for (i = 0; i < s; i++)
	{
		p = data[i];

		if (p == 255)
		{
			alpha = true;
			trans[i] = 0;
		}
		else
		{
			((byte *) &trans[i])[0] = ((byte *) &d_8to24table[p])[2];
			((byte *) &trans[i])[1] = ((byte *) &d_8to24table[p])[1];
			((byte *) &trans[i])[2] = ((byte *) &d_8to24table[p])[0];
			((byte *) &trans[i])[3] = 255;
		}
	}

	WriteTGA (va ("textures/%s.tga", identifier), width, height, 32, (byte *) trans);
}


int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha)
{
	int	i;
	gltexture_t	*glt;
	unsigned short crc;

	crc = CRC_Block (data, (width * height));

	// see if the texture is allready present
	if (identifier[0])
	{
		for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
		{
			if (!strcmp (identifier, glt->identifier))
			{
				// the data doesn't match (or we're reloading it anyway) so reload it
				if (crc != glt->crc || glt->reload)
				{
					// free the texture memory
					glDeleteTextures (1, &glt->texnum);
					goto CacheMismatch;
				}

				return gltextures[i].texnum;
			}
		}
	}

	glt = &gltextures[numgltextures];
	numgltextures++;

	strcpy (glt->identifier, identifier);
	glt->texnum = texture_extension_number++;

CacheMismatch:;
	glt->width = width;
	glt->height = height;
	glt->mipmap = mipmap;

	// this was *MISSING* from previous releases!!!
	// there's the cause of yer slow map loads!!!
	glt->crc = crc;

	// don't reload unless explicitly asked to
	glt->reload = false;

	glBindTexture (GL_TEXTURE_2D, glt->texnum);

	// external texture support
	if (!GL_UploadExternalTexture (identifier, mipmap))
		GL_Upload8 (data, width, height, mipmap, alpha);

	if (identifier[0])
	{
		//MakeTGA (identifier, data, width, height, mipmap, alpha);
	}

	// make sure that the next one is lightscaled unless explicitly told not to
	lightscale = true;

	return glt->texnum;
}


void GL_DeleteTextures (void)
{
	int i;

	for (i = 1; i <= texture_extension_number; i++)
	{
		glDeleteTextures (1, &i);
	}
}


/*
================
GL_LoadPicTexture
================
*/
int GL_LoadPicTexture (qpic_t *pic, char *picname)
{
	return GL_LoadTexture (picname, pic->width, pic->height, pic->data, false, true);
}


/*
================
FindFullbrightTexture

Determine if a given texture has fullbright components
================
*/
qboolean FindFullbrightTexture (byte *pixels, int num_pix)
{
    int i;

    for (i = 0; i < num_pix; i++)
        if (pixels[i] > 223 && pixels[i] != 255)
            return true;

    return false;
}


/*
================
ConvertPixels

Converts a textures non-fullbright components to a specified alpha colour
================
*/
void ConvertPixels (byte *pixels, int num_pixels, int alphaColour)
{
    int i;

    for (i = 0; i < num_pixels; i++)
        if (pixels[i] < 224)
            pixels[i] = alphaColour;
}



/*
================
RemoveFullBrights

Removes all fullbright components from a texture
================
*/
void RemoveFullBrights (byte *pixels, int num_pixels, int alphaColour)
{
    int i;

    for (i = 0; i < num_pixels; i++)
        if (pixels[i] > 223 && pixels[i] != 255)
            pixels[i] = alphaColour;
}



/*
================
AverageColour

Determine the average colour (expressed in RGB rather than palletted) of all colours in
a texture
================
*/
float avg[3];

void AverageColour (byte *pixels, int num_pixels)
{
    int i;
	unsigned rgba;
	int count;

	avg[0] = avg[1] = avg[2] = 0;

    for (i = 0, count = 0; i < num_pixels; i++)
	{
		// ignore alpha
		if (pixels[i] == 255) continue;

		rgba = d_8to24table[pixels[i]];
		count++;

		avg[0] += ((byte *) &rgba)[0];
		avg[1] += ((byte *) &rgba)[1];
		avg[2] += ((byte *) &rgba)[2];
	}

	if (!count) return;

	avg[0] /= count;
	avg[1] /= count;
	avg[2] /= count;
}


/*
================
GL_LoadMapTexture

Given a pre-filled in texture_t this will do the GL stuff on it
================
*/
void GL_LoadMapTexture (texture_t *tx)
{
	char fbr_mask_name[64];
	static byte temp[1024 * 1024];
	int pixels = tx->width * tx->height;

	if (tx->name[0] != '*')
	{
		lerptex = true;
		lightscale = true;
	}
	else
	{
		lerptex = false;
		lightscale = false;
	}

	tx->gl_texturenum = GL_LoadTexture (tx->name,
										tx->width,
										tx->height,
										(byte *) (tx+1), 
										true,
										false);

	// check for fullbright pixels in the texture - but do not check liquid textures,
	// since they will be rendered fullbright anyway...
	if ((tx->name[0] != '*') && (FindFullbrightTexture ((byte *) (tx + 1), pixels)))
	{
		// first off, let's be nice and non-destructive to the original pixels
		memcpy ((byte *) temp, (byte *) (tx + 1), pixels);

		// convert any non-fullbright pixel to fully transparent
		ConvertPixels (temp, pixels, 255);

		// derive a new name for the texture to prevent mismatches
		sprintf (fbr_mask_name, "%s_luma", tx->name);

		// and now load it - we need alpha for the non-fullbright components
		lerptex = true;

		tx->fullbright = GL_LoadTexture (fbr_mask_name,
										 tx->width, 
										 tx->height, 
										 temp, 
										 true, 
										 true);
	}
	else tx->fullbright = 0;

	// if it's a liquid texture we find an underwater fog colour for it
	if (tx->name[0] == '*')
	{
		AverageColour ((byte *) (tx + 1), pixels);

		tx->uwater_fog[0] = avg[0] / 192;
		tx->uwater_fog[1] = avg[1] / 192;
		tx->uwater_fog[2] = avg[2] / 192;
		tx->uwater_fog[3] = 0.2;
	}
}


/****************************************/


// oh what a steal!!!
void fractalnoise (byte *noise, int size, int startgrid)
{
	int x, y, g, g2, amplitude, min, max, size1 = size - 1, sizepower, gridpower;
	int *noisebuf;

#define n(x, y) noisebuf[((y) & size1) * size + ((x) & size1)]

	for (sizepower = 0; (1 << sizepower) < size; sizepower++);

	if (size != (1 << sizepower))
		Sys_Error ("fractalnoise: size must be power of 2\n");

	for (gridpower = 0; (1 << gridpower) < startgrid; gridpower++);

	if (startgrid != (1 << gridpower))
		Sys_Error ("fractalnoise: grid must be power of 2\n");

	startgrid = bound (0, startgrid, size);

	amplitude = 0xFFFF; // this gets halved before use

	noisebuf = malloc (size * size * sizeof (int));
	memset (noisebuf, 0, size * size * sizeof (int));

	for (g2 = startgrid; g2; g2 >>= 1)
	{
		// brownian motion (at every smaller level there is random behavior)
		amplitude >>= 1;

		for (y = 0; y < size; y += g2)
		{
			for (x = 0; x < size; x += g2)
			{
				n(x, y) += (rand () & amplitude);
			}
		}

		g = g2 >> 1;

		if (g)
		{
			// subdivide, diamond-square algorithm (really this has little to do with squares)
			// diamond
			for (y = 0; y < size; y += g2)
			{
				for (x = 0; x < size; x += g2)
				{
					n(x + g, y + g) = (n(x, y) + n(x + g2, y) + n(x, y + g2) + n(x + g2, y + g2)) >> 2;
				}
			}

			// square
			for (y = 0; y < size; y += g2)
			{
				for (x = 0; x < size; x += g2)
				{
					n(x + g, y) = (n(x, y) + n(x + g2, y) + n(x + g, y - g) + n(x + g, y + g)) >> 2;
					n(x, y + g) = (n(x, y) + n(x, y + g2) + n(x - g, y + g) + n(x + g,y + g)) >> 2;
				}
			}
		}
	}

	// find range of noise values
	min = max = 0;

	for (y = 0; y < size; y++)
	{
		for (x = 0; x < size; x++)
		{
			if (n(x, y) < min) min = n(x, y);
			if (n(x, y) > max) max = n(x, y);
		}
	}

	max -= min;
	max++;

	// normalize noise and copy to output
	for (y = 0;y < size;y++)
	{
		for (x = 0; x < size; x++)
		{
			*noise++ = (byte) (((n(x, y) - min) * 256) / max);
		}
	}

	free (noisebuf);
#undef n
}


void LoadDetailTexture (void)
{
#define DETAILRESOLUTION 256
	int i, x, y, light;
	float vc[3], vx[3], vy[3], vn[3], lightdir[3];
	byte data[DETAILRESOLUTION][DETAILRESOLUTION];
	byte noise[DETAILRESOLUTION][DETAILRESOLUTION];

	lightdir[0] = 0.5;
	lightdir[1] = 1;
	lightdir[2] = -0.25;

	VectorNormalize (lightdir);

	fractalnoise (&noise[0][0], DETAILRESOLUTION, DETAILRESOLUTION >> 4);

	for (y = 0; y < DETAILRESOLUTION; y++)
	{
		for (x = 0; x < DETAILRESOLUTION; x++)
		{
			vc[0] = x;
			vc[1] = y;

			vc[2] = noise[y][x] * (1.0f / 32.0f);

			vx[0] = x + 1;
			vx[1] = y;

			vx[2] = noise[y][(x + 1) % DETAILRESOLUTION] * (1.0f / 32.0f);

			vy[0] = x;
			vy[1] = y + 1;

			vy[2] = noise[(y + 1) % DETAILRESOLUTION][x] * (1.0f / 32.0f);

			VectorSubtract (vx, vc, vx);
			VectorSubtract (vy, vc, vy);
			CrossProduct (vx, vy, vn);
			VectorNormalize (vn);

			light = 128 - DotProduct (vn, lightdir) * 128;

			// adjust the detail texture colour range to get a match with the colour range
			// when they're turned off (otherwise things will look mondo darko...)
			light *= 13;
			light /= 16;

			data[y][x] = light > 255 ? 255 : (light < 0 ? 0 : light);
		}
	}

	// make it into a texture
	if (detailtexture == -1)
	{
		detailtexture = texture_extension_number;
		texture_extension_number++;
	}

	glBindTexture (GL_TEXTURE_2D, detailtexture);

	// hardware mipmapping (OpenGL 1.4 or higher an absolute must)
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexImage2D (GL_TEXTURE_2D, 
				  0, 
				  GL_LUMINANCE, 
				  DETAILRESOLUTION, 
				  DETAILRESOLUTION, 
				  0, 
				  GL_LUMINANCE, 
				  GL_UNSIGNED_BYTE, 
				  data);

	// sexy filtering modes are computationally very expensive on detail textures as they will
	// always use the minification filter and are of such a nature that mipmap artefacts are not
	// as noticeable as with standard world textures.
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
#undef DETAILRESOLUTION
}


void LoadCausticTexture (void)
{
	int i;
	int j;
	HINSTANCE hInstance;
	HANDLE hBitmap;
	BITMAPINFO *bmInfo;
	byte *bmData;
	unsigned tex[64 * 64];
	byte twopix;

	int CausticBitmaps[] =
	{
		IDB_BITMAP4, IDB_BITMAP5, IDB_BITMAP6, IDB_BITMAP7, IDB_BITMAP8, IDB_BITMAP9, IDB_BITMAP10, IDB_BITMAP11, 
		IDB_BITMAP12, IDB_BITMAP13, IDB_BITMAP14, IDB_BITMAP15, IDB_BITMAP16, IDB_BITMAP17, IDB_BITMAP18, IDB_BITMAP19, 
		IDB_BITMAP20, IDB_BITMAP21, IDB_BITMAP22, IDB_BITMAP23, IDB_BITMAP24, IDB_BITMAP25, IDB_BITMAP26, IDB_BITMAP27, 
		IDB_BITMAP28, IDB_BITMAP29, IDB_BITMAP30, IDB_BITMAP31, IDB_BITMAP32, IDB_BITMAP33, IDB_BITMAP34, IDB_BITMAP35,
		IDB_BITMAP36, IDB_BITMAP37, IDB_BITMAP38, IDB_BITMAP39, IDB_BITMAP40, IDB_BITMAP41, IDB_BITMAP42, IDB_BITMAP43, 
		IDB_BITMAP44, IDB_BITMAP45, IDB_BITMAP46, IDB_BITMAP47, IDB_BITMAP48, IDB_BITMAP49, IDB_BITMAP50, IDB_BITMAP51, 
		IDB_BITMAP52, IDB_BITMAP53, IDB_BITMAP54, IDB_BITMAP55, IDB_BITMAP56, IDB_BITMAP57, IDB_BITMAP58, IDB_BITMAP59, 
		IDB_BITMAP60, IDB_BITMAP61, IDB_BITMAP62, IDB_BITMAP63, IDB_BITMAP64, IDB_BITMAP65, IDB_BITMAP66, IDB_BITMAP67, 
		IDB_BITMAP68, IDB_BITMAP69, IDB_BITMAP70, IDB_BITMAP71, IDB_BITMAP72, IDB_BITMAP73, IDB_BITMAP74, IDB_BITMAP75
	};

	if (caustictexture == -1)
	{
		// make enough space for our 72 texes
		caustictexture = texture_extension_number;
		texture_extension_number += 72;
	}

	hInstance = GetModuleHandle (NULL);

	for (j = 0; j < 72; j++)
	{
		memset (tex, 0, sizeof (tex));

		hBitmap = LoadResource (hInstance, FindResource (hInstance, MAKEINTRESOURCE (CausticBitmaps[j]), RT_BITMAP));

		// if we can't load the bitmap just use it all black
		if (hBitmap == NULL)
			goto nullBMP;

		bmData = (byte *) LockResource (hBitmap);

		bmInfo = (BITMAPINFO *) bmData;

		// locate the pixel data - use (sizeof (RGBQUAD) * 15) instead of 16 cos the 
		// BITMAPINFO struct already holds one entry
		bmData += sizeof (BITMAPINFO) + (sizeof (RGBQUAD) * 15);

		// 4 BPP = / 2
		for (i = 0; i < bmInfo->bmiHeader.biWidth * bmInfo->bmiHeader.biHeight / 2; i++)
		{
			// get 2 pixels
			twopix = bmData[i];

			((byte *) &tex[i * 2])[0] = bmInfo->bmiColors[twopix >> 4].rgbRed;
			((byte *) &tex[i * 2])[1] = bmInfo->bmiColors[twopix >> 4].rgbGreen;
			((byte *) &tex[i * 2])[2] = bmInfo->bmiColors[twopix >> 4].rgbBlue;
			((byte *) &tex[i * 2])[3] = 255;

			// second pixel
			((byte *) &tex[(i * 2) + 1])[0] = bmInfo->bmiColors[twopix % 16].rgbRed;
			((byte *) &tex[(i * 2) + 1])[1] = bmInfo->bmiColors[twopix % 16].rgbGreen;
			((byte *) &tex[(i * 2) + 1])[2] = bmInfo->bmiColors[twopix % 16].rgbBlue;
			((byte *) &tex[(i * 2) + 1])[3] = 255;
		}

nullBMP:;
		// turn it into a texture
		glBindTexture (GL_TEXTURE_2D, caustictexture + j);

		// hardware mipmapping (OpenGL 1.4 or higher an absolute must)
		glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

		glTexImage2D (GL_TEXTURE_2D, 
					  0, 
					  GL_RGBA, 
					  64, 
					  64, 
					  0, 
					  GL_RGBA, 
					  GL_UNSIGNED_BYTE, 
					  tex);

		// sexy filtering
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
}


// after changing resolution, any texture which is current will be lost, so reload them all
void GL_ReloadLightmap (int lightmapnum);
void GL_ReloadStainmap (int stainmapnum);
void Scrap_Upload (void);
void R_ClearMenuCachePics (void);
void Sbar_Preload3DModels (void);
void Mod_ReloadAliasModel (model_t *mod);
extern model_t *conQuad;
extern qboolean SbarPreload;
extern qboolean WorldModelLoaded;

void GL_ReloadTextures (void)
{
	int i;

	// kill all texes
	GL_DeleteTextures ();

	// tell GL_LoadTexture that we're reloading
	GLT_SetReload (true);

	// do the console stuff first
	GL_LoadCharset ();

	// the MH logo - VERY important...
	MHLoadLogo ();

	// console quad
	Mod_ReloadAliasModel (conQuad);

	// menu pics
	R_ClearMenuCachePics ();

	// scrap stuff
	Scrap_Upload ();

	// HUD models
	WorldModelLoaded = true;
	Sbar_Preload3DModels ();
	SbarPreload = false;

	// particles
	R_InitParticleTexture ();

	// detail
	LoadDetailTexture ();
	LoadCausticTexture ();

	// map textures - if we have a map loaded, that is...!!!
	if (r_worldentity.model && cl.worldmodel)
	{
		int x;
		model_t *mod;

		for (x = 1; x < MAX_MODELS; x++)
		{
			mod = cl.model_precache[x];

			if (!mod) continue;

			switch (mod->type)
			{
			case mod_brush:
				// these will be loaded with the world
				if (mod->name[0] == '*') break;

				for (i = 0; i < mod->numtextures; i++)
				{
					if (mod->textures[i])
					{
						if (!Q_strncmp (mod->textures[i]->name,"sky",3))
						{
							R_InitSky (mod->textures[i]);
						}
						else
						{
							GL_LoadMapTexture (mod->textures[i]);
						}
					}
				}

				break;

			case mod_alias:
				Mod_ReloadAliasModel (mod);
				break;

			case mod_sprite:
				// MHQuake doesn't use sprites...!
				break;

			default:
				break;
			}
		}

		// lightmaps - the data is kept in main memory so we just reload from that
		for (i = 0; i < 64; i++)
		{
			GL_ReloadLightmap (i);
			GL_ReloadStainmap (i);
		}
	}

	// this will be fraudulently set to "true" after reloading the sbar brush models
	WorldModelLoaded = false;
}


/*
===============
GL_TexmanInit

Initialize the texture management system
===============
*/
void GL_TexmanInit (void)
{
	memset (gltextures, 0, sizeof (gltextures));

	GLT_SetReload (true);

	LoadDetailTexture ();
	LoadCausticTexture ();

	// set up the intensity table for texture lightscaling
	BuildIntensityTable ();
}


/*
===============
GL_TexManMapLoadFinish

Does some housekeeping after loading the map
===============
*/
void GL_TexManMapLoadFinish (void)
{
}
