/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

#if 0

/*
================================================================================================================================

				MEMORY MANAGEMENT

	Here we use seperate memory management routines for this operation rather than putting stuff on Quake's run-time hunk.

================================================================================================================================
*/

static qboolean mem_initialized = false;

typedef struct chunkheader_s
{
	// next and previous memheaders in chain belonging to pool
	struct chunkheader_s *next;
	struct chunkheader_s *prev;

	// size of the memory after the header (excluding header and sentinel2)
	size_t size;

	// immediately followed by data
} chunkheader_t;

static chunkheader_t *chunks;
static size_t total_allocated;
static size_t total_active, total_peakactive;

void *qmalloc (size_t size)
{
	void *data;
	chunkheader_t *chunk;

	if (size <= 0) return NULL;

	if (mem_initialized)
		data = malloc (sizeof (chunkheader_t) + size);
	else
		data = malloc (size);

	if (!data) Sys_Error ("Failed on allocating %d bytes\n", size);

	if (!mem_initialized) return data;

	chunk = (chunkheader_t *) data;
	chunk->size = size;
	chunk->next = chunks;
	chunk->prev = NULL;
	chunks = chunk;

	if (chunk->next) chunk->next->prev = chunk;

	total_allocated += size;
	total_active += size;

	if (total_active > total_peakactive) total_peakactive = total_active;

	return (void *) ((unsigned char *) data + sizeof (chunkheader_t));
}


void qfree (void *data)
{
	chunkheader_t *chunk;

	if (!data) return;

	if (!mem_initialized)
	{
		free (data);
		return;
	}

	chunk = (chunkheader_t *) ((unsigned char *) data - sizeof (chunkheader_t));

	if (chunk->prev)
		chunk->prev->next = chunk->next;
	else
		chunks = chunk->next;

	if (chunk->next) chunk->next->prev = chunk->prev;

	total_active -= chunk->size;

	free (chunk);
}


void Q_InitMem (void)
{
	if (mem_initialized) return;

	chunks = NULL;
	mem_initialized = true;
	total_allocated = 0;
	total_active = 0;
	total_peakactive = 0;
}


void Q_PrintMem (void)
{
	if(!mem_initialized) return;

	Con_Printf ("Active memory: %f MB (%u bytes)\n", total_active  / 1048576.0, (unsigned int) total_active);
	Con_Printf ("Peak active memory: %f MB (%u bytes)\n", total_peakactive  / 1048576.0, (unsigned int) total_peakactive);
	Con_Printf ("Total memory allocated: %f MB (%u bytes)\n", total_allocated  / 1048576.0, (unsigned int) total_allocated);
}


void Q_ShutdownMem (void)
{
	size_t total;

	if (!mem_initialized) return;

	total = 0;

	while (chunks)
	{
		total += chunks->size;
		qfree ((void *) ((unsigned char *) chunks + sizeof(chunkheader_t)));
	}

	mem_initialized = false;
}


/*
================================================================================================================================

				PORTAL FILE GENERATION FOR AUTO-WATER TRANSLUCENCY

	Inspiration taken from LordHavoc's hmap utility (cos the source to the original BSP2PRT was never made available),
	rewritten by me to make compatible with map load time usage.

	This is done at load time in gl_model.c rather than at run-time either client side or server side, so that the data only
	needs to be gotten once and is available to both the client and the server.  If the server is remote it won't be available
	whichever way it's done, so it doesn't really matter doing it this way.

================================================================================================================================
*/

typedef struct
{
	int		numpoints;
	vec3_t	points[8];			// variable sized
} winding_t;


typedef struct visfacet_s
{
	struct visfacet_s	*next;
	int					planenum;
	int					planeside;		// which side is the front of the face
	int					texturenum;
	int					contents[2];	// 0 = front side
	struct visfacet_s	*original;		// face on node
	int					outputnumber;	// only valid for original faces after write surfaces
	winding_t			*winding;
} face_t;

typedef struct surface_s
{
	struct surface_s	*next;

	int					planenum;
	vec3_t				mins, maxs;
	qboolean			onnode;			// true if surface has already been used as a splitting node
	face_t				*faces;			// links to all the faces on either side of the surf
} surface_t;

typedef struct brush_s
{
	struct brush_s		*next;
	vec3_t				mins, maxs;
	face_t				*faces;
	int					contents;
} brush_t;

#define	PLANENUM_LEAF		-1

typedef struct node_s
{
	vec3_t				mins, maxs;		// bounding volume, not just points inside
	int					planenum;		// PLANENUM_LEAF = leaf node

	// information for decision nodes
	int					outputplanenum;	// only valid after EmitNodePlanes
	int					firstface;		// decision node only
	int					numfaces;		// decision node only
	struct node_s		*children[2];	// only valid for decision nodes
	face_t				*faces;			// decision nodes only, list for both sides

	// information for leafs
	int					contents;		// leaf nodes (0 for decision nodes)
	face_t				**markfaces;	// leaf nodes only, point to node faces
	struct portal_s		*portals;
	int					visleafnum;		// -1 = solid
	int					valid;			// for flood filling
	int					occupied;		// light number in leaf for outside filling
} node_t;

typedef struct
{
	vec3_t		mins, maxs;		// bounding box
	brush_t		*brushes;		// NULL terminated list
	face_t		*validfaces[MAX_MAP_PLANES];	// build surfaces is also used by GatherNodeFaces
	surface_t	*surfaces;
	node_t		*headnode;
} tree_t;

// BSP on disk representations
dvertex_t	*dvertexes;
dedge_t		*dedges;
int			*dsurfedges;
dplane_t	*dplanes;
dface_t		*dfaces;
dleaf_t		*dleafs;
dnode_t		*dnodes;
dmodel_t	*dmodels;


void ClearBounds (vec3_t mins, vec3_t maxs)
{
	int i;

	for (i = 0; i < 3; i++)
	{
		mins[i] = 999999999;
		maxs[i] = -999999999;
	}
}


void FreeBrush (brush_t *b)
{
	qfree (b);
}

tree_t *AllocTree (void)
{
	tree_t	*t;

	t = qmalloc (sizeof (tree_t));
	memset (t, 0, sizeof (tree_t));
	ClearBounds (t->mins, t->maxs);

	return t;
}

void FreeTree (tree_t *t)
{
	brush_t		*b, *next;

	for (b = t->brushes; b; b = next)
	{
		next = b->next;
		FreeBrush (b);
	}

	qfree (t);
}


node_t *AllocNode (void)
{
	node_t	*n;

	n = qmalloc (sizeof (node_t));
	memset (n, 0, sizeof (node_t));

	return n;
}

void FreeNode (node_t *n)
{
	qfree (n);
}


void AddPointToBounds (vec3_t v, vec3_t mins, vec3_t maxs)
{
	int		i;

	for (i = 0; i < 3; i++)
	{
		if (v[i] < mins[i]) mins[i] = v[i];
		if (v[i] > maxs[i]) maxs[i] = v[i];
	}
}


void GetWorldBounds (vec3_t mins, vec3_t maxs)
{
	int			i, j, e;
	dface_t		*face;
	dvertex_t	*v;
	vec3_t		point;

	ClearBounds (mins, maxs);

	for (i = 0, face = dfaces; i < dmodels[0].numfaces; i++, face++)
	{
		for (j = 0; j < face->numedges; j++)
		{
			e = dsurfedges[face->firstedge + j];

			if (e >= 0)
				v = dvertexes + dedges[e].v[0];
			else
				v = dvertexes + dedges[-e].v[1];

			VectorCopy (v->point, point);
			AddPointToBounds (point, mins, maxs);
		}
	}
}

int	FindPlane (dplane_t *dplane, int *side)
{
	int			i;
	dplane_t	*dp, pl;
	vec_t		dot;

	pl = *dplane;
	NormalizePlane (&pl);

	if (DotProduct (pl.normal, dplane->normal) > 0)
		*side = 0;
	else
		*side = 1;

	for (i = 0, dp = dplanes; i < nummapplanes; i++, dp++)
	{
		if (DotProduct (dp->normal, pl.normal) > 1.0 - ANGLE_EPSILON && fabs (dp->dist - pl.dist) < DIST_EPSILON)
			return i; // regular match
	}

	if (nummapplanes == MAX_MAP_PLANES) Sys_Error ("FindPlane: nummapplanes == MAX_MAP_PLANES");

	dot = VectorLength (dplane->normal);

	if (dot < 1.0 - ANGLE_EPSILON || dot > 1.0 + ANGLE_EPSILON)
		Sys_Error ("FindPlane: normalization error (%f %f %f, length %f)", dplane->normal[0], dplane->normal[1], dplane->normal[2], dot);

	mapplanes[nummapplanes] = pl;

	return nummapplanes++;
}

node_t *RecursiveBuildTree (int nodenum)
{
	node_t		*n;

	if (nodenum < 0)
	{
		dleaf_t *leaf = &dleafs[-1 - nodenum];

		n = AllocNode ();
		n->planenum = PLANENUM_LEAF;
		n->contents = leaf->contents;
	}
	else
	{
		int		side;
		dplane_t plane;
		dnode_t *node = &dnodes[nodenum];

		n = AllocNode ();
		plane.dist = dplanes[node->planenum].dist;
		VectorCopy (dplanes[node->planenum].normal, plane.normal);
		n->planenum = FindPlane (&plane, &side);

		if (side) Sys_Error ("Bad node plane");

		n->children[0] = RecursiveBuildTree (node->children[0]);
		n->children[1] = RecursiveBuildTree (node->children[1]);
	}

	return n;
}
#endif

void BSP2PRTInit (model_t *mod, dheader_t *header)
{
	/*
	tree_t *tree;

	// set up the BSP structs
	// no need to error-check these cos it's already been done in Mod_LoadBrushModel!!!
	// likewise, the header will come in here already byte-swapped, so we can leave that bit out too.
	// we can also leave out some of the BSP data we're not interested in getting for this procedure
	// (textures, lighting, etc)
	dvertexes = (dvertex_t *) (((byte *) header) + header->lumps[LUMP_VERTEXES].fileofs);
	dedges = (dedge_t *) (((byte *) header) + header->lumps[LUMP_EDGES].fileofs);
	dsurfedges = (int *) (((byte *) header) + header->lumps[LUMP_SURFEDGES].fileofs);
	dplanes = (dplane_t *) (((byte *) header) + header->lumps[LUMP_PLANES].fileofs);
	dfaces = (dface_t *) (((byte *) header) + header->lumps[LUMP_FACES].fileofs);
	dleafs = (dleaf_t *) (((byte *) header) + header->lumps[LUMP_LEAFS].fileofs);
	dnodes = (dnode_t *) (((byte *) header) + header->lumps[LUMP_NODES].fileofs);
	dmodels = (dmodel_t *) (((byte *) header) + header->lumps[LUMP_MODELS].fileofs);

	// init memory
	Q_InitMem ();


	// set up the tree
	tree = AllocTree ();

	// get the bounds
	GetWorldBounds (tree->mins, tree->maxs);

	// build the tree recursively
	tree->headnode = RecursiveBuildTree (0);

	// shut down memory
	Q_ShutdownMem ();
	*/
}

