/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_state.h"


/*
================================================================================================================

											WORLD VOLUME CARVING

	Divide the world into discrete volumes.  A volume is a seperate contiguous underwater area with the
	same contents type.  These can then be used for setting correct underwater fog and caustics, and
	similar joyful stuff.

	This is nowhere near as elegant as I'd like - in the end I resorted to pure brute force to do some of
	the matching up job.  There are multiple passes that are most likely not even needed.

	To be honest, it's messy and confusing, but at least it's as near to reliable as I can get.

	You need to add volume and volumetexture to the msurface_t and mleaf_t structs, then call the
	MH_WorldVolumeCarve function.  See the call to it in my gl_model.c for context, etc.

================================================================================================================
*/


void MarkSurfaceVolumesInLeaf (mleaf_t *leaf, msurface_t *surf)
{
	int j;
	msurface_t **marksurf;

	// copy the data into the leaf
	leaf->volume = surf->volume;
	leaf->volumetexture = surf->volumetexture;

	for (j = 0, marksurf = leaf->firstmarksurface; j < leaf->nummarksurfaces; j++, marksurf++)
	{
		// already marked
		if ((*marksurf)->volume != -1) continue;

		// copy the data into the surfs
		(*marksurf)->volume = surf->volume;
		(*marksurf)->volumetexture = surf->volumetexture;
	}
}


void MarkLeafsForSurf (model_t *mod, msurface_t *surf)
{
	int i;
	int j;
	msurface_t **marksurf;
	mleaf_t *leaf;

	for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
	{
		// already marked
		if (leaf->volume != -1) continue;

		for (j = 0, marksurf = leaf->firstmarksurface; j < leaf->nummarksurfaces; j++, marksurf++)
		{
			// is it the same surf?  and has it been marked yet?
			if ((*marksurf)->surfnum == surf->surfnum)
			{
				// yes
				MarkSurfaceVolumesInLeaf (leaf, surf);

				// no need to check any more surfs in this leaf
				break;
			}
		}
	}
}


void MarkPVSForLeaf (model_t *mod, mleaf_t *leaf)
{
	// find all leafs in this PVS
	byte *vis = Mod_LeafPVS (leaf, mod);
	mleaf_t *markLeaf;
	msurface_t **surf;
	int i;
	int j;

	// go through them and mark them
	for (i = 0; i < mod->numleafs; i++)
	{
		// in this pvs
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			markLeaf = &mod->leafs[i + 1];

			// already marked
			if (markLeaf->volume != -1) continue;

			// invalid contents
			if (markLeaf->contents != leaf->contents) continue;

			// it's in the PVS and it has the same contents type, so set the volume number
			markLeaf->volume = leaf->volume;
			markLeaf->volumetexture = leaf->volumetexture;

			// mark all surfs in the leaf with this volume number
			for (j = 0, surf = markLeaf->firstmarksurface; j < markLeaf->nummarksurfaces; j++, surf++)
			{
				(*surf)->volume = leaf->volume;
				(*surf)->volumetexture = leaf->volumetexture;
			}

			// recurse for each leaf in the PVS
			MarkPVSForLeaf (mod, markLeaf);
		}
	}
}


void MarkStragglePVSForLeaf (model_t *mod, mleaf_t *leaf)
{
	// find all leafs in this PVS
	byte *vis = Mod_LeafPVS (leaf, mod);
	mleaf_t *markLeaf;
	msurface_t **surf;
	int i;
	int j;

	// go through them and mark them
	for (i = 0; i < mod->numleafs; i++)
	{
		// in this pvs
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			markLeaf = &mod->leafs[i + 1];

			// not marked so ignore it
			if (markLeaf->volume == -1) continue;

			// empty
			if (markLeaf->volume == 0) continue;

			// invalid contents
			if (markLeaf->contents != leaf->contents) continue;

			// it's in the PVS and it has the same contents type, so copy in the data
			leaf->volume = markLeaf->volume;
			leaf->volumetexture = markLeaf->volumetexture;

			// mark all surfs in the original leaf
			for (j = 0, surf = leaf->firstmarksurface; j < leaf->nummarksurfaces; j++, surf++)
			{
				(*surf)->volume = leaf->volume;
				(*surf)->volumetexture = leaf->volumetexture;
			}

			// we only need one hit cos it's only one leaf we're checking
			return;
		}
	}
}


void MH_WorldVolumeCarve (model_t *mod)
{
	int i;
	int j;
	msurface_t *surf;
	msurface_t **marksurf;
	mleaf_t *leaf;
	int runaway;
	int currentvolume;

	// mark all surfs underwater as volume -1 and all above water as 0
	for (i = 0, surf = mod->surfaces; i < mod->numsurfaces; i++, surf++)
	{
		surf->volume = -1;
		surf->volumetexture = NULL;

		// set the liquid type
		if (!surf->texinfo->texture->gl_texture)
		{
			// catch e2m3 invalid textures so it doesn't crash
		}
		else if (surf->flags & SURF_DRAWTURB)
		{
			if (surf->flags & SURF_DRAWLAVA)
				surf->texinfo->texture->gl_texture->liquidtype = LIQUIDTYPE_LAVA;
			else if (surf->flags & SURF_DRAWTELE)
				surf->texinfo->texture->gl_texture->liquidtype = LIQUIDTYPE_TELE;
			else if (surf->flags & SURF_DRAWSLIME)
				surf->texinfo->texture->gl_texture->liquidtype = LIQUIDTYPE_SLIME;
			else surf->texinfo->texture->gl_texture->liquidtype = LIQUIDTYPE_WATER;
		}
		else if (surf->flags & SURF_DRAWSKY)
		{
			// catch sky so it doesn't crash
		}
		else surf->texinfo->texture->gl_texture->liquidtype = LIQUIDTYPE_NOTLIQUID;

		if (surf->flags & SURF_DRAWTELE)
		{
			// don't mark volumes in teleports
			surf->volume = 0;
			continue;
		}

		if (surf->flags & SURF_UNDERWATER) continue;

		surf->volume = 0;
	}

	// mark all leafs underwater as volume -1 and all above water as 0
	for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
	{
		leaf->volume = -1;
		leaf->volumetexture = NULL;

		if (leaf->nummarksurfaces == 0)
		{
			leaf->volume = 0;
			continue;
		}

		if (leaf->contents == CONTENTS_WATER) continue;
		if (leaf->contents == CONTENTS_SLIME) continue;
		if (leaf->contents == CONTENTS_LAVA) continue;

		leaf->volume = 0;
	}

	// first pass sets the initial volumes on underwater leafs containing turbulent surfs
	// this gives us our base to work on
	currentvolume = 0;

	for (i = 0, surf = mod->surfaces; i < mod->numsurfaces; i++, surf++)
	{
		// find all surfs which are turbulent and underwater
		if (surf->volume == -1 && (surf->flags & SURF_DRAWTURB))
		{
			// set the surfs volume and volumetexture
			surf->volume = ++currentvolume;
			surf->volumetexture = surf->texinfo->texture;

			// find all leafs which contain this surf and mark them
			MarkLeafsForSurf (mod, surf);
		}
	}

	// infinite loop cos we may have to run this a number of times to pick up stragglers
	runaway = 0;

	while (1)
	{
		qboolean straggle = false;

		// if it takes more than this number of passes, we're highly unlikely to ever get it
		// (to be honest, at least 9 times out of 10 a single pass here will get us all we're going to get)
		if (runaway > 10)
		{
			Con_DPrintf ("Runaway...\n");
			break;
		}

		runaway++;

		// we have a set of marked leafs now, so for all subsequent passes we'll get the visibility for each of these
		for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
		{
			// we're only interested in leafs which were marked in a previous pass
			if (leaf->volume < 1) continue;

			// we have a leaf which was marked, so now get the PVS
			MarkPVSForLeaf (mod, leaf);
		}

		// check for stragglers in the leafs
		for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
		{
			if (leaf->volume == -1)
			{
				straggle = true;
				break;
			}
		}

		if (straggle) continue;

		// check for stragglers in the surfs
		for (i = 0, surf = mod->surfaces; i < mod->numsurfaces; i++, surf++)
		{
			if (surf->volume == -1)
			{
				straggle = true;
				break;
			}
		}

		if (straggle) continue;

		// no more stragglers
		break;
	}

	// next pass
	// if we have any leafs left with a volume of -1 they're not showing up in anybody elses PVS, but a valid leaf may
	// show up in theirs, so check for that.  again, we do an infinite loop.
	runaway = 0;

	while (1)
	{
		qboolean straggle = false;

		// if it takes more than this number of passes, we're highly unlikely to ever get it
		if (runaway > 10)
		{
			Con_DPrintf ("Runaway...\n");
			break;
		}

		runaway++;

		for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
		{
			if (leaf->volume == -1)
			{
				MarkStragglePVSForLeaf (mod, leaf);
			}
		}

		// check for stragglers in the leafs
		for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
		{
			if (leaf->volume == -1)
			{
				straggle = true;
				break;
			}
		}

		if (straggle) continue;

		// check for stragglers in the surfs
		for (i = 0, surf = mod->surfaces; i < mod->numsurfaces; i++, surf++)
		{
			if (surf->volume == -1)
			{
				straggle = true;
				break;
			}
		}

		if (straggle) continue;

		// no more stragglers
		break;
	}

	// next pass
	// if we have any surfs left with a volume of -1, they may be contained in a leaf that has a valid volume, so go through all
	// leafs that have a valid volume and set their marksurfaces
	for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
	{
		// invalid volume for marking
		if (leaf->volume == -1) continue;
		if (leaf->volume == 0) continue;

		for (j = 0, marksurf = leaf->firstmarksurface; j < leaf->nummarksurfaces; j++, marksurf++)
		{
			// already marked
			if ((*marksurf)->volume != -1) continue;

			(*marksurf)->volume = leaf->volume;
			(*marksurf)->volumetexture = leaf->volumetexture;
		}
	}

	// final pass!!!
	// anything left over defies the laws of quake as I understand them.  it's likely a leaf or surf tucked away
	// in some totally obscure place: e.g. a totally contained underwater box with no liquid surfs in it.
	// even on the most extremely craply compiled maps (e2m3) the whole process is reliable, so it's a fair assumption.
	// just mark anything left then with 0 and NULL...
	for (i = 0, leaf = mod->leafs; i < mod->numleafs; i++, leaf++)
	{
		if (leaf->volume == -1)
		{
			leaf->volume = 0;
			leaf->volumetexture = NULL;
		}
	}

	for (i = 0, surf = mod->surfaces; i < mod->numsurfaces; i++, surf++)
	{
		if (surf->volume == -1)
		{
			surf->volume = 0;
			surf->volumetexture = NULL;
		}
	}
}



/*
================================================================================================================

										AUTOMATIC WATER TRANSLUCENCY

	On maps that haven't been vis'ed for translucent water, the result is horrible and so 1996.  Unfortunately,
	there are still quite a few "old favourites" around from the Pentium 1 days when raw speed was considered
	an absolute priority.  Here, we make a reasonable stab at doing the necessary.  The correct way is to
	reportalise the BSP and do a full re-vis of it, but that may require who knows how long to do, and will
	also require the new version to be redistributed either as a new map or as a vispatch.  This method is
	best suited to ID1 size maps where it is reasonable to make a hack of this nature, and it won't take too
	long to do or cause the map to run slowly on modern hardware.  It's not perfect, but it sure does the job.

================================================================================================================
*/

