/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// models.c -- model loading and caching

// models are the only shared resource between a client and server running
// on the same machine.

#include "gl_state.h"

model_t	*loadmodel;
char	loadname[32];	// for hunk tags

void MH_WorldVolumeCarve (model_t *mod);


void Mod_LoadSpriteModel (model_t *mod, void *buffer);
void Mod_LoadBrushModel (model_t *mod, void *buffer);
void Mod_LoadAliasModel (model_t *mod, void *buffer);
model_t *Mod_LoadModel (model_t *mod, qboolean crash);

byte	mod_novis[MAX_MAP_LEAFS/8];

#define	MAX_MOD_KNOWN	512
model_t	mod_known[MAX_MOD_KNOWN];
int		mod_numknown;

char ModGameDir[64];

/*
===============
Mod_Init
===============
*/
void Mod_Init (void)
{
	memset (mod_novis, 0xff, sizeof(mod_novis));
}

/*
===============
Mod_Init

Caches the data if needed
===============
*/
void *Mod_Extradata (model_t *mod)
{
	void	*r;
	
	r = Cache_Check (&mod->cache);
	if (r)
		return r;

	Mod_LoadModel (mod, true);
	
	if (!mod->cache.data)
		Sys_Error ("Mod_Extradata: caching failed");
	return mod->cache.data;
}

/*
===============
Mod_PointInLeaf
===============
*/
mleaf_t *Mod_PointInLeaf (vec3_t p, model_t *model)
{
	mnode_t		*node;
	float		d;
	mplane_t	*plane;
	
	if (!model || !model->nodes)
		Sys_Error ("Mod_PointInLeaf: bad model");

	node = model->nodes;
	while (1)
	{
		if (node->contents < 0)
			return (mleaf_t *)node;
		plane = node->plane;
		d = DotProduct (p,plane->normal) - plane->dist;
		if (d > 0)
			node = node->children[0];
		else
			node = node->children[1];
	}
	
	return NULL;	// never reached
}


/*
===================
Mod_DecompressVis
===================
*/
byte *Mod_DecompressVis (byte *in, model_t *model)
{
	static byte	decompressed[MAX_MAP_LEAFS/8];
	int		c;
	byte	*out;
	int		row;

	row = (model->numleafs+7)>>3;	
	out = decompressed;

#if 0
	memcpy (out, in, row);
#else
	if (!in)
	{	// no vis info, so make all visible
		while (row)
		{
			*out++ = 0xff;
			row--;
		}
		return decompressed;		
	}

	do
	{
		if (*in)
		{
			*out++ = *in++;
			continue;
		}
	
		c = in[1];
		in += 2;
		while (c)
		{
			*out++ = 0;
			c--;
		}
	} while (out - decompressed < row);
#endif
	
	return decompressed;
}

/*
byte *Mod_LeafPVS (mleaf_t *leaf, model_t *model)
{
	if (leaf == model->leafs)
		return mod_novis;
	return Mod_DecompressVis (leaf->compressed_vis, model);
}
*/
// mh - auto water trans begin
byte *Mod_LeafPVS (mleaf_t *leaf, model_t *model)
{
	// leaf 0 is the solid leaf
	if (leaf == model->leafs) return mod_novis;

	// no visdata
	if (!leaf->decompressed_vis) return mod_novis;

	// return the decompressed vis
	return leaf->decompressed_vis;
}
// mh - auto water trans end


/*
===================
Mod_ClearAll
===================
*/
void Mod_ClearAll (void)
{
	int		i;
	model_t	*mod;
	
	for (i=0 , mod=mod_known ; i<mod_numknown ; i++, mod++)
		if (mod->type != mod_alias)
			mod->needload = true;
}

/*
==================
Mod_FindName

==================
*/
model_t *Mod_FindName (char *name)
{
	int		i;
	model_t	*mod;
	
	if (!name[0])
		Sys_Error ("Mod_ForName: NULL name");
		
//
// search the currently loaded models
//
	for (i=0 , mod=mod_known ; i<mod_numknown ; i++, mod++)
		if (!strcmp (mod->name, name) )
			break;
			
	if (i == mod_numknown)
	{
		if (mod_numknown == MAX_MOD_KNOWN)
			Sys_Error ("mod_numknown == MAX_MOD_KNOWN");
		strcpy (mod->name, name);
		mod->needload = true;
		mod_numknown++;
	}

	return mod;
}

/*
==================
Mod_TouchModel

==================
*/
void Mod_TouchModel (char *name)
{
	model_t	*mod;
	
	mod = Mod_FindName (name);
	
	if (!mod->needload)
	{
		if (mod->type == mod_alias)
			Cache_Check (&mod->cache);
	}
}

/*
==================
Mod_LoadModel

Loads a model into the cache
==================
*/
model_t *Mod_LoadModel (model_t *mod, qboolean crash)
{
	void	*d;
	unsigned *buf;
	byte	stackbuf[1024];

	if (!mod->needload)
	{
		if (mod->type == mod_alias)
		{
			d = Cache_Check (&mod->cache);

			if (d) return mod;
		}
		else return mod;
	}

	// because the world is so huge, load it one piece at a time
	if (!crash)
	{
	}

	// load the file
	buf = (unsigned *) COM_LoadStackFile (mod->name, stackbuf, sizeof (stackbuf));

	if (!buf)
	{
		if (crash) Sys_Error ("Mod_NumForName: %s not found", mod->name);

		return NULL;
	}

	// allocate a new model
	COM_FileBase (mod->name, loadname);

	loadmodel = mod;

	// fill it in
	mod->needload = false;

	// call the apropriate loader
	switch (LittleLong (*(unsigned *) buf))
	{
	case IDPOLYHEADER:
		Mod_LoadAliasModel (mod, buf);
		break;

	case IDSPRITEHEADER:
		Mod_LoadSpriteModel (mod, buf);
		break;

	default:
		Mod_LoadBrushModel (mod, buf);
		break;
	}

	return mod;
}

/*
==================
Mod_ForName

Loads in a model for the given name
==================
*/
model_t *Mod_ForName (char *name, qboolean crash)
{
	model_t	*mod;
	
	mod = Mod_FindName (name);
	
	return Mod_LoadModel (mod, crash);
}


/*
===============================================================================

					BRUSHMODEL LOADING

===============================================================================
*/

byte	*mod_base;


/*
=================
Mod_LoadTextures
=================
*/
void Mod_LoadTextures (lump_t *l)
{
	int		i, j, pixels, num, max, altmax;
	miptex_t	*mt;
	texture_t	*tx, *tx2;
	texture_t	*anims[10];
	texture_t	*altanims[10];
	dmiptexlump_t *m;

	if (!l->filelen)
	{
		loadmodel->textures = NULL;
		return;
	}
	m = (dmiptexlump_t *)(mod_base + l->fileofs);
	
	m->nummiptex = LittleLong (m->nummiptex);
	
	loadmodel->numtextures = m->nummiptex;
	loadmodel->textures = Hunk_AllocName (m->nummiptex * sizeof(*loadmodel->textures) , loadname);

	for (i=0 ; i<m->nummiptex ; i++)
	{
		m->dataofs[i] = LittleLong(m->dataofs[i]);

		if (m->dataofs[i] == -1) continue;

		mt = (miptex_t *)((byte *)m + m->dataofs[i]);
		mt->width = LittleLong (mt->width);
		mt->height = LittleLong (mt->height);

		if ( (mt->width & 15) || (mt->height & 15) )
			Sys_Error ("Texture %s is not 16 aligned", mt->name);

		pixels = mt->width*mt->height/64*85;
		tx = Hunk_AllocName (sizeof(texture_t) +pixels, loadname );

		loadmodel->textures[i] = tx;

		// store the texture number for volume carving
		tx->num = i;

		memcpy (tx->name, mt->name, sizeof (tx->name));

		tx->width = mt->width;
		tx->height = mt->height;

		// HACK HACK HACK
		// Note - CRC doesn't guarantee uniqueness, but it'll do for a hack
		if (!strcmp (mt->name, "shot1sid") && mt->width == 32 && mt->height == 32 && CRC_Block ((byte *)(mt + 1), mt->width * mt->height) == 65393)
		{
			// This texture in b_shell1.bsp has some of the first 32 pixels painted white.
			// They are invisible in software, but look really ugly in GL. So we just copy
			// 32 pixels from the bottom to make it look nice.
			memcpy (mt + 1, (byte *) (mt + 1) + 32 * 31, 32);
		}

		// the pixels immediately follow the structures
		memcpy (tx+1, mt+1, pixels);

		// set the data pointer for loading
		tx->data = (byte *) (tx + 1);

		if (!Q_strncmp (mt->name, "sky", 3))	
			R_InitSky (tx);
		else if (mt->name[0] == '*')
			tx->gl_texture = GL_LoadQuakeTexture (NULL, tx->name, tx->data, tx->width, tx->height, (TEX_MIPMAP | TEX_BSP));
		else tx->gl_texture = GL_LoadQuakeTexture (NULL, tx->name, tx->data, tx->width, tx->height, (TEX_MIPMAP | TEX_BSP | TEX_MAKENORMAL));
	}

	// sequence the animations
	for (i=0 ; i<m->nummiptex ; i++)
	{
		tx = loadmodel->textures[i];
		if (!tx || tx->name[0] != '+')
			continue;
		if (tx->anim_next)
			continue;	// allready sequenced

		// find the number of frames in the animation
		memset (anims, 0, sizeof(anims));
		memset (altanims, 0, sizeof(altanims));

		max = tx->name[1];
		altmax = 0;
		if (max >= 'a' && max <= 'z')
			max -= 'a' - 'A';
		if (max >= '0' && max <= '9')
		{
			max -= '0';
			altmax = 0;
			anims[max] = tx;
			max++;
		}
		else if (max >= 'A' && max <= 'J')
		{
			altmax = max - 'A';
			max = 0;
			altanims[altmax] = tx;
			altmax++;
		}
		else
			Sys_Error ("Bad animating texture %s", tx->name);

		for (j=i+1 ; j<m->nummiptex ; j++)
		{
			tx2 = loadmodel->textures[j];
			if (!tx2 || tx2->name[0] != '+')
				continue;
			if (strcmp (tx2->name+2, tx->name+2))
				continue;

			num = tx2->name[1];
			if (num >= 'a' && num <= 'z')
				num -= 'a' - 'A';
			if (num >= '0' && num <= '9')
			{
				num -= '0';
				anims[num] = tx2;
				if (num+1 > max)
					max = num + 1;
			}
			else if (num >= 'A' && num <= 'J')
			{
				num = num - 'A';
				altanims[num] = tx2;
				if (num+1 > altmax)
					altmax = num+1;
			}
			else
				Sys_Error ("Bad animating texture %s", tx->name);
		}
		
#define	ANIM_CYCLE	2
	// link them all together
		for (j=0 ; j<max ; j++)
		{
			tx2 = anims[j];
			if (!tx2)
				Sys_Error ("Missing frame %i of %s",j, tx->name);
			tx2->anim_total = max * ANIM_CYCLE;
			tx2->anim_min = j * ANIM_CYCLE;
			tx2->anim_max = (j+1) * ANIM_CYCLE;
			tx2->anim_next = anims[ (j+1)%max ];
			if (altmax)
				tx2->alternate_anims = altanims[0];
		}
		for (j=0 ; j<altmax ; j++)
		{
			tx2 = altanims[j];
			if (!tx2)
				Sys_Error ("Missing frame %i of %s",j, tx->name);
			tx2->anim_total = altmax * ANIM_CYCLE;
			tx2->anim_min = j * ANIM_CYCLE;
			tx2->anim_max = (j+1) * ANIM_CYCLE;
			tx2->anim_next = altanims[ (j+1)%altmax ];
			if (max)
				tx2->alternate_anims = anims[0];
		}
	}
}

/*
=================
Mod_LoadLighting

Contains some evil hackery to rescale the colour balance of lightmaps so that they preserve the same
intensity as the original BSP lighting.  This is needed cos various versions of MHColour didn't quite
do the job properly, and it's too late to go back now...
=================
*/
void Mod_LoadBSPLighting (lump_t *l)
{
	int	i;				// loop control
	int j;
	byte *light_base;	// base offset of light data
	FILE *LITFile;
	char LITFileName[1024];
	litheader_t	LITHeader;
	byte *light_buffer;
	int bspintens;
	int litintens;
	int templ;
	int hit;
	int ntscscales[3] = {30, 59, 11};

	if (!l->filelen)
	{
		loadmodel->lightdata = NULL;
		return;
	}

	// load from the BSP and expand to 3 component
	loadmodel->lightdata = Hunk_AllocName (l->filelen * 3, loadname);
	light_base = mod_base + l->fileofs;

	for (i = 0; i < l->filelen; i++)
	{
		// unnecessary brackets improve legibility and clarity
		loadmodel->lightdata[(i * 3) + 0] = light_base[i];
		loadmodel->lightdata[(i * 3) + 1] = light_base[i];
		loadmodel->lightdata[(i * 3) + 2] = light_base[i];
	}

	// build the name of the lit file
	strcpy (LITFileName, loadmodel->name);

	// we can't just search for '.' cos there may be one in a folder name!!!
	for (i = 0; LITFileName[i] != '\0'; i++);

	LITFileName[i - 3] = 'l';
	LITFileName[i - 2] = 'i';
	LITFileName[i - 1] = 't';

	// use my custom MH version so that LIT files are only read from the correct game dir
	// (e.g. hipnotic, rogue and id1 might all have a start.lit.  it would be a shame if 
	// id1's version got loaded if hipnotic had no start.lit)
	COM_FOpenFileInFolder (LITFileName, ModGameDir, &LITFile);

	// no lit file
	if (!LITFile) return;

	// read and validate the header
	fread (&LITHeader, sizeof (LITHeader), 1, LITFile);

	if (!(LITHeader.version == 1 && LITHeader.ident[0] == 'Q' && LITHeader.ident[1] == 'L' && LITHeader.ident[2] == 'I' && LITHeader.ident[3] == 'T'))
	{
		// bad lit file
		fclose (LITFile);
		return;
	}

	// it's valid
	light_buffer = (byte *) malloc (l->filelen * 3);

	// read it in
	fread (light_buffer, l->filelen * 3, 1, LITFile);

	// now compare with the light from the BSP - incorrect lightmap sizes in
	// mhcolour can cause some weird effects in some areas - the nailgun room
	// in e1m1 is a perfect example.  this is a hack - the real solution
	// would be to rewrite mhcolour to fix this.  too late for that though...
	for (i = 0; i < l->filelen; i++)
	{
		bspintens = 0;
		litintens = 0;
		hit = 0;

		for (j = 0; j < 3; j++)
		{
			// skip LIT loading on these conditions
			if (light_buffer[(i * 3) + j] == loadmodel->lightdata[(i * 3) + j]) continue;
			if (light_buffer[(i * 3) + j] == 0) continue;
			if (loadmodel->lightdata[(i * 3) + j] == 0) continue;
			if (light_buffer[(i * 3) + j] == 255) continue;
			if (loadmodel->lightdata[(i * 3) + j] == 255) continue;

			// load it in, and store the scales
			bspintens += (int) loadmodel->lightdata[(i * 3) + j] * ntscscales[j];
			litintens += (int) light_buffer[(i * 3) + j] * ntscscales[j];
			hit++;

			loadmodel->lightdata[(i * 3) + j] = light_buffer[(i * 3) + j];
		}

		// we need 3 hits for the next stage
		if (hit != 3) continue;

		// other continue conditions
		if (bspintens == litintens) continue;
		if (!bspintens) continue;
		if (!litintens) continue;

		// rescale the lit file to match the original in the bsp
		for (j = 0; j < 3; j++)
		{
			// scale
			templ = loadmodel->lightdata[(i * 3) + j];
			templ *= bspintens;
			templ /= litintens;

			// bound
			if (templ > 255) templ = 255;
			if (templ < 0) templ = 0;

			// store
			loadmodel->lightdata[(i * 3) + j] = (byte) templ;
		}
	}

	free (light_buffer);

	fclose (LITFile);
}


void Mod_LoadLighting (lump_t *l)
{
	int i;
	unsigned short *out;

	// call to Mod_LoadBSPLighting (which was the previous lighting loader) as a seperate func
	// so that we can set stainmaps correctly irrespective of the return point from that one.
	Mod_LoadBSPLighting (l);

	// set stainmaps to NULL if there's no lightdata.  this is unfortunate, but we really do need
	// the light offsets in the BSP.
	if (!l->filelen)
	{
		loadmodel->staindata = NULL;
		return;
	}

	// now set up staindata
	loadmodel->staindata = Hunk_AllocName (l->filelen * 3 * sizeof (short), loadname);
	out = loadmodel->staindata;

	// clear to nothing
	for (i = 0; i < l->filelen; i++)
	{
		*out++ = 0;
		*out++ = 0;
		*out++ = 0;
	}
}


/*
=================
Mod_LoadVisibility
=================
*/
void Mod_LoadVisibility (lump_t *l)
{
	if (!l->filelen)
	{
		loadmodel->visdata = NULL;
		return;
	}
	loadmodel->visdata = Hunk_AllocName ( l->filelen, loadname);	
	memcpy (loadmodel->visdata, mod_base + l->fileofs, l->filelen);
}


/*
=================
Mod_LoadEntities
=================
*/
void Mod_LoadEntities (lump_t *l)
{
	if (!l->filelen)
	{
		loadmodel->entities = NULL;
		loadmodel->numentities = 0;
		return;
	}

	loadmodel->entities = Hunk_AllocName ( l->filelen, loadname);
	memcpy (loadmodel->entities, mod_base + l->fileofs, l->filelen);

	loadmodel->numentities = l->filelen;
}


/*
=================
Mod_LoadVertexes
=================
*/
void Mod_LoadVertexes (lump_t *l)
{
	dvertex_t	*in;
	mvertex_t	*out;
	int			i, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->vertexes = out;
	loadmodel->numvertexes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->position[0] = LittleFloat (in->point[0]);
		out->position[1] = LittleFloat (in->point[1]);
		out->position[2] = LittleFloat (in->point[2]);
	}
}

/*
=================
Mod_LoadSubmodels
=================
*/
void Mod_LoadSubmodels (lump_t *l)
{
	dmodel_t	*in;
	dmodel_t	*out;
	int			i, j, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->submodels = out;
	loadmodel->numsubmodels = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{	// spread the mins / maxs by a pixel
			out->mins[j] = LittleFloat (in->mins[j]) - 1;
			out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
			out->origin[j] = LittleFloat (in->origin[j]);
		}
		for (j=0 ; j<MAX_MAP_HULLS ; j++)
			out->headnode[j] = LittleLong (in->headnode[j]);
		out->visleafs = LittleLong (in->visleafs);
		out->firstface = LittleLong (in->firstface);
		out->numfaces = LittleLong (in->numfaces);
	}
}

/*
=================
Mod_LoadEdges
=================
*/
void Mod_LoadEdges (lump_t *l)
{
	dedge_t *in;
	medge_t *out;
	int 	i, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( (count + 1) * sizeof(*out), loadname);	

	loadmodel->edges = out;
	loadmodel->numedges = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->v[0] = (unsigned short)LittleShort(in->v[0]);
		out->v[1] = (unsigned short)LittleShort(in->v[1]);
	}
}

/*
=================
Mod_LoadTexinfo
=================
*/
void Mod_LoadTexinfo (lump_t *l)
{
	texinfo_t *in;
	mtexinfo_t *out;
	int i, j, count;
	int miptex;
	float len1, len2;

	in = (void *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*in)) Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = Hunk_AllocName (count * sizeof (*out), loadname);

	loadmodel->texinfo = out;
	loadmodel->numtexinfo = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 8; j++) out->vecs[0][j] = LittleFloat (in->vecs[0][j]);

		len1 = Length (out->vecs[0]);
		len2 = Length (out->vecs[1]);
		len1 = (len1 + len2) / 2;

		if (len1 < 0.32)
			out->mipadjust = 4;
		else if (len1 < 0.49)
			out->mipadjust = 3;
		else if (len1 < 0.99)
			out->mipadjust = 2;
		else
			out->mipadjust = 1;

		miptex = LittleLong (in->miptex);
		out->flags = LittleLong (in->flags);

		if (!loadmodel->textures)
		{
			// checkerboard texture
			out->texture = r_notexture_mip;
			out->flags = 0;
		}
		else
		{
			// MH - this was a Sys_Error
			if (miptex >= loadmodel->numtextures) goto SetNoTexture;

			out->texture = loadmodel->textures[miptex];

			if (!out->texture)
			{
SetNoTexture:;
				// texture not found
				out->texture = r_notexture_mip;
				out->flags = 0;
			}
		}
	}
}

/*
================
CalcSurfaceExtents

Fills in s->texturemins[] and s->extents[]
================
*/
void CalcSurfaceExtents (msurface_t *s)
{
	float	mins[2], maxs[2], val;
	int		i,j, e;
	mvertex_t	*v;
	mtexinfo_t	*tex;
	int		bmins[2], bmaxs[2];

	mins[0] = mins[1] = 999999;
	maxs[0] = maxs[1] = -99999;

	tex = s->texinfo;
	
	for (i=0 ; i<s->numedges ; i++)
	{
		e = loadmodel->surfedges[s->firstedge+i];
		if (e >= 0)
			v = &loadmodel->vertexes[loadmodel->edges[e].v[0]];
		else
			v = &loadmodel->vertexes[loadmodel->edges[-e].v[1]];
		
		for (j=0 ; j<2 ; j++)
		{
			val = v->position[0] * tex->vecs[j][0] + 
				v->position[1] * tex->vecs[j][1] +
				v->position[2] * tex->vecs[j][2] +
				tex->vecs[j][3];
			if (val < mins[j])
				mins[j] = val;
			if (val > maxs[j])
				maxs[j] = val;
		}
	}

	for (i=0 ; i<2 ; i++)
	{	
		bmins[i] = floor(mins[i]/16);
		bmaxs[i] = ceil(maxs[i]/16);

		s->texturemins[i] = bmins[i] * 16;
		s->extents[i] = (bmaxs[i] - bmins[i]) * 16;
		if ( !(tex->flags & TEX_SPECIAL) && s->extents[i] > 2048 )
			Sys_Error ("Bad surface extents");
	}
}


/*
=================
Mod_LoadFaces
=================
*/
void Mod_LoadFaces (lump_t *l)
{
	dface_t		*fa;
	msurface_t 	*surf;
	int			i, count, surfnum;
	int			planenum, side;

	fa = (void *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*fa))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*fa);
	surf = Hunk_AllocName (count * sizeof (*surf), loadname);	

	loadmodel->surfaces = surf;
	loadmodel->numsurfaces = count;

	for ( surfnum=0 ; surfnum<count ; surfnum++, fa++, surf++)
	{
		surf->firstedge = LittleLong(fa->firstedge);
		surf->numedges = LittleShort(fa->numedges);		
		surf->flags = 0;

		planenum = LittleShort(fa->planenum);
		side = LittleShort(fa->side);

		if (side) surf->flags |= SURF_PLANEBACK;			

		surf->plane = loadmodel->planes + planenum;

		// hacky fix.  Quake's normal is used for BSP tree traversal, but does not take backfacing
		// into account, and so is inappropriate for OpenGL use.  Here we calculate one that does.
		// (MHQuake doesn't actually use this for anything yet, but it's useful to have in case I ever do need it)
		if (side)
		{
			surf->truenormal[0] = surf->plane->normal[0] * -1;
			surf->truenormal[1] = surf->plane->normal[1] * -1;
			surf->truenormal[2] = surf->plane->normal[2] * -1;
		}
		else
		{
			surf->truenormal[0] = surf->plane->normal[0];
			surf->truenormal[1] = surf->plane->normal[1];
			surf->truenormal[2] = surf->plane->normal[2];
		}

		// make it unit length
		VectorNormalize (surf->truenormal);

		surf->texinfo = loadmodel->texinfo + LittleShort (fa->texinfo);

		CalcSurfaceExtents (surf);

		// set up light styles
		for (i = 0; i < MAXLIGHTMAPS; i++)
			surf->styles[i] = fa->styles[i];

		i = LittleLong (fa->lightofs);

		// replicated surf->samples for surf->stainsamples
		// also fixed up where loadmodel->lightdata may be NULL but i may have a value from
		// either a crappy BSP compiler or something else going wrong somewhere
		if (i == -1 || !loadmodel->lightdata)
		{
			surf->samples = NULL;
			surf->stainsamples = NULL;
		}
		else
		{
			surf->samples = loadmodel->lightdata + (i * 3);
			surf->stainsamples = loadmodel->staindata + (i * 3);
		}

		// store the surface number
		surf->surfnum = surfnum;

		// set an initial valid lightmap pointer.
		// as stainmaps will be stored in the same struct, this will suffice for both
		surf->lightmap = NULL;

		surf->volume = -1;
		surf->volumetexture = NULL;

		// precalculate the surface warp indexes
		// this is valid for liquid and for normal surfs (needed for caustics)
		// we really should be setting a plane coefficient for either the texture matrix or some
		// texgen stuff based on surf->truenormal here, but bleagh.  maybe someday...
		switch (surf->plane->type)
		{
		case PLANE_Y:
			// vertical in line with the world grid
			surf->warpindex[0] = 0;
			surf->warpindex[1] = 2;
			break;

		case PLANE_X:
			// vertical in line with the world grid
			surf->warpindex[0] = 2;
			surf->warpindex[1] = 1;
			break;

		case PLANE_ANYX:
		case PLANE_ANYY:
			// vertical at an angle to the world grid
			surf->warpindex[0] = 2;
			surf->warpindex[1] = 0;
			break;

		case PLANE_Z:
			// flat
		default:
			// same as flat
			surf->warpindex[0] = 0;
			surf->warpindex[1] = 1;
			break;
		}

		// catch e2m3 crapness
		if (surf->texinfo->texture->name[0] == 0)
			surf->texinfo->texture->gl_texture = &r_black;

		// set the drawing flags flag
		if (!Q_strncmp(surf->texinfo->texture->name, "sky", 3))	// sky
		{
			surf->flags |= (SURF_DRAWSKY | SURF_DRAWTILED);
			//GL_SubdivideSurface (out);	// cut up polygon for warps
			continue;
		}

		if (surf->texinfo->texture->gl_texture->flags & TEX_LIQUID)
		{
			surf->flags |= (SURF_DRAWTURB | SURF_DRAWTILED);

			// set the contents type here
			if (surf->texinfo->texture->name[1] == 'l' &&
				surf->texinfo->texture->name[2] == 'a' &&
				surf->texinfo->texture->name[3] == 'v' &&
				surf->texinfo->texture->name[4] == 'a')
				surf->flags |= SURF_DRAWLAVA;
			else if (surf->texinfo->texture->name[1] == 't' &&
				surf->texinfo->texture->name[2] == 'e' &&
				surf->texinfo->texture->name[3] == 'l' &&
				surf->texinfo->texture->name[4] == 'e')
				surf->flags |= SURF_DRAWTELE;
			else if (surf->texinfo->texture->name[1] == 's' &&
				surf->texinfo->texture->name[2] == 'l' &&
				surf->texinfo->texture->name[3] == 'i' &&
				surf->texinfo->texture->name[4] == 'm' &&
				surf->texinfo->texture->name[5] == 'e')
				surf->flags |= SURF_DRAWSLIME;
			else surf->flags |= SURF_DRAWWATER;

			for (i = 0; i < 2; i++)
			{
				surf->extents[i] = 16384;
				surf->texturemins[i] = -8192;
			}

			//GL_SubdivideSurface (out);	// cut up polygon for warps
			continue;
		}
	}
}


/*
=================
Mod_SetParent
=================
*/
void Mod_SetParent (mnode_t *node, mnode_t *parent)
{
	node->parent = parent;
	if (node->contents < 0)
		return;
	Mod_SetParent (node->children[0], node);
	Mod_SetParent (node->children[1], node);
}

/*
=================
Mod_LoadNodes
=================
*/
void Mod_LoadNodes (lump_t *l)
{
	int			i, j, count, p;
	dnode_t		*in;
	mnode_t 	*out;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->nodes = out;
	loadmodel->numnodes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{
			out->minmaxs[j] = LittleShort (in->mins[j]);
			out->minmaxs[3+j] = LittleShort (in->maxs[j]);
		}
	
		p = LittleLong(in->planenum);
		out->plane = loadmodel->planes + p;

		out->firstsurface = LittleShort (in->firstface);
		out->numsurfaces = LittleShort (in->numfaces);
		
		for (j=0 ; j<2 ; j++)
		{
			p = LittleShort (in->children[j]);
			if (p >= 0)
				out->children[j] = loadmodel->nodes + p;
			else
				out->children[j] = (mnode_t *)(loadmodel->leafs + (-1 - p));
		}
	}
	
	Mod_SetParent (loadmodel->nodes, NULL);	// sets nodes and leafs
}

/*
=================
Mod_LoadLeafs
=================
*/
//void Mod_LoadLeafs (lump_t *l)
// mh - auto water trans
void Mod_LoadLeafs (lump_t *l, lump_t *v)
{
	dleaf_t 	*in;
	mleaf_t 	*out;
	int			i, j, count, p;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->leafs = out;
	loadmodel->numleafs = count;

	// mh - auto water trans begin
	// Mod_LoadVisibility stuff
	// we know how many leafs there are in the map, so we can set up the visdata now.
	if (v->filelen)
	{
		// add 7 to prevent round down errors with int division by 8.
		loadmodel->vissize = (loadmodel->numleafs + 7) / 8;

		loadmodel->visdata = Hunk_AllocName (loadmodel->numleafs * loadmodel->vissize, loadname);

		// set to all solid initially
		memset (loadmodel->visdata, 0xff, loadmodel->numleafs * loadmodel->vissize);
	}
	else
	{
		loadmodel->vissize = 0;
		loadmodel->visdata = NULL;
	}
	// mh - auto water trans end

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{
			out->minmaxs[j] = LittleShort (in->mins[j]);
			out->minmaxs[3+j] = LittleShort (in->maxs[j]);
		}

		p = LittleLong(in->contents);
		out->contents = p;

		out->firstmarksurface = loadmodel->marksurfaces +
			LittleShort(in->firstmarksurface);
		out->nummarksurfaces = LittleShort(in->nummarksurfaces);
		
		/*
		p = LittleLong(in->visofs);
		if (p == -1)
			out->compressed_vis = NULL;
		else
			out->compressed_vis = loadmodel->visdata + p;
			*/

		// mh - auto water trans begin
		p = LittleLong (in->visofs);

		// if the leaf has a valid visoffset and the model has visdata and the
		// leaf is one of the valid types for having visdata, calculate visdata
		// this is likely multiple redundant checking...
		if (p != -1 && loadmodel->visdata && (out->contents == CONTENTS_EMPTY || 
			out->contents == CONTENTS_WATER || out->contents == CONTENTS_LAVA || 
			out->contents == CONTENTS_SLIME || out->contents == CONTENTS_SKY))
		{
			// paranoid about brackets!!!
			out->decompressed_vis = loadmodel->visdata + (i * loadmodel->vissize);

			// copy it in
			memcpy (out->decompressed_vis, 
				Mod_DecompressVis (mod_base + v->fileofs + p, loadmodel), 
				loadmodel->vissize);
		}
		else out->decompressed_vis = NULL;
		// mh - auto water trans end

		out->efrags = NULL;

		out->volume = -1;
		out->volumetexture = NULL;
		
		for (j=0 ; j<4 ; j++)
			out->ambient_sound_level[j] = in->ambient_level[j];

		// gl underwater warp
		if (out->contents != CONTENTS_EMPTY)
		{
			for (j=0 ; j<out->nummarksurfaces ; j++)
				out->firstmarksurface[j]->flags |= SURF_UNDERWATER;
		}
	}	
}

/*
=================
Mod_LoadClipnodes
=================
*/
void Mod_LoadClipnodes (lump_t *l)
{
	dclipnode_t *in, *out;
	int			i, count;
	hull_t		*hull;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->clipnodes = out;
	loadmodel->numclipnodes = count;

	hull = &loadmodel->hulls[1];
	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count-1;
	hull->planes = loadmodel->planes;
	hull->clip_mins[0] = -16;
	hull->clip_mins[1] = -16;
	hull->clip_mins[2] = -24;
	hull->clip_maxs[0] = 16;
	hull->clip_maxs[1] = 16;
	hull->clip_maxs[2] = 32;

	hull = &loadmodel->hulls[2];
	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count-1;
	hull->planes = loadmodel->planes;
	hull->clip_mins[0] = -32;
	hull->clip_mins[1] = -32;
	hull->clip_mins[2] = -24;
	hull->clip_maxs[0] = 32;
	hull->clip_maxs[1] = 32;
	hull->clip_maxs[2] = 64;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->planenum = LittleLong(in->planenum);
		out->children[0] = LittleShort(in->children[0]);
		out->children[1] = LittleShort(in->children[1]);
	}
}

/*
=================
Mod_MakeHull0

Deplicate the drawing hull structure as a clipping hull
=================
*/
void Mod_MakeHull0 (void)
{
	mnode_t		*in, *child;
	dclipnode_t *out;
	int			i, j, count;
	hull_t		*hull;
	
	hull = &loadmodel->hulls[0];	
	
	in = loadmodel->nodes;
	count = loadmodel->numnodes;
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count-1;
	hull->planes = loadmodel->planes;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->planenum = in->plane - loadmodel->planes;
		for (j=0 ; j<2 ; j++)
		{
			child = in->children[j];
			if (child->contents < 0)
				out->children[j] = child->contents;
			else
				out->children[j] = child - loadmodel->nodes;
		}
	}
}

/*
=================
Mod_LoadMarksurfaces
=================
*/
void Mod_LoadMarksurfaces (lump_t *l)
{	
	int		i, j, count;
	short		*in;
	msurface_t **out;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->marksurfaces = out;
	loadmodel->nummarksurfaces = count;

	for ( i=0 ; i<count ; i++)
	{
		j = LittleShort(in[i]);
		if (j >= loadmodel->numsurfaces)
			Sys_Error ("Mod_ParseMarksurfaces: bad surface number");
		out[i] = loadmodel->surfaces + j;
	}
}

/*
=================
Mod_LoadSurfedges
=================
*/
void Mod_LoadSurfedges (lump_t *l)
{	
	int		i, count;
	int		*in, *out;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->surfedges = out;
	loadmodel->numsurfedges = count;

	for ( i=0 ; i<count ; i++)
		out[i] = LittleLong (in[i]);
}


/*
=================
Mod_LoadPlanes
=================
*/
void Mod_LoadPlanes (lump_t *l)
{
	int			i, j;
	mplane_t	*out;
	dplane_t 	*in;
	int			count;
	int			bits;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*2*sizeof(*out), loadname);	
	
	loadmodel->planes = out;
	loadmodel->numplanes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		bits = 0;
		for (j=0 ; j<3 ; j++)
		{
			out->normal[j] = LittleFloat (in->normal[j]);
			if (out->normal[j] < 0)
				bits |= 1<<j;
		}

		out->dist = LittleFloat (in->dist);
		out->type = LittleLong (in->type);
		out->signbits = bits;
	}
}


// mh - auto water trans begin
// detect if a model has been vised for translucent water
qboolean Mod_DetectWaterTrans (model_t *mod)
{
	int i;

	// no visdata
	if (!mod->visdata) return true;

	for (i = 0; i < mod->numleafs; i++)
	{
		// leaf 0 is the solid leaf, leafs go to numleafs + 1
		mleaf_t *leaf = &mod->leafs[i + 1];
		byte *vis;
		msurface_t **surf;
		int j;

		// not interested in these leafs
		if (leaf->contents >= CONTENTS_EMPTY) continue;
		if (leaf->contents == CONTENTS_SOLID) continue;
		if (leaf->contents == CONTENTS_SKY) continue;
		if (leaf->contents == CONTENTS_ORIGIN) continue;
		if (leaf->contents == CONTENTS_CLIP) continue;

		// check marksurfaces for a water texture
		surf = leaf->firstmarksurface;

		for (j = 0; j < leaf->nummarksurfaces; j++, surf++)
		{
			// bad surf/texinfo/texture (some old maps have this from a bad qbsp)
			if (!surf) continue;
			if (!(*surf)) continue;
			if (!(*surf)->texinfo) continue;
			if (!(*surf)->texinfo->texture) continue;

			// not interested in teleports
			if ((*surf)->texinfo->texture->name[0] == '*' &&
				(*surf)->texinfo->texture->name[1] != 't' &&
				(*surf)->texinfo->texture->name[2] != 'e' &&
				(*surf)->texinfo->texture->name[3] != 'l' &&
				(*surf)->texinfo->texture->name[4] != 'e')
			{
				goto LeafOK;
			}
		}

		// no water/etc textures here
		continue;

LeafOK:;
		// get the decompressed vis
		//vis = Mod_DecompressVis (leaf->compressed_vis, mod);
		vis = leaf->decompressed_vis;
		if (!vis) continue;

		// check the other leafs
		for (j = 0; j < mod->numleafs; j++)
		{
			// in the PVS
			if (vis[j >> 3] & (1 << (j & 7)))
			{
				mleaf_t *visleaf = &mod->leafs[j + 1];

				// the leaf we hit originally was under water/slime/lava, and a
				// leaf in it's pvs is above water/slime/lava.
				if (visleaf->contents == CONTENTS_EMPTY) return true;
			}
		}
	}

	// found nothing
	return false;
}
// mh - auto water trans end


/*
=================
RadiusFromBounds
=================
*/
float RadiusFromBounds (vec3_t mins, vec3_t maxs)
{
	int		i;
	vec3_t	corner;

	for (i=0 ; i<3 ; i++)
	{
		corner[i] = fabs(mins[i]) > fabs(maxs[i]) ? fabs(mins[i]) : fabs(maxs[i]);
	}

	return Length (corner);
}


// reliable detection of a world map
qboolean isWorldMap (model_t *mod)
{
	int i;
	char *entstr;

	// not sure if the -11 is 1 too many, but it doesn't matter anyway
	for (i = 0; i < mod->numentities - 11; i++)
	{
		// catch an info_player_* entity
		// this is suitable for all gameplay types.
		if (mod->entities[i] == 'i' &&
			mod->entities[i + 1] == 'n' &&
			mod->entities[i + 2] == 'f' &&
			mod->entities[i + 3] == 'o' &&
			mod->entities[i + 4] == '_' &&
			mod->entities[i + 5] == 'p' &&
			mod->entities[i + 6] == 'l' &&
			mod->entities[i + 7] == 'a' &&
			mod->entities[i + 8] == 'y' &&
			mod->entities[i + 9] == 'e' &&
			mod->entities[i + 10] == 'r') return true;
	}

	return false;
}


/*
=================
Mod_LoadBrushModel
=================
*/
int mapcount = 20;
void BSP2PRTInit (model_t *mod, dheader_t *header);

void Mod_LoadBrushModel (model_t *mod, void *buffer)
{
	int			i, j;
	dheader_t	*header;
	dmodel_t 	*bm;
	qboolean	IsWorld;

	// this can only fail if the bsp could not be opened!!!
	COM_FindGameDirForFile (ModGameDir, mod->name);

	loadmodel->type = mod_brush;

	header = (dheader_t *)buffer;

	i = LittleLong (header->version);

	if (i != BSPVERSION)
		Sys_Error ("Mod_LoadBrushModel: %s has wrong version number (%i should be %i)", mod->name, i, BSPVERSION);

	// swap all the lumps
	mod_base = (byte *)header;

	for (i = 0; i < sizeof (dheader_t) / 4; i++)
		((int *) header)[i] = LittleLong (((int *) header)[i]);

	// load entities before anything else so we can see if it's the world
	Mod_LoadEntities (&header->lumps[LUMP_ENTITIES]);

	IsWorld = isWorldMap (mod);

	// texture cache flushing.  to kkep things clean, the texture cache is periodically flushed.
	// this, and the various safeguards i've built into the loading, should ensure it never
	// runs out.  of course, a dynamically allocated cache would be better, and i did have one
	// for a while, but management of it just became far too messy
	if (IsWorld && mapcount > 10)
	{
		// flush every 10 or so maps
		mapcount = 0;
		GL_FlushTextureCache ();
	}
	else if (!strcmp (loadmodel->name, "maps/start.bsp"))
	{
		// flush on start map load
		mapcount = 0;
		GL_FlushTextureCache ();
	}
	else if (IsWorld) mapcount++;

	// now load the rest
	Mod_LoadVertexes (&header->lumps[LUMP_VERTEXES]);
	Mod_LoadEdges (&header->lumps[LUMP_EDGES]);
	Mod_LoadSurfedges (&header->lumps[LUMP_SURFEDGES]);
	Mod_LoadTextures (&header->lumps[LUMP_TEXTURES]);
	Mod_LoadLighting (&header->lumps[LUMP_LIGHTING]);
	Mod_LoadPlanes (&header->lumps[LUMP_PLANES]);
	Mod_LoadTexinfo (&header->lumps[LUMP_TEXINFO]);
	Mod_LoadFaces (&header->lumps[LUMP_FACES]);
	Mod_LoadMarksurfaces (&header->lumps[LUMP_MARKSURFACES]);
	//Mod_LoadVisibility (&header->lumps[LUMP_VISIBILITY]);
	//Mod_LoadLeafs (&header->lumps[LUMP_LEAFS]);

	// mh - auto water trans begin
	Mod_LoadLeafs (&header->lumps[LUMP_LEAFS], &header->lumps[LUMP_VISIBILITY]);
	Con_DPrintf ("Vis Memory Overhead: %i [was %i]\n", 
		mod->numleafs * mod->vissize, header->lumps[LUMP_VISIBILITY].filelen);
	// mh - auto water trans end

	Mod_LoadNodes (&header->lumps[LUMP_NODES]);
	Mod_LoadClipnodes (&header->lumps[LUMP_CLIPNODES]);
	Mod_LoadSubmodels (&header->lumps[LUMP_MODELS]);

	Mod_MakeHull0 ();

	// mh - auto water trans begin
	mod->watertrans = Mod_DetectWaterTrans (mod);
	// mh - auto water trans end

	mod->numframes = 2;		// regular and alternate animation

	if (IsWorld)
	{
		// made this a developer print now that i know it works...
		Con_DPrintf ("Map %s is the world\n", mod->name);

		// world volume carving
		MH_WorldVolumeCarve (mod);

		// here we portalize the BSP and do a quick 'n' dirty re-vis of parts of it for auto water translucency
		BSP2PRTInit (mod, header);
	}

	// set up the submodels (FIXME: this is confusing)
	for (i=0 ; i<mod->numsubmodels ; i++)
	{
		bm = &mod->submodels[i];

		mod->hulls[0].firstclipnode = bm->headnode[0];
		for (j=1 ; j<MAX_MAP_HULLS ; j++)
		{
			mod->hulls[j].firstclipnode = bm->headnode[j];
			mod->hulls[j].lastclipnode = mod->numclipnodes-1;
		}
		
		mod->firstmodelsurface = bm->firstface;
		mod->nummodelsurfaces = bm->numfaces;
		
		VectorCopy (bm->maxs, mod->maxs);
		VectorCopy (bm->mins, mod->mins);

		mod->radius = RadiusFromBounds (mod->mins, mod->maxs);

		mod->numleafs = bm->visleafs;

		if (i < mod->numsubmodels-1)
		{
			// duplicate the basic information
			char	name[10];

			sprintf (name, "*%i", i+1);
			loadmodel = Mod_FindName (name);
			*loadmodel = *mod;
			strcpy (loadmodel->name, name);
			mod = loadmodel;
		}
	}
}

/*
==============================================================================

ALIAS MODELS

==============================================================================
*/

aliashdr_t	*pheader;

stvert_t	stverts[MAXALIASVERTS];
mtriangle_t	triangles[MAXALIASTRIS];

// a pose is a single set of vertexes.  a frame may be
// an animating sequence of poses
trivertx_t	*poseverts[MAXALIASFRAMES];
int			posenum;

byte		**player_8bit_texels_tbl;
byte		*player_8bit_texels;

/*
=================
Mod_LoadAliasFrame
=================
*/


// alias bbox fix
int aliasbboxmins[3], aliasbboxmaxs[3];


void * Mod_LoadAliasFrame (void * pin, maliasframedesc_t *frame)
{
	trivertx_t		*pframe, *pinframe;
	int				i, j;
	daliasframe_t	*pdaliasframe;
	
	pdaliasframe = (daliasframe_t *)pin;

	strcpy (frame->name, pdaliasframe->name);
	frame->firstpose = posenum;
	frame->numposes = 1;

	for (i = 0; i < 3; i++)
	{
		// these are byte values, so we don't have to worry about
		// endianness
		frame->bboxmin.v[i] = pdaliasframe->bboxmin.v[i];
		frame->bboxmax.v[i] = pdaliasframe->bboxmax.v[i];	// bug - was min in the lvalue

		aliasbboxmins[i] = frame->bboxmin.v[i] > aliasbboxmins[i] ? aliasbboxmins[i] : frame->bboxmin.v[i];
		aliasbboxmaxs[i] = frame->bboxmax.v[i] < aliasbboxmaxs[i] ? aliasbboxmaxs[i] : frame->bboxmax.v[i];
	}

	pinframe = (trivertx_t *)(pdaliasframe + 1);

	poseverts[posenum] = pinframe;
	posenum++;

	pinframe += pheader->numverts;

	return (void *)pinframe;
}


/*
=================
Mod_LoadAliasGroup
=================
*/
void *Mod_LoadAliasGroup (void * pin,  maliasframedesc_t *frame)
{
	daliasgroup_t		*pingroup;
	int					i, numframes;
	daliasinterval_t	*pin_intervals;
	void				*ptemp;
	
	pingroup = (daliasgroup_t *)pin;

	numframes = LittleLong (pingroup->numframes);

	frame->firstpose = posenum;
	frame->numposes = numframes;

	for (i=0 ; i<3 ; i++)
	{
		// these are byte values, so we don't have to worry about endianness
		frame->bboxmin.v[i] = pingroup->bboxmin.v[i];
		frame->bboxmax.v[i] = pingroup->bboxmax.v[i];

		aliasbboxmins[i] = frame->bboxmin.v[i] > aliasbboxmins[i] ? aliasbboxmins[i] : frame->bboxmin.v[i];
		aliasbboxmaxs[i] = frame->bboxmax.v[i] < aliasbboxmaxs[i] ? aliasbboxmaxs[i] : frame->bboxmax.v[i];
	}

	pin_intervals = (daliasinterval_t *)(pingroup + 1);

	frame->interval = LittleFloat (pin_intervals->interval);

	pin_intervals += numframes;

	ptemp = (void *)pin_intervals;

	for (i=0 ; i<numframes ; i++)
	{
		poseverts[posenum] = (trivertx_t *)((daliasframe_t *)ptemp + 1);
		posenum++;

		ptemp = (trivertx_t *)((daliasframe_t *)ptemp + 1) + pheader->numverts;
	}

	return ptemp;
}

//=========================================================

/*
=================
Mod_FloodFillSkin

Fill background pixels so mipmapping doesn't have haloes - Ed
=================
*/

typedef struct
{
	short		x, y;
} floodfill_t;

extern unsigned d_8to24table[];

// must be a power of 2
#define FLOODFILL_FIFO_SIZE 0x1000
#define FLOODFILL_FIFO_MASK (FLOODFILL_FIFO_SIZE - 1)

#define FLOODFILL_STEP( off, dx, dy ) \
{ \
	if (pos[off] == fillcolor) \
	{ \
		pos[off] = 255; \
		fifo[inpt].x = x + (dx), fifo[inpt].y = y + (dy); \
		inpt = (inpt + 1) & FLOODFILL_FIFO_MASK; \
	} \
	else if (pos[off] != 255) fdc = pos[off]; \
}

void Mod_FloodFillSkin( byte *skin, int skinwidth, int skinheight )
{
	byte				fillcolor = *skin; // assume this is the pixel to fill
	floodfill_t			fifo[FLOODFILL_FIFO_SIZE];
	int					inpt = 0, outpt = 0;
	int					filledcolor = -1;
	int					i;

	if (filledcolor == -1)
	{
		filledcolor = 0;
		// attempt to find opaque black
		for (i = 0; i < 256; ++i)
			if (d_8to24table[i] == (255 << 0)) // alpha 1.0
			{
				filledcolor = i;
				break;
			}
	}

	// can't fill to filled color or to transparent color (used as visited marker)
	if ((fillcolor == filledcolor) || (fillcolor == 255))
	{
		//printf( "not filling skin from %d to %d\n", fillcolor, filledcolor );
		return;
	}

	fifo[inpt].x = 0, fifo[inpt].y = 0;
	inpt = (inpt + 1) & FLOODFILL_FIFO_MASK;

	while (outpt != inpt)
	{
		int			x = fifo[outpt].x, y = fifo[outpt].y;
		int			fdc = filledcolor;
		byte		*pos = &skin[x + skinwidth * y];

		outpt = (outpt + 1) & FLOODFILL_FIFO_MASK;

		if (x > 0)				FLOODFILL_STEP( -1, -1, 0 );
		if (x < skinwidth - 1)	FLOODFILL_STEP( 1, 1, 0 );
		if (y > 0)				FLOODFILL_STEP( -skinwidth, 0, -1 );
		if (y < skinheight - 1)	FLOODFILL_STEP( skinwidth, 0, 1 );
		skin[x + skinwidth * y] = fdc;
	}
}

/*
===============
Mod_LoadAllSkins
===============
*/
void *Mod_LoadAllSkins (int numskins, daliasskintype_t *pskintype)
{
	int		i, j, k;
	char	name[32];
	int		s;
	byte	*copy;
	byte	*skin;
	byte	*texels;
	daliasskingroup_t		*pinskingroup;
	int		groupskins;
	daliasskininterval_t	*pinskinintervals;
	char loadmodelname[32];
	
	skin = (byte *)(pskintype + 1);

	if (numskins < 1 || numskins > MAX_SKINS)
		Sys_Error ("Mod_LoadAliasModel: Invalid # of skins: %d\n", numskins);

	s = pheader->skinwidth * pheader->skinheight;

	// make the base name by ripping out the extension
	for (i = 0; ; i++)
	{
		loadmodelname[i] = loadmodel->name[i];

		if (loadmodelname[i] == '.') loadmodelname[i] = 0;
		if (loadmodelname[i] == 0) break;
	}

	for (i = 0; i < numskins; i++)
	{
		if (pskintype->type == ALIAS_SKIN_SINGLE)
		{
			Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

			texels = Hunk_AllocName (s, loadname);
			pheader->texels[i] = texels - (byte *) pheader;
			memcpy (texels, (byte *) (pskintype + 1), s);

			sprintf (name, "%s_%i", loadmodelname, i);

			pheader->gl_texture[i][0] =
			pheader->gl_texture[i][1] =
			pheader->gl_texture[i][2] =
			pheader->gl_texture[i][3] =
				GL_LoadQuakeTexture (NULL, name, (byte *) (pskintype + 1), pheader->skinwidth, pheader->skinheight, (TEX_MIPMAP | TEX_MDL | TEX_MAKENORMAL));

			pskintype = (daliasskintype_t *) ((byte *) (pskintype + 1) + s);
		}
		else
		{
			// animating skin group.  yuck.
			pskintype++;
			pinskingroup = (daliasskingroup_t *) pskintype;
			groupskins = LittleLong (pinskingroup->numskins);
			pinskinintervals = (daliasskininterval_t *) (pinskingroup + 1);

			pskintype = (void *)(pinskinintervals + groupskins);

			for (j = 0; j < groupskins; j++)
			{
					Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

					if (j == 0)
					{
						texels = Hunk_AllocName (s, loadname);
						pheader->texels[i] = texels - (byte *) pheader;
						memcpy (texels, (byte *) (pskintype), s);
					}

					sprintf (name, "%s_%i_%i", loadmodelname, i, j);

					pheader->gl_texture[i][j & 3] = GL_LoadQuakeTexture (NULL, name, (byte *) (pskintype), pheader->skinwidth, pheader->skinheight, (TEX_MIPMAP | TEX_MDL | TEX_MAKENORMAL));

					pskintype = (daliasskintype_t *) ((byte *) (pskintype) + s);
			}

			k = j;

			for (; j < 4; j++)
			{
				pheader->gl_texture[i][j & 3] = pheader->gl_texture[i][j - k]; 
			}
		}
	}

	return (void *) pskintype;
}

//=========================================================================


/*
=================
Mod_LoadAliasModel
=================
*/
void Mod_LoadAliasModel (model_t *mod, void *buffer)
{
	int					i, j;
	mdl_t				*pinmodel;
	stvert_t			*pinstverts;
	dtriangle_t			*pintriangles;
	int					version, numframes, numskins;
	int					size;
	daliasframetype_t	*pframetype;
	daliasskintype_t	*pskintype;
	int					start, end, total;
	
	start = Hunk_LowMark ();

	pinmodel = (mdl_t *) buffer;

	version = LittleLong (pinmodel->version);

	if (version != ALIAS_VERSION)
	{
		Sys_Error ("%s has wrong version number (%i should be %i)", mod->name, version, ALIAS_VERSION);
	}

	// allocate space for a working header, plus all the data except the frames,
	// skin and group info
	size = 	sizeof (aliashdr_t) + (LittleLong (pinmodel->numframes) - 1) * sizeof (pheader->frames[0]);

	pheader = Hunk_AllocName (size, loadname);
	
	mod->flags = LittleLong (pinmodel->flags);

	// endian-adjust and copy the data, starting with the alias model header
	pheader->boundingradius = LittleFloat (pinmodel->boundingradius);
	pheader->numskins = LittleLong (pinmodel->numskins);
	pheader->skinwidth = LittleLong (pinmodel->skinwidth);
	pheader->skinheight = LittleLong (pinmodel->skinheight);

	if (pheader->skinheight > MAX_LBM_HEIGHT)
	{
		Sys_Error ("model %s has a skin taller than %d", mod->name, MAX_LBM_HEIGHT);
	}

	pheader->numverts = LittleLong (pinmodel->numverts);

	if (pheader->numverts <= 0) Sys_Error ("model %s has no vertices", mod->name);
	if (pheader->numverts > MAXALIASVERTS) Sys_Error ("model %s has too many vertices", mod->name);

	pheader->numtris = LittleLong (pinmodel->numtris);

	if (pheader->numtris <= 0) Sys_Error ("model %s has no triangles", mod->name);

	pheader->numframes = LittleLong (pinmodel->numframes);
	numframes = pheader->numframes;

	if (numframes < 1) Sys_Error ("Mod_LoadAliasModel: Invalid # of frames: %d\n", numframes);

	pheader->size = LittleFloat (pinmodel->size) * ALIAS_BASE_SIZE_RATIO;
	mod->synctype = LittleLong (pinmodel->synctype);
	mod->numframes = pheader->numframes;

	for (i = 0; i < 3; i++)
	{
		pheader->scale[i] = LittleFloat (pinmodel->scale[i]);
		pheader->scale_origin[i] = LittleFloat (pinmodel->scale_origin[i]);
		pheader->eyeposition[i] = LittleFloat (pinmodel->eyeposition[i]);
	}

	// load the skins
	pskintype = (daliasskintype_t *) &pinmodel[1];
	pskintype = Mod_LoadAllSkins (pheader->numskins, pskintype);

	// load base s and t vertices
	pinstverts = (stvert_t *) pskintype;

	for (i = 0; i < pheader->numverts; i++)
	{
		stverts[i].onseam = LittleLong (pinstverts[i].onseam);
		stverts[i].s = LittleLong (pinstverts[i].s);
		stverts[i].t = LittleLong (pinstverts[i].t);
	}

	// load triangle lists
	pintriangles = (dtriangle_t *)&pinstverts[pheader->numverts];

	for (i = 0; i < pheader->numtris; i++)
	{
		triangles[i].facesfront = LittleLong (pintriangles[i].facesfront);

		for (j = 0; j < 3; j++)
		{
			triangles[i].vertindex[j] = LittleLong (pintriangles[i].vertindex[j]);
		}
	}

	// load the frames
	posenum = 0;
	pframetype = (daliasframetype_t *) &pintriangles[pheader->numtris];

	aliasbboxmins[0] = aliasbboxmins[1] = aliasbboxmins[2] = 99999;
	aliasbboxmaxs[0] = aliasbboxmaxs[1] = aliasbboxmaxs[2] = -99999;

	for (i = 0; i < numframes; i++)
	{
		aliasframetype_t	frametype;

		frametype = LittleLong (pframetype->type);

		if (frametype == ALIAS_SINGLE)
		{
			pframetype = (daliasframetype_t *) Mod_LoadAliasFrame (pframetype + 1, &pheader->frames[i]);
		}
		else
		{
			pframetype = (daliasframetype_t *) Mod_LoadAliasGroup (pframetype + 1, &pheader->frames[i]);
		}
	}

	pheader->numposes = posenum;

	mod->type = mod_alias;

	// FIXME: do this right
	// MH - consider it done.
	for (i = 0; i < 3; i++)
	{
		mod->mins[i] = aliasbboxmins[i] * pheader->scale[i] + pheader->scale_origin[i];
		mod->maxs[i] = aliasbboxmaxs[i] * pheader->scale[i] + pheader->scale_origin[i];
	}

	// set the midpoint
	for (i = 0; i < 3; i++)
	{
		// add half the size to mins rather than taking a straight average cos the mins might be negative
		mod->mids[i] = mod->mins[i] + ((mod->maxs[i] - mod->mins[i]) / 2.0f);
	}

	mod->radius = RadiusFromBounds (mod->mins, mod->maxs);

	// build the draw lists
	GL_MakeAliasModelDisplayLists (mod, pheader);

	// set special effects - default to everything off
	pheader->bolt = false;
	pheader->flicker = false;
	pheader->glow = false;
	pheader->interpolate = false;
	pheader->missile = false;
	pheader->powerup = false;
	pheader->pulse = false;
	pheader->shambler = false;
	pheader->torch = false;
	pheader->trans = false;
	pheader->translevel = 1;

	if (!strncmp (mod->name, "progs/flame", 11))
	{
		pheader->glow = true;
		pheader->flicker = true;
		pheader->torch = true;
		pheader->colour = 107;
		pheader->radius = 30.0;
	}
	else if (!strncmp (mod->name, "progs/bolt", 10))
	{
	}
	else if (!strcmp  (mod->name, "progs/missile.mdl"))
	{
		pheader->glow = true;
		pheader->flicker = true;
		pheader->missile = true;
		pheader->colour = 107;
		pheader->radius = 30;
	}
	else if (!strcmp  (mod->name, "progs/laser.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/w_spike.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/v_spike.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/k_spike.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/beam.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/candle.mdl"))
	{
		pheader->glow = true;
		pheader->flicker = true;
		pheader->torch = true;
		pheader->colour = 107;
		pheader->radius = 30.0;
	}
	else if (!strcmp  (mod->name, "progs/lantern.mdl"))
	{
		pheader->glow = true;
		pheader->flicker = true;
		pheader->torch = true;
		pheader->colour = 107;
		pheader->radius = 30.0;
	}
	else if (!strcmp  (mod->name, "progs/lspike.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/lasrspik.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/plasma.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/s_light.mdl"))
	{
		pheader->glow = true;
		pheader->shambler = true;
		pheader->radius = 150;
		pheader->colour = 246;
	}
	else if (!strcmp  (mod->name, "progs/quaddama.mdl"))
	{
	}
	else if (!strcmp  (mod->name, "progs/invulner.mdl"))
	{
	}

	// move the complete, relocatable alias model to the cache
	end = Hunk_LowMark ();
	total = end - start;
	
	Cache_Alloc (&mod->cache, total, loadname);

	if (!mod->cache.data) return;

	memcpy (mod->cache.data, pheader, total);

	Hunk_FreeToLowMark (start);
}


//=============================================================================

/*
=================
Mod_LoadSpriteFrame
=================
*/
void * Mod_LoadSpriteFrame (void * pin, mspriteframe_t **ppframe, int framenum)
{
	dspriteframe_t		*pinframe;
	mspriteframe_t		*pspriteframe;
	int					i, width, height, size, origin[2];
	unsigned short		*ppixout;
	byte				*ppixin;
	char				name[64];

	pinframe = (dspriteframe_t *)pin;

	width = LittleLong (pinframe->width);
	height = LittleLong (pinframe->height);
	size = width * height;

	pspriteframe = Hunk_AllocName (sizeof (mspriteframe_t),loadname);

	Q_memset (pspriteframe, 0, sizeof (mspriteframe_t));

	*ppframe = pspriteframe;

	pspriteframe->width = width;
	pspriteframe->height = height;
	origin[0] = LittleLong (pinframe->origin[0]);
	origin[1] = LittleLong (pinframe->origin[1]);

	pspriteframe->up = origin[1];
	pspriteframe->down = origin[1] - height;
	pspriteframe->left = origin[0];
	pspriteframe->right = width + origin[0];

	sprintf (name, "%s_%i", loadmodel->name, framenum);
	pspriteframe->gl_texture = GL_LoadQuakeTexture (NULL, name, (byte *) (pinframe + 1), width, height, (TEX_TEXTURE));

	return (void *)((byte *)pinframe + sizeof (dspriteframe_t) + size);
}


/*
=================
Mod_LoadSpriteGroup
=================
*/
void * Mod_LoadSpriteGroup (void * pin, mspriteframe_t **ppframe, int framenum)
{
	dspritegroup_t		*pingroup;
	mspritegroup_t		*pspritegroup;
	int					i, numframes;
	dspriteinterval_t	*pin_intervals;
	float				*poutintervals;
	void				*ptemp;

	pingroup = (dspritegroup_t *)pin;

	numframes = LittleLong (pingroup->numframes);

	pspritegroup = Hunk_AllocName (sizeof (mspritegroup_t) +
				(numframes - 1) * sizeof (pspritegroup->frames[0]), loadname);

	pspritegroup->numframes = numframes;

	*ppframe = (mspriteframe_t *)pspritegroup;

	pin_intervals = (dspriteinterval_t *)(pingroup + 1);

	poutintervals = Hunk_AllocName (numframes * sizeof (float), loadname);

	pspritegroup->intervals = poutintervals;

	for (i=0 ; i<numframes ; i++)
	{
		*poutintervals = LittleFloat (pin_intervals->interval);
		if (*poutintervals <= 0.0)
			Sys_Error ("Mod_LoadSpriteGroup: interval<=0");

		poutintervals++;
		pin_intervals++;
	}

	ptemp = (void *)pin_intervals;

	for (i=0 ; i<numframes ; i++)
	{
		ptemp = Mod_LoadSpriteFrame (ptemp, &pspritegroup->frames[i], framenum * 100 + i);
	}

	return ptemp;
}


/*
=================
Mod_LoadSpriteModel
=================
*/
void Mod_LoadSpriteModel (model_t *mod, void *buffer)
{
	int					i;
	int					version;
	dsprite_t			*pin;
	msprite_t			*psprite;
	int					numframes;
	int					size;
	dspriteframetype_t	*pframetype;
	
	pin = (dsprite_t *)buffer;

	version = LittleLong (pin->version);
	if (version != SPRITE_VERSION)
		Sys_Error ("%s has wrong version number "
				 "(%i should be %i)", mod->name, version, SPRITE_VERSION);

	numframes = LittleLong (pin->numframes);

	size = sizeof (msprite_t) +	(numframes - 1) * sizeof (psprite->frames);

	psprite = Hunk_AllocName (size, loadname);

	mod->cache.data = psprite;

	psprite->type = LittleLong (pin->type);
	psprite->maxwidth = LittleLong (pin->width);
	psprite->maxheight = LittleLong (pin->height);
	psprite->beamlength = LittleFloat (pin->beamlength);
	mod->synctype = LittleLong (pin->synctype);
	psprite->numframes = numframes;

	mod->mins[0] = mod->mins[1] = -psprite->maxwidth/2;
	mod->maxs[0] = mod->maxs[1] = psprite->maxwidth/2;
	mod->mins[2] = -psprite->maxheight/2;
	mod->maxs[2] = psprite->maxheight/2;

	// check for sprite replacement.  even if it's replaced we'll load it anyway to keep
	// the cache clean and consistent at all times...
	if (!strcmp (mod->name, "progs/s_bubble.spr"))
	{
		// underwater bubbles
		psprite->replace = P_BUBBLE;
	}
	else if (!strcmp (mod->name, "progs/s_explod.spr"))
	{
		// explosions.  these aren't replaced directly in the sprite drawing routines, but
		// are flagged as replaced anyway so that they are not drawn
		psprite->replace = P_BLAST;
	}
	else if (!strcmp (mod->name, "progs/s_rain.spr"))
	{
		// rain
		psprite->replace = P_RAIN;
	}
	else if (!strcmp (mod->name, "progs/s_light.spr"))
	{
		// light globes
		psprite->replace = P_FLARE2;
	}
	else
	{
		psprite->replace = P_NONE;
	}

	// load the frames
	if (numframes < 1)
		Sys_Error ("Mod_LoadSpriteModel: Invalid # of frames: %d\n", numframes);

	mod->numframes = numframes;

	pframetype = (dspriteframetype_t *)(pin + 1);

	for (i=0 ; i<numframes ; i++)
	{
		spriteframetype_t	frametype;

		frametype = LittleLong (pframetype->type);
		psprite->frames[i].type = frametype;

		if (frametype == SPR_SINGLE)
		{
			pframetype = (dspriteframetype_t *)
					Mod_LoadSpriteFrame (pframetype + 1,
										 &psprite->frames[i].frameptr, i);
		}
		else
		{
			pframetype = (dspriteframetype_t *)
					Mod_LoadSpriteGroup (pframetype + 1,
										 &psprite->frames[i].frameptr, i);
		}
	}

	mod->type = mod_sprite;
}

//=============================================================================

/*
================
Mod_Print
================
*/
void Mod_Print (void)
{
	int		i;
	model_t	*mod;

	Con_Printf ("Cached models:\n");
	for (i=0, mod=mod_known ; i < mod_numknown ; i++, mod++)
	{
		Con_Printf ("%8p : %s\n",mod->cache.data, mod->name);
	}
}


/*
==============================================================================

MODELS CACHED IN DISPLAY LISTS

==============================================================================
*/

// immediate mode drawing routine for cached brush model.  this is ONLY EVER intended to be stored
// in a display list, and should not be called during the normal refresh, as certain assumptions
// are or may be made
void GL_BuildPolygonFromSurface (model_t *mod, msurface_t *surf);

void R_DrawBrushModelImmediate (entity_t *e)
{
	model_t *mod = e->model;
	int i;
	int j;
	msurface_t *surf;
	float *v;
	float boxverts[3];

	float modHeight = mod->maxs[2] - mod->mins[2];
	float modWidth = mod->maxs[1] - mod->mins[1];
	float modDepth = mod->maxs[0] - mod->mins[0];

	qboolean lumablend = false;

	// get the first surf
	surf = &mod->surfaces[mod->firstmodelsurface];

	// darken it a bit
	glColor3f (0.7, 0.7, 0.7);

	allocverts = 0;

	// these are normally 6 surfs with 4 verts per surf, but we can't assume that's a
	// general case...
	for (i = 0; i < mod->numsurfaces; i++, surf++)
	{
		// these shouldn't be present here but let's not mess things up anyway
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		// reconstruct the polygon
		GL_BuildPolygonFromSurface (mod, surf);

		// hack - texture management will age these textures so ensure that they are not
		// marked as map textures to avoid this
		surf->texinfo->texture->gl_texture->flags |= TEX_NOAGE;

		if (surf->texinfo->texture->gl_texture->lumatexnum) lumablend = true;

		glBindTexture (GL_TEXTURE_2D, surf->texinfo->texture->gl_texture->gltexnum);

		glBegin (GL_TRIANGLE_FAN);

		// because we don't have polys set for anything here (so as not to mess up the vertex
		// array building) we must instead reconstruct the data from scratch)
		for (j = 0; j < surf->polys->numverts; j++)
		{
			// scale the verts so that it's sized correctly for the sbar
			boxverts[0] = ((surf->polys->verts[j].v[0] * 24) / modDepth) - 12;
			boxverts[1] = ((surf->polys->verts[j].v[1] * 24) / modWidth) - 12;
			boxverts[2] = ((surf->polys->verts[j].v[2] * 24) / modHeight) - 12;

			// draw it
			glTexCoord2fv (surf->polys->verts[j].st);
			glVertex3fv (boxverts);
		}

		glEnd ();
	}

	// back to normal colouring
	glColor3f (1, 1, 1);

	if (!lumablend) return;

	// blend mode for lumas
	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE);

	// get the first surf again
	surf = &mod->surfaces[mod->firstmodelsurface];

	// luma luma luma
	for (i = 0; i < mod->numsurfaces; i++, surf++)
	{
		// these shouldn't be present here but let's not mess things up anyway
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		if (!surf->texinfo->texture->gl_texture->lumatexnum) continue;

		glBindTexture (GL_TEXTURE_2D, surf->texinfo->texture->gl_texture->lumatexnum);

		glBegin (GL_TRIANGLE_FAN);

		// because we don't have polys set for anything here (so as not to mess up the vertex
		// array building) we must instead reconstruct the data from scratch)
		for (j = 0; j < surf->polys->numverts; j++)
		{
			// scale the verts so that it's sized correctly for the sbar
			boxverts[0] = ((surf->polys->verts[j].v[0] * 24) / modDepth) - 12;
			boxverts[1] = ((surf->polys->verts[j].v[1] * 24) / modWidth) - 12;
			boxverts[2] = ((surf->polys->verts[j].v[2] * 24) / modHeight) - 12;

			// draw it
			glTexCoord2fv (surf->polys->verts[j].st);
			glVertex3fv (boxverts);
		}

		glEnd ();
	}

	// restore the original blend mode
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
}


void GL_DrawAliasFrameImmediate (aliashdr_t *paliashdr, int posenum)
{
	drawvertx_t	*verts;
	int		*order;
	int		count;

	verts = (drawvertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	glColor3f (1, 1, 1);

	while (1)
	{
		count = *order++;

		if (!count)
			break;

		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else glBegin (GL_TRIANGLE_STRIP);

		do
		{
			glTexCoord2fv ((float *) order);
			glVertex3fv (verts->v);

			order += 2;
			verts++;
		} while (--count);

		glEnd ();
	}
}


extern qboolean TranslateMDLSkin;
extern int setup_top;
extern int setup_bottom;
gltexture_t r_translatedskin = {0, "***translateskin***"};

void R_SetupAliasFrameImmediate (int frame, aliashdr_t *paliashdr, gltexture_t *glt)
{
	int pose;
	int numposes;

	float interval;

	if ((frame >= paliashdr->numframes) || (frame < 0)) frame = 0;

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[frame].interval;
		pose += (int)(realtime / interval) % numposes;
	}

	// this condition is only valid for the player mdl in the mplayer setup menu
	if (TranslateMDLSkin && frame)
	{
		// some vile ninja hackery going on in here...
		gltexture_t tempglt;
		byte translate[256];
		int i;
		byte *original;
		byte translated[320 * 200];

		int top = setup_top * 16;
		int bottom = setup_bottom * 16;
		int s = paliashdr->skinwidth * paliashdr->skinheight;

		for (i = 0; i < 256; i++)
			translate[i] = i;

		for (i = 0; i < 16; i++)
		{
			if (top < 128)	// the artists made some backwards ranges.  sigh.
				translate[TOP_RANGE + i] = top + i;
			else
				translate[TOP_RANGE + i] = top + 15 - i;
					
			if (bottom < 128)
				translate[BOTTOM_RANGE + i] = bottom + i;
			else
				translate[BOTTOM_RANGE + i] = bottom + 15 - i;
		}

		// skinnum is always 0 here
		original = (byte *) paliashdr + paliashdr->texels[0];

		for (i = 0; i < s; i += 4)
		{
			translated[i] = translate[original[i]];
			translated[i + 1] = translate[original[i + 1]];
			translated[i + 2] = translate[original[i + 2]];
			translated[i + 3] = translate[original[i + 3]];
		}

		// make a texture object the first time through here
		if (!r_translatedskin.gltexnum) GL_MakeTextureObjects (1, &r_translatedskin.gltexnum);

		glBindTexture (GL_TEXTURE_2D, r_translatedskin.gltexnum);

		// upload it
		GL_LoadQuakeTexture (&r_translatedskin, "***translateskin***", translated, paliashdr->skinwidth, paliashdr->skinheight, (TEX_TEXTURE | TEX_NOGEN));

		TranslateMDLSkin = false;
	}
	else if (frame && r_translatedskin.gltexnum)
	{
		// only valid for the player mdl
		glBindTexture (GL_TEXTURE_2D, r_translatedskin.gltexnum);
	}
	else glBindTexture (GL_TEXTURE_2D, glt->gltexnum);

	// never interpolate these cos we're using temporary entities
	GL_DrawAliasFrameImmediate (paliashdr, pose);

	if (glt->lumatexnum)
	{
		// blend mode for lumas
		glDepthMask (GL_FALSE);
		glEnable (GL_BLEND);
		glBlendFunc (GL_ONE, GL_ONE);

		// bind
		glBindTexture (GL_TEXTURE_2D, glt->lumatexnum);

		// never interpolate these cos we're using temporary entities
		GL_DrawAliasFrameImmediate (paliashdr, pose);

		// restore the original blend mode
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable (GL_BLEND);
		glDepthMask (GL_TRUE);
	}
}


// used for the player mdl
void Mod_Draw3DSkinnedWidget (char *WidgetName, int SkinNum, int framenum)
{
	model_t *WidgetModel;
	entity_t WidgetEntity;
	aliashdr_t *paliashdr;

	// load the model through the standard routine
	// don;t crash if it fails, we'll just either not draw it or replace it
	// with a default model
	WidgetModel = Mod_ForName (WidgetName, false);

	// did we get a model???
	if (!WidgetModel) return;

	// only allow alias models cos brushes can cause weird shit
	if (WidgetModel->type != mod_alias) return;

	// set everything in the dummy entity to 0
	memset (&WidgetEntity, 0, sizeof (WidgetEntity));

	// set the model
	WidgetEntity.model = WidgetModel;

	// set the skin (only valid for alias mdls but does no harm)
	WidgetEntity.skinnum = SkinNum;

	// get the alias model header
	paliashdr = (aliashdr_t *) Mod_Extradata (WidgetModel);

	// force a translation first time
	if (!r_translatedskin.gltexnum) TranslateMDLSkin = true;

	// draw
	R_SetupAliasFrameImmediate (framenum, paliashdr, paliashdr->gl_texture[WidgetEntity.skinnum][0]);
}


// load a widget with a variable skin.
// skins must be valid for the model if it's an alias, they are ignored for brushes
int Mod_Load3DSkinnedWidget (char *WidgetName, int SkinNum, int framenum)
{
	int DisplayListNum;
	model_t *WidgetModel;
	entity_t WidgetEntity;

	// never valid for display lists
	TranslateMDLSkin = false;

	// load the model through the standard routine
	// don;t crash if it fails, we'll just either not draw it or replace it
	// with a default model
	WidgetModel = Mod_ForName (WidgetName, false);

	// did we get a model???
	if (!WidgetModel) return 0;

	// bad model type (MHQuake doesn't support sprites)
	if (WidgetModel->type != mod_alias && WidgetModel->type != mod_brush) return 0;

	// make a new display list object
	DisplayListNum = GL_MakeDisplayListObjects (1);
	glNewList (DisplayListNum, GL_COMPILE);

	// set everything in the dummy entity to 0
	memset (&WidgetEntity, 0, sizeof (WidgetEntity));

	// set the model
	WidgetEntity.model = WidgetModel;

	// set the skin (only valid for alias mdls but does no harm)
	WidgetEntity.skinnum = SkinNum;

	// call the appropriate drawing routine.
	// because vertex array pointers cannot be stored in display lists,
	// and the generic routines are unsuitable cos they require a worldmodel,
	// i've written some simplified immediate mode ones for this instead.
	if (WidgetModel->type == mod_alias)
	{
		// get the alias model header
		aliashdr_t *paliashdr = (aliashdr_t *) Mod_Extradata (WidgetModel);

		// get the correct skin
		glBindTexture (GL_TEXTURE_2D, paliashdr->gl_texture[WidgetEntity.skinnum][0]->gltexnum);

		// draw
		R_SetupAliasFrameImmediate (framenum, paliashdr, paliashdr->gl_texture[WidgetEntity.skinnum][0]);
	}
	else
	{
		// it's a brush so draw a brush
		R_DrawBrushModelImmediate (&WidgetEntity);
	}

	// finish the display list
	glEndList ();

	return DisplayListNum;
}


int Mod_Load3DWidget (char *WidgetName)
{
	// wouldn't overloading be so fantastic here?  oh well...
	return Mod_Load3DSkinnedWidget (WidgetName, 0, 0);
}


