/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#include "quakedef.h"

byte		*draw_chars;				// 8*8 graphic characters
qpic_t		*draw_disc;
qpic_t		*draw_backtile;

int			translate_texture;
gltexture_t *char_texture;
gltexture_t *symb_texture;
gltexture_t *menu_charset;
qboolean	usehiresmenu;
qboolean useLoading;

gltexture_t *watercaustics[32];

typedef struct
{
	gltexture_t *gl_texture;
	float	sl, tl, sh, th;
} glpic_t;

byte		conback_buffer[sizeof(qpic_t) + sizeof(glpic_t)];
qpic_t		*conback = (qpic_t *)&conback_buffer;

extern float frametime;

// quad drawing
glvertex_t drawverts[4];

//=============================================================================
/* Support Routines */

typedef struct cachepic_s
{
	char		name[MAX_QPATH];
	qpic_t		pic;
	byte		padding[32];	// for appended glpic
} cachepic_t;

#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

int		pic_texels;
int		pic_count;


qpic_t *Draw_PicFromWad (char *name, ...)
{
	qpic_t	*p;
	char newname[64];
	glpic_t	*gl;
	va_list argptr;
	char string[1024];

	va_start (argptr, name);
	vsprintf (string, name, argptr);
	va_end (argptr);

	p = W_GetLumpName (string);

	sprintf (newname, "wad/%s", string);

	gl = (glpic_t *) p->data;
	gl->gl_texture = GL_LoadTexture (newname, p->width, p->height, p->data, 0);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;

	return p;
}


/*
================
Draw_CachePic
================
*/
qpic_t	*Draw_CachePic (char *path)
{
	cachepic_t	*pic;
	int			i;
	qpic_t		*dat;
	glpic_t		*gl;

	for (pic = menu_cachepics, i = 0; i < menu_numcachepics; pic++, i++)
		if (!strcmp (path, pic->name))
			return &pic->pic;

	if (menu_numcachepics == MAX_CACHED_PICS)
		Sys_Error ("menu_numcachepics == MAX_CACHED_PICS");

	menu_numcachepics++;
	strcpy (pic->name, path);

	// load the pic from disk
	dat = (qpic_t *) COM_LoadTempFile (path);

	if (!dat) Sys_Error ("Draw_CachePic: failed to load %s", path);
	SwapPic (dat);

	// HACK HACK HACK --- we need to keep the bytes for
	// the translatable player picture just for the menu
	// configuration dialog
	if (!strcmp (path, "gfx/menuplyr.lmp"))
		memcpy (menuplyr_pixels, dat->data, dat->width * dat->height);

	pic->pic.width = dat->width;
	pic->pic.height = dat->height;

	gl = (glpic_t *) pic->pic.data;
	gl->gl_texture = GL_LoadTexture (path, dat->width, dat->height, dat->data, TEX_MIPMAP);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;

	return &pic->pic;
}


void Draw_CharToConback (int num, byte *dest)
{
	int		row, col;
	byte	*source;
	int		drawline;
	int		x;

	row = num>>4;
	col = num&15;
	source = draw_chars + (row<<10) + (col<<3);

	drawline = 8;

	while (drawline--)
	{
		for (x=0 ; x<8 ; x++)
			if (source[x] != 255)
				dest[x] = 0x60 + source[x];
		source += 128;
		dest += 320;
	}

}

/*
===============
Draw_Init
===============
*/
void GDI_MakeFont (int FontTexnum);

void Draw_Init (void)
{
	int i;

	// set correct dimensions (307200 == 640 * 480)
	if (vid.height < 384)
	{
		vid.width = 196608 / vid.height;
		vid.height = 384;
	}
	else if (vid.width < 512)
	{
		vid.height = 196608 / vid.width;
		vid.width = 512;
	}

	// load a dummy charset to set up the texture
	draw_chars = W_GetLumpName ("conchars");
	char_texture = GL_LoadTexture ("charset", 128, 128, draw_chars, 0);

	// now create the proper texture
	GDI_MakeFont (char_texture->texnum);

	// set the correct alpha colour
	for (i = 0; i < 128 * 128; i++)
		if (draw_chars[i] == 0)
			draw_chars[i] = 255;

	// now load the charset again as the symbol texture
	symb_texture = GL_LoadTexture ("wad/charset", 128, 128, draw_chars, 0);

	// load the console background
	conback = Draw_CachePic ("gfx/conback.lmp");

	// set width and height so it displays properly
	conback->width = vid.width;
	conback->height = vid.height;

	// save a texture slot for translated picture
	GL_AllocTexture (&translate_texture);

	// get the other pics we need
	draw_disc = Draw_PicFromWad ("disc");
	draw_backtile = Draw_PicFromWad ("backtile");
	menu_charset = GL_LoadTexture ("wad/mcharset", 1024, 1024, NULL, 0);

	if (!menu_charset)
	{
		Con_DPrintf ("Failed to load menu charset\n");
		usehiresmenu = false;
	}
	else
	{
		Con_DPrintf ("Loaded menu charset OK\n");
		usehiresmenu = true;
	}

	for (i = 0; i < 32; i++)
	{
		char causticname[64];

		sprintf (causticname, "caustic/caust_%02i", i);
		watercaustics[i] = GL_LoadTexture (causticname, 128, 128, NULL, TEX_MIPMAP);

		if (!watercaustics[i]) break;
	}

	// identify no caustics
	if (i != 32) watercaustics[0] = 0;
}


void Draw_Crosshair (int texnum)
{
	GL_BindTexture (texnum);

	drawverts[0].st[0] = 0;
	drawverts[0].st[1] = 0;
	drawverts[0].v[0] = vid.width / 2 - 12;
	drawverts[0].v[1] = vid.height / 2 - 4;

	drawverts[1].st[0] = 1;
	drawverts[1].st[1] = 0;
	drawverts[1].v[0] = vid.width / 2 + 12;
	drawverts[1].v[1] = vid.height / 2 - 4;

	drawverts[2].st[0] = 1;
	drawverts[2].st[1] = 1;
	drawverts[2].v[0] = vid.width / 2 + 12;
	drawverts[2].v[1] = vid.height / 2 + 20;

	drawverts[3].st[0] = 0;
	drawverts[3].st[1] = 1;
	drawverts[3].v[0] = vid.width / 2 - 12;
	drawverts[3].v[1] = vid.height / 2 + 20;

	glDrawArrays (GL_QUADS, 0, 4);
}


/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Character (int x, int y, int num)
{
	int				row, col;
	float			frow, fcol, xsize, ysize;
	int xadd;

	if (num == 32)
		return;		// space

	y += 1;

	num &= 255;

	// totally off screen
	if (y <= -8) return;

	// hack - swap these two textures around to prevent SHITE from happening
	if (num == 130)
		num = 131;
	else if (num == 131)
		num = 130;

	row = num>>4;
	col = num&15;

	frow = row * 0.0625;
	fcol = col * 0.0625;

	// use an alternate texture for the symbol chars
	switch (num)
	{
	case 29:
	case 30:
	case 31:
	case 128:
	case 129:
	case 130:
	case 131:
	case 157:
	case 158:
	case 159:
		GL_BindTexture (symb_texture->texnum);
		xsize = 0.0625;
		ysize = 0.0625;
		xadd = 8;

		break;

	default:
		GL_BindTexture (char_texture->texnum);
		xsize = 0.05;
		ysize = 0.05;
		xadd = 12;

		break;
	}

	drawverts[0].st[0] = fcol;
	drawverts[0].st[1] = frow;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = fcol + xsize;
	drawverts[1].st[1] = frow;
	drawverts[1].v[0] = x + xadd;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = fcol + xsize;
	drawverts[2].st[1] = frow + ysize;
	drawverts[2].v[0] = x + xadd;
	drawverts[2].v[1] = y + 8;

	drawverts[3].st[0] = fcol;
	drawverts[3].st[1] = frow + ysize;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + 8;

	glDrawArrays (GL_QUADS, 0, 4);
}


/*
================
Draw_String
================
*/
void Draw_String (int x, int y, char *str)
{
	while (*str)
	{
		Draw_Character (x, y, *str);
		str++;
		x += 8;
	}
}


/*
=============
Draw_AlphaPic
=============
*/
void Draw_AlphaPic (int x, int y, qpic_t *pic, float alpha)
{
	glpic_t			*gl;

	gl = (glpic_t *)pic->data;

	glColor4f (1, 1, 1, alpha);

	GL_BindTexture (gl->gl_texture->texnum);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	drawverts[0].st[0] = gl->sl;
	drawverts[0].st[1] = gl->tl;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = gl->sh;
	drawverts[1].st[1] = gl->tl;
	drawverts[1].v[0] = x + pic->width;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = gl->sh;
	drawverts[2].st[1] = gl->th;
	drawverts[2].v[0] = x + pic->width;
	drawverts[2].v[1] = y + pic->height;

	drawverts[3].st[0] = gl->sl;
	drawverts[3].st[1] = gl->th;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + pic->height;

	glDrawArrays (GL_QUADS, 0, 4);

	glColor4f (1, 1, 1, 1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}


/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
	glpic_t			*gl;

	gl = (glpic_t *)pic->data;

	// nothing to draw
	if (!gl->gl_texture) return;

	glColor4f (1,1,1,1);

	GL_BindTexture (gl->gl_texture->texnum);

	drawverts[0].st[0] = gl->sl;
	drawverts[0].st[1] = gl->tl;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = gl->sh;
	drawverts[1].st[1] = gl->tl;
	drawverts[1].v[0] = x + pic->width;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = gl->sh;
	drawverts[2].st[1] = gl->th;
	drawverts[2].v[0] = x + pic->width;
	drawverts[2].v[1] = y + pic->height;

	drawverts[3].st[0] = gl->sl;
	drawverts[3].st[1] = gl->th;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + pic->height;

	glDrawArrays (GL_QUADS, 0, 4);
}


void Draw_GLTexturePic (int x, int y, int w, int h, gltexture_t *glt)
{
	// nothing to draw
	if (!glt) return;

	GL_BindTexture (glt->texnum);

	drawverts[0].st[0] = 0;
	drawverts[0].st[1] = 0;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = 1;
	drawverts[1].st[1] = 0;
	drawverts[1].v[0] = x + w;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = 1;
	drawverts[2].st[1] = 1;
	drawverts[2].v[0] = x + w;
	drawverts[2].v[1] = y + h;

	drawverts[3].st[0] = 0;
	drawverts[3].st[1] = 1;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + h;

	glDrawArrays (GL_QUADS, 0, 4);
}


/*
=============
Draw_TransPic
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{
	Draw_Pic (x, y, pic);
}


/*
=============
Draw_TransPicTranslate

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
	int				v, u, c;
	unsigned		trans[64*64], *dest;
	byte			*src;
	int				p;

	GL_BindTexture (translate_texture);

	c = pic->width * pic->height;

	dest = trans;
	for (v=0 ; v<64 ; v++, dest += 64)
	{
		src = &menuplyr_pixels[ ((v*pic->height)>>6) *pic->width];
		for (u=0 ; u<64 ; u++)
		{
			p = src[(u*pic->width)>>6];
			if (p == 255)
				dest[u] = p;
			else
				dest[u] =  d_8to24tableRGBA[translation[p]];
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glColor3f (1,1,1);

	drawverts[0].st[0] = 0;
	drawverts[0].st[1] = 0;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = 1;
	drawverts[1].st[1] = 0;
	drawverts[1].v[0] = x + pic->width;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = 1;
	drawverts[2].st[1] = 1;
	drawverts[2].v[0] = x + pic->width;
	drawverts[2].v[1] = y + pic->height;

	drawverts[3].st[0] = 0;
	drawverts[3].st[1] = 1;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + pic->height;

	glDrawArrays (GL_QUADS, 0, 4);
}


/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
	int y = (vid.height * 3) >> 2;

	if (lines > y)
		Draw_Pic (0, lines - vid.height, conback);
	else
		Draw_AlphaPic (0, lines - vid.height, conback, (float)(1.2 * lines)/y);
}


/*
=============
Draw_TileClear

This repeats a 64*64 tile graphic to fill the screen around a sized down
refresh window.
=============
*/
void Draw_TileClear (int x, int y, int w, int h)
{
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	glDisable (GL_TEXTURE_2D);

	glColor3f (host_basepal[c * 3] / 255.0, host_basepal[c * 3 + 1] / 255.0, host_basepal[c * 3 + 2] / 255.0);

	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].v[0] = x + w;
	drawverts[1].v[1] = y;

	drawverts[2].v[0] = x + w;
	drawverts[2].v[1] = y + h;

	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + h;

	glDrawArrays (GL_QUADS, 0, 4);

	glColor3f (1, 1, 1);
	glEnable (GL_TEXTURE_2D);
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);
	glColor4f (0, 0, 0, 0.5);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glColor4f (1,1,1,1);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	Sbar_Changed();
}


//=============================================================================

/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.
================
*/
void Draw_BeginDisc (void)
{
}


/*
================
Draw_EndDisc

Erases the disc icon.
Call after completing any disc IO
================
*/
void Draw_EndDisc (void)
{
}


/*
========================================================================================

	BBB   L     OO    OO   DDD      AA   N  N  DDD     DDD   EEEE   AA  TTTTT  H  H
	B  B  L    O  O  O  O  D  D    A  A  NN N  D  D    D  D  E     A  A   T    H  H
	BBB   L    O  O  O  O  D  D    A  A  N NN  D  D    D  D  EEE   A  A   T    HHHH
	B  B  L    O  O  O  O  D  D    AAAA  N  N  D  D    D  D  E     AAAA   T    H  H
	BBB   LLLL  OO    OO   DDD     A  A  N  N  DDD     DDD   EEEE  A  A   T    H  H

========================================================================================
*/


float DeathAlpha = 0;
float DeathRed = 0.666;

void Draw_DeathScreen (void)
{
	// console is full screen
	if (scr_con_current == vid.height) return;

	// no death in intermission
	if (cl.intermission) return;

	// not connected
	if (cls.state != ca_connected) return;

	if (cl.stats[STAT_HEALTH] > 0)
	{
		DeathAlpha = 0;
		DeathRed = 0.666;
		return;
	}

	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);
	glColor4f (DeathRed, 0, 0, DeathAlpha);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glColor4f (1,1,1,1);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	Sbar_Changed();

	DeathAlpha += frametime / 5.0;
	DeathRed -= frametime / 8.0;
}


extern int NumScreenBloods;

#define MAX_SCREEN_BLOODS		128

typedef struct screenblood_s
{
	int x;
	int y;
	int w;
	int h;
	float fade;
	int particlenumber;
} screenblood_t;

extern gltexture_t *BloodTextures[];

screenblood_t ScreenBloods[MAX_SCREEN_BLOODS];

void R_InitBlood (void)
{
	int i;

	for (i = 0; i < MAX_SCREEN_BLOODS; i++)
	{
		ScreenBloods[i].fade = -1;
	}
}
extern gltexture_t *MuzzleFlash;
void R_DrawBlood (void)
{
	int i;
	int j;

	// make sure we have blood available
	if (!BloodTextures[0]) return;
	if (!BloodTextures[1]) return;
	if (!BloodTextures[2]) return;
	if (!BloodTextures[3]) return;

	// console is full screen
	if (scr_con_current == vid.height) return;

	// no blood in intermission
	if (cl.intermission) return;

	// not connected
	if (cls.state != ca_connected) return;

	if (NumScreenBloods)
	{
		// calculate a random offset to prevent them from being clustered around the center of the screen
		int xofs = (rand () % 192) - 96;
		int yofs = (rand () % 192) - 96;

		// add each
		for (i = 0; i < NumScreenBloods; i++)
		{
			// find a free slot
			for (j = 0; j < MAX_SCREEN_BLOODS; j++)
			{
				if (ScreenBloods[j].fade < 0)
				{
					// found one
					ScreenBloods[j].fade = 0.3f + (float) (rand () % 8) / 10.0f;
					ScreenBloods[j].w = 256 + (rand () % 128);
					ScreenBloods[j].h = ScreenBloods[j].w;
					ScreenBloods[j].x = (rand () % (vid.width - ScreenBloods[j].w)) + xofs;
					ScreenBloods[j].y = (rand () % (vid.height - ScreenBloods[j].h)) + yofs;
					ScreenBloods[j].particlenumber = (rand () % 4);

					break;
				}
			}
		}
	}

	glEnable (GL_BLEND);

	// enable this blendfunc for particles!
	//glBlendFunc (GL_ONE, GL_ONE);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// draw each that is active
	for (i = 0; i < MAX_SCREEN_BLOODS; i++)
	{
		// inactive
		if (ScreenBloods[i].fade < 0) continue;

		GL_BindTexture (BloodTextures[ScreenBloods[i].particlenumber]->texnum);

		glColor4f (1, 1, 1, ScreenBloods[i].fade);

		drawverts[0].st[0] = 0;
		drawverts[0].st[1] = 0;
		drawverts[0].v[0] = ScreenBloods[i].x;
		drawverts[0].v[1] = ScreenBloods[i].y;

		drawverts[1].st[0] = 1;
		drawverts[1].st[1] = 0;
		drawverts[1].v[0] = ScreenBloods[i].x + ScreenBloods[i].w;
		drawverts[1].v[1] = ScreenBloods[i].y;

		drawverts[2].st[0] = 1;
		drawverts[2].st[1] = 1;
		drawverts[2].v[0] = ScreenBloods[i].x + ScreenBloods[i].w;
		drawverts[2].v[1] = ScreenBloods[i].y + ScreenBloods[i].h;

		drawverts[3].st[0] = 0;
		drawverts[3].st[1] = 1;
		drawverts[3].v[0] = ScreenBloods[i].x;
		drawverts[3].v[1] = ScreenBloods[i].y + ScreenBloods[i].h;

		glDrawArrays (GL_QUADS, 0, 4);


		// attenuate fade
		ScreenBloods[i].fade -= frametime;
	}

	glColor4f (1, 1, 1, 1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	NumScreenBloods = 0;
}


/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value) return;
	if (!v_blend[3]) return;

	// console is full screen
	if (scr_con_current == vid.height) return;

	// no polyblend in intermission
	if (cl.intermission) return;

	// not connected
	if (cls.state != ca_connected) return;

	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);

	glColor4fv (v_blend);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glEnable (GL_TEXTURE_2D);

	glColor4f (1, 1, 1, 1);
}


void Draw_Brightness (void)
{
	float brightval;

	// no need
	if (gl_brightness.value > 0.98 && gl_brightness.value < 1.02) return;

	// clamp
	if (gl_brightness.value < 0) gl_brightness.value = 0;
	if (gl_brightness.value > 2) gl_brightness.value = 2;

	brightval = gl_brightness.value / 2.0f;

	glEnable (GL_BLEND);
	glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);
	glDisable (GL_TEXTURE_2D);

	glColor3f (brightval, brightval, brightval);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glEnable (GL_TEXTURE_2D);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f (1, 1, 1, 1);
}


/*
================
GL_Set2D

Setup as if the screen was 320*200
================
*/
void GL_Set2D (void)
{
	glViewport (glx, gly, glwidth, glheight);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	glOrtho  (0, vid.width, vid.height, 0, -99999, 99999);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);

	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glColor4f (1, 1, 1, 1);

	// bring up vertex pointer
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (2, GL_FLOAT, sizeof (glvertex_t), drawverts[0].v);

	// bring up image texture array
	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), drawverts[0].st);

	// do polyblend here rather than in 3d
	R_PolyBlend ();

	// do blood and death before anything else
	R_DrawBlood ();
	Draw_DeathScreen ();
}

//====================================================================


void LoadingDisplay (void)
{
	if (!useLoading) return;
}

