/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "quakedef.h"

void BuildSurfaceDisplayList (msurface_t *surf);

byte BlockLightTable[262144] = {255};
byte TargTable[256];


/*
==================
R_AnimateLight
==================
*/
void R_AnimateLight (void)
{
	int			j, k;
	float		l;
	int			flight;
	int			clight;
	float		lerpfrac;
	float		backlerp;

	// light animations
	// 'm' is normal light, 'a' is no light, 'z' is double bright
	flight = (int) floor (cl.time * 10);
	clight = (int) ceil (cl.time * 10);
	lerpfrac = (cl.time * 10) - flight;
	backlerp = 1.0f - lerpfrac;

	for (j = 0; j < MAX_LIGHTSTYLES; j++)
	{
		if (!cl_lightstyle[j].length)
		{
			// was 256, changed to 264 for consistency
			d_lightstylevalue[j] = 264;
			continue;
		}
		else if (cl_lightstyle[j].length == 1)
		{
			// single length style so don't bother interpolating
			d_lightstylevalue[j] = 22 * (cl_lightstyle[j].map[0] - 'a');
			continue;
		}

		// interpolate animating light
		// frame just gone
		k = flight % cl_lightstyle[j].length;
		k = cl_lightstyle[j].map[k] - 'a';
		l = (float) (k * 22) * backlerp;

		// upcoming frame
		k = clight % cl_lightstyle[j].length;
		k = cl_lightstyle[j].map[k] - 'a';
		l += (float) (k * 22) * lerpfrac;

		d_lightstylevalue[j] = (int) l;
	}	
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

/*
=============
R_MarkLights

=============
*/
void R_MarkLights (dlight_t *light, int bit, mnode_t *node)
{
}


void R_LightVisibility (vec3_t org)
{
}


/*
=============
R_PushDlights
=============
*/
void R_PushDlights (mnode_t *headnode)
{
}


void R_PushDlightsNOVIS (mnode_t *headnode)
{
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;
static vec3_t	pointcolor;


int RecursiveLightPoint (vec3_t color, mnode_t *node, vec3_t start, vec3_t end)
{
	float		front, back, frac;
	vec3_t		mid;

loc0:;
	// hit nothing
	if (node->contents < 0) return false;

	// calculate mid point
	if (node->plane->type < 3)
	{
		front = start[node->plane->type] - node->plane->dist;
		back = end[node->plane->type] - node->plane->dist;
	}
	else
	{
		front = DotProduct (start, node->plane->normal) - node->plane->dist;
		back = DotProduct (end, node->plane->normal) - node->plane->dist;
	}

	// LordHavoc: optimized recursion
	if ((back < 0) == (front < 0))
	{
		node = node->children[front < 0];
		goto loc0;
	}

	frac = front / (front-back);
	mid[0] = start[0] + (end[0] - start[0]) * frac;
	mid[1] = start[1] + (end[1] - start[1]) * frac;
	mid[2] = start[2] + (end[2] - start[2]) * frac;

	// go down front side
	if (RecursiveLightPoint (color, node->children[front < 0], start, mid))
	{
		// hit something
		return true;
	}
	else
	{
		int i, ds, dt;
		msurface_t *surf;

		// check for impact on this node
		VectorCopy (mid, lightspot);
		lightplane = node->plane;

		surf = cl.worldmodel->surfaces + node->firstsurface;

		for (i = 0; i < node->numsurfaces; i++, surf++)
		{
			// no lightmaps on these
			if (surf->flags & SURF_DRAWTURB) continue;
			if (surf->flags & SURF_DRAWSKY) continue;

			ds = (int) ((float) DotProduct (mid, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]);
			dt = (int) ((float) DotProduct (mid, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]);

			// out of range
			if (ds < surf->texturemins[0] || dt < surf->texturemins[1]) continue;

			ds -= surf->texturemins[0];
			dt -= surf->texturemins[1];

			// out of range
			if (ds > surf->extents[0] || dt > surf->extents[1]) continue;

			if (surf->samples)
			{
				// LordHavoc enhanced to interpolate lighting
				byte *lightmap;
				int maps, line3, dsfrac = ds & 15, dtfrac = dt & 15, r00 = 0, g00 = 0, b00 = 0, r01 = 0, g01 = 0, b01 = 0, r10 = 0, g10 = 0, b10 = 0, r11 = 0, g11 = 0, b11 = 0;
				float scale;
				line3 = surf->smax * 3;

				lightmap = surf->samples + ((dt >> 4) * surf->smax + (ds >> 4)) * 3;

				for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
				{
					scale = (float) d_lightstylevalue[surf->styles[maps]] * 0.003921568627451;

					r00 += (float) lightmap[0] * scale;
					g00 += (float) lightmap[1] * scale;
					b00 += (float) lightmap[2] * scale;

					r01 += (float) lightmap[3] * scale;
					g01 += (float) lightmap[4] * scale;
					b01 += (float) lightmap[5] * scale;

					r10 += (float) lightmap[line3 + 0] * scale;
					g10 += (float) lightmap[line3 + 1] * scale;
					b10 += (float) lightmap[line3 + 2] * scale;

					r11 += (float) lightmap[line3 + 3] * scale;
					g11 += (float) lightmap[line3 + 4] * scale;
					b11 += (float) lightmap[line3 + 5] * scale;

					lightmap += surf->smax * surf->tmax * 3;
				}

				// i'm pretty sure that LH added a lot of these extra brackets for no reason other than to obfuscate his code.
				// i've left them in anyway, but spaced things out a bit better
				color[0] += (float) ((int) ((((((((r11 - r10) * dsfrac) >> 4) + r10) - 
					((((r01 - r00) * dsfrac) >> 4) + r00)) * dtfrac) >> 4) + ((((r01 - r00) * dsfrac) >> 4) + r00)));

				color[1] += (float) ((int) ((((((((g11 - g10) * dsfrac) >> 4) + g10) - 
					((((g01 - g00) * dsfrac) >> 4) + g00)) * dtfrac) >> 4) + ((((g01 - g00) * dsfrac) >> 4) + g00)));

				color[2] += (float) ((int) ((((((((b11 - b10) * dsfrac) >> 4) + b10) - 
					((((b01 - b00) * dsfrac) >> 4) + b00)) * dtfrac) >> 4) + ((((b01 - b00) * dsfrac) >> 4) + b00)));
			}

			// success
			return true;
		}

		// go down back side
		return RecursiveLightPoint (color, node->children[front >= 0], mid, end);
	}
}


void R_LightPoint (vec3_t p, vec3_t colour)
{
	vec3_t		end;

	if (!cl.worldmodel->lightdata)
	{
		// no light data
		pointcolor[0] = pointcolor[1] = pointcolor[2] = 255;
		VectorCopy (colour, pointcolor);
		return;
	}

	// end point increased to full map scale
	end[0] = p[0];
	end[1] = p[1];
	end[2] = p[2] - 8192;

	// initial colour
	pointcolor[0] = pointcolor[1] = pointcolor[2] = 0;

	// get it
	RecursiveLightPoint (pointcolor, cl.worldmodel->nodes, p, end);

	// copy across
	VectorCopy (pointcolor, colour);

	// adjust for lighting table lookup to keep consistent with the world
	colour[0] = (colour[0] > 255) ? 255 : (colour[0] < 0) ? 0 : TargTable[(int) colour[0]];
	colour[1] = (colour[1] > 255) ? 255 : (colour[1] < 0) ? 0 : TargTable[(int) colour[1]];
	colour[2] = (colour[2] > 255) ? 255 : (colour[2] < 0) ? 0 : TargTable[(int) colour[2]];
}


/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

int 	skytexturenum;

unsigned int blocklights[3][18*18];

gllightmap_t *gllightmaps = NULL;


/*
===============
R_AddDynamicLights

A lot of this code is now duplicated in R_MarkLights, but (a) it makes R_MarkLights more effective, (b) it
prevents unnecessary lightmap uploads at runtime, and (c) it won't make a blind bit of difference when I
move the engine to 3D attenuation maps for dynamic lights (when I'm going to really need the more precise info
in R_MarkLights anyway...)
===============
*/
void R_AddDynamicLights (msurface_t *surf)
{
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	mtexinfo_t	*tex;
	float		rgb[3];

	tex = surf->texinfo;

	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		// dead light
		// id forgot this...!!!
		if (cl_dlights[lnum].die < cl.time) continue;

		// not marked for checking by this light
		if (!(surf->dlightbits[(lnum >> 3)] & (1 << (lnum & 7)))) continue;

		rad = cl_dlights[lnum].radius;
		dist = DotProduct (cl_dlights[lnum].origin, surf->plane->normal) - surf->plane->dist;
		rad -= fabs (dist);

		minlight = cl_dlights[lnum].minlight;

		if (rad < minlight) continue;
		minlight = rad - minlight;

		for (i = 0; i < 3; i++)
			impact[i] = cl_dlights[lnum].origin[i] - surf->plane->normal[i] * dist;

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];

		rgb[0] = 255.0f * cl_dlights[lnum].rgb[0];
		rgb[1] = 255.0f * cl_dlights[lnum].rgb[1];
		rgb[2] = 255.0f * cl_dlights[lnum].rgb[2];

		for (t = 0; t < surf->tmax; t++)
		{
			td = local[1] - t * 16;
			if (td < 0) td = -td;

			for (s = 0; s < surf->smax; s++)
			{
				sd = local[0] - s * 16;
				if (sd < 0) sd = -sd;

				if (sd > td)
					dist = sd + (td >> 1);
				else
					dist = td + (sd >> 1);

				if (dist < minlight)
				{
					blocklights[0][t * surf->smax + s] += (rad - dist) * rgb[0];
					blocklights[1][t * surf->smax + s] += (rad - dist) * rgb[1];
					blocklights[2][t * surf->smax + s] += (rad - dist) * rgb[2];
				}
			}
		}
	}
}


void R_InitBLTable (void)
{
	int i;
	int t;
	float newt;
	float mult;

	mult = 0.5;

	// build a higher dynamic range for the lighting
	for (i = 0; i < 256; i++)
	{
		newt = (float) i * mult;

		if (newt > 255) newt = 255;

		TargTable[i] = (byte) newt;

		mult += 0.00625;
	}

	for (i = 0; i < 262144; i++)
	{
		t = i / 256;

		if (t > 255) t = 255;

		BlockLightTable[i] = (byte) TargTable[t];
	}
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
void R_BuildLightMap (msurface_t *surf, int stride)
{
	int			i, j, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	unsigned int *bl[3];
	byte		*dest = surf->base;

	surf->cached_dlight = (surf->dlightframe == r_framecount);

	size = surf->smax * surf->tmax;
	lightmap = surf->samples;

	if (BlockLightTable[0] == 255) R_InitBLTable ();

	// set to full bright if no light data
	if (r_fullbright.value || !cl.worldmodel->lightdata)
	{
		for (i = 0; i < size; i++)
		{
			blocklights[0][i] = 255 * 256;
			blocklights[1][i] = 255 * 256;
			blocklights[2][i] = 255 * 256;
		}

		goto store;
	}

	// clear to no light
	for (i = 0; i < size; i++)
	{
		blocklights[0][i] = 0;
		blocklights[1][i] = 0;
		blocklights[2][i] = 0;
	}

	// add all the lightmaps
	if (lightmap)
	{
		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps]!= 255 ; maps++)
		{
			scale = d_lightstylevalue[surf->styles[maps]];
			surf->cached_light[maps] = scale;

			for (i = 0, j = 0; i < size; i++)
			{
				blocklights[0][i] += lightmap[j++] * scale;
				blocklights[1][i] += lightmap[j++] * scale;
				blocklights[2][i] += lightmap[j++] * scale;
			}

			// skip to next lightmap
			lightmap += size * 3;
		}
	}

	// add all the dynamic lights
	//if (surf->dlightframe == r_framecount) R_AddDynamicLights (surf);

	// bound, invert, and shift
store:
	stride -= (surf->smax << 2);

	bl[0] = blocklights[0];
	bl[1] = blocklights[1];
	bl[2] = blocklights[2];

	for (i = 0; i < surf->tmax; i++, dest += stride)
	{
		for (j = 0; j < surf->smax; j++)
		{
			// fill in from the lookup table
			dest[0] = BlockLightTable[(*bl[0]++)];
			dest[1] = BlockLightTable[(*bl[1]++)];
			dest[2] = BlockLightTable[(*bl[2]++)];
			dest[3] = 255;

			dest += LIGHTMAP_BYTES;
		}
	}
}


void R_BlendLightmap (gllightmap_t *lm)
{
	glRect_t	*theRect;

	if (lm->Modified)
	{
		lm->Modified = false;
		theRect = &lm->RectChange;

		glTexSubImage2D (GL_TEXTURE_2D, 
						 0, 
						 0, 
						 theRect->t, 
						 LM_BLOCK_WIDTH, 
						 theRect->h, 
						 GL_RGBA, 
						 GL_UNSIGNED_BYTE,
						 lm->Data + (theRect->t * LM_BLOCK_WIDTH * LIGHTMAP_BYTES));

		theRect->l = LM_BLOCK_WIDTH;
		theRect->t = LM_BLOCK_HEIGHT;
		theRect->h = 0;
		theRect->w = 0;
	}
}


void R_ModifySurfaceLightmap (msurface_t *surf)
{
	int			maps;
	glRect_t    *theRect;

	// no lightmaps on these surface types
	if (surf->flags & SURF_DRAWTURB) return;
	if (surf->flags & SURF_DRAWSKY) return;

	// check for lightmap modification
	for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
	{
		if (d_lightstylevalue[surf->styles[maps]] != surf->cached_light[maps])
		{
			surf->lightmap->Modified = true;
			theRect = &surf->lightmap->RectChange;

			if (surf->light_t < theRect->t)
			{
				if (theRect->h)
					theRect->h += theRect->t - surf->light_t;
				theRect->t = surf->light_t;
			}

			if (surf->light_s < theRect->l)
			{
				if (theRect->w)
					theRect->w += theRect->l - surf->light_s;
				theRect->l = surf->light_s;
			}

			if ((theRect->w + theRect->l) < (surf->light_s + surf->smax))
				theRect->w = (surf->light_s-theRect->l)+surf->smax;
			if ((theRect->h + theRect->t) < (surf->light_t + surf->tmax))
				theRect->h = (surf->light_t-theRect->t)+surf->tmax;

			R_BuildLightMap (surf, LM_BLOCK_WIDTH * LIGHTMAP_BYTES);

			return;
		}
	}
}


void R_SurfaceLightmapTexCoords (msurface_t *surf, float *verts, float *vec)
{
	float s;
	float t;

	// lightmap texture coordinates
	s = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
	s -= surf->texturemins[0];
	s += surf->light_s*16;
	s += 8;
	s /= LM_BLOCK_WIDTH*16; //surf->texinfo->texture->width;

	t = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
	t -= surf->texturemins[1];
	t += surf->light_t*16;
	t += 8;
	t /= LM_BLOCK_HEIGHT*16; //surf->texinfo->texture->height;

	verts[0] = s;
	verts[1] = t;
}


// returns a texture number and the position inside it
gllightmap_t *AllocBlock (int w, int h, int *x, int *y)
{
	int i, j;
	int best, best2;
	gllightmap_t *lm;

	for (lm = gllightmaps; lm; lm = lm->next)
	{
		best = LM_BLOCK_HEIGHT;

		for (i = 0; i < LM_BLOCK_WIDTH - w; i++)
		{
			best2 = 0;

			for (j = 0; j < w; j++)
			{
				if (lm->Allocated[i + j] >= best) break;

				if (lm->Allocated[i + j] > best2) best2 = lm->Allocated[i + j];
			}

			if (j == w)
			{
				// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > LM_BLOCK_HEIGHT)
			continue;

		for (i = 0; i < w; i++)
			lm->Allocated[*x + i] = best + h;

		return lm;
	}

	// allocate a new one
	lm = (gllightmap_t *) malloc (sizeof (gllightmap_t));
	memset (lm, 0, sizeof (gllightmap_t));

	// link it in
	lm->next = gllightmaps;
	gllightmaps = lm;

	// clear allocation
	memset (lm->Allocated, 0, sizeof (lm->Allocated));

	// call recursively to get the new lightmap
	return AllocBlock (w, h, x, y);
}


/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	// set lightmap size
	surf->smax = (surf->extents[0] >> 4) + 1;
	surf->tmax = (surf->extents[1] >> 4) + 1;

	// allocate a lightmap block
	surf->lightmap = AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t);

	// set data pointer
	surf->base = surf->lightmap->Data;
	surf->base += (surf->light_t * LM_BLOCK_WIDTH + surf->light_s) * LIGHTMAP_BYTES;

	// build the data
	R_BuildLightMap (surf, LM_BLOCK_WIDTH * LIGHTMAP_BYTES);
}


/*
==================
GL_BuildLightmaps

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
extern mvertex_t	*r_pcurrentvertbase;
extern model_t		*currentmodel;

void GL_BuildLightmaps (void)
{
	int		i, j;
	model_t	*m;
	gllightmap_t *lm;

	// no dlightcache
	r_framecount = 1;

	// clear down lightmaps
	if (gllightmaps)
	{
		i = 0;

		for (;;)
		{
			lm = gllightmaps->next;
			glDeleteTextures (1, (GLuint *) &gllightmaps->texnum);
			free (gllightmaps);
			gllightmaps = lm;

			i++;

			if (!gllightmaps) break;
		}

		Con_DPrintf ("Released %i lightmaps\n", i);
	}

	// build lightmaps and surfaces for all models
	for (j = 1; j < MAX_MODELS; j++)
	{
		m = cl.model_precache[j];

		// no more models
		if (!m) break;

		// don't build lightmaps for these model types
		if (m->type != mod_brush) continue;

		if (m->name[0] == '*') continue;

		r_pcurrentvertbase = m->vertexes;
		currentmodel = m;

		for (i = 0; i < m->numsurfaces; i++)
		{
			// no lightmaps on these surfs (polys have already been built)
			if (m->surfaces[i].flags & SURF_DRAWTURB) continue;
			if (m->surfaces[i].flags & SURF_DRAWSKY) continue;

			GL_CreateSurfaceLightmap (m->surfaces + i);
			BuildSurfaceDisplayList (m->surfaces + i);
		}
	}

	// upload all lightmaps that were filled
	i = 0;

	for (lm = gllightmaps; lm; lm = lm->next)
	{
		// no more used
		if (!lm->Allocated[0]) break;

		// set base data
		lm->Modified = false;

		lm->RectChange.l = LM_BLOCK_WIDTH;
		lm->RectChange.t = LM_BLOCK_HEIGHT;
		lm->RectChange.w = 0;
		lm->RectChange.h = 0;

		// create a texture object for it
		GL_AllocTexture (&lm->texnum);

		// upload
		GL_BindTexture (lm->texnum);

		glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE);

		glTexImage2D (GL_TEXTURE_2D, 
					  0, 
					  GL_RGBA, 
					  LM_BLOCK_WIDTH, 
					  LM_BLOCK_HEIGHT, 
					  0, 
					  GL_RGBA, 
					  GL_UNSIGNED_BYTE, 
					  lm->Data);

		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		// no polys
		lm->Polys = NULL;

		i++;
	}

	Con_DPrintf ("Uploaded %i lightmaps\n", i);
}

