/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

entity_t	r_worldentity;

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

int			currenttexture = -1;		// to avoid unnecessary texture sets

int			cnttextures[2] = {-1, -1};     // cached

gltexture_t	playertextures[16];		// up to 16 color translated skins

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf;
mleaf_t		*r_oldviewleaf = NULL;

texture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value


void R_MarkLeaves (void);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_fullbright = {"r_fullbright","0"};
cvar_t	r_shadows = {"r_shadows","0"};
cvar_t	r_novis = {"r_novis","0"};

cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels","1"};
cvar_t	gl_affinemodels = {"gl_affinemodels","0"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_playermip = {"gl_playermip","0"};
cvar_t	gl_nocolors = {"gl_nocolors","0"};
cvar_t	gl_reporttjunctions = {"gl_reporttjunctions","0"};
cvar_t	gl_doubleeyes = {"gl_doubleeys", "1"};
cvar_t	gl_brightness = {"gl_brightness", "1", true};

void R_LightPoint (vec3_t p, vec3_t colour);
void RotatePointAroundVector( vec3_t dst, const vec3_t dir, const vec3_t point, float degrees );
void R_AnimateLight (void);
void V_CalcBlend (void);
void R_DrawWorld (void);
void R_DrawWaterSurfaces (void);
void R_DrawParticles (void);

float			frametime;
float			pickupsign = 1;

/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	int		i;

	for (i=0 ; i<4 ; i++)
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
			return true;
	return false;
}


/*
=============================================================

  SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_Printf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

	// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
	// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t		*psprite;

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{
		// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{
		// normal sprite
		up = vup;
		right = vright;
	}

	glColor3f (1, 1, 1);

	GL_BindTexture (frame->gl_texture->texnum);

	surfverts[0].st[0] = 0;
	surfverts[0].st[1] = 0;
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	surfverts[0].v[0] = point[0];
	surfverts[0].v[1] = point[1];
	surfverts[0].v[2] = point[2];

	surfverts[1].st[0] = 1;
	surfverts[1].st[1] = 0;
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	surfverts[1].v[0] = point[0];
	surfverts[1].v[1] = point[1];
	surfverts[1].v[2] = point[2];

	surfverts[2].st[0] = 1;
	surfverts[2].st[1] = 1;
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	surfverts[2].v[0] = point[0];
	surfverts[2].v[1] = point[1];
	surfverts[2].v[2] = point[2];

	surfverts[3].st[0] = 0;
	surfverts[3].st[1] = 1;
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	surfverts[3].v[0] = point[0];
	surfverts[3].v[1] = point[1];
	surfverts[3].v[2] = point[2];

	glDrawArrays (GL_QUADS, 0, 4);
}

/*
=============================================================

  ALIAS MODELS

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
};

vec3_t	shadevector;
vec3_t	shadecolour;

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
;

float	*shadedots1 = r_avertexnormal_dots[0];
float	*shadedots2 = r_avertexnormal_dots[0];
float	lightLerpOffset;

int	lastposenum;
extern gltexture_t *blacktexture;


void GL_DrawAliasVerts (glvertex_t *cache, aliashdr_t *paliashdr, int pose1, int pose2, float backlerp)
{
	float l;
	float ldiff;
	float lquant[2];

	drawvertx_t	*verts1;
	drawvertx_t	*verts2;

	int		*order;
	int		count;

	float frontlerp = 1.0f - backlerp;

	lastposenum = pose1;

	verts1 = (drawvertx_t *) ((byte *) paliashdr + paliashdr->posedata);
	verts2 = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	order = (int *) ((byte *) paliashdr + paliashdr->commands);

	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		if (count < 0) count = -count;

		do
		{
			// texture coordinates come from the draw list
			cache->st[0] = ((float *) order)[0];
			cache->st[1] = ((float *) order)[1];

			// hack - we can't derive caustic coords too easily for these, so lets just set them to the regular texcoords
			// and hope for the best!
			cache->wc[0] = cache->st[0] / 2.0f;
			cache->wc[1] = cache->st[1] / 2.0f;

			// set up the lighting
			lquant[0] = (shadedots1[verts1->lightnormalindex] + (backlerp * (shadedots1[verts2->lightnormalindex] - shadedots1[verts1->lightnormalindex])));
			lquant[1] = (shadedots2[verts1->lightnormalindex] + (backlerp * (shadedots2[verts2->lightnormalindex] - shadedots2[verts1->lightnormalindex])));

			if (lquant[0] != lquant[1])
			{
				if (lquant[0] > lquant[1])
				{
					ldiff = lquant[0] - lquant[1];
					ldiff *= lightLerpOffset;
					l = lquant[0] - ldiff;
				}
				else
				{
					ldiff = lquant[1] - lquant[0];
					ldiff *= lightLerpOffset;
					l = lquant[0] + ldiff;
				}
			}
			else l = lquant[0];

			// reduce l to keep it consistent with the world lighting
			l *= 0.666f;

			cache->rgba[0] = l;
			cache->rgba[1] = l;
			cache->rgba[2] = l;
			cache->rgba[3] = 1;

			// compose vertexes
			cache->v[0] = verts1->v[0] * frontlerp + verts2->v[0] * backlerp;
			cache->v[1] = verts1->v[1] * frontlerp + verts2->v[1] * backlerp;
			cache->v[2] = verts1->v[2] * frontlerp + verts2->v[2] * backlerp;

			// advance pointers
			verts1++;
			verts2++;

			cache++;
			order += 2;
		} while (--count);
	}
}


/*
=============
GL_DrawAliasShadow
=============
*/
extern	vec3_t			lightspot;

void GL_DrawAliasShadow (aliashdr_t *paliashdr, int posenum)
{
}



void R_SetupAliasVerts (entity_t *e, glvertex_t *cache)
{
	int				pose, numposes;
	float			interval;
	float			blend;
	qboolean		lerpmdl = true;
	aliashdr_t		*paliashdr = (aliashdr_t *) Mod_Extradata (e->model);

	if ((e->frame >= paliashdr->numframes) || (e->frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", e->frame);
		e->frame = 0;
	}

	pose = paliashdr->frames[e->frame].firstpose;
	numposes = paliashdr->frames[e->frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[e->frame].interval;
		pose += (int)(cl.time / interval) % numposes;

		e->frame_interval = interval;
	}
	else e->frame_interval = 0.1;

	// conditions for turning lerping off
	if (e->pose2 != pose) lerpmdl = false;					// don't interpolate between different poses
	if (paliashdr->numposes == 1) lerpmdl = false;			// only one pose
	if (e == &cl.viewent && e->frame == 0) lerpmdl = false;	// super-nailgun spinning down bug
	if (r_framecount < 100) lerpmdl = false;					// reliving their dying throes

	// entities with these special flags aren't interpolated either
	if (e->model->flags & EF_GIB) lerpmdl = false;
	if (e->model->flags & EF_ZOMGIB) lerpmdl = false;
	if (e->model->flags & EF_ROCKET) lerpmdl = false;
	if (e->model->flags & EF_GRENADE) lerpmdl = false;
	if (e->model->flags & EF_ROTATE) lerpmdl = false;
	if (e->model->flags & EF_TRACER) lerpmdl = false;
	if (e->model->flags & EF_TRACER2) lerpmdl = false;
	if (e->model->flags & EF_TRACER3) lerpmdl = false;

	// wait till now to do this so we can correctly switch interpolation off
	// (viewent can switch mdls but keep the same pose, and therefore have invalid data cached in the poses)
	if (!lerpmdl)
	{
		e->frame_start_time = realtime;
		e->pose1 = e->pose2;
		e->pose2 = pose;
		blend = 0;
	}
	else blend = (realtime - e->frame_start_time) / e->frame_interval;

	// don't let blend pass 1
	if (cl.paused || blend > 1.0) blend = 1.0;

	GL_DrawAliasVerts (cache, paliashdr, e->pose1, e->pose2, blend);
}


//==================================================================================


void R_DrawEntitiesOnList (void)
{
	int		i;
	qboolean drawsprite = false;

	if (!r_drawentities.value) return;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_sprite:
			//drawsprite = true;
			break;

		default:
			break;
		}
	}

	if (drawsprite)
	{
		glEnable (GL_BLEND);
		R_ConfigureVertexArrays (VA_VERTEX | VA_TEXTURE0_ST);

		for (i = 0; i < cl_numvisedicts; i++)
		{
			currententity = cl_visedicts[i];

			if (currententity->model->type == mod_sprite) R_DrawSpriteModel (currententity);
		}

		R_ConfigureVertexArrays (VA_KILL);
		glDisable (GL_BLEND);
	}
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}


void R_SetFrustum (void)
{
	int		i;

	if (r_refdef.fov_x == 90) 
	{
		// front side is visible

		VectorAdd (vpn, vright, frustum[0].normal);
		VectorSubtract (vpn, vright, frustum[1].normal);

		VectorAdd (vpn, vup, frustum[2].normal);
		VectorSubtract (vpn, vup, frustum[3].normal);
	}
	else
	{
		// rotate VPN right by FOV_X/2 degrees
		RotatePointAroundVector( frustum[0].normal, vup, vpn, -(90-r_refdef.fov_x / 2 ) );
		// rotate VPN left by FOV_X/2 degrees
		RotatePointAroundVector( frustum[1].normal, vup, vpn, 90-r_refdef.fov_x / 2 );
		// rotate VPN up by FOV_X/2 degrees
		RotatePointAroundVector( frustum[2].normal, vright, vpn, 90-r_refdef.fov_y / 2 );
		// rotate VPN down by FOV_X/2 degrees
		RotatePointAroundVector( frustum[3].normal, vright, vpn, -( 90 - r_refdef.fov_y / 2 ) );
	}

	for (i=0 ; i<4 ; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
// don't allow cheats in multiplayer
	if (cl.maxclients > 1)
		Cvar_Set ("r_fullbright", "0");

	R_AnimateLight ();

	r_framecount++;

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);

	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;

	c_brush_polys = 0;
	c_alias_polys = 0;

}


void MYgluPerspective( GLdouble fovy, GLdouble aspect,
		     GLdouble zNear, GLdouble zFar )
{
   GLdouble xmin, xmax, ymin, ymax;

   ymax = zNear * tan( fovy * M_PI / 360.0 );
   ymin = -ymax;

   xmin = ymin * aspect;
   xmax = ymax * aspect;

   glFrustum( xmin, xmax, ymin, ymax, zNear, zFar );
}


/*
=============
R_SetupGL
=============
*/
void R_SetupGL (void)
{
	float	screenaspect;
	extern	int glwidth, glheight;
	int		x, x2, y2, y, w, h;

	// set up viewpoint
	x = r_refdef.vrect.x * glwidth / vid.width;
	x2 = (r_refdef.vrect.x + r_refdef.vrect.width) * glwidth / vid.width;
	y = (vid.height - r_refdef.vrect.y) * glheight / vid.height;
	y2 = (vid.height - (r_refdef.vrect.y + r_refdef.vrect.height)) * glheight / vid.height;

	// fudge around because of frac screen scale
	if (x > 0) x--;
	if (x2 < glwidth) x2++;
	if (y2 < 0) y2--;
	if (y < glheight) y++;

	w = x2 - x;
	h = y - y2;

	glViewport (glx + x, gly + y2, w, h);

    screenaspect = (float) r_refdef.vrect.width / r_refdef.vrect.height;

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();

	gluPerspective (r_refdef.fov_y, screenaspect, 4, r_farclip);

	glCullFace(GL_FRONT);

	// make sure we have identity in the world
	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	// now do it in software
	swLoadIdentity ();
	swRotatef (-90,  1, 0, 0);	    // put Z going up
	swRotatef (90,  0, 0, 1);	    // put Z going up
	swRotatef (-r_refdef.viewangles[2],  1, 0, 0);
	swRotatef (-r_refdef.viewangles[0],  0, 1, 0);
	swRotatef (-r_refdef.viewangles[1],  0, 0, 1);
	swTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

	swGetMatrixf (r_worldentity.matrix);

	// set drawing parms
	if (gl_cull.value)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
}

/*
================
R_RenderScene

r_refdef must be set before the first call
================
*/
void R_RenderScene (void)
{
	R_SetupFrame ();

	R_SetFrustum ();

	R_SetupGL ();

	R_MarkLeaves ();	// done here so we know if we're in water

	R_DrawWorld ();		// adds static entities to the list

	R_DrawEntitiesOnList ();

	R_DrawWaterSurfaces ();

	S_ExtraUpdate ();	// don't let sound get messed up if going slow

	R_DrawParticles ();


}


/*
=============
R_Clear
=============
*/
void R_Clear (void)
{
	glDepthFunc (GL_LEQUAL);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glDepthRange (0, 1);
}


/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	double	time1 = 0;
	double time2;

	frametime = cl.time - cl.oldtime;

	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

	if (r_speeds.value)
	{
		time1 = Sys_FloatTime ();
		c_brush_polys = 0;
		c_alias_polys = 0;
	}

	glFinish ();

	R_Clear ();

	// render normal view
	R_RenderScene ();

	if (r_speeds.value)
	{
		time2 = Sys_FloatTime ();
		Con_Printf ("%3i ms  %4i wpoly %4i epoly\n", (int)((time2-time1)*1000), c_brush_polys, c_alias_polys); 
	}
}
