/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"


void R_InitParticles (void);
void R_ClearParticles (void);
void GL_BuildLightmaps (void);
void R_InitParticleTextures (void);

extern int vertcount;
extern	gltexture_t	playertextures[];
extern glvertex_t *vertcache;



/*
==========================================================================================
testing of blend funcs
----------------------
this is a good way to experiment with different blending modes to see what looks
best - just replace a call to glBlendFunc with this.  be careful to only test one
blend at a time (e.g. water/light/caustic/etc) or you may get confused!!!
==========================================================================================
*/
GLenum bFactors[11] =
{
	GL_ZERO,					// 0
	GL_ONE,						// 1
	GL_SRC_COLOR,				// 2
	GL_ONE_MINUS_SRC_COLOR,		// 3
	GL_DST_COLOR,				// 4
	GL_ONE_MINUS_DST_COLOR,		// 5
	GL_SRC_ALPHA,				// 6
	GL_ONE_MINUS_SRC_ALPHA,		// 7
	GL_DST_ALPHA,				// 8
	GL_ONE_MINUS_DST_ALPHA,		// 9
	GL_SRC_ALPHA_SATURATE		// 10
};


cvar_t gl_blend_src = {"gl_blend_src", "0"};
cvar_t gl_blend_dst = {"gl_blend_dst", "0"};

void GL_BlendTest (void)
{
	if (gl_blend_src.value < 0) Cvar_SetValue ("gl_blend_src", 0);
	if (gl_blend_src.value > 10) Cvar_SetValue ("gl_blend_src", 10);

	if (gl_blend_dst.value < 0) Cvar_SetValue ("gl_blend_dst", 0);
	if (gl_blend_dst.value > 9) Cvar_SetValue ("gl_blend_dst", 9);

	glBlendFunc (bFactors[(int) gl_blend_src.value], bFactors[(int) gl_blend_dst.value]);
}


// keep a cache of model surfaces globally as hunk compression will mess up with dynamic allocation
msurface_t ModelSurfaces[MAX_MODELS];

void R_InitModelSurfaces (void)
{
	int i;

	for (i = 0; i < MAX_MODELS; i++)
	{
		// this is all we need to do here...
		ModelSurfaces[i].polys = NULL;
	}
}


/*
===============
R_Init
===============
*/
void R_InitSinTable (void);
void R_InitSkySphere (void);

void R_Init (void)
{	
	extern byte *hunk_base;
	int i;

	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);

	Cvar_RegisterVariable (&r_norefresh);
	Cvar_RegisterVariable (&r_fullbright);
	Cvar_RegisterVariable (&r_drawentities);
	Cvar_RegisterVariable (&r_drawviewmodel);
	Cvar_RegisterVariable (&r_shadows);
	Cvar_RegisterVariable (&r_novis);
	Cvar_RegisterVariable (&r_speeds);

	Cvar_RegisterVariable (&gl_cull);
	Cvar_RegisterVariable (&gl_smoothmodels);
	Cvar_RegisterVariable (&gl_affinemodels);
	Cvar_RegisterVariable (&gl_polyblend);
	Cvar_RegisterVariable (&gl_playermip);
	Cvar_RegisterVariable (&gl_nocolors);

	Cvar_RegisterVariable (&gl_reporttjunctions);

	Cvar_RegisterVariable (&gl_doubleeyes);
	Cvar_RegisterVariable (&r_skyspeed);

	Cvar_RegisterVariable (&gl_blend_src);
	Cvar_RegisterVariable (&gl_blend_dst);

	Cvar_RegisterVariable (&gl_brightness);

	R_InitParticles ();
	R_InitParticleTextures ();

	// set up player textures
	for (i = 0; i < 16; i++)
		GL_AllocTexture (&playertextures[i]);

	// other initialisation that doesn't fit in elsewhere
	R_InitSinTable ();
	R_InitSkySphere ();
	R_InitModelSurfaces ();
}

/*
===============
R_TranslatePlayerSkin

Translates a skin texture by the per-player color lookup
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int		top, bottom;
	byte	translate[256];
	int		i, s;
	model_t	*model;
	aliashdr_t *paliashdr;
	byte	*original;
	int			inwidth, inheight;
	byte	translated[320*200];

	top = cl.scores[playernum].colors & 0xf0;
	bottom = (cl.scores[playernum].colors & 0x0f) << 4;

	for (i=0 ; i<256 ; i++)
		translate[i] = i;

	for (i=0 ; i<16 ; i++)
	{
		if (top < 128)	// the artists made some backwards ranges.  sigh.
			translate[TOP_RANGE+i] = top+i;
		else
			translate[TOP_RANGE+i] = top+15-i;
				
		if (bottom < 128)
			translate[BOTTOM_RANGE+i] = bottom+i;
		else
			translate[BOTTOM_RANGE+i] = bottom+15-i;
	}

	//
	// locate the original skin pixels
	//
	currententity = &cl_entities[1+playernum];
	model = currententity->model;
	if (!model)
		return;		// player doesn't have a model yet
	if (model->type != mod_alias)
		return; // only translate skins on alias models

	paliashdr = (aliashdr_t *)Mod_Extradata (model);
	s = paliashdr->skinwidth * paliashdr->skinheight;
	if (currententity->skinnum < 0 || currententity->skinnum >= paliashdr->numskins) {
		Con_Printf("(%d): Invalid player skin #%d\n", playernum, currententity->skinnum);
		original = (byte *)paliashdr + paliashdr->texels[0];
	} else
		original = (byte *)paliashdr + paliashdr->texels[currententity->skinnum];
	if (s & 3)
		Sys_Error ("R_TranslateSkin: s&3");

	inwidth = paliashdr->skinwidth;
	inheight = paliashdr->skinheight;

	for (i = 0; i < s; i += 4)
	{
		translated[i] = translate[original[i]];
		translated[i+1] = translate[original[i+1]];
		translated[i+2] = translate[original[i+2]];
		translated[i+3] = translate[original[i+3]];
	}

	// set up a dummy GLT
	// make the name invalid so that it won't go through any external texture loaders
	sprintf (playertextures[playernum].identifier, "playertexture_%i", playernum);
	playertextures[playernum].width = paliashdr->skinwidth;
	playertextures[playernum].height = paliashdr->skinheight;
	playertextures[playernum].bpp = 8;
	playertextures[playernum].data = translated;
	playertextures[playernum].flags = TEX_MIPMAP;

	// upload it
	GL_UploadTexture (&playertextures[playernum]);
}


/*
===============
R_NewMap
===============
*/
void GL_DeleteTextures (void);
extern int maxverts;
void R_InitBlood (void);

extern vec3_t farmin;
extern vec3_t farmax;

void R_BuildModelSurfaces (void)
{
	int i;
	model_t *m;
	glpoly_t *p;
	int *order;
	int count;
	aliashdr_t *pheader;

	// clear down all allocated polys first
	for (i = 1; i < MAX_MODELS; i++)
	{
		// check has it any polys allocated
		if (ModelSurfaces[i].polys)
		{
			// free allocated polys
			for (;;)
			{
				p = ModelSurfaces[i].polys->next;
				free (ModelSurfaces[i].polys);
				ModelSurfaces[i].polys = p;

				if (!ModelSurfaces[i].polys) break;
			}
		}
	}

	// now go through the precached list and set the polys
	for (i = 1; i < MAX_MODELS; i++)
	{
		m = cl.model_precache[i];

		// no more models
		if (!m) break;

		// don't build lightmaps for these model types
		if (m->type != mod_alias) continue;

		// take a surface
		m->surfaces = &ModelSurfaces[i];
		m->numsurfaces = 1;

		// point header at the cached data
		pheader = (aliashdr_t *) m->cache.data;

		if (!pheader) continue;

		order = (int *) ((byte *) pheader + pheader->commands);

		for (;;)
		{
			// get the vertex count and primitive type
			count = *order++;

			// done
			if (!count) break;

			p = (glpoly_t *) malloc (sizeof (glpoly_t));

			p->next = m->surfaces->polys;
			m->surfaces->polys = p;

			if (count < 0) count = -count;

			// get the correct order
			do
			{
				order += 2;
			} while (--count);
		}

		// now we go through it again and set up the polys.  this is required cos the method above allocates the
		// polys in the reverse of the order they're drawn.  it's ugly, it's vile, it's a hack, but it works and
		// it was too late at night to think of a more elegant method.
		order = (int *) ((byte *) pheader + pheader->commands);
		p = m->surfaces->polys;

		for (;;)
		{
			// get the vertex count and primitive type
			count = *order++;

			// done
			if (!count) break;

			if (count < 0)
			{
				count = -count;
				p->PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
			}
			else p->PRIMITIVE_TYPE = GL_TRIANGLE_STRIP;

			// this doesn't need verts allocated as it will use the vertcache for drawing
			p->numverts = count;

			// get the correct order
			do
			{
				order += 2;
			} while (--count);

			p = p->next;
		}
	}

	Con_DPrintf ("*** MODEL SURFACES COMPLETE ***\n");
}



lightent_t *lightents = NULL;
int numlights;


void StringToVector (char *TheString, float *TheVector)
{
	char *v, *w;
	int i;

	v = TheString;
	w = TheString;

	for (i = 0; i < 3; i++)
	{
		while (*v && *v != ' ')
			v++;

		*v = 0;

		TheVector[i] = atof (w);
		w = v = v + 1;
	}
}


char *R_ParseEntity (char *entities)
{
	char classname[64];
	char origin[64];
	char light[64];
	char style[64];
	char key[64];
	lightent_t *le;

	classname[0] = 0;
	origin[0] = 0;
	style[0] = 0;

	for (;;)
	{
		entities = COM_Parse (entities);

		if (com_token[0] == '}') break;

		strcpy (key, com_token);

		entities = COM_Parse (entities);

		// save out the fields we want to use
		if (!strnicmp (key, "classname", 9)) strcpy (classname, com_token);
		if (!strnicmp (key, "origin", 6)) strcpy (origin, com_token);
		if (!strnicmp (key, "style", 5)) strcpy (style, com_token);
		if (!strnicmp (key, "light", 5)) strcpy (light, com_token);
	}

	// check for a light
	if (!strnicmp (classname, "light", 5))
	{
		// spawn a new light ent
		le = (lightent_t *) malloc (sizeof (lightent_t));
		le->next = lightents;
		lightents = le;

		// fill in the light entity
		StringToVector (origin, le->origin);
		le->intensity = atoi (light);
		le->style = atoi (style);
		le->lightleaf = Mod_PointInLeaf (le->origin, cl.worldmodel);

		// set default intensity
		if (!le->intensity) le->intensity = 200;

		// initial colour
		le->colour[0] = 1;
		le->colour[1] = 1;
		le->colour[2] = 1;

		if (!le->style)
		{
			// hack in a style based on the light name
			if (!strnicmp (classname, "light_flame", 11)) le->style = 1;
			if (!strnicmp (classname, "light_torch", 11)) le->style = 1;
		}

		// hack in colour for certain light types
		if (!strnicmp (classname, "light_flame", 11))
		{
			le->colour[0] = 1.5f;
			le->colour[1] = 1.0f;
			le->colour[2] = 0.5f;
		}
		else if (!strnicmp (classname, "light_torch", 11))
		{
			le->colour[0] = 1.5f;
			le->colour[1] = 1.0f;
			le->colour[2] = 0.5f;
		}
		else if (le->style == 1 || le->style == 6 || le->style == 3 || le->style == 7 || le->style == 8)
		{
			le->colour[0] = 1.5f;
			le->colour[1] = 1.0f;
			le->colour[2] = 0.5f;
		}
		else
		{
			msurface_t **mark = le->lightleaf->firstmarksurface;
			int c = le->lightleaf->nummarksurfaces;
			float lumaaccum = 0;

			if (c)
			{
				le->colour[0] = 0;
				le->colour[1] = 0;
				le->colour[2] = 0;

				do
				{
					if (!(*mark)->texinfo->texture)
					{
						mark++;
						continue;
					}

					if (!(*mark)->texinfo->texture->gl_texture)
					{
						mark++;
						continue;
					}

					if ((*mark)->texinfo->texture->gl_texture->luma)
					{
						le->colour[0] += (*mark)->texinfo->texture->gl_texture->colour[0];
						le->colour[1] += (*mark)->texinfo->texture->gl_texture->colour[1];
						le->colour[2] += (*mark)->texinfo->texture->gl_texture->colour[2];

						lumaaccum += 1.0f;
					}

					mark++;
				} while (--c);

				if (lumaaccum)
				{
					le->colour[0] /= lumaaccum;
					le->colour[1] /= lumaaccum;
					le->colour[2] /= lumaaccum;

					lumaaccum = le->colour[0] + le->colour[1] + le->colour[2];
					lumaaccum = 3.0f / lumaaccum;

					le->colour[0] *= lumaaccum;
					le->colour[1] *= lumaaccum;
					le->colour[2] *= lumaaccum;
				}
				else
				{
					le->colour[0] = 1;
					le->colour[1] = 1;
					le->colour[2] = 1;
				}
			}
			else
			{
				le->colour[0] = 1;
				le->colour[1] = 1;
				le->colour[2] = 1;
			}
		}

		// increment the light counter
		numlights++;
	}

	return entities;
}


// the entities have already been validated in PR_Edict.c, so we'll be loose on the error checking
void R_ReadLights (void)
{
	char *entities = cl.worldmodel->entities;
	int i;
	lightent_t *le;

	if (lightents)
	{
		for (i = 0; ;)
		{
			le = lightents->next;
			free (lightents);
			lightents = le;

			i++;

			if (!lightents) break;
		}
	}

	numlights = 0;

	for (;;)
	{
		entities = COM_Parse (entities);
		if (!entities) break;

		entities = R_ParseEntity (entities);
	}

	Con_DPrintf ("read %i lights\n", numlights);
}

extern int r_lightframecount;

void R_NewMap (void)
{
	int		i;
	vec3_t	dist;
	
	for (i=0 ; i<256 ; i++)
		d_lightstylevalue[i] = 264;		// normal light value

	memset (&r_worldentity, 0, sizeof(r_worldentity));
	r_worldentity.model = cl.worldmodel;

	// clear out efrags in case the level hasn't been reloaded
	// FIXME: is this one short?
	// yes
	for (i=0 ; i<cl.worldmodel->numleafs + 1; i++)
		cl.worldmodel->leafs[i].efrags = NULL;

	r_viewleaf = NULL;
	R_ClearParticles ();

	// set up the far clipping plane dimensions
	farmin[0] = farmin[1] = farmin[2] = 999999999;
	farmax[0] = farmax[1] = farmax[2] = -999999999;

	GL_DeleteTextures ();
	GL_BuildLightmaps ();

	// build surfaces and polys for all the alias models
	R_BuildModelSurfaces ();

	// identify sky texture
	skytexturenum = -1;

	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		if (!cl.worldmodel->textures[i])
			continue;
		if (!Q_strncmp(cl.worldmodel->textures[i]->name,"sky",3) )
			skytexturenum = i;
	}

	Con_DPrintf ("Used %i vertexes (vert size; %i bytes) (ram: %i k)\n", vertcount, sizeof (glvertex_t), ((vertcount * sizeof (glvertex_t)) / 1024) + 1);
	Con_DPrintf ("Max verts per poly: %i\n", maxverts);

	// clear the vertex cache if we have one
	if (vertcache) free (vertcache);

	// allocate an extra slot for the view entity
	vertcache = (glvertex_t *) malloc (maxverts * sizeof (glvertex_t) * (MAX_VISEDICTS + 1));

	// report back.  most maps this will hover around 8 to 10 MB.
	Con_DPrintf ("Vertex Cache: %i KB\n", maxverts * sizeof (glvertex_t) * (MAX_VISEDICTS + 1) / 1024);

	// set far clipping plane - give it an extra 512 for headroom
	dist[0] = farmax[0] - farmin[0];
	dist[1] = farmax[1] - farmin[1];
	dist[2] = farmax[2] - farmin[2];

	r_farclip = sqrt (dist[0] * dist[0] + dist[1] * dist[1] + dist[2] * dist[2]) + 512.0f;

	Con_DPrintf ("Far Clipping Plane: %f\n", r_farclip);

	R_ReadLights ();

	vertcount = 0;
	R_InitBlood ();

	r_lightframecount = 0;
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
	int			i;
	float		start, stop, time;

	glDrawBuffer  (GL_FRONT);
	glFinish ();

	start = Sys_FloatTime ();
	for (i=0 ; i<128 ; i++)
	{
		r_refdef.viewangles[1] = i/128.0*360.0;
		R_RenderView ();
	}

	glFinish ();
	stop = Sys_FloatTime ();
	time = stop-start;
	Con_Printf ("%f seconds (%f fps)\n", time, 128/time);

	glDrawBuffer  (GL_BACK);
	GL_EndRendering ();
}

void D_FlushCaches (void)
{
}


