/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __MODEL__
#define __MODEL__

#include "modelgen.h"
#include "spritegn.h"

/*

d*_t structures are on-disk representations
m*_t structures are in-memory

*/

// entity effects

#define	EF_BRIGHTFIELD			1
#define	EF_MUZZLEFLASH 			2
#define	EF_BRIGHTLIGHT 			4
#define	EF_DIMLIGHT 			8


/*
==============================================================================

BRUSH MODELS

==============================================================================
*/


//
// in memory representation
//
// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	vec3_t		position;
} mvertex_t;

#define	SIDE_FRONT	0
#define	SIDE_BACK	1
#define	SIDE_ON		2


// plane_t structure
// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct mplane_s
{
	vec3_t	normal;
	float	dist;
	byte	type;			// for texture axis selection and fast side tests
	byte	signbits;		// signx + signy<<1 + signz<<1
	byte	pad[2];
} mplane_t;

typedef struct texture_s
{
	// base texture properties
	char		name[16];
	unsigned	width, height;

	// texture objects used
	struct gltexture_s *gl_texture;

	// animation data
	int			anim_total;				// total tenths in sequence ( 0 = no)
	int			anim_min, anim_max;		// time for this frame min <=time< max

	// textures in the animation sequence
	struct texture_s *anim_next;		// in the animation sequence
	struct texture_s *alternate_anims;	// bmodels in frmae 1 use these

	// software quake legacy
	unsigned	offsets[MIPLEVELS];		// four mip maps stored
} texture_t;


// base surface types
#define	SURF_PLANEBACK		2
#define	SURF_DRAWSKY		4
#define SURF_DRAWSPRITE		8
#define SURF_DRAWTURB		16
#define SURF_DRAWTILED		32
#define SURF_DRAWBACKGROUND	64
#define SURF_UNDERWATER		128
#define SURF_BMODEL			4096

// extra water surface types
#define SURF_DRAWLAVA		256
#define SURF_DRAWSLIME		512
#define SURF_DRAWTELE		1024
#define SURF_DRAWWATER		2048

void GL_AllocTexture (int *texnum);

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	unsigned short	v[2];
	unsigned int	cachededgeoffset;
} medge_t;

typedef struct
{
	float		vecs[2][4];
	float		mipadjust;
	texture_t	*texture;
	int			flags;
} mtexinfo_t;


// define the vertex data we're going to use
typedef struct glvertex_s
{
	// vertex positions
	float v[3];

	// texture coords
	float st[2];

	// lightmap coords
	float lm[2];

	// water caustic coords
	float wc[2];

	// colour
	float rgba[4];
} glvertex_t;

typedef struct glpoly_s
{
	// chain for surfaces with multiple polys
	struct	glpoly_s	*next;

	// chain for lightmaps
	struct	glpoly_s	*chain;

	// number of verts in this poly
	int		numverts;

	// drawing flags
	int		flags;

	// mid point of the poly (for underwater checking on models)
	vec3_t	midpoint;

	// type of OpenGL primitive used for drawing (triangles, quads or triangle fan)
	int		PRIMITIVE_TYPE;

	// matrix for transformation
	float	*matrix;

	// pointer to verts used by the poly
	glvertex_t *verts;
} glpoly_t;


// texture loading
typedef struct gltexture_s
{
	// texture properties
	int		texnum;
	char	identifier[64];
	int		width;
	int		height;
	int		bpp;

	// pointer to data so we only need to pass this struct into our loading routine
	byte	*data;

	// flags
	int		flags;

	// checksum
	byte	md5digest[16];

	qboolean	resident;

	// texture chain for drawing
	struct msurface_s *texturechain;

	// identify if the texture is used in the current pass
	int			passflags;

	// next in list
	struct gltexture_s *next;

	// pointer to luma for the texture
	struct gltexture_s *luma;

	// colour of the texture (for fullbrights and liquids)
	float colour[3];
} gltexture_t;

// chained list of all textures in use
extern gltexture_t	*gltextures;


// texture upload flags
#define TEX_MIPMAP		1
#define TEX_LUMA		2
#define TEX_32BPP		4
#define TEX_BSP			8
#define TEX_KILL		16

// pass flags for surface refresh
#define PASS_NODRAW			0
#define PASS_TEXTURED		1
#define PASS_LIGHTMAPPED	2
#define PASS_COLOURED		4
#define PASS_LUMA			8
#define PASS_CAUSTIC		16
#define PASS_SKY			32
#define PASS_LIQUID			64
#define PASS_LUMACAUSTIC	128
#define PASS_DYNAMIC		256
#define PASS_ENTITY			512


#define	LM_BLOCK_WIDTH	128
#define	LM_BLOCK_HEIGHT	128

#define LIGHTMAP_BYTES 4

typedef struct glRect_s
{
	unsigned char l,t,w,h;
} glRect_t;


typedef struct gllightmap_s
{
	glpoly_t *Polys;
	qboolean Modified;
	glRect_t RectChange;
	int Allocated[LM_BLOCK_WIDTH];
	byte Data[LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * LIGHTMAP_BYTES];

	int texnum;

	struct gllightmap_s *next;
} gllightmap_t;


typedef struct msurface_s
{
	int			visframe;		// should be drawn when node is crossed

	mplane_t	*plane;
	int			flags;

	int			firstedge;	// look up in model->surfedges[], negative numbers
	int			numedges;	// are backwards edges
	
	short		texturemins[2];
	short		extents[2];

	int			light_s, light_t;	// gl lightmap coordinates
	int			smax, tmax;

	glpoly_t	*polys;				// multiple if warped
	byte		warpindex[2];		// warp texcoords

	struct	msurface_s	*texturechain;

	// void pointer (to be cast to an entity_t * at run time) to the entity using this surf
	// (surfs drawn through the main render only have one entity)
	void		*entity;

	// flags for which passes this surf uses
	int			passflags;

	// != r_framecount the first time we visit the surf each frame
	int			passframe;

	mtexinfo_t	*texinfo;

// lighting info
	int			dlightframe;
	byte		dlightbits[(MAX_DLIGHTS + 7) / 8];

	struct gllightmap_s *lightmap;
	byte		*base;

	byte		styles[MAXLIGHTMAPS];
	int			cached_light[MAXLIGHTMAPS];	// values currently used in lightmap
	qboolean	cached_dlight;				// true if dynamic light in cache
	byte		*samples;		// [numstyles*surfsize]
} msurface_t;


typedef struct mnode_s
{
// common with leaf
	int			contents;		// 0, to differentiate from leafs
	int			visframe;		// node needs to be traversed if current
	int			lightvisframe;
	
	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

// node specific
	mplane_t	*plane;
	struct mnode_s	*children[2];

	unsigned short		firstsurface;
	unsigned short		numsurfaces;
} mnode_t;



typedef struct mleaf_s
{
// common with node
	int			contents;		// wil be a negative contents number
	int			visframe;		// node needs to be traversed if current
	int			lightvisframe;

	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

// leaf specific
	byte		*compressed_vis;
	efrag_t		*efrags;

	msurface_t	**firstmarksurface;
	int			nummarksurfaces;
	int			key;			// BSP sequence number for leaf's contents
	byte		ambient_sound_level[NUM_AMBIENTS];
} mleaf_t;

// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct
{
	dclipnode_t	*clipnodes;
	mplane_t	*planes;
	int			firstclipnode;
	int			lastclipnode;
	vec3_t		clip_mins;
	vec3_t		clip_maxs;
} hull_t;

/*
==============================================================================

SPRITE MODELS

==============================================================================
*/


// FIXME: shorten these?
typedef struct mspriteframe_s
{
	int		width;
	int		height;
	float	up, down, left, right;
	gltexture_t *gl_texture;
} mspriteframe_t;

typedef struct
{
	int				numframes;
	float			*intervals;
	mspriteframe_t	*frames[1];
} mspritegroup_t;

typedef struct
{
	spriteframetype_t	type;
	mspriteframe_t		*frameptr;
} mspriteframedesc_t;

typedef struct
{
	int					type;
	int					maxwidth;
	int					maxheight;
	int					numframes;
	float				beamlength;		// remove?
	void				*cachespot;		// remove?
	mspriteframedesc_t	frames[1];
} msprite_t;


/*
==============================================================================

ALIAS MODELS

Alias models are position independent, so the cache manager can move them.
==============================================================================
*/

typedef struct drawvertx_s
{
	float	v[3];
	byte	lightnormalindex;
	qboolean nodraw;
} drawvertx_t;

typedef struct
{
	int					firstpose;
	int					numposes;
	float				interval;
	trivertx_t			bboxmin;
	trivertx_t			bboxmax;
	int					frame;
	char				name[16];
} maliasframedesc_t;

typedef struct
{
	trivertx_t			bboxmin;
	trivertx_t			bboxmax;
	int					frame;
} maliasgroupframedesc_t;

typedef struct
{
	int						numframes;
	int						intervals;
	maliasgroupframedesc_t	frames[1];
} maliasgroup_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct mtriangle_s {
	int					facesfront;
	int					vertindex[3];
} mtriangle_t;


#define	MAX_SKINS	32
typedef struct {
	int			ident;
	int			version;
	vec3_t		scale;
	vec3_t		scale_origin;
	float		boundingradius;
	vec3_t		eyeposition;
	int			numskins;
	int			skinwidth;
	int			skinheight;
	int			numverts;
	int			numtris;
	int			numframes;
	synctype_t	synctype;
	int			flags;
	float		size;

	int					numposes;
	int					poseverts;
	int					posedata;	// numposes*poseverts trivert_t
	int					commands;	// gl command list with embedded s/t
	gltexture_t			*gl_texture[MAX_SKINS][4];
	int					texels[MAX_SKINS];	// only for player skins
	maliasframedesc_t	frames[1];	// variable sized
} aliashdr_t;

#define	MAXALIASVERTS	2000
#define	MAXALIASFRAMES	256
#define	MAXALIASTRIS	2048
extern	aliashdr_t	*pheader;
extern	stvert_t	stverts[MAXALIASVERTS];
extern	mtriangle_t	triangles[MAXALIASTRIS];
extern	trivertx_t	*poseverts[MAXALIASFRAMES];

//===================================================================

//
// Whole model
//


// worldtypes
#define WORLDTYPE_WIZARD	0
#define WORLDTYPE_METAL		1
#define WORLDTYPE_BASE		2


typedef enum {mod_brush, mod_sprite, mod_alias} modtype_t;

#define	EF_ROCKET	1			// leave a trail
#define	EF_GRENADE	2			// leave a trail
#define	EF_GIB		4			// leave a trail
#define	EF_ROTATE	8			// rotate (bonus items)
#define	EF_TRACER	16			// green split trail
#define	EF_ZOMGIB	32			// small blood trail
#define	EF_TRACER2	64			// orange split trail + rotate
#define	EF_TRACER3	128			// purple trail

typedef struct model_s
{
	char		name[MAX_QPATH];
	qboolean	needload;		// bmodels and sprites don't cache normally

	modtype_t	type;
	int			numframes;
	synctype_t	synctype;
	
	int			flags;

//
// volume occupied by the model graphics
//		
	vec3_t		mins, maxs;
	float		radius;

//
// solid volume for clipping 
//
	qboolean	clipbox;
	vec3_t		clipmins, clipmaxs;

//
// brush model
//
	int			firstmodelsurface, nummodelsurfaces;

	int			numsubmodels;
	dmodel_t	*submodels;

	int			numplanes;
	mplane_t	*planes;

	int			numleafs;		// number of visible leafs, not counting 0
	mleaf_t		*leafs;

	int			numvertexes;
	mvertex_t	*vertexes;

	int			numedges;
	medge_t		*edges;

	int			numnodes;
	mnode_t		*nodes;

	int			numtexinfo;
	mtexinfo_t	*texinfo;

	int			numsurfaces;
	msurface_t	*surfaces;

	int			numsurfedges;
	int			*surfedges;

	int			numclipnodes;
	dclipnode_t	*clipnodes;

	int			nummarksurfaces;
	msurface_t	**marksurfaces;

	hull_t		hulls[MAX_MAP_HULLS];

	int			numtextures;
	texture_t	**textures;

	byte		*visdata;
	byte		*lightdata;
	char		*entities;

	// world type of the model for setting various blah client-side
	int			worldtype;

//
// additional model data
//
	cache_user_t	cache;		// only access through Mod_Extradata

} model_t;

//============================================================================

void	Mod_Init (void);
void	Mod_ClearAll (void);
model_t *Mod_ForName (char *name, qboolean crash);
void	*Mod_Extradata (model_t *mod);	// handles caching
void	Mod_TouchModel (char *name);

mleaf_t *Mod_PointInLeaf (float *p, model_t *model);
byte	*Mod_LeafPVS (mleaf_t *leaf, model_t *model);

#endif	// __MODEL__
