/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cvar.c -- dynamic variable tracking

#include "quakedef.h"

cvar_t	*cvar_vars;
char	*cvar_null_string = "";

int		num_cvars = 0;

extern qboolean force_vis;

/*
============
Cvar_FindVar
============
*/
cvar_t *Cvar_FindVar (char *var_name)
{
	cvar_t	*var;
	
	for (var=cvar_vars ; var ; var=var->next)
		if (!strcmp (var_name, var->name))
			return var;

	return NULL;
}

/*
============
Cvar_VariableValue
============
*/
float	Cvar_VariableValue (char *var_name)
{
	cvar_t	*var;
	
	var = Cvar_FindVar (var_name);
	if (!var)
		return 0;
	return atof (var->string);
}


/*
============
Cvar_VariableString
============
*/
char *Cvar_VariableString (char *var_name)
{
	cvar_t *var;
	
	var = Cvar_FindVar (var_name);
	if (!var)
		return cvar_null_string;
	return var->string;
}


/*
============
Cvar_CompleteVariable
============
*/
char *Cvar_CompleteVariable (int search_num, char *partial)
{
	cvar_t		*cvar;
	int			len;

	len = strlen (partial);

	if (!len)
		return NULL;

	// check functions
	for (cvar = cvar_vars; cvar; cvar = cvar->next)
	{
		if (!strncmp (partial,cvar->name, len))
		{
			if (search_num == 0)
				return cvar->name;
			else search_num--;
		}
	}

	return NULL;
}


int Cvar_SearchVariables (char *partial)
{
	cvar_t		*cvar;
	int			len;
	int cvar_found;
	int first_find;
	
	len = strlen(partial);
	
	if (!len)
		return 0;

	cvar_found = 0;
	first_find = 1;

	// check functions
	for (cvar = cvar_vars; cvar; cvar = cvar->next)
	{
		if (!strncmp (partial, cvar->name, len))
		{
			if (first_find) 
			{
				Con_Printf ("]%s\n\n", partial);
				first_find = 0;
			}

			cvar_found++;
			Con_Printf ("%s   ", cvar->name);
		}
	}

	return cvar_found;
}


/*
============
Cvar_Set
============
*/
void Cvar_Set (char *var_name, char *value)
{
	cvar_t	*var;
	qboolean changed;

	var = Cvar_FindVar (var_name);

	if (!var)
	{
		// there is an error in C code if this happens
		Con_Printf ("Cvar_Set: variable %s not found\n", var_name);
		return;
	}

	// r_novis setting from 1 to 0 should make force_vis = true
	if (!strcmp (var_name, "r_novis"))
		if (r_novis.value) force_vis = true;

	changed = strcmp(var->string, value);
	
	Z_Free (mainzone, var->string);	// free the old value string
	
	var->string = Z_Malloc (mainzone, strlen(value)+1);
	strcpy (var->string, value);
	var->value = atof (var->string);

	if (var->server && changed)
	{
		if (sv.active)
			SV_BroadcastPrintf ("\"%s\" changed to \"%s\"\n", var->name, var->string);
	}
}

/*
============
Cvar_SetValue
============
*/
void Cvar_SetValue (char *var_name, float value)
{
	char	val[32];
	
	sprintf (val, "%f",value);
	Cvar_Set (var_name, val);
}


/*
============
Cvar_RegisterVariable

Adds a freestanding variable to the variable list.
============
*/
void Cvar_RegisterVariable (cvar_t *variable)
{
	char	*oldstr;
	
	// first check to see if it has allready been defined
	if (Cvar_FindVar (variable->name))
	{
		Con_Printf ("Can't register variable %s, allready defined\n", variable->name);
		return;
	}
	
	// check for overlap with a command
	if (Cmd_Exists (variable->name))
	{
		Con_Printf ("Cvar_RegisterVariable: %s is a command\n", variable->name);
		return;
	}
		
	// copy the value off, because future sets will Z_Free it
	num_cvars++;	// we need this for tab completion
	oldstr = variable->string;
	variable->string = Z_Malloc (mainzone, strlen(variable->string)+1);	
	strcpy (variable->string, oldstr);
	variable->value = atof (variable->string);
	
	// link the variable in
	// FIXME - it's just a simple poxy linked list, so sort them alphabetically...
	variable->next = cvar_vars;
	cvar_vars = variable;
}

/*
============
Cvar_Command

Handles variable inspection and changing from the console
============
*/
qboolean Cvar_Command (void)
{
	cvar_t			*v;

	// check variables
	v = Cvar_FindVar (Cmd_Argv(0));
	if (!v)
		return false;
		
	// perform a variable print or set
	if (Cmd_Argc() == 1)
	{
		Con_Printf ("\"%s\" is \"%s\"\n", v->name, v->string);
		return true;
	}

	Cvar_Set (v->name, Cmd_Argv(1));
	return true;
}


/*
============
Cvar_WriteVariables

Writes lines containing "set variable value" for all variables
with the archive flag set to true.
============
*/
void Cvar_WriteVariables (FILE *f)
{
	cvar_t	*var;
	
	for (var = cvar_vars ; var ; var = var->next)
		if (var->archive)
			fprintf (f, "%s \"%s\"\n", var->name, var->string);
}



void Cmd_Cvarlist_f (void)
{
	FILE *f;
	cvar_t *var;
	char cvarfile[1024];

	if (Cmd_Argc () == 2)
	{
		sprintf (cvarfile, "%s/%s", com_gamedir, Cmd_Argv (1));

		f = fopen (cvarfile, "w");

		if (!f) return;
	}

	for (var = cvar_vars; var; var = var->next)
	{
		if (Cmd_Argc () == 2)
			fprintf (f, "%s\n", var->name);
		else if (Cmd_Argc () == 1)
			Con_Printf ("%s\n", var->name);
		else
		{
			Con_Printf ("cvarlist : prints a list of cvars to screen or file\n");
			return;
		}
	}

	if (Cmd_Argc () == 2) fclose (f);
}


