/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "gl_edicts.h"

// some of this stuff is only done on the worldmodel, other stuff is done on all bsp models
// (eg. the WAD)
void ParseEntityLump (lump_t *l)
{
	byte *in;

	int i, j;
	int entnum;

	qboolean in_entity;

	char in_string[1024];
	char last_name[1024];

	qboolean reading_name;
	qboolean reading_value;

	// stuff we need to store out for an entity - although the max value is 1024, most of
	// these are never going to be that big, so we can reduce the sizes accordingly.
	char e_classname[100];
	int e_worldtype;
	char e_skybox[100];
	char e_wad[1024];

	// initialize wadstring to a zero length string - same for skybox
	wadstring[0] = '\0';
	e_skybox[0] = '\0';

	// default draw distance for halflife maps
	if (hlbsp) Cvar_SetValue ("r_drawdist", 4096);

	if (!l->filelen)
	{
		// the bsp will not load anyway if there is no entities in it - we need at least
		// worldspawn and a start point - so we'll just do nothing and let quake handle it
		// in it's normal fashion (probably by crashing) - although the build tools should
		// catch all these problems anyway.
		loadmodel->entities = NULL;
		return;
	}

	in = mod_base + l->fileofs;

	entnum = 0;
	in_entity = false;

	reading_name = true;
	reading_value = false;

	// i is always the current position of the pointer into the entity lump - hence the fact
	// that i do some strange looking stuff with it below...
	for (i = 0; i < l->filelen;)
	{
		if (in_entity && *in == '"')
		{
			// read and store out a key/value pair - incrementing 'i' here is intentional!
			// no need for an end condition cos the build tools will have already tested the
			// maximum lengths for us
			for (j = 0;;)
			{
				in++; i++;
				if (*in == '"') break;
				in_string[j++] = *in;
			}

			in_string[j] = '\0';

			// only quake bsps have a worldtype key - this is not 100% reliable (especially
			// if using worldcraft) but is quite close - we can perform other tests later.
			if (!strcmp (in_string, "worldtype") && hlbsp)
			{
				hlbsp = false;
				qsg30 = true;
			}

			if (!worldtype_found)
			{
				// switch whether we're currently reading the name or the value
				if (reading_name)
				{
					reading_name = false;
					reading_value = true;

					// store out the name cos what we do with the value depends on it
					strcpy (last_name, in_string);
				}
				else if (reading_value)
				{
					reading_name = true;
					reading_value = false;

					// act according to the value of last_name - there is a limited amount of
					// stuff here we're actually interested in - but it can be expanded later
					if (!strcmp (last_name, "classname"))
						strcpy (e_classname, in_string);
					else if (!strcmp (last_name, "worldtype"))
						e_worldtype = atoi (in_string);
					else if (!strcmp (last_name, "_skybox"))
						strcpy (e_skybox, in_string);
					else if (!strcmp (last_name, "_skyname"))
						strcpy (e_skybox, in_string);
					else if (!strcmp (last_name, "skyname"))
						strcpy (e_skybox, in_string);
					else if (!strcmp (last_name, "wad"))
						strcpy (e_wad, in_string);
					else if (!strcmp (last_name, "MaxRange"))
						Cvar_SetValue ("r_drawdist", atof (in_string));
				}
			}
			else
			{
				// switch whether we're currently reading the name or the value
				if (reading_name)
				{
					reading_name = false;
					reading_value = true;

					// store out the name cos what we do with the value depends on it
					strcpy (last_name, in_string);
				}
				else if (reading_value)
				{
					reading_name = true;
					reading_value = false;

					// act according to the value of last_name - there is a limited amount of
					// stuff here we're actually interested in - but it can be expanded later
					// we also do wad here in case of hl bsp models
					if (!strcmp (last_name, "classname"))
						strcpy (e_classname, in_string);
					else if (!strcmp (last_name, "wad"))
						strcpy (e_wad, in_string);
				}
			}
		}

		if (*in == '{')
		{
			// start a new entity
			in_entity = true;

			e_wad[0] = '\0';

			if (!worldtype_found)
			{
				// initialize all the stuff we're saving out as zero length strings
				e_classname[0] = '\0';
				strcpy (e_skybox, "none");
				e_worldtype = 0;
			}
		}
		else if (*in == '}')
		{
			// finish current entity
			entnum++;
			in_entity = false;

			if (e_wad[0] != '\0')
			{
				// save out the wad string for hl wad3 loading
				strcpy (wadstring, e_wad);
			}

			if (!worldtype_found)
			{
				// set variables, etc depending on the entity - in practice, we'll only need
				// worldspawn - but this still does the entire entity lump in case we need to
				// expand it later...
				if (!strcmp (e_classname, "worldspawn"))
				{
					// worldtype
					global_worldtype = e_worldtype;

					// skybox
					Con_DPrintf ("Name is %s\n", e_skybox);

					// persistent skybox
					if (!strcmp (e_skybox, "none"))
					{
						if (sb_persist.value)
							gl_useskybox = 1;
						else
						{
							gl_useskybox = 0;
							Cvar_Set ("_r_skybox", "none");
						}
					}
					else if (e_skybox[0] == '\0')
					{
						gl_useskybox = 0;
						Cvar_Set ("_r_skybox", "none");
					}
					else
					{
						gl_useskybox = 1;
						Cvar_Set ("_r_skybox", e_skybox);
					}
				}
			}
		}

		in++; i++;
	}

	Con_DPrintf ("ParseEntityLump : %i Entities read from BSP\n", entnum);

	// we might as well load the entities while we're here...
	loadmodel->entities = Hunk_AllocName ( l->filelen, loadname);	
	memcpy (loadmodel->entities, mod_base + l->fileofs, l->filelen);
}


