// lightmap building, modification, regeneration and uploading
#include "quakedef.h"
#include "gl_frontend_shared.h"
#include "glext.h"


int		lightmap_bytes = 4;
int		lightmap_textures = 0;

unsigned 	blocklightcolours[3][18*18];


// for lightmaps
int	current_lightmap_texture;

int			allocated[BLOCK_WIDTH];

// the lightmap texture data needs to be kept in
// main memory so texsubimage can update properly
byte		lightmaps[4 * BLOCK_WIDTH * BLOCK_HEIGHT * MAX_LIGHTMAPS];

typedef struct glRect_s
{
	unsigned char l,t,w,h;
} glRect_t;



glpoly_t	*lightmap_polys[MAX_LIGHTMAPS];
qboolean	lightmap_modified[MAX_LIGHTMAPS];
glRect_t	lightmap_rectchange[MAX_LIGHTMAPS];

//int			allocated[MAX_LIGHTMAPS][BLOCK_WIDTH];

// the lightmap texture data needs to be kept in
// main memory so texsubimage can update properly
byte		lightmaps[4*MAX_LIGHTMAPS*BLOCK_WIDTH*BLOCK_HEIGHT];



/*
=============================================================================

  DYNAMIC LIGHTS

=============================================================================
*/

/*
=============
R_MarkLights

For Brush Models cos the visibility code screws them up
=============
*/
void R_MarkLights (dlight_t *light, int bit, mnode_t *node)
{
	mplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;
	int			i, sidebit;

	if (node->contents < 0)
		return;

	splitplane = node->plane;

	switch (splitplane->type)
	{
	case PLANE_X:
		dist = light->origin[0] - splitplane->dist;
		break;
	case PLANE_Y:
		dist = light->origin[1] - splitplane->dist;
		break;
	case PLANE_Z:
		dist = light->origin[2] - splitplane->dist;
		break;
	default:
		dist = DotProduct (light->origin, splitplane->normal) - splitplane->dist;
		break;
	}

	if (dist > light->radius)
	{
		// outside the radius of the dlight
		R_MarkLights (light, bit, node->children[0]);
		return;
	}

	if (dist < -light->radius)
	{
		// outside the radius of the dlight
		R_MarkLights (light, bit, node->children[1]);
		return;
	}

	// mark the polygons
	surf = cl.worldmodel->surfaces + node->firstsurface;

	if (dist >= 0)
		sidebit = 0;
	else sidebit = SURF_PLANEBACK;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		if ((surf->flags & SURF_PLANEBACK) != sidebit)
			continue;	// wrong side

		if (surf->dlightframe != r_framecount)
		{
			// hit by this light
			surf->dlightbits = 0;
			surf->dlightframe = r_framecount;
		}

		surf->dlightbits |= bit;
	}

	R_MarkLights (light, bit, node->children[0]);
	R_MarkLights (light, bit, node->children[1]);
}


/*
=============
R_MarkWorldLights

for the world, so we can take visibility into account.
two types of visibility are valid when marking surfs - is the surf visible to the client?
if not don't mark it.  is the surf visible to the dlight?  if not, don't mark it.
=============
*/
void R_MarkWorldLights (dlight_t *light, int bit, mnode_t *node)
{
	mplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;
	int			i, sidebit, side;

	// prevent evil promiscuous dlight spamming
	if (node->contents == CONTENTS_SOLID) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;
	if (node->dlightframe != r_framecount) return;
	if (node->visframe != r_visframecount) return;
	if (node->contents < 0) return;

	splitplane = node->plane;

	switch (splitplane->type)
	{
	case PLANE_X:
		dist = light->origin[0] - splitplane->dist;
		break;
	case PLANE_Y:
		dist = light->origin[1] - splitplane->dist;
		break;
	case PLANE_Z:
		dist = light->origin[2] - splitplane->dist;
		break;
	default:
		dist = DotProduct (light->origin, splitplane->normal) - splitplane->dist;
		break;
	}

	if (dist > light->radius)
	{
		// outside the radius of the dlight on front side
		R_MarkWorldLights (light, bit, node->children[0]);
		return;
	}

	if (dist < -light->radius)
	{
		// outside the radius of the dlight on back side
		R_MarkWorldLights (light, bit, node->children[1]);
		return;
	}

	if (dist >= 0)
	{
		side = 0;
		sidebit = 0;
	}
	else
	{
		side = 1;
		sidebit = SURF_PLANEBACK;
	}

	R_MarkWorldLights (light, bit, node->children[side]);

	// mark the polygons
	surf = cl.worldmodel->surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		// prevent evil promiscuous dlight spamming
		if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;
		if (surf->flags & (SURF_DRAWTURB | SURF_DRAWSKY)) continue;
		if (surf->visframe != r_framecount) continue;

		if (surf->dlightframe != r_framecount)
		{
			// hit by this light
			surf->dlightbits = 0;
			surf->dlightframe = r_framecount;
		}

		surf->dlightbits |= bit;
	}

	R_MarkWorldLights (light, bit, node->children[!side]);
}


// dlight visibility for the world model
// -------------------------------------
// gives a fairly reliable prevention of dlight shine through.  the speed hit
// incurred here is made up for by not uploading as many new lightmaps - especially
// on slower machines.  visually it's worth it anyway...
void DLightVis (dlight_t *dl)
{
	byte *vis;
	mleaf_t *dleaf = Mod_PointInLeaf (dl->origin, cl.worldmodel);
	int i;
	mnode_t *node;

	vis = Mod_LeafPVS (dleaf, cl.worldmodel);

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			// mark the hit nodes
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->dlightframe == r_framecount) break;

				node->dlightframe = r_framecount;
				node = node->parent;
			}
			while (node);
		}
	}
}



/*
=============
R_PushDlights
=============
*/
void R_PushDlights (void)
{
	int		i;
	dlight_t	*dl;

	dl = cl_dlights;

	for (i = 0; i < MAX_DLIGHTS; i++, dl++)
	{
		if (dl->die < realtime || !dl->radius)
			continue;

		// calculate visibility from the perspective 
		// of the light to kill evil dlight spamming
		DLightVis (dl);
		R_MarkWorldLights (dl, 1 << i, cl.worldmodel->nodes);
	}
}





/*
===============
R_AddDynamicLights
===============
*/
// add a single dynamic light
void R_AddDynamicLight (msurface_t *surf, vec3_t colour, vec3_t origin, float radius, float minlight)
{
	int			sd, td;
	float		dist;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	float		dlr;
	float		dlg;
	float		dlb;
	int			add;
	int			sacc, tacc;

	dlr = colour[0] * 256;
	dlg = colour[1] * 256;
	dlb = colour[2] * 256;

	dist = DotProduct (origin, surf->plane->normal) - surf->plane->dist;
	radius -= fabs (dist);

	if (radius < minlight) return;

	minlight = radius - minlight;

	for (i = 0; i < 3; i++)
		impact[i] = origin[i] - surf->plane->normal[i] * dist;

	local[0] = DotProduct (impact, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
	local[1] = DotProduct (impact, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];

	local[0] -= surf->texturemins[0];
	local[1] -= surf->texturemins[1];
		
	for (t = 0, tacc = 0; t < surf->tmax; t++, tacc += 16)
	{
		td = local[1] - tacc;

		if (td < 0) td = -td;

		add = t * surf->smax;

		for (s = 0, sacc = 0; s < surf->smax; s++, sacc += 16)
		{
			sd = local[0] - sacc;

			if (sd < 0) sd = -sd;

			if (sd > td)
				dist = sd + (td >> 1);
			else dist = td + (sd >> 1);

			if (dist < minlight)
			{
				blocklightcolours[0][add + s] += (radius - dist) * dlr;
				blocklightcolours[1][add + s] += (radius - dist) * dlg;
				blocklightcolours[2][add + s] += (radius - dist) * dlb;
			}
		}
	}
}


void R_AddDynamicLights (msurface_t *surf)
{
	int			lnum;
	dlight_t	*dl;

	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		dl = &cl_dlights[lnum];

		if (dl->die < realtime || !dl->radius) continue;
		if (!(surf->dlightbits & (1 << lnum))) continue;

		R_AddDynamicLight (surf, dl->colour, dl->origin, dl->radius, dl->minlight);
	}
}






/*
==========================================================================================

LIGHTMAP MODIFICATION UPDATING

==========================================================================================
*/


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
#define CDIV  200
#define CMULT 255


byte BLTable[262144];

qboolean BLTableInit = false;

void InitBLTable (void)
{
	int i, j;
	float k;

	memset (BLTable, 255, sizeof (BLTable));

	for (i = 0; i < 65535; i++)
	{
		j = i >> 7;

		if (j > 255) j = 255;

		k = ((float) j) / CDIV;
		k *= k;
		j = k * CMULT;

		if (j > 255) j = 255;

		BLTable[i] = j;
	}

	BLTableInit = true;
}


void R_BuildLightMap (msurface_t *surf, byte *dest, int stride)
{
	int			t;
	int			i, j, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	int			lightadj[4];
	unsigned	*bl;
	unsigned	*blcr, *blcg, *blcb;
	int			r, g, b, a;
	int			style;

	if (!BLTableInit) InitBLTable ();

	surf->cached_dlight = (surf->dlightframe == r_framecount);

	size = surf->smax * surf->tmax;
	lightmap = surf->samples;

	// set to full bright if no light data
	if (r_fullbright.value)
	{
		for (i = 0; i < size; i++)
			blocklightcolours[0][i] = 
			blocklightcolours[1][i] = 
			blocklightcolours[2][i] = 255 * 256;

		goto store;
	}

	// clear to no light
	for (i = 0; i < size; i++)
		blocklightcolours[0][i] = 
		blocklightcolours[1][i] = 
		blocklightcolours[2][i] = 0;

	// add all the lightmaps
	if (lightmap)
	{
		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
		{
			style = surf->styles[maps];
			scale = d_lightstylevalue[style];
			surf->cached_light[maps] = scale;	// 8.8 fraction

			// scale can NEVER be 1... (it can be 0 though...)
			if (scale)
			{
				for (i = 0, j = 0; i < size; i++)
				{
					blocklightcolours[0][i] += lightmap[j++] * scale;
					blocklightcolours[1][i] += lightmap[j++] * scale;
					blocklightcolours[2][i] += lightmap[j++] * scale;
				}
			}

			lightmap += size * 3;	// skip to next lightmap
		}
	}

	// add all the dynamic lights
	if (surf->dlightframe == r_framecount)
		R_AddDynamicLights (surf);

	// bound, invert, and shift
store:;
	stride -= (surf->smax << 2);

	blcr = blocklightcolours[0];
	blcg = blocklightcolours[1];
	blcb = blocklightcolours[2];

	// RGBA lightmaps
	for (i = 0; i < surf->tmax ; i++, dest += stride)
	{
		for (j = 0; j < surf->smax; j++)
		{
			// we're taking a BIIIIG risk here, but in practice they never get
			// higher than 170,000 or so, so we can probably afford to
			dest[0] = BLTable[*blcr++];
			dest[1] = BLTable[*blcg++];
			dest[2] = BLTable[*blcb++];
			dest[3] = 255;

			dest += 4;
		}
	}
}


void R_UpdateSurfaceLightmap (msurface_t *surf)
{
	int map;
	glRect_t *theRect;
	byte *base;

	if ((surf->dlightframe == r_framecount || surf->cached_dlight) && r_dynamic.value)
		goto remakeLM;

	// this was MAX_LIGHTMAPS which is a totally different thing altogether!!!
	for (map = 0; map < MAXLIGHTMAPS && surf->styles[map] != 255; map++)
	{
		if (d_lightstylevalue[surf->styles[map]] != surf->cached_light[map])
		{
remakeLM:;
			lightmap_modified[surf->lightmaptexturenum] = true;

			// calculate how much of the lightmap has changed so we only upload as much
			// as we need to
			theRect = &lightmap_rectchange[surf->lightmaptexturenum];

			if (surf->light_t < theRect->t)
			{
				if (theRect->h) theRect->h += theRect->t - surf->light_t;

				theRect->t = surf->light_t;
			}

			if (surf->light_s < theRect->l)
			{
				if (theRect->w) theRect->w += theRect->l - surf->light_s;

				theRect->l = surf->light_s;
			}

			if ((theRect->w + theRect->l) < (surf->light_s + surf->smax))
				theRect->w = (surf->light_s - theRect->l) + surf->smax;

			if ((theRect->h + theRect->t) < (surf->light_t + surf->tmax))
				theRect->h = (surf->light_t - theRect->t) + surf->tmax;

			R_BuildLightMap (surf, surf->base, BLOCK_WIDTH * lightmap_bytes);
		}
	}

	// add it to the chain for drawing
	surf->polys->chain = lightmap_polys[surf->lightmaptexturenum];
	lightmap_polys[surf->lightmaptexturenum] = surf->polys;
}


/*
=============================================================================

  POLYGON RECONSTRUCTION

  MUST BE CALLED *BEFORE* LIGHTMAPS ARE BUILT

=============================================================================
*/


void GL_BuildPolygonFromSurface (model_t *mod, msurface_t *fa)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;

	// reconstruct the polygon
	pedges = mod->edges;
	lnumverts = fa->numedges;

	// draw texture
	poly = Hunk_Alloc (sizeof(glpoly_t) + (lnumverts-4) * VERTEXSIZE * sizeof (float));
	poly->next = fa->polys;
	poly->flags = fa->flags;
	fa->polys = poly;
	poly->numverts = lnumverts;

	for (i=0 ; i<lnumverts ; i++)
	{
		lindex = mod->surfedges[fa->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = mod->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = mod->vertexes[r_pedge->v[1]].position;
		}

		// poly verts
		poly->verts[i][0] = vec[0];
		poly->verts[i][1] = vec[1];
		poly->verts[i][2] = vec[2];

		// texture s and t
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s /= fa->texinfo->texture->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t /= fa->texinfo->texture->height;

		poly->verts[i][3] = s;
		poly->verts[i][4] = t;

		// allocate 2 new verts for a 64 * 64 version of the texture, and set their data
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s /= 64;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t /= 64;

		poly->verts[i][7] = s;
		poly->verts[i][8] = t;
	}
}


/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/


static void LM_InitBlock (void)
{
	// clear the block to no light
	memset (allocated, 0, sizeof (allocated));
}


static void LM_UploadBlock (void)
{
	glBindTexture (GL_TEXTURE_2D, lightmap_textures + current_lightmap_texture);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glTexImage2D (GL_TEXTURE_2D,
				  0,
				  lightmap_bytes,
				  BLOCK_WIDTH,
				  BLOCK_HEIGHT,
				  0,
				  GL_RGBA,
				  GL_UNSIGNED_BYTE,
				  lightmaps + (current_lightmap_texture * BLOCK_WIDTH * BLOCK_HEIGHT * lightmap_bytes));

	lightmap_modified[current_lightmap_texture] = false;

	lightmap_rectchange[current_lightmap_texture].l = BLOCK_WIDTH;
	lightmap_rectchange[current_lightmap_texture].t = BLOCK_HEIGHT;
	lightmap_rectchange[current_lightmap_texture].w = 0;
	lightmap_rectchange[current_lightmap_texture].h = 0;

	if (++current_lightmap_texture >= MAX_LIGHTMAPS)
		Host_Error ("Error: MAX_LIGHTMAPS exceeded\n");
}


static qboolean LM_AllocBlock (int w, int h, int *x, int *y)
{
	int i, j;
	int best, best2;

	best = BLOCK_HEIGHT;

	for (i = 0; i < BLOCK_WIDTH - w; i++)
	{
		best2 = 0;

		for (j = 0; j < w; j++)
		{
			if (allocated[i + j] >= best) break;

			if (allocated[i + j] > best2)
				best2 = allocated[i + j];
		}

		if (j == w)
		{
			// this is a valid spot
			*x = i;
			*y = best = best2;
		}
	}

	if (best + h > BLOCK_HEIGHT)
		return false;

	for (i = 0; i < w; i++)
		allocated[*x + i] = best + h;

	return true;
}


void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	byte *base;
	vec3_t vec;
	float s, t;
	int i;

	// is this check redundant???
	if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB))
		return;

	// allocate space in a lightmap (also fills in light_s and light_t)
	if (!LM_AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t))
	{
		LM_UploadBlock ();
		LM_InitBlock ();

		if (!LM_AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t))
			Sys_Error ("Consecutive calls to LM_AllocBlock failed");
	}

	// calculate lightmap texture coordinates
	for (i = 0; i < surf->polys->numverts; i++)
	{
		vec[0] = surf->polys->verts[i][0];
		vec[1] = surf->polys->verts[i][1];
		vec[2] = surf->polys->verts[i][2];

		s = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
		s -= surf->texturemins[0];
		s += surf->light_s * 16;
		s += 8;
		s /= BLOCK_WIDTH * 16;

		t = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
		t -= surf->texturemins[1];
		t += surf->light_t * 16;
		t += 8;
		t /= BLOCK_HEIGHT * 16;

		surf->polys->verts[i][5] = s;
		surf->polys->verts[i][6] = t;
	}

	// build the lightmap
	surf->lightmaptexturenum = current_lightmap_texture;

	base = lightmaps + (surf->lightmaptexturenum * lightmap_bytes * BLOCK_WIDTH * BLOCK_HEIGHT);
	base += (surf->light_t * BLOCK_WIDTH + surf->light_s) * lightmap_bytes;

	// save out the base to avoid recalculating it at run time
	surf->base = base;

	R_BuildLightMap (surf, base, BLOCK_WIDTH * lightmap_bytes);
}


void GL_BeginBuildingLightmaps (void)
{
	LM_InitBlock ();

	r_framecount = 1;

	Con_DPrintf ("GL_BeginBuildingLightmaps\n");
}


void GL_EndBuildingLightmaps (void)
{
	Con_DPrintf ("GL_EndBuildingLightmaps\n");

	LM_UploadBlock ();
}


void GL_BuildAllLightmaps (void)
{
	int i;
	int j;
	model_t *m;
	msurface_t *surf;

	if (!lightmap_textures)
	{
		lightmap_textures = texture_extension_number;
		texture_extension_number += MAX_LIGHTMAPS;
	}

	current_lightmap_texture = 0;
	memset (lightmap_modified, 0, sizeof (lightmap_modified));
	memset (lightmap_rectchange, 0, sizeof (lightmap_rectchange));
	memset (lightmap_polys, 0, sizeof (lightmap_polys));

	GL_BeginBuildingLightmaps ();

	for (i = 1; i < MAX_MODELS; i++)
	{
		m = cl.model_precache[i];

		if (!m) break;
		if (m->name[0] == '*') continue;
		if (m->type != mod_brush) continue;

		for (j = 0, surf = m->surfaces; j < m->numsurfaces; j++, surf++)
		{
			if (surf->flags & (SURF_DRAWTURB | SURF_DRAWSKY)) continue;

			GL_BuildPolygonFromSurface (m, surf);
			GL_CreateSurfaceLightmap (surf);
		}
	}

	GL_EndBuildingLightmaps ();
}



/*
=============================================================================

  LIGHTMAP BLEND RENDERING

=============================================================================
*/



void R_BlendLightmaps (void)
{
	int			i, j;
	glpoly_t	*p;
	float		*v;
	glRect_t	*theRect;

	if (r_fullbright.value) return;

	glDepthMask (0);

	glBlendFunc (GL_ZERO, GL_SRC_COLOR);
	glEnable (GL_BLEND);

	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		p = lightmap_polys[i];

		if (!p) continue;

		glBindTexture (GL_TEXTURE_2D, lightmap_textures + i);

		if (lightmap_modified[i])
		{
			// for the next frame
			lightmap_modified[i] = false;

			theRect = &lightmap_rectchange[i];

			// only upload as much as we need to
			glTexSubImage2D (GL_TEXTURE_2D,
							 0, 
							 0, 
							 theRect->t, 
							 BLOCK_WIDTH, 
							 theRect->h, 
							 GL_RGBA, 
							 GL_UNSIGNED_BYTE, 
							 lightmaps + (i * BLOCK_HEIGHT + theRect->t) * BLOCK_WIDTH * lightmap_bytes);

			theRect->l = BLOCK_WIDTH;
			theRect->t = BLOCK_HEIGHT;
			theRect->h = 0;
			theRect->w = 0;
		}

		for (; p; p = p->chain)
		{
			{
				glBegin (GL_POLYGON);
				v = p->verts[0];
				for (j = 0; j < p->numverts; j++, v += VERTEXSIZE)
				{
					glTexCoord2f (v[5], v[6]);
					glVertex3fv (v);
				}
				glEnd ();
			}
		}
	}


	glDisable (GL_BLEND);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glDepthMask (1);
}


