// visibility handling for the world model


#include "quakedef.h"
#include "gl_frontend_shared.h"
#include "glext.h"


/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			i, c, side, sidebit, *pindex;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		d, dot;
	vec3_t		mins, maxs;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;
	
	// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *)node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		if (pleaf->efrags)
			R_StoreEfrags (&pleaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;

	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;

	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;

	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0)
	{
		side = 0;
		sidebit = 0;
	}
	else
	{
		side = 1;
		sidebit = SURF_PLANEBACK;
	}

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	c = node->numsurfaces;

	for (c = node->numsurfaces, surf = cl.worldmodel->surfaces + node->firstsurface; c; c--, surf++)
	{
		if (surf->visframe != r_framecount) continue;
		if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

		if (surf->flags & SURF_DRAWSKY)
		{
			// we don't chain sky as it is not drawn using surfs from the BSP
			skyRender = true;
			continue;
		}
		else if (surf->flags & SURF_DRAWTURB)
		{
			surf->texturechain = waterchain;
			waterchain = surf;
			liquidRender = true;
			continue;
		}
		else
		{
			// add to the correct texture chain
			surf->texturechain = surf->texinfo->texture->texturechain;
			surf->texinfo->texture->texturechain = surf;
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}



/*
===============
AddVisibleLeaf

Adds the specified leaf to the PVS
This is tolerant of adding the same leaf twice...
===============
*/
void AddVisibleLeaf (mleaf_t *VLeaf)
{
	byte	*vis;
	mnode_t	*node;
	int		i;
	byte	solid[4096];

	// already been added this frame
//	if (VLeaf->waterVisFrame == r_visframecount) return;

	if (r_novis.value)
	{
		vis = solid;

		// set it to 255 so the bits always match
		memset (solid, 0xff, (cl.worldmodel->numleafs + 7) >> 3);
	}
	else vis = Mod_LeafPVS (VLeaf, cl.worldmodel);

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			// mark the visible nodes
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				// already added
				if (node->visframe == r_visframecount) break;

				node->visframe = r_visframecount;
				node = node->parent;
			}
			while (node);
		}
	}

	// mark it as having been checked
//	VLeaf->waterVisFrame = r_visframecount;
}


/*
===============
R_MarkLeaves
===============
*/
//extern cvar_t gl_lockpvs;

void R_MarkLeaves (void)
{
	// still in the same leaf
	if (r_oldviewleaf == r_viewleaf) return;

//	if (gl_lockpvs.value) return;

	// this only increments when the viewleaf changes
	r_visframecount++;

	// add the viewleaf (we can add PVS for any leaf we want with this function)
	AddVisibleLeaf (r_viewleaf);

	r_oldviewleaf = r_viewleaf;
}



